// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _StringSgExpressionAVLMap_h
#ifdef __GNUG__
// #pragma interface
#endif
#define _StringSgExpressionAVLMap_h 1

#include "expressionMap.h"

struct StringSgExpressionAVLNode
{
  StringSgExpressionAVLNode*      lt;
  StringSgExpressionAVLNode*      rt;
  String                 item;
  SgExpression                 cont;
  char                stat;
                      StringSgExpressionAVLNode(String& h, SgExpression& c, 
                                    StringSgExpressionAVLNode* l=0, StringSgExpressionAVLNode* r=0);
                      ~StringSgExpressionAVLNode();
};

inline StringSgExpressionAVLNode::StringSgExpressionAVLNode(String& h, SgExpression& c, 
                                    StringSgExpressionAVLNode* l, StringSgExpressionAVLNode* r)
     :item(h), cont(c), lt(l), rt(r), stat(0) {}

inline StringSgExpressionAVLNode::~StringSgExpressionAVLNode() {}

typedef StringSgExpressionAVLNode* StringSgExpressionAVLNodePtr;


class StringSgExpressionAVLMap : public StringSgExpressionMap
{
protected:
  StringSgExpressionAVLNode*   root;

  StringSgExpressionAVLNode*   leftmost();
  StringSgExpressionAVLNode*   rightmost();
  StringSgExpressionAVLNode*   pred(StringSgExpressionAVLNode* t);
  StringSgExpressionAVLNode*   succ(StringSgExpressionAVLNode* t);
  void            _kill(StringSgExpressionAVLNode* t);
  void            _add(StringSgExpressionAVLNode*& t);
  void            _del(StringSgExpressionAVLNode* p, StringSgExpressionAVLNode*& t);

public:
                StringSgExpressionAVLMap(SgExpression& dflt);
                StringSgExpressionAVLMap(StringSgExpressionAVLMap& a);
                ~StringSgExpressionAVLMap();

  SgExpression&          operator [] (String& key);

  void          del(String& key);

  Pix           first();
  void          next(Pix& i);
  String&          key(Pix i);
  SgExpression&          contents(Pix i);

  Pix           seek(String& key);
  int           contains(String& key);

  void          clear(); 

  Pix           last();
  void          prev(Pix& i);

  int           OK();
};

inline StringSgExpressionAVLMap::~StringSgExpressionAVLMap()
{
  _kill(root);
}

inline StringSgExpressionAVLMap::StringSgExpressionAVLMap(SgExpression& dflt) :StringSgExpressionMap(dflt)
{
  root = 0;
}

inline Pix StringSgExpressionAVLMap::first()
{
  return Pix(leftmost());
}

inline Pix StringSgExpressionAVLMap::last()
{
  return Pix(rightmost());
}

inline void StringSgExpressionAVLMap::next(Pix& i)
{
  if (i != 0) i = Pix(succ((StringSgExpressionAVLNode*)i));
}

inline void StringSgExpressionAVLMap::prev(Pix& i)
{
  if (i != 0) i = Pix(pred((StringSgExpressionAVLNode*)i));
}

inline String& StringSgExpressionAVLMap::key(Pix i)
{
  if (i == 0) error("null Pix");
  return ((StringSgExpressionAVLNode*)i)->item;
}

inline SgExpression& StringSgExpressionAVLMap::contents(Pix i)
{
  if (i == 0) error("null Pix");
  return ((StringSgExpressionAVLNode*)i)->cont;
}

inline void StringSgExpressionAVLMap::clear()
{
  _kill(root);
  count = 0;
  root = 0;
}

inline int StringSgExpressionAVLMap::contains(String& key)
{
  return seek(key) != 0;
}

#endif
