// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _StringFMProgBlockAVLMap_h
#ifdef __GNUG__
// #pragma interface
#endif
#define _StringFMProgBlockAVLMap_h 1

#include "fmProgBlkMap.h"

struct StringFMProgBlockAVLNode
{
  StringFMProgBlockAVLNode*      lt;
  StringFMProgBlockAVLNode*      rt;
  String                 item;
  FMProgBlock                 cont;
  char                stat;
                      StringFMProgBlockAVLNode(String& h, FMProgBlock& c, 
                                    StringFMProgBlockAVLNode* l=0, StringFMProgBlockAVLNode* r=0);
                      ~StringFMProgBlockAVLNode();
};

inline StringFMProgBlockAVLNode::StringFMProgBlockAVLNode(String& h, FMProgBlock& c, 
                                    StringFMProgBlockAVLNode* l, StringFMProgBlockAVLNode* r)
     :item(h), cont(c), lt(l), rt(r), stat(0) {}

inline StringFMProgBlockAVLNode::~StringFMProgBlockAVLNode() {}

typedef StringFMProgBlockAVLNode* StringFMProgBlockAVLNodePtr;


class StringFMProgBlockAVLMap : public StringFMProgBlockMap
{
protected:
  StringFMProgBlockAVLNode*   root;

  StringFMProgBlockAVLNode*   leftmost();
  StringFMProgBlockAVLNode*   rightmost();
  StringFMProgBlockAVLNode*   pred(StringFMProgBlockAVLNode* t);
  StringFMProgBlockAVLNode*   succ(StringFMProgBlockAVLNode* t);
  void            _kill(StringFMProgBlockAVLNode* t);
  void            _add(StringFMProgBlockAVLNode*& t);
  void            _del(StringFMProgBlockAVLNode* p, StringFMProgBlockAVLNode*& t);

public:
                StringFMProgBlockAVLMap(FMProgBlock& dflt);
                StringFMProgBlockAVLMap(StringFMProgBlockAVLMap& a);
                ~StringFMProgBlockAVLMap();

  FMProgBlock&          operator [] (String& key);

  void          del(String& key);

  Pix           first();
  void          next(Pix& i);
  String&          key(Pix i);
  FMProgBlock&          contents(Pix i);

  Pix           seek(String& key);
  int           contains(String& key);

  void          clear(); 

  Pix           last();
  void          prev(Pix& i);

  int           OK();
};

inline StringFMProgBlockAVLMap::~StringFMProgBlockAVLMap()
{
  _kill(root);
}

inline StringFMProgBlockAVLMap::StringFMProgBlockAVLMap(FMProgBlock& dflt) :StringFMProgBlockMap(dflt)
{
  root = 0;
}

inline Pix StringFMProgBlockAVLMap::first()
{
  return Pix(leftmost());
}

inline Pix StringFMProgBlockAVLMap::last()
{
  return Pix(rightmost());
}

inline void StringFMProgBlockAVLMap::next(Pix& i)
{
  if (i != 0) i = Pix(succ((StringFMProgBlockAVLNode*)i));
}

inline void StringFMProgBlockAVLMap::prev(Pix& i)
{
  if (i != 0) i = Pix(pred((StringFMProgBlockAVLNode*)i));
}

inline String& StringFMProgBlockAVLMap::key(Pix i)
{
  if (i == 0) error("null Pix");
  return ((StringFMProgBlockAVLNode*)i)->item;
}

inline FMProgBlock& StringFMProgBlockAVLMap::contents(Pix i)
{
  if (i == 0) error("null Pix");
  return ((StringFMProgBlockAVLNode*)i)->cont;
}

inline void StringFMProgBlockAVLMap::clear()
{
  _kill(root);
  count = 0;
  root = 0;
}

inline int StringFMProgBlockAVLMap::contains(String& key)
{
  return seek(key) != 0;
}

#endif
