// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _StringSgStatementAVLMap_h
#ifdef __GNUG__
// #pragma interface
#endif
#define _StringSgStatementAVLMap_h 1

#include "statementMap.h"

struct StringSgStatementAVLNode
{
  StringSgStatementAVLNode*      lt;
  StringSgStatementAVLNode*      rt;
  String                 item;
  SgStatement                 cont;
  char                stat;
                      StringSgStatementAVLNode(String& h, SgStatement& c, 
                                    StringSgStatementAVLNode* l=0, StringSgStatementAVLNode* r=0);
                      ~StringSgStatementAVLNode();
};

inline StringSgStatementAVLNode::StringSgStatementAVLNode(String& h, SgStatement& c, 
                                    StringSgStatementAVLNode* l, StringSgStatementAVLNode* r)
     :item(h), cont(c), lt(l), rt(r), stat(0) {}

inline StringSgStatementAVLNode::~StringSgStatementAVLNode() {}

typedef StringSgStatementAVLNode* StringSgStatementAVLNodePtr;


class StringSgStatementAVLMap : public StringSgStatementMap
{
protected:
  StringSgStatementAVLNode*   root;

  StringSgStatementAVLNode*   leftmost();
  StringSgStatementAVLNode*   rightmost();
  StringSgStatementAVLNode*   pred(StringSgStatementAVLNode* t);
  StringSgStatementAVLNode*   succ(StringSgStatementAVLNode* t);
  void            _kill(StringSgStatementAVLNode* t);
  void            _add(StringSgStatementAVLNode*& t);
  void            _del(StringSgStatementAVLNode* p, StringSgStatementAVLNode*& t);

public:
                StringSgStatementAVLMap(SgStatement& dflt);
                StringSgStatementAVLMap(StringSgStatementAVLMap& a);
                ~StringSgStatementAVLMap();

  SgStatement&          operator [] (String& key);

  void          del(String& key);

  Pix           first();
  void          next(Pix& i);
  String&          key(Pix i);
  SgStatement&          contents(Pix i);

  Pix           seek(String& key);
  int           contains(String& key);

  void          clear(); 

  Pix           last();
  void          prev(Pix& i);

  int           OK();
};

inline StringSgStatementAVLMap::~StringSgStatementAVLMap()
{
  _kill(root);
}

inline StringSgStatementAVLMap::StringSgStatementAVLMap(SgStatement& dflt) :StringSgStatementMap(dflt)
{
  root = 0;
}

inline Pix StringSgStatementAVLMap::first()
{
  return Pix(leftmost());
}

inline Pix StringSgStatementAVLMap::last()
{
  return Pix(rightmost());
}

inline void StringSgStatementAVLMap::next(Pix& i)
{
  if (i != 0) i = Pix(succ((StringSgStatementAVLNode*)i));
}

inline void StringSgStatementAVLMap::prev(Pix& i)
{
  if (i != 0) i = Pix(pred((StringSgStatementAVLNode*)i));
}

inline String& StringSgStatementAVLMap::key(Pix i)
{
  if (i == 0) error("null Pix");
  return ((StringSgStatementAVLNode*)i)->item;
}

inline SgStatement& StringSgStatementAVLMap::contents(Pix i)
{
  if (i == 0) error("null Pix");
  return ((StringSgStatementAVLNode*)i)->cont;
}

inline void StringSgStatementAVLMap::clear()
{
  _kill(root);
  count = 0;
  root = 0;
}

inline int StringSgStatementAVLMap::contains(String& key)
{
  return seek(key) != 0;
}

#endif
