// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>
#include <string.h>

void match_variable_shape(SgExpression *,SgExpression *);

SgPortTypeExp *getPortType(SgExpression *port)
{
    char *ECStr;
    SgExpression *e;
    SgStatement *s;


    s = getPortStatement(port);

    assert(s);

    if (s->variant() == INPORT_DECL) {
	SgInportStmt *inportStmt = isSgInportStmt(s);

	return (inportStmt->porttypes());
    }
    if (s->variant() == OUTPORT_DECL) {
	SgOutportStmt *outportStmt = isSgOutportStmt(s);
	
	return(outportStmt->porttypes());
    } 

    fm_internal_error("illegal port, neither inport nor outport");
    return((SgPortTypeExp *) 0);
}

SgStatement *getPortStatement(SgExpression *port)
{
    char *ECStr;
    SgExpression *e;
    SgStatement *s;
    SgDataImpliedDoExp *dtIpDo;
    SgDataEltExp *eltExp;

//    printVariantName(port->variant());

    //
    // some times arguments of a send/receive
    // come in
    //
    if (! is_possible_port_expression(port)) {
	return (SgStatement *)0;
    }

    if (port->variant() != VAR_REF 
	&& port->variant() != ARRAY_REF
	&& port->variant() != EXPR_LIST)
    {
	port = portExp(port);
    }

    switch (port->variant()) {
      case VAR_REF:
      case ARRAY_REF:
	ECStr = portDictKeyEncode(port->symbol());
	break;
      case EXPR_LIST:
	e = (isSgExprListExp(port))->value();
	ECStr = portDictKeyEncode(e->symbol());
	break;
      case DATA_IMPL_DO:
	dtIpDo = isSgDataImpliedDoExp(port);
	eltExp = isSgDataEltExp(dtIpDo->dataelt(0));
	ECStr = portDictKeyEncode(eltExp->name());
	break;
      default:
	printf("get a port type: ");
	printVariantName(port->variant());
	printf("\n");
	fm_internal_error("illegal port type");
	break;
    }
    
    s = portMapGet(portVarDict, ECStr);
    return (s);
}

char * getPortSetKeyFromPortRef(SgExpression *port)
{
    SgStatement *s;
    s           = getPortStatement(port);

    if (!s ) { return (char *) 0; }

    return(portSetKeyEncode(s));
}


int getPortVariant(SgExpression *port)
{
    SgStatement *s = getPortStatement(port);
    if (s) { 
	return s->variant(); 
    } else { 
	return -1; 
    }
}

int getNPortTypes(SgExpression *port)
{
    SgStatement *s = getPortStatement(port);

    if (s->variant() == INPORT_DECL) {
	SgInportStmt *inportStmt = isSgInportStmt(s);
	
	return(inportStmt->numberOfPortTypes());
    } 
    if (s->variant() == OUTPORT_DECL) {
	SgOutportStmt *outportStmt = isSgOutportStmt(s);
	
	return(outportStmt->numberOfPortTypes());
    }
    return(-1);
}

int eqPortTypeCiList(SgExpression *port_1, SgExpression *port_2)
{
    SgPortTypeExp   *p_type_1, *p_type_2, *p1, *p2;
    SgExpression    *r1, *r2;
    int              n_port_types_1, n_port_types_2;
    SgValueExp      *int_value;
    int             i;

    n_port_types_1 = getNPortTypes(port_1);
    n_port_types_2 = getNPortTypes(port_2);

    if (n_port_types_1 != n_port_types_2) {
	fm_error(currentStatement, "\
port type mismatch: %s has %d type fields whereas %s has %d type fields",
		 (portExp(port_1)->symbol())->identifier(), n_port_types_1,
		 (portExp(port_2)->symbol())->identifier(), n_port_types_2);
//	return (0);
    }

    p1 = p_type_1       = getPortType(port_1);
    p2 = p_type_2       = getPortType(port_2);
    r1                  = p_type_1->ref();
    r2                  = p_type_2->ref();

    portTypeMapClear();
    i = 1;

    while (p1 || p2) {
	int_value = new SgValueExp(i);
	if (r1 && 
	    r1->variant() != PORT_TYPE_OP &&
	    r1->variant() != INPORT_TYPE_OP &&
	    r1->variant() != OUTPORT_TYPE_OP)
		   
	{
	    SgSymbol *sb1   = r1->symbol();
	    char *name1 = (char *)malloc(strlen(sb1->identifier()) + 2);
	    
	    assert(name1);
	    sprintf(name1, "1%s", sb1->identifier());
	    portTypeMapSet(portTypeMap, name1, int_value);
	}

	if (r2 && 
	    r2->variant() != PORT_TYPE_OP &&
	    r2->variant() != INPORT_TYPE_OP &&
	    r2->variant() != OUTPORT_TYPE_OP )

	{
	    SgSymbol *sb2   = r2->symbol();

	    char *name2 = (char *)malloc(strlen(sb2->identifier()) + 2);
	    assert(name2);
	    sprintf(name2, "2%s", sb2->identifier());
	    portTypeMapSet(portTypeMap, name2, int_value);
	}
	i++;
	p1 = p1->next();
	p2 = p2->next();
        if (p1) {
            r1 = p1->ref();
        }
        if (p2) {
            r2 = p2->ref();
        }
    }
    if ((p1 && ! p2) || (!p1 && p2)) {
        return 0;
    }
    p1 = p_type_1;
    p2 = p_type_2;
	
    i = 1;
    while (p1 || p2) {
	int eq      = eqPortType(p1, p2, i);

	if (eq) {
	    p1 = p1->next();
	    p2 = p2->next();
	    i++;
	} else {
	    return(0);
	}
    }
    return (1);
}

int eqPortType(SgPortTypeExp *p_type_1, SgPortTypeExp *p_type_2, int index)
{
    SgType       *t1 = p_type_1->type();
    SgType       *t2 = p_type_2->type();
    SgExpression *r1 = p_type_1->ref();
    SgExpression *r2 = p_type_2->ref();

    // type info
//    if (t1->variant() != t2->variant()) {

//    printf("t1 = ");
//    if (t1->symbol()) {
//	printf("%s", t1->symbol()->identifier());
//    }
//    printf("\n");

    if (! t1->equivalentToType(*t2)) {
	fm_error(currentStatement, "\
port type mismatch at field %d: one is %s, the other is %s",
		 index,
		 getVariantByName(t1->variant()),
		 getVariantByName(t2->variant()));

//		 (t1->symbol())->identifier(), 
//		 (t2->symbol())->identifier());
	return (0);
    }
    // both do not have refs
    if ( (! r1) && (! r2)) {
	return (1);
    }
    
    if ((! r1) && r2 || r1 && (! r2))  {
	fm_error(currentStatement, "\
port type mismatch at field %d: one has variable reference the other does not",
		 index);
	return (0);
    }

    if (r1->variant() != r2->variant()) {
	fm_error(currentStatement, "\
port type mismatch at field %d: different type of variables\n", index);
	return(0);
    }

    if (r1->variant() == VAR_REF) {
	return(1);
    }
    // array ref:
    // the index must point to the right ref
    if (r1->variant() == ARRAY_REF) {
	SgArrayRefExp *a1 = isSgArrayRefExp(r1);
	SgArrayRefExp *a2 = isSgArrayRefExp(r2);
	
	if (a1->numberOfSubscripts() != a2->numberOfSubscripts()) {
	    fm_error(currentStatement, "\
port type mismatch at field %d: %s has %d dimensions whereas %s has %d dimensions", 
		     (a1->symbol())->identifier(), a1->numberOfSubscripts(),
		     (a2->symbol())->identifier(), a2->numberOfSubscripts());

	    return(0);
	} 
	
	for (int i = 0; i < a1->numberOfSubscripts(); i++) {
	    SgExpression *e1 = a1->subscript(i);
	    SgExpression *e2 = a2->subscript(i);

	    if (e1->variant() != e2->variant()) {
		fm_error(currentStatement, "\
port type mismatch at field %d: different variable references for subscript %d",
			 index, i);
		return(0);
	    }
	    
	    if (e1->variant() == CONST_REF) {
		if (strcmp((e1->symbol())->identifier(),
			   (e2->symbol())->identifier())) 
		{
		    fm_error(currentStatement, "\
port type mismatch at field %d: two arrays differ in dimension %d",
			     index, i);
		    return(0);
		}
		return(1);
	    }

	    if (e1->variant() == VAR_REF) {
		char *str1 = 
		  (char *)malloc(strlen((e1->symbol())->identifier()) + 2);
		char *str2 =
		  (char *)malloc(strlen((e2->symbol())->identifier()) + 2);
		SgValueExp  *ind1, *ind2;
		int ind1_value_int, ind2_value_int;

		sprintf(str1, "1%s", (e1->symbol())->identifier());
		sprintf(str2, "2%s", (e2->symbol())->identifier());
		
		ind1 = isSgValueExp(portTypeMapGet(portTypeMap, str1));
		ind2 = isSgValueExp(portTypeMapGet(portTypeMap, str2));
		
		if ((!ind1) && ind2 || ind1 && (! ind2)) {
		    return (0);
		}
		
		ind1_value_int = -1;
		ind2_value_int = -2;

		if (isSgValueExp(ind1)) {
		    ind1_value_int = ind1->intValue();
		}
		if (isSgValueExp(ind2)) {
		    ind2_value_int = ind2->intValue();
		}
		if (ind1_value_int != ind2_value_int) {
		    return(0);
		}
		free(str1);
		free(str2);
	    }
	}
    }
    return (1);
}

/* 
 * checkPortShape()
 *
 * Determine whether the port reference conforms to the
 * port shape defined in the corresponding port statement
 *
 */

void checkPortShape(SgExpression *port)
{
    SgSymbol            *port_symbol;
    SgExpression        *port_declared;
    SgExpression        *port_reference;

    port_reference      = portExp(port);
    if (port_reference->variant() == EXPR_LIST) {
        port_reference = (isSgExprListExp(port_reference))->elem(0);
    }
    port_symbol         = port_reference->symbol();

    printf("port = ");
    port_reference->unparsestdout();
    printf(" variant == %s\n", getVariantByName(port_reference->variant()));
    if (port_symbol) {
	port_declared = findSymbInDeclStmt(port_symbol);
	if (! port_declared) {
	    //
	    // this port has not been declared
	    // but this has been complained before
	    // so be quiet here
	    //
	    return;
	} else {
	    match_variable_shape(port_reference, port_declared);
	}
    } else {
        fm_error(currentStatement, "\
Illegal port reference: %s", port_reference->unparse());
    }
}

/*
 * match_variable_shape()
 *
 * Determine whether variable reference R conforms to what
 * it is defined shapewise as given by D.
 *
 */
	    
void match_variable_shape(SgExpression *r, SgExpression *d)
{
    if (r->variant() != d->variant()) {
	fm_error(currentStatement, "\
illegal use of variable %s as it is defined as %s",
		 symbName(r), d->unparse());
	return;
    }
    
    if (r->variant() == ARRAY_REF) {
	SgArrayRefExp  *a_r;
	SgArrayRefExp  *a_d;

	a_r            = isSgArrayRefExp(r);
	a_d            = isSgArrayRefExp(d);

	if (a_r->numberOfSubscripts() != a_d->numberOfSubscripts()) {
	    fm_error(currentStatement, "\
Illegal expression %s as variable %s is defined as a %d dimensional array",
		     a_r->unparse(),
		     symbName(a_r),
		     a_d->numberOfSubscripts());
	}
    }
}



