// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include <assert.h>

void pairing_ports();
int inport_type_has_ports(SgInportStmt *);
int outport_type_has_ports(SgOutportStmt *);

#define FOR_EACH_PROGRAM(progStmt,progName) \
    for (Pix _ind = fmProgBlockMap->first();_ind;fmProgBlockMap->next(_ind)) \
    { \
	FMProgBlock *block = &(fmProgBlockMap->contents(_ind)); \
	SgStatement *stmt = block->proc();  \
	if (!stmt) { \
	    fm_internal_error("Expect a non-null FM program block at %s:%d", __FILE__, __LINE__); \
	} \
	if (stmt->variant() == PROG_HEDR) { \
            SgProgHedrStmt  *progStmt; \
            char            *progName; \
            switchToFMProgBlock(stmt->symbol()->identifier()); \
	    progStmt  = isSgProgHedrStmt(stmt); \
            progName  = progStmt->symbol()->identifier();


#define END_FOR_EACH_PROGRAM() }}


#define FOR_EACH_PROCESS(prosStmt,prosName) \
    for (Pix _ind = fmProgBlockMap->first();_ind;fmProgBlockMap->next(_ind)) \
    { \
	FMProgBlock *block = &(fmProgBlockMap->contents(_ind)); \
	SgStatement *stmt = block->proc();  \
	if (!stmt) { \
	    fm_internal_error("Expect a non-null FM program block at %s:%d", __FILE__, __LINE__); \
	} \
	if (stmt->variant() == PROS_HEDR) { \
            SgProsHedrStmt  *prosStmt; \
            char            *prosName; \
            switchToFMProgBlock(stmt->symbol()->identifier()); \
	    prosStmt  = isSgProsHedrStmt(stmt); \
            prosName  = prosStmt->symbol()->identifier();

#define END_FOR_EACH_PROCESS() } }

#define FOR_EACH_PARAMETER_AND_TYPE(prosStmt,symb,type) \
    for (int i = 0; i < prosStmt->numberOfParameters(); i++) { \
	SgSymbol     *symb; \
	SgType       *type; \
	SgExpression *var_declared; \
	symb         = prosStmt->parameter(i); \
	type         = getProsArgType(prosName,symb);
 
#define END_FOR_EACH_PARAMETER_AND_TYPE() }

#define FOR_EACH_PORT_ARG(prefix,prosStmt,port,symb) \
    for (int prefix##_i = 0; \
        prefix##_i < prosStmt->numberOfParameters(); \
        prefix##_i++) { \
        SgSymbol     *symb; \
        SgType       *type; \
	SgStatement  *stmt; \
        SgExpression *var_declared; \
        symb         = prosStmt->parameter(prefix##_i); \
        stmt         = stmtMapGet(prosMap,prosMapKeyEncode(prosStmt->symbol()->identifier(),symb->identifier())); \
	if (stmt->variant() == INPORT_DECL||stmt->variant()== OUTPORT_DECL) { \
            SgStatement *port = stmt;
     
#define END_FOR_EACH_PORT_ARG() }}

#define FOR_EACH_PORT_DECL(stmt,pkey) \
    for (ind = portSet->first(); ind; portSet->next(ind)) { \
        SgStatement *stmt = &(portSet->contents(ind)); \
	String      *pkey  = &(portSet->key(ind)); \
	if (!stmt) { \
	    fm_internal_error("Expect a non-null FM program block at %s:%d", __FILE__, __LINE__); \
	}

#define END_FOR_EACH_PORT_DECL() }

#define FOR_EACH_INPORT_DECL(stmt,key) \
    FOR_EACH_PORT_DECL(_port_stmt,key) \
        if (_port_stmt->variant() == INPORT_DECL) { \
	    SgInportStmt *stmt = isSgInportStmt(_port_stmt); 

#define END_FOR_EACH_INPORT_DECL() \
    END_FOR_EACH_PORT_DECL() }

#define FOR_EACH_OUTPORT_DECL(stmt,key) \
    FOR_EACH_PORT_DECL(_port_stmt,key) \
        if (_port_stmt->variant() == OUTPORT_DECL) { \
	    SgOutportStmt *stmt = isSgOutportStmt(_port_stmt); 

#define END_FOR_EACH_OUTPORT_DECL() \
    END_FOR_EACH_PORT_DECL() }

#define FOR_EACH_PORT_TYPE(i,stmt,p_type,port_ref) \
    for (int i = 0;i<stmt->numberOfPortTypes();i++) { \
        SgPortTypeExp *port_type; \
        SgExpression  *port_ref; \
        SgType        *p_type; \
        port_type     = stmt->porttype(i); \
        p_type        = port_type->type(); \
        port_ref      = port_type->ref();

#define END_FOR_EACH_PORT_TYPE() }

#define FOR_EACH_MANGLED_FUNC(mg_func_name, func_name) \
    for(Pix _ind = portMangleMap->first();_ind;portMangleMap->next(_ind)) \
    { \
        SgExpression *func_exp = &(portMangleMap->contents(_ind)); \
        String       *mg_key   = &(portMangleMap->key(_ind)); \
        char         *mg_func_name = mg_key->chars(); \
        char         *func_name    = isSgValueExp(func_exp)->stringValue(); 

#define END_FOR_EACH_MANGLE_FUNC() }


SgType *              getProsArgType(char *, SgSymbol *);
char *                str2upper(char *);
char *                str2lower(char *);
SgExpression *        getFormalArgInProsHedrMap(SgExpression *);
void                  gen_compute_IN_buffer_size(FILE *, SgProsHedrStmt *);
void                  gen_compute_OUT_buffer_size(FILE *, SgProsHedrStmt *);
void                  gen_pack_IN_buffer(FILE *, SgProsHedrStmt *);
void                  gen_pack_OUT_buffer(FILE *, SgProsHedrStmt *);
void                  gen_unpack_handler_buffer(FILE *, SgProsHedrStmt *);
int                   handler_id(char *);

extern "C" {
    char *strerror(int);
}

//
// non-zero means current module has handler table
// zero otherwise.
//
static int this_module_has_handler_table;

void gen_preamble(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * This file is generated by fm\n\
 */\n\n\
#include \"fm.h\"\n\
#ifdef FM_SIMPLE_INST\n\
#ifdef FM_SIMPLE_INST_INCLUDE_FILE\n\
#include FM_SIMPLE_INST_INCLUDE_FILE\n\
#endif\n\
#endif\n\n");

   if (opt_i8) 
       fprintf(fp, "\
#undef  nexus_sizeof_int\n\
#define nexus_sizeof_int nexus_sizeof_long\n\
#undef  nexus_put_int\n\
#define nexus_put_int nexus_put_long\n\
#undef  nexus_get_int\n\
#define nexus_get_int nexus_get_long\n\
#undef  nexus_get_stashed_int\n\
#define nexus_get_stashed_int nexus_get_stashed_long\n\
#define int long\n\n");

}

void gen_fm_args_t_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * FM_ARGS_T\n\
 *\n\
 * Argument buffer. For each argument to a process we have a field\n\
 * wherein the data for that argument is located. The field for argument\n\
 * i is named A_<i>.\n\
 *\n\
 * Scalars are represented as scalars of the appropriate type.\n\
 *\n\
 * Arrays whose size are known at compile time are represented as\n\
 * fixed-size arrays.\n\
 *\n\
 * Arrays whose size aren't known until runtime are represented as\n\
 * pointers of the appropriate data type. A field A_<i>_len is also\n\
 * generated which is an integer which contains the length of the\n\
 * variable.\n\
 *\n\
 * Also stored here is a global pointer to the process block entry\n\
 * on the caller side.\n\
 */\n");
}

void gen_fm_args_t(FILE *fp)
{
    int                i;
    SgExpression       *var_declared;

//    gen_fm_args_t_comments(fp);
    
    FOR_EACH_PROCESS(s,prosName);
    
    if (s->numberOfParameters() == 0) {
	fprintf(fp, "\
typedef struct\n{\n\
    f_integer_t vc_offset;\n\
    int dummy;\n\
    nexus_global_pointer_t entry_gp;\n\
} FM_ARGS_T(%s);\n\n", 
		prosName);
	continue;
    }

    fprintf(fp, "\
typedef struct\n{\n\
    f_integer_t vc_offset;\n");
	    
    for (i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol                *symb;
	SgStatement             *stmt;
	char                    *varName;
	
	symb = s->parameter(i);
	if (! symb) {
	    fm_internal_error("\
Expect a non-null symbol as argument %d for process %s, but failed",
			      i, prosName);
	}
	
	varName = symb->identifier();
	stmt = stmtMapGet(prosMap, prosMapKeyEncode(prosName,varName));
	
	if (!stmt) {
	    //
	    // This means the i-th parameter has not
	    // been defined,  it has been complained
	    // already so keep it quiet here
	    //
	    continue;
	}
	
	switch (stmt->variant()) {
	  case VAR_DECL:
	    SgVarDeclStmt       *varDeclStmt;
	    SgType              *t;
	    int                 t_variant;
	    
	    varDeclStmt         = isSgVarDeclStmt(stmt);
	    t                   = varDeclStmt->type();
	    t_variant           = t->variant();
	    
	    var_declared        = findSymbInDeclStmt(symb);
	    
	    if (var_declared->variant() == ARRAY_REF) {
		SgArrayRefExp   *a;
		int              j;
		SgExpression    *a_size;
		
		a      = isSgArrayRefExp(var_declared);
		a_size = &(arraySizeExp(a));
		if (IsInteger(a_size)) {
		    fprintf(fp, "\
    f_%s_t A_%d[%d];\n",  
			    getTypeString(t_variant), i,
			    integerValue(a_size));
		} else {
		    fprintf(fp, "\
    f_%s_t *A_%d;\n",  getTypeString(t_variant), i);
		    fprintf(fp, "\
    int A_%d_len;\n", i);
		}
	    } else if (var_declared->variant() == VAR_REF) {
		fprintf(fp, "\
    f_%s_t A_%d;\n",  getTypeString(t_variant), i);
	    }
	    break;
	  case INPORT_DECL:
	  case OUTPORT_DECL:
	    var_declared  = findSymbInDeclStmt(symb);
	    
	    
	    if (var_declared->variant() == ARRAY_REF) {
		SgArrayRefExp   *a;
		int              j;
		SgExpression    *a_size;
		
		a      = isSgArrayRefExp(var_declared);
		a_size = &(arraySizeExp(a));
		if (IsInteger(a_size)) {
		    fprintf(fp, "\
    f_integer_t A_%d[%d];\n", i, integerValue(a_size));
		} else {
		    fprintf(fp, "\
    f_integer_t *A_%d;\n", i);
		    fprintf(fp, "\
    int A_%d_len;\n", i);
		}
	    } else if (var_declared->variant() == VAR_REF) {
		fprintf(fp, "\
    f_integer_t A_%d", i);
		fprintf(fp, ";\n");
	    }
	    break;
	  default:
	    fm_internal_error("\
Do not expect statement type %s at %s:%d",
			      getVariantByName(stmt->variant()),
			      __FILE__, __LINE__);
	    break;
	}
    }
	    
    fprintf(fp, "\
    nexus_global_pointer_t entry_gp;\n\
} FM_ARGS_T(%s);\n\n", 
		    prosName);
    END_FOR_EACH_PROCESS();
}	
    
void gen_fm_return_args_t_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * FM_RETURN_ARGS_T\n\
 *\n\
 * Saves output variable pointers. A pointer to each INTENT(OUT) and\n\
 * INTENT(INOUT) argument is stored by the caller of the process in the\n\
 * appropriate field in this structure.\n\
 *\n\
 */\n\n");
}

/*
 * gen_fm_return_args_t()
 *
 * Write the fm_returns_args_t struct definition to
 * the file designated by FP.
 *
 */

void gen_fm_return_args_t(FILE *fp)
{
    int n_return_args;
//    gen_fm_return_args_t_comments(fp);

    FOR_EACH_PROCESS(s,prosName);
    n_return_args = 0;
    //
    // generate a dummy argument if 
    // the process doesn't have any argument
    //
    if (s->numberOfParameters() == 0) {
	fprintf(fp, "\
typedef struct\n{\n\
    int dummy;\n\
} FM_RETURN_ARGS_T(%s);\n\n", prosName);
	continue;
    }

    fprintf(fp, "\
typedef struct\n{\n");
    
    for (int i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol            *symb;
	SgExpression        *attribute;
	int                  should_be_copied_out;
	
	should_be_copied_out = 0;
	symb                 = s->parameter(i);
	if (! symb) {
	    fm_internal_error("\
Expect a non-null symbol at %s:%d", __FILE__, __LINE__);
	}
	
	//
	// get the attribute of this argument
	//
	attribute = intentMapGet(intentMap, symb->identifier());
	if (! attribute) {
	    //
	    // by default == intent(INOUT)
	    //
	    should_be_copied_out = 1;
	} else if (attribute->variant() == OUT_OP 
		   || attribute->variant() == INOUT_OP)
	{
	    should_be_copied_out = 1;
	}
	
	if (should_be_copied_out) {
	    // get type information
	    SgType          *t;
	    
	    t               = getProsArgType(prosName, symb);
	    if (!t) {
		//
		// unknown type caused by a implicit none
		// statement, which has been complained
		// already, so keep it quiet here
		//
		continue;
	    } 
	    
	    fprintf(fp, "\
    f_%s_t *A_%d;\n", 
		    getTypeString(t->variant()), i);
	    n_return_args++;
	}
    }

    if (n_return_args == 0) {
	fprintf(fp, "\
    int dummy;\n");
    }

    fprintf(fp, "\
} FM_RETURN_ARGS_T(%s);\n\n", prosName);

    END_FOR_EACH_PROCESS();
}
	
/* 
 * 
 */

void gen_handler_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * Handler definitions.\n\
 *\n\
 * Note the name and hash value for the handlers. The compiler needs\n\
 * to compute the hash value for itself. Look in src/nexus/hash.c\n\
 * for the definition of the hash function.\n\
 *\n\
 * The forward decl of the handlers can be eliminated by putting\n\
 * the handler table and the module init function at the end of this\n\
 * file.\n\
 */\n");
}

void gen_handler_defs(FILE *fp)
{
    Pix           ind;

//    gen_handler_comments(fp);
    for (int phase = 0; phase < 2; phase++) {
	if (phase == 1) {
	    fprintf(fp, "\n\
static nexus_handler_t handlers[] =\n{\n");
	}
	
	for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	    FMProgBlock      *block;
	    SgStatement      *s;
	    
	    block            = &(fmProgBlockMap->contents(ind));
	    if (! block) {
		fm_internal_error("\
Expect a non-null fmProgBlock data at %s:%d", __FILE__, __LINE__);
	    }
	    
	    s = block->proc();
	    if (s->variant() == PROS_HEDR) {
		char     *prosName;
		char     *handler_func_name;
		char *lowerName;

		prosName = s->symbol()->identifier();
		lowerName = str2lower(prosName);

		this_module_has_handler_table++;

		handler_func_name = (char *)malloc(20+strlen(prosName)+1);
		
		sprintf(handler_func_name, "_FMH_invoke_%s", lowerName);

		if (phase == 0) {
		    fprintf(fp, "\
static void invoke_%s_handler(void *address, nexus_buffer_t *buf);\n",
			    lowerName);
		}
		if (phase == 1) {
		    fprintf(fp, "\
    {\"_FMH_invoke_%s\", %d, FM_INVOKE_HANDLER_TYPE, (nexus_handler_func_t) invoke_%s_handler},\n",
			    lowerName, handler_id(handler_func_name), lowerName);
		}
		sprintf(handler_func_name, "_FMH_ShouldShare_%s", lowerName);

		if (phase == 0) {
		    fprintf(fp, "\
static void FM_SHOULD_SHARE_HANDLER_FUNC(%s)(void *address, nexus_buffer_t *buf);\n",
			    lowerName);
		}
		if (phase == 1) {
		    fprintf(fp, "\
    {\"%s\", %d, FM_SHOULD_SHARE_HANDLER_TYPE, (nexus_handler_func_t) FM_SHOULD_SHARE_HANDLER_FUNC(%s)},\n",
			    handler_func_name,
			    handler_id(handler_func_name), lowerName);
		}
		free(lowerName);
	    }
	}
    }
    fprintf(fp, "\
    {(char *) NULL, 0, NEXUS_HANDLER_TYPE_NON_THREADED, (nexus_handler_func_t) NULL},\n");
    fprintf(fp, "};\n\n");
}

void gen_compiler_flavours_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * Defines for various compiler flavors\n\
 *\n\
 */\n\n");
}


void gen_compiler_flavours(FILE *fp)
{
    Pix        ind;
    int        phase;
//    gen_compiler_flavours_comments(fp);
    fprintf(fp, "\
#ifdef FM_NEED_TRAILING_UNDERSCORES\n");

    for (phase = 0; phase < 4; phase++) {
	for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	    FMProgBlock *block = &(fmProgBlockMap->contents(ind));
	    SgStatement *s = block->proc();
	    
	    if (s->variant() == PROS_HEDR) {
		char            *prosName;
		char            *prosNameCapital;
		SgProsHedrStmt  *prosHedr;
		int             n_args;

		prosHedr        = isSgProsHedrStmt(s);
		n_args          = prosHedr->numberOfParameters();
		prosName        = s->symbol()->identifier();
		prosNameCapital = str2upper(prosName);
		
		
		if (phase == 0) {
		    fprintf(fp, "\
#define %s %s_\n", 
			    prosNameCapital, prosName);
		} else if (phase == 1) {
		    fprintf(fp, "\
#define FM_%s_SHOULD_SHARE fm_%s_should_share_\n", 
			    prosNameCapital, prosName);
		} else if (phase == 2) {
		    fprintf(fp, "\
#define FM_%s_PORT_INIT fm_%s_port_init_\n",
			    prosNameCapital, prosName);
		} else if (phase == 3) {
		    fprintf(fp, "\
#define FM_%s_%d_INVOKE fm_%s_%d_invoke_\n", 
			    prosNameCapital, n_args, prosName, n_args);
		}
	    }
	}
    }
    //
    // output send/receive function macros
    //
    FOR_EACH_PORT_DECL(portDecl,portKey);

    if (portDecl->variant() == INPORT_DECL) {
	//
	// generate receive function call
	//
	char *rcv_func_name 
 	  = ppString("fm_", str2lower((char *) portKey->chars()), 
 		     (opt_debug)? "_rcv_d" : "_rcv");
				    
	char *rcv_func_name_with_underscore 
 	  = ppString("fm_", str2lower((char *) portKey->chars()), 
 		     (opt_debug)? "_rcv_d_" : "_rcv_");
			 
	SgValueExp *rcv_mg_func_name_exp 
	  = isSgValueExp(exprMapGet(portMangleMap, rcv_func_name));

	char *rcv_func_full_name;

	rcv_func_full_name = rcv_func_name;
	if (rcv_mg_func_name_exp) {
	    /* Need to map the full function name to the 
	       mangled function name */

	    rcv_func_name = rcv_mg_func_name_exp->stringValue();
	    rcv_func_name_with_underscore = ppString("", rcv_func_name, "_");
	}

	fprintf(fp, "\
#define %s %s\n",
		str2upper(rcv_func_full_name),
		rcv_func_name_with_underscore);
    } else if (portDecl->variant() == OUTPORT_DECL) {
	//
	// generate send function call
	//
	char *snd_func_name                 
 	  = ppString("fm_", str2lower((char *) portKey->chars()), 
 		     (opt_debug)? "_snd_d" : "_snd");

	char *snd_func_name_with_underscore 
 	  = ppString("fm_", str2lower((char *) portKey->chars()), 
 		     (opt_debug)? "_snd_d_" : "_snd_");

	char *snd_func_full_name;

	SgValueExp *snd_mg_func_name_exp
	  = isSgValueExp(exprMapGet(portMangleMap, snd_func_name));

	char *port_mg_init_func_name;
	char *port_full_init_func_name;
	SgValueExp *port_mg_init_func_exp;

	snd_func_full_name = snd_func_name;
	if (snd_mg_func_name_exp) {
	    /* Need to map the full function name to the 
	       mangled function name */
	    snd_func_name = snd_mg_func_name_exp->stringValue();
	    snd_func_name_with_underscore = ppString("", snd_func_name, "_");
	} 	    

	fprintf(fp, "\
#define %s %s\n",
		str2upper(snd_func_full_name),
		snd_func_name_with_underscore);

	port_full_init_func_name
	  = ppString("fm_", str2lower((char*)portKey->chars()),"_init");
	port_mg_init_func_exp 
	  = isSgValueExp(exprMapGet(portMangleMap, port_full_init_func_name));

	if (port_mg_init_func_exp) {
	    port_mg_init_func_name = port_mg_init_func_exp->stringValue();
	} else {
	    port_mg_init_func_name = port_full_init_func_name;
	}

	fprintf(fp, "\
#define FM_%s_INIT %s_\n",
		str2upper((char*)portKey->chars()),
		port_mg_init_func_name);
    } else {
	fm_internal_error("\
Do not expect port type to be other than INPORT/OUTPORT: %s:%d",
			  __FILE__, __LINE__);
    }	

    END_FOR_EACH_PORT_DECL();

    fprintf(fp, "\n\
#else /* FM_NEED_TRAILING_UNDERSCORES */\n\n");
    fprintf(fp, "\
#ifndef FM_NEED_UPPERCASE\n\n");
    
    for (phase = 0; phase < 4; phase++) {
	for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	    FMProgBlock *block = &(fmProgBlockMap->contents(ind));
	    SgStatement *s = block->proc();
	    
	    if (s->variant() == PROS_HEDR) {
		char            *prosName;
		char            *prosNameCapital;
		SgProsHedrStmt  *prosHedr;
		int             n_args;

		prosHedr        = isSgProsHedrStmt(s);
		n_args          = prosHedr->numberOfParameters();
		prosName        = s->symbol()->identifier();
		prosNameCapital = str2upper(prosName);

		if (phase == 0) {
		    fprintf(fp, "\
#define %s %s\n",
			    prosNameCapital, prosName);
		} else if (phase == 0) {
		    fprintf(fp, "\
#define FM_%s_SHOULD_SHARE fm_%s_should_share\n", 
			    prosNameCapital, prosName);
		} else if (phase == 1) {
		    fprintf(fp, "\
#define FM_%s_PORT_INIT fm_%s_port_init\n",
			    prosNameCapital, prosName);
		} else if (phase == 2) {
		    fprintf(fp, "\
#define FM_%s_%d_INVOKE fm_%s_%d_invoke\n", 
			    prosNameCapital, n_args, prosName, n_args);
		}
	    }
	}
    }

    FOR_EACH_PORT_DECL(portDecl,portKey);

    if (portDecl->variant() == INPORT_DECL) {
	//
	// generate receive function call
	//
	char *rcv_func_name           
 	  = ppString("fm_", str2lower((char *)portKey->chars()), 
 		     (opt_debug)? "_rcv_d" : "_rcv");
	char *rcv_mg_func_name;
	SgValueExp *rcv_mg_func_exp;

	rcv_mg_func_exp 
	  = isSgValueExp(exprMapGet(portMangleMap, rcv_func_name));
	
	if (rcv_mg_func_exp) {
	    rcv_mg_func_name = rcv_mg_func_exp->stringValue();
	} else {
	    rcv_mg_func_name = rcv_func_name;
	}

	fprintf(fp, "\
#define %s %s\n",
		str2upper(rcv_func_name),
		rcv_mg_func_name);
    } else if (portDecl->variant() == OUTPORT_DECL) {
	//
	// generate send function call
	//
	char *snd_func_name           
 	  = ppString("fm_", str2lower((char *)portKey->chars()), 
 		     (opt_debug)? "_snd_d" : "_snd");
	char *snd_func_full_name;

	SgValueExp *snd_mg_func_name_exp
	  = isSgValueExp(exprMapGet(portMangleMap, snd_func_name));

	char *port_mg_init_func_name;
	char *port_full_init_func_name;
	SgValueExp *port_mg_init_func_exp;

	snd_func_full_name = snd_func_name;
	if (snd_mg_func_name_exp) {
	    /* Need to map the full function name to the 
	       mangled function name */
	    snd_func_name = snd_mg_func_name_exp->stringValue();
	}

	fprintf(fp, "\
#define %s %s\n",
		str2upper(snd_func_full_name),
		snd_func_name);


	port_full_init_func_name
	  = ppString("fm_", str2lower((char*)portKey->chars()),"_init");
	port_mg_init_func_exp 
	  = isSgValueExp(exprMapGet(portMangleMap, port_full_init_func_name));

	if (port_mg_init_func_exp) {
	    port_mg_init_func_name = port_mg_init_func_exp->stringValue();
	} else {
	    port_mg_init_func_name = port_full_init_func_name;
	}

	fprintf(fp, "\
#define FM_%s_INIT %s\n",
		str2upper((char*)portKey->chars()),
		port_mg_init_func_name);

    } else {
	fm_internal_error("\
Do not expect port type to be other than INPORT/OUTPORT: %s:%d",
			  __FILE__, __LINE__);
    }	

    END_FOR_EACH_PORT_DECL();

    fprintf(fp, "\n\
#endif /* FM_NEED_UPPERCASE */\n");
    fprintf(fp, "\n\
#endif /* FM_NEED_TRAILING_UNDERSCORES */\n\n");
}    

void gen_outport_init_func_proto(FILE *fp)
{
    Pix ind;

    FOR_EACH_OUTPORT_DECL(outportStmt,key)
    {
	fprintf(fp, "\
void FM_%s_INIT();\n",
		str2upper((char *)key->chars()));
    }
    END_FOR_EACH_OUTPORT_DECL();
}

void gen_process_func_proto(FILE *fp)
{
    Pix ind;

    FOR_EACH_PROCESS(prosStmt,prosName)
    {
	fprintf(fp, "\
extern void %s(",
		str2upper(prosName));

	FOR_EACH_PARAMETER_AND_TYPE(prosStmt,symb,type)
	{
	    fprintf(fp, "f_%s_t *", getTypeString(type->variant()));
	    if (i != prosStmt->numberOfParameters()-1) {
		fprintf(fp, ", ");
	    }
	}
	END_FOR_EACH_PARAMETER_AND_TYPE();
	fprintf(fp, ");\n\n");
    }
    END_FOR_EACH_PROCESS();
}


void gen_module_init_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * Module initialization\n\
 *\n\
 */\n");
}

void gen_module_init(FILE *fp)
{
    Pix     ind;

//    gen_module_init_comments(fp);
#if 0
    for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	FMProgBlock *block = &(fmProgBlockMap->contents(ind));
	SgStatement *s = block->proc();
	
	if (s->variant() == PROG_HEDR ||
	    s->variant() == PROS_HEDR ) {
	    char            *progName;
	    progName        = s->symbol()->identifier();
	    
	}
    }
#endif /* 0 */
    
    if (this_module_has_handler_table) {
	fprintf(fp, "\
void _fm_module_init_%s()\n{\n\
    nexus_register_handlers(handlers);\n\
}\n\n", currentFileBase);
    }
}

void gen_fm_should_share_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * FM_<PROC>_SHOULD_SHARE()\n\
 *\n\
 * This is an integer-valued function returning non-zero if the\n\
 * context created for PROC should share a process with other\n\
 * contexts.\n\
 */\n\n");
}
   
void gen_fm_should_share(FILE *fp)
{
    Pix      ind;
    
//    gen_fm_should_share_comments(fp);
    for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	FMProgBlock *block = &(fmProgBlockMap->contents(ind));
	SgStatement *s = block->proc();
	char        *prosName;
	SgStatement *threadable_stmt;

	prosName    = s->symbol()->identifier();
	switchToFMProgBlock(prosName);

	threadable_stmt = stmtMapGet(prosMap, directiveMapKeyEncode(prosName,"threadableProcess"));

	if (s->variant() == PROS_HEDR) {
	    fprintf(fp, "FM_SHOULD_SHARE_HANDLER(%s, %d)\n",
		    str2lower(prosName),
		    (threadable_stmt)? 1 : 0);
	}
    }
}

void gen_fm_return_copy_func_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * FM_RETURN_COPY_FUNC\n\
 *\n\
 * This purpose of this function is to copy the output variables from the\n\
 * nexus message buffer into the addresses kept in the return args\n\
 * structure.\n\
 *\n\
 * Two flavors of this function are generated: one for nexus_buffer_t,\n\
 * and one for nexus_stashed_buffer_t.\n\
 *\n\
 * The format of the return values buffer is:\n\
 *\n\
 * Scalars are encoded as single values.\n\
 *\n\
 * Fixed-length arrays are encoded as vectors of the proper length.\n\
 *\n\
 * Variable-length arrays are encoded as length-vector pairs.\n\
 */\n\n");
}

void gen_fm_return_copy_func(FILE *fp)
{
    Pix     ind;
    int     i;

//    gen_fm_return_copy_func_comments(fp);
    for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	FMProgBlock *block = &(fmProgBlockMap->contents(ind));
	SgStatement *stmt = block->proc();
	
	if (stmt->variant() == PROS_HEDR) {
	    SgProsHedrStmt  *s;
	    char            *prosName;

	    s               = isSgProsHedrStmt(stmt);
	    prosName        = s->symbol()->identifier();
	    switchToFMProgBlock(prosName);
	    fprintf(fp, "\
FM_RETURN_COPY_FUNC_DECL(%s)\n{\n", prosName);
	    //
	    // find out the (OUT/INOUT) args whose sizes 
	    // are determined at run time
	    //
	    for (i = 0; i < s->numberOfParameters(); i++) {
		SgSymbol            *symb;
		SgExpression        *attribute;
		int                  should_be_copied_out;
		
		should_be_copied_out = 0;
		symb                 = s->parameter(i);
		if (! symb) {
		    fm_internal_error("\
Expect a non-null symbol at %s:%d", __FILE__, __LINE__);
		}
		
		//
		// get the attribute of this argument
		//
		attribute = intentMapGet(intentMap, symb->identifier());
		if (! attribute) {
		    //
		    // by default == intent(INOUT)
		    //
		    should_be_copied_out = 1;
		} else if (attribute->variant() == OUT_OP 
			   || attribute->variant() == INOUT_OP)
		{
		    should_be_copied_out = 1;
		}
		
		if (should_be_copied_out) {
		    // get type information
		    SgType          *t;
		    SgExpression    *var_declared;
		    SgStatement     *declStmt;
		    
		    t               = getProsArgType(prosName, symb);
		    if (!t) {
			//
			// unknown type caused by a implicit none
			// statement, which has been complained
			// already, so keep it quiet here
			//
			continue;
		    } 
		    
		    var_declared    = findSymbInDeclStmt(symb);
		    if (var_declared->variant() == ARRAY_REF) {
			SgArrayRefExp   *a;
			SgExpression    *a_size;
			
			a      = isSgArrayRefExp(var_declared);
			a_size = &(arraySizeExp(a));
			if (! IsInteger(a_size)) {
			    fprintf(fp, "\
    int A_%d_len;\n", i);
			}
		    }
		}
	    }
	    fprintf(fp, "\n\
    FM_RETURN_COPY_FUNC_CODE(%s);\n", prosName);
	    
	    for (i = 0; i < s->numberOfParameters(); i++) {
		SgSymbol            *symb;
		SgExpression        *attribute;
		int                  should_be_copied_out;
		
		should_be_copied_out = 0;
		symb                 = s->parameter(i);
		if (! symb) {
		    fm_internal_error("\
Expect a non-null symbol at %s:%d", __FILE__, __LINE__);
		}
		
		//
		// get the attribute of this argument
		//
		attribute = intentMapGet(intentMap, symb->identifier());
		if (! attribute) {
		    //
		    // by default == intent(INOUT)
		    //
		    should_be_copied_out = 1;
		} else if (attribute->variant() == OUT_OP 
			   || attribute->variant() == INOUT_OP)
		{
		    should_be_copied_out = 1;
		}
		
		if (should_be_copied_out) {
		    // get type information
		    SgType          *t;
		    SgExpression    *var_declared;
		    SgStatement     *declStmt;
		    
		    t               = getProsArgType(prosName, symb);
		    if (!t) {
			//
			// unknown type caused by a implicit none
			// statement, which has been complained
			// already, so keep it quiet here
			//
			continue;
		    } 

		    var_declared    = findSymbInDeclStmt(symb);
		    if (var_declared->variant() == ARRAY_REF) {
			SgArrayRefExp   *a;
			SgExpression    *a_size;
			
			a      = isSgArrayRefExp(var_declared);
			a_size = &(arraySizeExp(a));
			if (! IsInteger(a_size)) {
			    fprintf(fp, "\
    nexus_get_int(buffer, &A_%d_len, 1);\n", i);
			    fprintf(fp, "\
    nexus_get_%s(buffer, args->A_%d, A_%d_len);\n", 
				    getTypeCString(t->variant()),
				    i, i);
			} else {
			    fprintf(fp, "\
    nexus_get_%s(buffer, args->A_%d, %d);\n", 
				    getTypeCString(t->variant()),
				    i, 
				    integerValue(a_size));
			}
		    } else if (var_declared->variant() == VAR_REF) {
			fprintf(fp, "\
    nexus_get_%s(buffer, args->A_%d, 1);\n", 
				getTypeCString(t->variant()),
				i);
		    }
		}
	    }
	    fprintf(fp, "}\n\n");
	} /* proc boundary */
    }
}
	
void gen_fm_return_stashed_copy_func(FILE *fp)
{
    Pix     ind;
    int i;

    for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	FMProgBlock *block = &(fmProgBlockMap->contents(ind));
	SgStatement *stmt = block->proc();
	
	if (stmt->variant() == PROS_HEDR) {
	    SgProsHedrStmt  *s;
	    char            *prosName;

	    s               = isSgProsHedrStmt(stmt);
	    prosName        = s->symbol()->identifier();
	    switchToFMProgBlock(prosName);
	    fprintf(fp, "\
FM_RETURN_STASHED_COPY_FUNC_DECL(%s)\n{\n", prosName);
	    //
	    // find out the (OUT/INOUT) args whose sizes 
	    // are determined at run time
	    //
	    for (i = 0; i < s->numberOfParameters(); i++) {
		SgSymbol            *symb;
		SgExpression        *attribute;
		int                  should_be_copied_out;
		
		should_be_copied_out = 0;
		symb                 = s->parameter(i);
		if (! symb) {
		    fm_internal_error("\
Expect a non-null symbol at %s:%d", __FILE__, __LINE__);
		}
		
		//
		// get the attribute of this argument
		//
		attribute = intentMapGet(intentMap, symb->identifier());
		if (! attribute) {
		    //
		    // by default == intent(INOUT)
		    //
		    should_be_copied_out = 1;
		} else if (attribute->variant() == OUT_OP 
			   || attribute->variant() == INOUT_OP)
		{
		    should_be_copied_out = 1;
		}
		
		if (should_be_copied_out) {
		    // get type information
		    SgType          *t;
		    SgExpression    *var_declared;
		    
		    t               = getProsArgType(prosName, symb);
		    if (!t) {
			//
			// unknown type caused by a implicit none
			// statement, which has been complained
			// already, so keep it quiet here
			//
			continue;
		    } 
		    
		    var_declared    = findSymbInDeclStmt(symb);
		    if (var_declared->variant() == ARRAY_REF) {
			SgArrayRefExp   *a;
			SgExpression    *a_size;
			
			a      = isSgArrayRefExp(var_declared);
			a_size = &(arraySizeExp(a));
			if (! IsInteger(a_size)) {
			    fprintf(fp, "\
    int A_%d_len;\n", i);
			}
		    }
		}
	    }
	    fprintf(fp, "\n\
    FM_RETURN_STASHED_COPY_FUNC_CODE(%s);\n\n", prosName);
	    
	    for (i = 0; i < s->numberOfParameters(); i++) {
		SgSymbol            *symb;
		SgExpression        *attribute;
		int                  should_be_copied_out;
		
		should_be_copied_out = 0;
		symb                 = s->parameter(i);
		if (! symb) {
		    fm_internal_error("\
Expect a non-null symbol at %s:%d", __FILE__, __LINE__);
		}
		
		//
		// get the attribute of this argument
		//
		attribute = intentMapGet(intentMap, symb->identifier());
		if (! attribute) {
		    //
		    // by default == intent(INOUT)
		    //
		    should_be_copied_out = 1;
		} else if (attribute->variant() == OUT_OP 
			   || attribute->variant() == INOUT_OP)
		{
		    should_be_copied_out = 1;
		}
		
		if (should_be_copied_out) {
		    // get type information
		    SgType          *t;
		    SgExpression    *var_declared;
		    
		    t               = getProsArgType(prosName, symb);
		    if (!t) {
			//
			// unknown type caused by a implicit none
			// statement, which has been complained
			// already, so keep it quiet here
			//
			continue;
		    } 
		    
		    var_declared    = findSymbInDeclStmt(symb);
		    if (var_declared->variant() == ARRAY_REF) {
			SgArrayRefExp   *a;
			SgExpression    *a_size;
			
			a      = isSgArrayRefExp(var_declared);
			a_size = &(arraySizeExp(a));
			if (! IsInteger(a_size)) {
			    fprintf(fp, "\
    nexus_get_stashed_int(buffer, &A_%d_len, 1);\n", i);
			    fprintf(fp, "\
    nexus_get_stashed_%s(buffer, args->A_%d, A_%d_len);\n", 
				    getTypeCString(t->variant()),
				    i, i);
			} else {
			    fprintf(fp, "\
    nexus_get_stashed_%s(buffer, args->A_%d, %d);\n", 
				    getTypeCString(t->variant()),
				    i, 
				    integerValue(a_size));
			}
		    } else if (var_declared->variant() == VAR_REF) {
			fprintf(fp, "\
    nexus_get_stashed_%s(buffer, args->A_%d, 1);\n", 
				getTypeCString(t->variant()),
				i);
		    }
		}
	    }
	    fprintf(fp, "}\n\n");
	} /* proc boundary */
    }
}

void gen_fm_port_init_func(FILE *fp)
{
    FOR_EACH_PROCESS(prosStmt,prosName)
    {
	int n_inports;
	int n_outports;
	int ip_i;
	int op_i;
	
	n_inports = n_outports = 0;
	ip_i = op_i = 0;

	fprintf(fp, "\
void FM_%s_PORT_INIT(\n",
		str2upper(prosName));
	FOR_EACH_PARAMETER_AND_TYPE(prosStmt,pros_symb,t)
	{
	    fprintf(fp, "\
                     f_%s_t *A_%d,\n",
		    getTypeString(t->variant()),
		    i);
	}
	END_FOR_EACH_PARAMETER_AND_TYPE();
	fprintf(fp,"\
                     f_integer_t *f_pblock)\n{\n");
	fprintf(fp,"\
    FM_PORT_INIT_CODE_1;\n");

	FOR_EACH_PORT_ARG(prefix,prosStmt,portStmt,symb)
	{
	    if (portStmt->variant() == INPORT_DECL) {
		n_inports++;
	    } 
	    if (portStmt->variant() == OUTPORT_DECL) {
		n_outports++;
	    }
	} 
	END_FOR_EACH_PORT_ARG();
	if (! n_inports) {
	    fprintf(fp, "\
    entry->ipls = (fm_ipl_t *) NULL;\n\
    entry->n_ipls = 0;\n\
    entry->n_total_inports = 0;\n\n");
	} else {
	    fprintf(fp, "\
    NexusMalloc(FM_%s_PORT_INIT(),entry->ipls, fm_ipl_t *,\n\
                sizeof(fm_ipl_t) * %d);\n",
		    str2upper((char *)prosName), n_inports);
	    fprintf(fp, "\
    entry->n_ipls = %d;\n",
		    n_inports);
	    fprintf(fp, "\
    entry->n_total_inports = 0;\n");

	    FOR_EACH_PORT_ARG(prefix,prosStmt,portStmt,portSymb)
	    {
		if (portStmt->variant() == INPORT_DECL) {
		    SgExpression *port;
		    fprintf(fp, "\
    entry->ipls[%d].entries = (fm_inport_t **) A_%d;\n",
			    ip_i, prefix_i);
		    port = findSymbInDeclStmt(portSymb);
		    if (! port) { continue; }
		    if (port->variant() == VAR_REF) {
			fprintf(fp, "\
    entry->ipls[%d].n_entries = %d;\n",
				ip_i, 1);
			fprintf(fp, "\
    entry->n_total_inports += 1;\n\n");

		    } else if (port->variant() == ARRAY_REF) {
			SgArrayRefExp *a;
			SgExpression  *a_size;
			a             = isSgArrayRefExp(port);
			a_size        = &(arraySizeExp(a));
			
			if (IsInteger(a_size)) {
			    fprintf(fp, "\
    entry->ipls[%d].n_entries = %d;\n",
				    ip_i, integerValue(a_size));
			    fprintf(fp, "\
    entry->n_total_inports += %d;\n\n",
				    integerValue(a_size));

			} else {
			    SgExpression *real_size;
			    real_size =&(replaceFormalArgConstInExpr(a_size));
			    fprintf(fp, "\
    entry->ipls[%d].n_entries = ",
				    ip_i);
			    printExprByDeref(fp, real_size);
			    fprintf(fp, ";\n");
			    fprintf(fp, "\
    entry->n_total_inports += ");
			    printExprByDeref(fp, real_size);
			    fprintf(fp, ";\n\n");
			}
		    } else {
			fm_internal_error("\
Do not expect arg type to be: %s at %s:%d", 
					  getVariantByName(port->variant()),
					  __FILE__,__LINE__);
		    }
		    ip_i++;
		}
	    } 
	    END_FOR_EACH_PORT_ARG();
	} 

	if (! n_outports) {
	    fprintf(fp, "\
    entry->opls = (fm_opl_t *) NULL;\n\
    entry->n_opls = 0;\n\
    entry->n_total_outports = 0;\n\n");

	} else {
	    fprintf(fp, "\
    NexusMalloc(FM_%s_PORT_INIT(),entry->opls, fm_opl_t *,\n\
                sizeof(fm_opl_t) * %d);\n",
		    str2upper((char *)prosName), n_outports);
	    fprintf(fp, "\
    entry->n_opls = %d;\n", n_outports);
	    fprintf(fp, "\
    entry->n_total_outports = 0;\n\n");

	    FOR_EACH_PORT_ARG(prefix,prosStmt,portStmt,portSymb)
	    {
		if (portStmt->variant() == OUTPORT_DECL) {
		    SgExpression *port;
		    fprintf(fp, "\
    entry->opls[%d].entries = (fm_outport_t **) A_%d;\n",
			    op_i, prefix_i);
		    port = findSymbInDeclStmt(portSymb);
		    if (! port) { continue; }
		    if (port->variant() == VAR_REF) {
			fprintf(fp, "\
    entry->opls[%d].n_entries = %d;\n",
				op_i, 1);
			fprintf(fp, "\
    entry->n_total_outports += 1;\n\n");

		    } else if (port->variant() == ARRAY_REF) {
			SgArrayRefExp *a;
			SgExpression  *a_size;
			a             = isSgArrayRefExp(port);
			a_size        = &(arraySizeExp(a));
			
			if (IsInteger(a_size)) {
			    fprintf(fp, "\
    entry->opls[%d].n_entries = %d;\n",
				    op_i, integerValue(a_size));
			    fprintf(fp, "\
    entry->n_total_outports += %d;\n\n",
				    integerValue(a_size));
			} else {
			    SgExpression *real_size;
//			    exprMapDump("prosHedrMap", prosHedrMap);
			    real_size =&(replaceFormalArgConstInExpr(a_size));
//			    printf("real_size = %s\n", real_size->unparse());
			    fprintf(fp, "\
    entry->opls[%d].n_entries = ",
				    op_i);
			    printExprByDeref(fp, real_size);
			    fprintf(fp, ";\n");
			    fprintf(fp, "\
    entry->n_total_outports += ");
			    printExprByDeref(fp, real_size);
			    fprintf(fp, ";\n\n");
			}
		    } else {
			fm_internal_error("\
Do not expect arg type to be: %s at %s:%d", 
					  getVariantByName(port->variant()),
					  __FILE__,__LINE__);
		    }
		    op_i++;
		}
	    } 
	    END_FOR_EACH_PORT_ARG();
	} 

	fprintf(fp, "\
    FM_PORT_INIT_CODE_2;\n",
		n_inports, n_outports);
	fprintf(fp, "}\n\n");
    }
    END_FOR_EACH_PROCESS();
}

		
	    

void gen_fm_invoke_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * FM_<PROC>_INVOKE()\n\
 *\n\
 * This is the function that is called from the compiled process\n\
 * block. It is passed all arguments to the process and a pointer\n\
 * to the process block.\n\
 *\n\
 * The compiler generates code to\n\
 *\n\
 *      Fill in the return_args structure with the addresses of the\n\
 *      real arguments\n\
 *\n\
 *      Compute the lengths of the variable-length array arguments.\n\
 *\n\
 *      Compute the size of the buffer needed for INTENT(IN) and\n\
 *      INTENT(INOUT) arguments.\n\
 *\n\
 *      Set the buffer size\n\
 *\n\
 *      Pack the buffer with the INTENT(IN) and INTENT(OUT) arguments\n\
 *\n\
 * The arguments passed in the message buffer are encoded as in\n\
 * the FM_RETURN_COPY_FUNC.\n\
 */\n");
}

void gen_fm_invoke(FILE *fp)
{
    int i;
    Pix ind;
//    gen_fm_invoke_comments(fp);
    for (ind = fmProgBlockMap->first(); ind; fmProgBlockMap->next(ind)) {
	FMProgBlock *block = &(fmProgBlockMap->contents(ind));
	SgStatement *stmt = block->proc();
	
	if (stmt->variant() == PROS_HEDR) {
	    SgProsHedrStmt    *s;
	    char              *prosName;
	    char              *handler_func_name;
 
	    s                 = isSgProsHedrStmt(stmt);
	    prosName          = s->symbol()->identifier();
	    handler_func_name = (char *)malloc(12+strlen(prosName)+1);
	    sprintf(handler_func_name, "_FMH_invoke_%s", prosName);

	    switchToFMProgBlock(prosName);

	    fprintf(fp, "\
void FM_%s_%d_INVOKE(\n", str2upper(prosName), s->numberOfParameters());
	    
	    for (i = 0; i < s->numberOfParameters(); i++) {
		SgSymbol      *symb;
		SgStatement   *stmt;
		SgType        *t;
		
		symb          = s->parameter(i);
		t             = getProsArgType(prosName, symb);
		if (! t) { continue; }
		
		fprintf(fp, "\
                   f_%s_t *A_%d,\n",
			getTypeString(t->variant()),
			i);
	    }
	    fprintf(fp, "\
                   f_integer_t *f_vc_offset,\n\
                   f_integer_t *f_pblock)\n");
	    
	    fprintf(fp, "{\n");
	    
	    //
	    // find out the (OUT/INOUT) args whose sizes 
	    // are determined at run time
	    //
	    for (i = 0; i < s->numberOfParameters(); i++) {
		SgSymbol            *symb;
		SgExpression        *attribute;
		int                  should_be_copied_out;
		
		should_be_copied_out = 0;
		symb                 = s->parameter(i);
		if (! symb) {
		    fm_internal_error("\
Expect a non-null symbol at %s:%d", __FILE__, __LINE__);
		}
		
		//
		// get the attribute of this argument
		//
		attribute = intentMapGet(intentMap, symb->identifier());
		if (! attribute) {
		    //
		    // by default == intent(INOUT)
		    //
		    should_be_copied_out = 1;
		} else if (attribute->variant() == OUT_OP 
			   || attribute->variant() == INOUT_OP)
		{
		    should_be_copied_out = 1;
		}
		
		if (should_be_copied_out || 1) {
		    // get type information
		    SgType          *t;
		    SgExpression    *var_declared;
		    
		    t               = getProsArgType(prosName, symb);
		    if (!t) {
			//
			// unknown type caused by a implicit none
			// statement, which has been complained
			// already, so keep it quiet here
			//
			continue;
		    } 

		    var_declared    = findSymbInDeclStmt(symb);
		    if (var_declared->variant() == ARRAY_REF) {
			SgArrayRefExp   *a;
			SgExpression    *a_size;
			
			a      = isSgArrayRefExp(var_declared);
			a_size = &(arraySizeExp(a));
			if (! IsInteger(a_size)) {
			    fprintf(fp, "\
    int A_%d_len;\n", i);
			}
		    }
		}
	    }
	    
	    fprintf(fp, "\
    FM_INVOKER_DECLS(%s);\n\n", 
		    prosName);
	    
	    fprintf(fp, "\
    FM_INVOKER_CODE_1(%s, \"_FMH_invoke_%s\", %d);\n",
		    prosName, prosName, handler_id(handler_func_name));
	    
	    //
	    // do return arguments
	    //
	    for (i = 0; i < s->numberOfParameters(); i++) {
		SgSymbol            *symb;
		SgExpression        *attribute;
		int                  should_be_copied_out;
		
		should_be_copied_out = 0;
		symb                 = s->parameter(i);
		if (! symb) {
		    fm_internal_error("\
Expect a non-null symbol at %s:%d", __FILE__, __LINE__);
		}
		
		//
		// get the attribute of this argument
		//
		attribute = intentMapGet(intentMap, symb->identifier());
		if (! attribute) {
		    //
		    // by default == intent(INOUT)
		    //
		    should_be_copied_out = 1;
		} else if (attribute->variant() == OUT_OP 
			   || attribute->variant() == INOUT_OP)
		{
		    should_be_copied_out = 1;
		}
		
		if (should_be_copied_out) {
		    fprintf(fp, "\
    return_args->A_%d = A_%d;\n", i, i);
		}
	    }
	    fprintf(fp, "\n\
    FM_INVOKER_CODE_2(%s);\n\n", prosName);
	    
	    // 
	    // match un-determined array size with
	    // with the argument
	    //
	    for (i = 0; i < s->numberOfParameters(); i++) {
		// find the var
		SgSymbol     *symb;
		SgExpression *var_declared;
		
		symb         = s->parameter(i);
		var_declared = findSymbInDeclStmt(symb);
		
		if (! var_declared || var_declared->variant() == VAR_REF) {
		    continue;
		}

		if (var_declared->variant() == ARRAY_REF) {
		    int            k;
		    SgArrayRefExp *a; 
		    SgExpression  *a_size;
                    SgExpression  *real_size;

		    a             = isSgArrayRefExp(var_declared);
		    a_size        = &(arraySizeExp(a));
		    
		    if (IsInteger(a_size)) {
			continue;
		    }
                    real_size = &(replaceFormalArgConstInExpr(a_size));
                    fprintf(fp, "\
    A_%d_len = ",
                            i);
		    printExprByDeref(fp, real_size);
		    fprintf(fp, ";\n");
		} else {
		    fm_internal_error("\
Do not expect var type (%s) at %s:%d", 
				      getVariantByName(var_declared->variant()), __FILE__, __LINE__);
		}
	    }
	    //
	    // Pack annotation vector
	    //
	    fprintf(fp, "\
    size += nexus_sizeof_int(&mbuf, 1);\n\
    n_elts++;\n");

	    // 
	    // Compute the size of buffer needed for
	    // INTENT(IN) and INTENT(INOUT) arguments
	    //
	    gen_compute_IN_buffer_size(fp, s);
	    //
	    // Pack the annotation vector
	    //
	    fprintf(fp, "\
    nexus_put_int(&mbuf, f_vc_offset, 1);\n");

	    gen_pack_IN_buffer(fp, s);

	    fprintf(fp, "\
    FM_INVOKER_CODE_3(%s);\n}\n\n", prosName);
	}
    }
}
		
void gen_invoke_thread_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * invoke_<proc>_thread()\n\
 * \n\
 * This is the routine that is running in the new thread created by \n\
 * the process invocation handler.\n\
 * \n\
 * It is responsible for calling the user subroutine, passing\n\
 * pointers into the args structure for the arguments to the \n\
 * subroutine. \n\
 * \n\
 * The compiler must generate code to compute the size of the\n\
 * INTENT(INOUT) and INTENT(OUT) arguments, allocate the buffer,\n\
 * and pack the buffer with those arguments. Remember that the\n\
 * lengths of the variable lengths arrays are stored in the \n\
 * args structure. \n\
 * \n\
 * The message buffer is encoded as in FM_RETURN_COPY_FUNC().\n\
 */\n");
}

void gen_invoke_thread(FILE *fp)
{
    SgProcessorsRefExp *subm_vec;
    
//    gen_invoke_thread_comments(fp);
    FOR_EACH_PROCESS(s,prosName)
    
    s        = isSgProsHedrStmt(stmt);
    prosName = stmt->symbol()->identifier();

//static void invoke_%s_thread(void *t_args)\n", prosName);

    fprintf(fp, "\
FM_INVOKE_THREAD_DECL(%s)\n{\n", prosName);

    fprintf(fp, "\
    FM_INVOKE_THREAD_CODE_1(%s);\n\n", prosName);

    //
    // set VC table offset
    //
    fprintf(fp, "\
    _fm_vc_table_offset(args->vc_offset);\n");

    fprintf(fp, "\
    %s(", str2upper(prosName));

    for (int i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol      *symb;
	SgExpression  *var_declared;

	symb          = s->parameter(i);
	var_declared  = findSymbInDeclStmt(symb);

	if (var_declared->variant() == VAR_REF) {
	    fprintf(fp, "&args->A_%d", i);
	} else {
	    fprintf(fp, "args->A_%d", i);
	}
	if (i != s->numberOfParameters() - 1) {
	    fprintf(fp, ", ");
	} 
    }
    fprintf(fp, ");\n\n");

    fprintf(fp, "\
    FM_INVOKE_THREAD_CODE_2(%s);\n\n", prosName);

    gen_compute_OUT_buffer_size(fp, s);
    
    gen_pack_OUT_buffer(fp, s);

    fprintf(fp, "\
    FM_INVOKE_THREAD_CODE_3(%s);\n}\n\n", prosName);

    END_FOR_EACH_PROCESS()
}

void gen_invoke_handler_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * FM_INVOKE_HANDLER()\n\
 * \b\
 * This is the handler that is invoked to create a new FM process.\n\
 * \n\
 * The compiler must generate code to unpack the message buffer containing \n\
 * the INTENT(IN) and INTENT(INOUT) arguments. This buffer is packed usingn\n\
 * the scheme described under FM_RETURN_COPY_FUNC().\n\
 * \n\
 * Note that for variable length arrays, the length is unpacked, \n\
 * FM_ALLOC_ARG_SPACE() is invoked to allocate the storage for that\n\
 * argument, and then the array itself is unpaced into the newly allocated\n\
 * storage.\n\
 * \n\
 */\n");
}

void gen_invoke_handler(FILE *fp)
{
//    gen_invoke_handler_comments(fp);
    FOR_EACH_PROCESS(s,prosName);

    fprintf(fp, "\
FM_INVOKE_HANDLER_DECL(%s)\n{\n", prosName);
    
    fprintf(fp, "\
    FM_INVOKE_HANDLER_CODE_1(%s);\n", prosName);

    //
    // get annotation vector
    //
    fprintf(fp, "\
    nexus_get_int(buf, &args->vc_offset, 1);\n");

    gen_unpack_handler_buffer(fp, s);

    fprintf(fp, "\
    FM_INVOKE_HANDLER_CODE_2(%s);\n}\n\n", prosName);

    END_FOR_EACH_PROCESS()
}

void gen_fm_port_init(FILE *fp)
{
    int             i;

    FOR_EACH_PROCESS(s,prosName);

    fprintf(fp, "\
void FM_%s_PORT_INIT(\n", str2upper(prosName));

    FOR_EACH_PARAMETER_AND_TYPE(s,symb,t);
      
    fprintf(fp, "\
                     f_%s_t *A_%d,\n",
	    getTypeString(t->variant()),
	    i);

    END_FOR_EACH_PARAMETER_AND_TYPE();

    fprintf(fp, "\
                     f_integer_t *f_pblock)\n{\n");
    fprintf(fp, "\
    fm_pblock_t *pblock;\n\
    fm_pblock_entry *entry;\n\n\
    pblock = (fm_pblock_t *) (*f_pblock);\n\n\
    NexusAssert(pblock->current_entry >= 0 && \n\
                pblock->current_entry < pblock->n_entries);\n\n\
    entry = &(pblock->entries[pblock->current_entry]);\n\n");
    
    END_FOR_EACH_PROCESS();
}

void gen_destination_struct_comments(FILE *fp)
{
    fprintf(fp, "\
/*\n\
 * Destination struct.\n\
 *\n\
 * This is used to save the pointers into user space of the arguments\n\
 * of a FM receive.\n\
 */\n");
}

void gen_destination_struct(FILE *fp)
{
    Pix       ind;
    int i;

//    gen_destination_struct_comments(fp);

    FOR_EACH_INPORT_DECL(inport_stmt,key);

    fprintf(fp, "\
typedef struct\n{\n");

    for (i = 0; i < inport_stmt->numberOfPortTypes(); i++) {
	SgPortTypeExp *port_type;
	SgType        *t;
	SgExpression  *ref;

	port_type     = inport_stmt->porttype(i);
	t             = port_type->type();
	ref           = port_type->ref();

	fprintf(fp, "\
    f_%s_t *A_%d;\n", 
		getTypeString(t->variant()),
		i);
	if (ref) {
	    if (ref->variant() == ARRAY_REF) {
		fprintf(fp, "\
    int A_%d_size;\n",
			i);
	    }
	}
    }

    if (opt_debug) {
	fprintf(fp, "\
    f_character_t f_file_name;\n\
    int f_file_line;\n");
    }
    
    fprintf(fp, "\
} %s_dest_t;\n\n", 
	    (char *)key->chars());

#if 0
    END_FOR_EACH_INPORT_DECL();
    /*
     * generate fwd_dest_t
     *
     */

    FOR_EACH_INPORT_DECL(inport_stmt,key);
#endif
    
//    if (inport_type_has_ports(inport_stmt)) {
//	continue;
//    }

    fprintf(fp, "\
typedef struct\n{\n");

    for (i = 0; i < inport_stmt->numberOfPortTypes(); i++) {
	SgPortTypeExp *port_type;
	SgType        *t;
	SgExpression  *ref;
	int           t_variant;

	port_type     = inport_stmt->porttype(i);
	t             = port_type->type();
	ref           = port_type->ref();
	t_variant     = t->variant();

	if (! ref || t_variant == T_INPORT || t_variant == T_OUTPORT) {
	    fprintf(fp, "\
    f_%s_t A_%d;\n",  getTypeString(t_variant), i);
	    continue;
	}

	if (ref->variant() == ARRAY_REF) {
	    SgArrayRefExp *a;
	    int            j;
	    SgExpression  *a_size;

	    a      = isSgArrayRefExp(ref);
	    a_size = &(arraySizeExp(a));
	    if (IsInteger(a_size)) {
		fprintf(fp, "\
    f_%s_t A_%d[%d];\n",  
			getTypeString(t_variant), i,
			integerValue(a_size));
	    } else {
		fprintf(fp, "\
    f_%s_t *A_%d;\n",  getTypeString(t_variant), i);
#if BIZZARE
		fprintf(fp, "\
    int A_%d_len;\n", i);
#endif /* 0 */
	    }
	} else if (ref->variant() == VAR_REF) {
	    fprintf(fp, "\
    f_%s_t A_%d;\n",  getTypeString(t_variant), i);
	} else {
	    fm_internal_error("\
Don't expect to have a port type reference to be type %s",
			      getVariantByName(t_variant));
	}
    }
#if 0
	fprintf(fp, "\
    f_%s_t A_%d;\n", 
		getTypeString(t->variant()),
		i);
	if (ref) {
	    if (ref->variant() == ARRAY_REF) {
		fprintf(fp, "\
    int A_%d_size;\n",
			i);
	    }
	}
    }
#endif

    fprintf(fp, "\
} %s_fwd_dest_t;\n\n", 
	    (char *)key->chars());


    END_FOR_EACH_INPORT_DECL();
}

/*
 * opt2ipt()
 * 
 * Return the string which is the key to a inport
 * compatible to the outport
 *
 */

char * opt2ipt(const char *key)
{
    /*
     * change each occurance of 'o' to 'i'
     * until the resulting string is the key
     * to an inport in portSet
     *
     */
    String port_key(key);
    String port_type = port_key.after(currentFileBase);
    char *n_key         = strdup(port_type.chars());
    char *string_buffer = n_key;

    while (1) {
	char *o_char = (char *)strchr((const char *)string_buffer,'o');
	
	if (! o_char) {
#if 0
	    fm_internal_error("\
Invalid outport key string encountered: %s", key);
#endif /* 0 */
	    return((char *)key);
	}

	if (*(o_char+1) == 'p' && *(o_char+2) == 't') {
	    *o_char       = 'i';
	    n_key = ppString(currentFileBase, "", n_key);
	    return n_key;
	}
	string_buffer = o_char+1;
    }
    return (char *)0;
}

char * ipt2opt(const char *key)
{
    /*
     * change each occurance of 'o' to 'i'
     * until the resulting string is the key
     * to an inport in portSet
     *
     */
    String port_key(key);
    String port_type = port_key.after(currentFileBase);
    char *n_key         = strdup(port_type.chars());
    char *string_buffer = n_key;

    while (1) {
	char *i_char = (char *)strchr((const char *)string_buffer,'i');
	
	if (! i_char) {
#if 0
	    fm_internal_error("\
Invalid inport key string encountered: %s", key);
#endif /* 0 */
	    return ((char *) key);
	}


	if (*(i_char+1) == 'p' && *(i_char+2) == 't') {
	    *i_char       = 'o';
	    n_key = ppString(currentFileBase, "", n_key);
	    return(n_key);
	}
	string_buffer = i_char+1;
    }
    return (char *)0;
}
    

/*
 * gen_porttype_init()
 *
 * Generate functions which initializes a channel of 
 * the given port types
 */

void gen_fm_outport_init(FILE *fp)
{
    Pix ind;

    FOR_EACH_OUTPORT_DECL(stmt,key)
    {
	fprintf(fp, "\
void FM_%s_INIT(fm_channel_t *channel)\n\
{\n\
    channel->copy_func = fm_%s_copy;\n\
    channel->stashed_copy_func = fm_%s_stashed_copy;\n",
                str2upper((char *)key->chars()), opt2ipt(key->chars()), opt2ipt(key->chars()));

//        if (! outport_type_has_ports(stmt)) {
            fprintf(fp, "\
    channel->forward_func = fm_%s_forward;\n",
                    opt2ipt(key->chars()));
//        }
        fprintf(fp, "}\n\n");
    }
    END_FOR_EACH_OUTPORT_DECL();
}

static void gen_send_debug(FILE *fp, int i,
                           SgType *t, SgExpression *p_type_ref)
{
    char *type;
    char *fmt;

    type = getTypeCString(t->variant());
    fmt = getTypePrintfFormat(t->variant());

    if ((p_type_ref == NULL) || (p_type_ref->variant() == VAR_REF))
    {
	fprintf(fp, "\
        if (FMDebug(UserSR, 2))\n\
	    nexus_printf(\"send(%%s:%%d): putting %s arg_%d = %s\\n\", file_name, *f_file_line, *arg_%d);\n",
			type, i, fmt, i);

    }
    else if (p_type_ref->variant() == ARRAY_REF)
    {
	fprintf(fp, "\
        if (FMDebug(UserSR, 1))\n\
	    nexus_printf(\"(%%s:%%d) fm_send: putting %s arg_%d len = %%d\\n\", file_name, *f_file_line, *arg_%d_len);\n",
		         type, i, i);
    }
}

/*
 * gen_porttype_send()
 *
 * Generate send functions
 *
 */


void gen_porttype_send(FILE *fp, SgOutportStmt *s, char *key)
{
    int i;
    char sendname[1024];
    int n_inports, n_outports;
    int ip_count, op_count;

    sprintf(sendname, "FM_%s_SEND", str2upper(key));

    assert(s);

    fprintf(fp, "\
void FM_%s_SND%s(\n\
                  f_integer_t *f_outport", 
	    str2upper(key), (opt_debug)? "_D" : "");
    
    for (i = 0; i < s->numberOfPortTypes(); i++) {
	SgPortTypeExp *p_type = s->porttype(i);
	SgExpression  *p_type_ref = p_type->ref();
	char *p_type_string = getTypeString((p_type->type())->variant());
	
	fprintf(fp,",\n\
                  f_%s_t *arg_%d", p_type_string, i);
	
	if (isSgArrayRefExp(p_type_ref)) {
	    fprintf(fp, ",\n\
                  f_integer_t *arg_%d_len", i);
	}	
    }
    fprintf(fp, ",\n\
                  f_integer_t *f_ioval");

    if (opt_debug) {
	fprintf(fp, ",\n\
                  f_character_t f_file_name,\n\
                  f_integer_t *f_file_line\n\
                  CHAR_LEN_DECL(f_file_name_len))\n");
    } else {
	fprintf(fp, ")\n");
    }
    
    fprintf(fp, "{\n");

    if (opt_debug) {
	fprintf(fp, "\
    char file_name[1024];\n\
    LOCAL_CHAR_LEN_DECL(f_file_name_len)\n");
    }

    fprintf(fp, "\
    nexus_buffer_t buffer;\n\
    fm_outport_t *outport = (fm_outport_t *) *f_outport;\n\n");

    n_inports = n_outports = 0;
    for (i = 0; i < s->numberOfPortTypes(); i++) {
	SgPortTypeExp     *p_type_exp;
	SgExpression      *p_type_ref;
	SgType            *t;
	
	p_type_exp        = s->porttype(i);
	p_type_ref        = p_type_exp->ref();
	t                 = p_type_exp->type();

	if (t->variant() == T_INPORT) {
	    n_inports++;
	}
	if (t->variant() == T_OUTPORT) {
	    n_outports++;
	}
    }

    /*
     * Port Migration Code
     *
     */
    
    if (n_inports || n_outports) {
	fprintf(fp, "\
    fm_synch_t synch;\n");
    }

    /*
     * Generate the inport and outport list
     * Data Structure
     *
     */

    if (n_inports) {
	fprintf(fp, "\
    fm_ipl_t ipls[%d];\n",
		n_inports);
    }

    if (n_outports) {
	fprintf(fp, "\
    fm_opl_t opls[%d];\n",
		n_outports);
    }

    fprintf(fp, "\n");

    ip_count = 0;
    if (n_inports) {
	for (i = 0; i < s->numberOfPortTypes(); i++) {
	    SgPortTypeExp     *p_type_exp;
	    SgExpression      *p_type_ref;
	    SgType            *t;
	    
	    p_type_exp        = s->porttype(i);
	    p_type_ref        = p_type_exp->ref();
	    t                 = p_type_exp->type();
	    
	    if (t->variant() == T_INPORT) {
		fprintf(fp, "\
    ipls[%d].entries = (fm_inport_t **) arg_%d;\n\
    ipls[%d].n_entries = 1;\n",
			ip_count,
			i,
			ip_count);
		ip_count++;
	    }
	}
    }

    fm_assert(ip_count == n_inports);

    op_count = 0;
    if (n_outports) {
	for (i = 0; i < s->numberOfPortTypes(); i++) {
	    SgPortTypeExp     *p_type_exp;
	    SgExpression      *p_type_ref;
	    SgType            *t;
	    
	    p_type_exp        = s->porttype(i);
	    p_type_ref        = p_type_exp->ref();
	    t                 = p_type_exp->type();
	    
	    if (t->variant() == T_OUTPORT) {
		fprintf(fp, "\
    opls[%d].entries = (fm_outport_t **) arg_%d;\n\
    opls[%d].n_entries = 1;\n",
			op_count,
			i,
			op_count);
		op_count++;
	    }
	}
    }

    fm_assert(op_count == n_outports);

#if ALLOW_ARRAY_INITIALIZER_ON_AUTOMATIC_ARRAYS
    if (n_inports) {
	fprintf(fp, "\
    fm_ipl_t ipls[] =\n\
    {\n");

	for (i = 0; i < s->numberOfPortTypes(); i++) {
	    SgPortTypeExp     *p_type_exp;
	    SgExpression      *p_type_ref;
	    SgType            *t;
	    
	    p_type_exp        = s->porttype(i);
	    p_type_ref        = p_type_exp->ref();
	    t                 = p_type_exp->type();
	    
	    if (t->variant() == T_INPORT) {
		fprintf(fp, "\
       { (fm_inport_t **) arg_%d, 1 }",
			i);
		if (i != s->numberOfPortTypes()-1) {
		    fprintf(fp, ",\n");
		} else {
		    fprintf(fp, "\n");
		}
	    }
	}
	fprintf(fp, "\
    };\n\n");
    }

    if (n_outports) {
	fprintf(fp, "\
    fm_opl_t opls[] =\n\
    {\n");

	for (i = 0; i < s->numberOfPortTypes(); i++) {
	    SgPortTypeExp     *p_type_exp;
	    SgExpression      *p_type_ref;
	    SgType            *t;
	    
	    p_type_exp        = s->porttype(i);
	    p_type_ref        = p_type_exp->ref();
	    t                 = p_type_exp->type();
	    
	    if (t->variant() == T_OUTPORT) {
		fprintf(fp, "\
       { (fm_outport_t **) arg_%d, 1 }",
			i);
		if (i != s->numberOfPortTypes()-1) {
		    fprintf(fp, ",\n");
		} else {
		    fprintf(fp, "\n");
		}
	    }
	}
	fprintf(fp, "\
    };\n\n");
    }
#endif /* 0 */
    /*
     * Perform the hairy Fortran to C string conversion
     *
     */

    if (opt_debug) {	    
	fprintf(fp, "\
    CONVERT_CHAR_FTN_TO_C_INPLACE(f_file_name, f_file_name_len, file_name);\n\n");
    }

    fprintf(fp, "\
    OutportMagicCheck(outport);\n\n");

    if (n_inports || n_outports) {
	fprintf(fp, "\
    _fm_synch_init(&synch);\n\n");

	fprintf(fp, "\
    _fm_create_ports(%d, %s, %d,\n\
                     %d, %s, %d,\n\
                     &outport->channel_gp,\n\
                     &synch);\n\n",
		n_inports,
		(n_inports)? "ipls" : "(fm_ipl_t *) NULL",
		n_inports,
		n_outports,
		(n_outports)? "opls" : "(fm_opl_t *) NULL",
		n_outports);

	fprintf(fp, "\
    _fm_synch_wait(&synch);\n\n\
    _fm_notify_ports(%d, (fm_ipl_t *) %s, %d,\n\
                     %d, (fm_opl_t *) %s, %d);\n\n",
		n_inports,
		(n_inports)? "ipls" : "NULL",
		n_inports,
		n_outports,
		(n_outports)? "opls" : "NULL",
		n_outports);
    }

    if (opt_debug) {
	fprintf(fp, "\
    _fm_init_send_debug(outport,&buffer,f_ioval,file_name,*f_file_line);\n\n");
    } else {
	fprintf(fp, "\
    _fm_init_send(outport,&buffer,f_ioval);\n\n");
    }

    fprintf(fp, "\
    if (*f_ioval == 0)\n\
    {\n\
        int size, n_elts, obj_n_elts;\n\n\
        size = n_elts = 0;\n\n");
    
    for (i = 0; i < s->numberOfPortTypes(); i++) {
	SgPortTypeExp     *p_type_exp;
	SgExpression      *p_type_ref;
	SgType            *t;
	
	p_type_exp        = s->porttype(i);
	p_type_ref        = p_type_exp->ref();
	t                 = p_type_exp->type();

	if (! t) {
	    fm_internal_error("\
Do not expect a null-type for a port type %s:%d", __FILE__, __LINE__);
	}

	/* 
	 * port over port
         *
         */

	if (t->variant() == T_INPORT ||
	    t->variant() == T_OUTPORT )
	{
	    fprintf(fp, "\
        size += _fm_sizeof_%s(&buffer, arg_%d, 1, &obj_n_elts);\n\
        n_elts += obj_n_elts;\n\n",
		    getTypeCString(t->variant()),
		    i);
	    continue;
	}
	
	if (! p_type_ref || t->variant() == T_INPORT || t->variant() == T_OUTPORT) {
	    fprintf(fp, "\
        size += nexus_sizeof_%s(&buffer, 1);\n\
        n_elts++;\n",
		    getTypeCString(t->variant()));
	} else {
	    if (p_type_ref->variant() == VAR_REF) {
		fprintf(fp, "\
        size += nexus_sizeof_%s(&buffer, 1);\n\
        n_elts++;\n",
		    getTypeCString(t->variant()));
	    } else if (p_type_ref->variant() == ARRAY_REF) {
		fprintf(fp, "\
        size += nexus_sizeof_int(&buffer, 1);\n\
        n_elts++;\n\n\
        size += nexus_sizeof_%s(&buffer, (int) *arg_%d_len);\n\
        n_elts += (int) *arg_%d_len;\n", 
			getTypeCString(t->variant()),
			i, i);
	    } else {
		fm_internal_error("\
Do not expect port type to be %s at %s:%d",
				  getVariantByName(p_type_ref->variant()),
				  __FILE__, __LINE__);
	    }
	}
    }
    
    fprintf(fp, "\
        nexus_set_buffer_size(&buffer, size, n_elts);\n\n");

    for (i = 0; i < s->numberOfPortTypes(); i++) {
	SgPortTypeExp     *p_type_exp;
	SgExpression      *p_type_ref;
	SgType            *t;
	
	p_type_exp        = s->porttype(i);
	p_type_ref        = p_type_exp->ref();
	t                 = p_type_exp->type();

	if (! t) {
	    fm_internal_error("\
Do not expect a null-type for a port type %s:%d", __FILE__, __LINE__);
	}

	if (opt_debug)
	    gen_send_debug(fp, i, t, p_type_ref);
	if (! p_type_ref || t->variant() == T_INPORT || t->variant() == T_OUTPORT) {
	    fprintf(fp, "\
        nexus_put_%s(&buffer, arg_%d, 1);\n", 
		    getTypeCString(t->variant()),
		    i);
	} else if (p_type_ref->variant() == VAR_REF) {
	    fprintf(fp, "\
        nexus_put_%s(&buffer, arg_%d, 1);\n", 
		    getTypeCString(t->variant()),
		    i);
	} else if (p_type_ref->variant() == ARRAY_REF) {
	    fprintf(fp, "\
        nexus_put_int(&buffer, arg_%d_len, 1);\n\
        nexus_put_%s(&buffer, arg_%d, (int) *arg_%d_len);\n",
		    i,
		    getTypeCString(t->variant()),
		    i, i);
	} else {
	    fm_internal_error("\
Do not expect port type to be %s at %s:%d",
			      getVariantByName(p_type_ref->variant()),
			      __FILE__, __LINE__);
	}
    }
    if (opt_debug) {
	fprintf(fp, "\
        _fm_send_debug(&buffer, f_ioval, file_name, *f_file_line);\n\
    }\n");
    } else {
	fprintf(fp, "\
        _fm_send(&buffer, f_ioval);\n\
    }\n");
    }

    if (n_inports || n_outports) {
	fprintf(fp, "\
    _fm_invalidate_ports(%d, (fm_ipl_t *) %s, %d,\n\
                         %d, (fm_opl_t *) %s, %d);\n\n",
		n_inports,
		(n_inports)? "ipls" : "NULL",
		n_inports,
		n_outports,
		(n_outports)? "opls" : "NULL",
		n_outports);
    }
    
    if (n_inports || n_outports)
        fprintf(fp, "\n\
    _fm_synch_destroy(&synch);");

    fprintf(fp, "\n}\n\n");
}	

void gen_fm_receive_copy(FILE *fp)
{
    Pix ind;
    char tail_str[1000];

    FOR_EACH_INPORT_DECL(inportStmt,key)
    {
	fprintf(fp, "\
static void fm_%s_copy(nexus_buffer_t *buffer, void *destination)\n{\n\
    %s_dest_t *dest;\n",
		key->chars(), key->chars());

	FOR_EACH_PORT_TYPE(_p_i,inportStmt,p_type,p_ref)
	{
	    if (! p_ref || p_type->variant() == T_INPORT || p_type->variant() == T_OUTPORT) {
		continue;
	    }

	    if (p_ref->variant() == ARRAY_REF) {
		fprintf(fp, "\
    int A_%d_len;\n\n", 
			_p_i);
	    }
	}
	END_FOR_EACH_PORT_TYPE();

	fprintf(fp, "\
    dest = (%s_dest_t *) destination;\n",
		key->chars());

	FOR_EACH_PORT_TYPE(_p_j,inportStmt,p_type,p_ref)
        {
	    if (p_type->variant() == T_INPORT)
	    {
		sprintf(tail_str, ", FM_%s_INIT",
			str2upper(ipt2opt(popTypeEncode(p_type,p_ref))));
	    }
	    else
		tail_str[0] = '\0';
	    
	    if (! p_ref || p_type->variant() == T_INPORT || p_type->variant() == T_OUTPORT)
	    {
		fprintf(fp, "\
    nexus_get_%s(buffer, dest->A_%d, 1%s);\n",
			getTypeCString(p_type->variant()),
			_p_j, tail_str);
		if (opt_debug)
		{
		    fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"(%%s:%%d) fm_copy(): got A_%d = %s\\n\", dest->f_file_name, dest->f_file_line, *dest->A_%d);\n",
			    _p_j,
			    getTypePrintfFormat(p_type->variant()),
			    _p_j);
		}
	    }
	    else if (p_ref->variant() == VAR_REF)
	    {
		fprintf(fp, "\
    nexus_get_%s(buffer, dest->A_%d, 1%s);\n",
			getTypeCString(p_type->variant()),
			_p_j, tail_str);
		if (opt_debug)
		{
		    fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"(%%s:%%d) fm_copy(): got A_%d = %s\\n\", dest->f_file_name, dest->f_file_line, *dest->A_%d);\n",
			    _p_j,
			    getTypePrintfFormat(p_type->variant()),
			    _p_j);
		}
	    }
	    else if (p_ref->variant() == ARRAY_REF)
	    {
		fprintf(fp, "\
    nexus_get_int(buffer, &A_%d_len, 1);\n",
			_p_j);
		if (opt_debug)
		{
		    fprintf(fp, "\
    if (FMDebug(UserSR, 1))\n\
        nexus_printf(\"(%%s:%%d) fm_copy(): got A_%d_len = %%d\\n\", dest->f_file_name, dest->f_file_line, A_%d_len);\n",
			    _p_j, _p_j);
		}
		if (opt_debug)
		{
		    fprintf(fp, "\
    if (dest->A_%d_size >= 0 && A_%d_len > dest->A_%d_size)\n\
         nexus_fatal(\"receive overrun (%%s:%%d). Destination size=%%d, message size=%%d\\n\", dest->f_file_name, dest->f_file_line, dest->A_%d_size, A_%d_len );\n\n",
 			    _p_j, _p_j, _p_j, _p_j, _p_j );
		}
//        nexus_fatal(\"receive overrun in fm_%s_copy(). Destination size=%%d, message size=%%d\\n\", dest->A_%d_size, A_%d_len );\n\n",
//			_p_j, _p_j, _p_j, key->chars(),_p_j, _p_j );
		
		fprintf(fp, "\
    nexus_get_%s(buffer, dest->A_%d, A_%d_len%s);\n",
			getTypeCString(p_type->variant()),
			_p_j, _p_j, tail_str);
	    }
	    else
	    {
		fm_internal_error("\
Do not expect a reference in a port type to be: %s",
				  getVariantByName(p_ref->variant()));
	    }
	} 
        END_FOR_EACH_PORT_TYPE();
	fprintf(fp, "}\n\n");
    }
    END_FOR_EACH_INPORT_DECL();
}

void gen_fm_receive_stashed_copy(FILE *fp)
{
    Pix ind;
    char tail_str[1000];

    FOR_EACH_INPORT_DECL(inportStmt, key)
    {

        fprintf(fp, "\
static void fm_%s_stashed_copy(nexus_stashed_buffer_t *buffer, void *destination)\n{\n\
    %s_dest_t *dest;\n",
                key->chars(), key->chars());

        FOR_EACH_PORT_TYPE(_p_i,inportStmt,p_type,p_ref)
        {
            if (! p_ref || p_type->variant() == T_INPORT || p_type->variant() == T_OUTPORT)
            {
                continue;
            } 
            
            if (p_ref->variant() == ARRAY_REF)
            {
                fprintf(fp, "\
    int A_%d_len;\n\n", 
                        _p_i);
            }
        }
        END_FOR_EACH_PORT_TYPE();

        fprintf(fp, "\
    dest = (%s_dest_t *) destination;\n",
                key->chars());
        
        FOR_EACH_PORT_TYPE(_p_j,inportStmt,p_type,p_ref)
        {
            if (p_type->variant() == T_INPORT)
            {
                sprintf(tail_str, ", FM_%s_INIT",
			str2upper(ipt2opt(popTypeEncode(p_type,p_ref))));
            }
            else
                tail_str[0] = '\0';
            
            if (! p_ref || p_type->variant() == T_INPORT || p_type->variant() == T_OUTPORT)
            {
                fprintf(fp, "\
    nexus_get_stashed_%s(buffer, dest->A_%d, 1%s);\n",
                        getTypeCString(p_type->variant()),
                        _p_j, tail_str);
                if (opt_debug)
                {
                    fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"(%%s:%%d) fm_copy(): got A_%d = %s\\n\", dest->f_file_name, dest->f_file_line, *dest->A_%d);\n",
			_p_j,
                            getTypePrintfFormat(p_type->variant()),
                            _p_j);
                }
            }
            else if (p_ref->variant() == VAR_REF)
            {
                fprintf(fp, "\
    nexus_get_stashed_%s(buffer, dest->A_%d, 1%s);\n",
			getTypeCString(p_type->variant()),
                        _p_j, tail_str);
                if (opt_debug)
                {
                    fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"(%%s:%%d) fm_copy(): got A_%d = %s\\n\", dest->f_file_name, dest->f_file_line, *dest->A_%d);\n",
                            _p_j,
                            getTypePrintfFormat(p_type->variant()),
                            _p_j);
                }
            }
            else if (p_ref->variant() == ARRAY_REF)
            {
                fprintf(fp, "\
    nexus_get_stashed_int(buffer, &A_%d_len, 1);\n",
                        _p_j);

                if (opt_debug)
                {
                    fprintf(fp, "\
    if (FMDebug(UserSR, 1))\n\
        nexus_printf(\"(%%s:%%d) fm_copy(): got A_%d_len = %%d\\n\", dest->f_file_name, dest->f_file_line, A_%d_len);\n",
                            _p_j, _p_j);
                }
                if (opt_debug)
                {
                    fprintf(fp, "\
    if (dest->A_%d_size >= 0 && A_%d_len > dest->A_%d_size)\n\
         nexus_fatal(\"receive overrun (%%s:%%d). Destination size=%%d, message size=%%d\\n\", dest->f_file_name, dest->f_file_line, dest->A_%d_size, A_%d_len );\n\n",
                            _p_j, _p_j, _p_j, _p_j, _p_j );
                }
                //        nexus_fatal(\"receive overrun in fm_%s_stashed_copy().  Destination size=%%d, message size=%%d\\n\", dest->A_%d_size, A_%d_len);\n\n",
                //			_p_j, _p_j, _p_j, key->chars(), _p_j, _p_j);

                fprintf(fp, "\
    nexus_get_stashed_%s(buffer, dest->A_%d, A_%d_len%s);\n",
                        getTypeCString(p_type->variant()),
                        _p_j, _p_j, tail_str);
            }
	    else
            {
                fm_internal_error("\
Do not expect a reference in a port type to be: %s",
                                  getVariantByName(p_ref->variant()));
            }
        }
        END_FOR_EACH_PORT_TYPE();
        fprintf(fp, "}\n\n");

    }
    END_FOR_EACH_INPORT_DECL();
}

void gen_fm_receive_forward(FILE *fp)
{
    Pix ind;
    int i;
    int n_inports;
    int n_outports;
    int ip_count;
    int op_count;

    FOR_EACH_INPORT_DECL(inportStmt, key);


    fprintf(fp, "\
static void fm_%s_forward(nexus_stashed_buffer_t *buffer,\n\
                          nexus_global_pointer_t *dest_gp)\n{\n\
    %s_fwd_dest_t dest;\n\
    int size, n_elts, obj_n_elts;\n\
    nexus_buffer_t mbuf;\n\n",
	    key->chars(), key->chars());

#if 0    
    if (inport_type_has_ports(inportStmt))
    {
	fprintf(fp, "\
    nexus_fatal(\"fm_%s_forward(): forwarding of channels containing ports not yet implemented\\n\");\n\
}\n\n",
		key->chars());
	continue;
    }
#endif /* 0 */

    n_inports = n_outports = 0;
    FOR_EACH_PORT_TYPE(_p_i,inportStmt,p_type,p_ref)
    {
	if (! p_ref ) {
	    continue;
	} else if (p_type->variant() == T_INPORT) {
	    n_inports++;
	} else if (p_type->variant() == T_OUTPORT) {
	    n_outports++;
	}
    
	if (p_ref->variant() == ARRAY_REF) {
	    SgArrayRefExp  *a;
	    SgExpression   *a_size;

	    a               = isSgArrayRefExp(p_ref);
	    a_size          = &(arraySizeExp(a));

	    if (! IsInteger(a_size)) {
		fprintf(fp, "\
    int A_%d_len;\n\n", 
			_p_i);
	    }
	}
    }
    END_FOR_EACH_PORT_TYPE();

    if (n_inports || n_outports) {
	fprintf(fp, "\
    fm_synch_t synch;\n");
    }

    if (n_inports) {
	fprintf(fp, "\
    fm_ipl_t ipls[%d];\n",
		n_inports);
    } 

    if (n_outports) {
	fprintf(fp, "\
    fm_opl_t opls[%d];\n",
		n_outports);
    } 

    ip_count = op_count = 0;
    FOR_EACH_PORT_TYPE(_p_j,inportStmt,p_type,p_ref);

/*    printf("inport=%s\n", portSetKeyEncode(inportStmt)); */
      if (! p_ref ) {
	  fprintf(fp, "\
    nexus_get_stashed_%s(buffer, &dest.A_%d, 1);\n",
		  getTypeCString(p_type->variant()),
		  _p_j);
	  if (opt_debug)
	      fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"fm_forward(): got A_%d = %s\\n\", dest.A_%d);\n",
			    _p_j,
			    getTypePrintfFormat(p_type->variant()),
			    _p_j);
      } else if (p_type->variant() == T_INPORT) {
	  fprintf(fp, "\
    _fm_get_stashed_%s(buffer, &dest.A_%d, 1, FM_%s_INIT);\n",
		  getTypeCString(p_type->variant()),
		  _p_j,
		  str2upper(ipt2opt(popTypeEncode(p_type, p_ref))));
	  fprintf(fp, "\
    ipls[%d].entries = (fm_inport_t **) &dest.A_%d;\n\
    ipls[%d].n_entries = 1;\n\n",
		  ip_count,
		  _p_j,
		  ip_count);

	  ip_count++;
	  
	  if (opt_debug)
	      fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"fm_forward(): got A_%d = %s\\n\", dest.A_%d);\n",
			    _p_j,
			    getTypePrintfFormat(p_type->variant()),
			    _p_j);
      } else if (p_type->variant() == T_OUTPORT) {
	  fprintf(fp, "\
    _fm_get_stashed_%s(buffer, &dest.A_%d, 1, FM_%s_INIT);\n",
		  getTypeCString(p_type->variant()),
		  _p_j,
		  str2upper(ipt2opt(popTypeEncode(p_type, p_ref))));


	  fprintf(fp, "\
    opls[%d].entries = (fm_outport_t **) &dest.A_%d;\n\
    opls[%d].n_entries = 1;\n\n",
		  op_count,
		  _p_j,
		  op_count);

	  op_count++;
	  
	  if (opt_debug)
	      fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"fm_forward(): got A_%d = %s\\n\", dest.A_%d);\n",
			    _p_j,
			    getTypePrintfFormat(p_type->variant()),
			    _p_j);
      } else if (p_ref->variant() == VAR_REF) {
	  fprintf(fp, "\
    nexus_get_stashed_%s(buffer, &dest.A_%d, 1);\n",
		  getTypeCString(p_type->variant()),
		  _p_j);
	  if (opt_debug)
	      fprintf(fp, "\
    if (FMDebug(UserSR, 2))\n\
        nexus_printf(\"fm_forward(): got A_%d = %s\\n\", dest.A_%d);\n",
			    _p_j,
			    getTypePrintfFormat(p_type->variant()),
			    _p_j);
      } else if (p_ref->variant() == ARRAY_REF) {
	  SgArrayRefExp  *a;
	  SgExpression   *a_size;

	  a               = isSgArrayRefExp(p_ref);
	  a_size          = &(arraySizeExp(a));

	  if (IsInteger(a_size)) {
	      fprintf(fp, "\
    nexus_get_stashed_%s(buffer, dest.A_%d, %d);\n",
		      getTypeCString(p_type->variant()),
		      _p_j, 
		      integerValue(a_size));
	  } else {
	      fprintf(fp, "\
    nexus_get_stashed_int(buffer, &A_%d_len, 1);\n",
		      _p_j);

	      fprintf(fp, "\
    NexusMalloc(fm_%s_forward, dest.A_%d, f_%s_t *,\n\
                sizeof(f_%s_t) * A_%d_len);\n",
		      key->chars(), 
		      _p_j, 
		      getTypeString(p_type->variant()),
		      getTypeString(p_type->variant()),
		      _p_j);

	      if (opt_debug)
		fprintf(fp, "\
    if (FMDebug(UserSR, 1))\n\
        nexus_printf(\"fm_forward(): got A_%d_len = %%d\\n\", A_%d_len);\n",
			_p_j, _p_j);
	      
	      fprintf(fp, "\
    nexus_get_stashed_%s(buffer, dest.A_%d, A_%d_len);\n",
		      getTypeCString(p_type->variant()),
		      _p_j, _p_j);
	  }
      } else {
	  fm_internal_error("\
Do not expect a reference in a port type to be: %s",
			    getVariantByName(p_ref->variant()));
      }
    END_FOR_EACH_PORT_TYPE();


    if (n_inports || n_outports) {
	fprintf(fp, "\
    _fm_synch_init(&synch);\n\n");

	fprintf(fp, "\
    _fm_create_ports(%d, %s, %d,\n\
                     %d, %s, %d,\n\
                     dest_gp,\n\
                     &synch);\n\n",
		n_inports,
		(n_inports)? "ipls" : "(fm_ipl_t *) NULL",
		n_inports,
		n_outports,
		(n_outports)? "opls" : "(fm_opl_t *) NULL",
		n_outports);

	fprintf(fp, "\
    _fm_synch_wait(&synch);\n\n\
    _fm_notify_ports(%d, (fm_ipl_t *) %s, %d,\n\
                     %d, (fm_opl_t *) %s, %d);\n\n",
		n_inports,
		(n_inports)? "ipls" : "NULL",
		n_inports,
		n_outports,
		(n_outports)? "opls" : "NULL",
		n_outports);
    }

    fprintf(fp, "\
    nexus_init_remote_service_request(&mbuf,\n\
                                      dest_gp,\n\
                                      FM_CHANNEL_FORWARD_BUFFER_NAME,\n\
                                      FM_CHANNEL_FORWARD_BUFFER_HASH);\n\n");

    fprintf(fp, "\
    size = n_elts = 0;\n");

    for (i = 0; i < inportStmt->numberOfPortTypes(); i++) {
	SgPortTypeExp     *p_type_exp;
	SgExpression      *p_type_ref;
	SgType            *t;
	
	p_type_exp        = inportStmt->porttype(i);
	p_type_ref        = p_type_exp->ref();
	t                 = p_type_exp->type();

	if (! t) {
	    fm_internal_error("\
Do not expect a null-type for a port type %s:%d", __FILE__, __LINE__);
	}

	if (! p_type_ref) {
	    fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, 1);\n\
    n_elts++;\n",
		    getTypeCString(t->variant()));
	} else if (t->variant() == T_INPORT || t->variant() == T_OUTPORT) {
	    fprintf(fp, "\
    size += _fm_sizeof_%s(&mbuf, &dest.A_%d, 1, &obj_n_elts);\n\
    n_elts += obj_n_elts;\n\n",
		    getTypeCString(t->variant()),
		    i);
	} else {
	    if (p_type_ref->variant() == VAR_REF) {
		fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, 1);\n\
    n_elts++;\n",
		    getTypeCString(t->variant()));
	    } else if (p_type_ref->variant() == ARRAY_REF) {
		SgArrayRefExp  *a;
		SgExpression   *a_size;
		
		a               = isSgArrayRefExp(p_type_ref);
		a_size          = &(arraySizeExp(a));
		
		if (IsInteger(a_size)) {
		    fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, %d);\n\
    n_elts += 1;\n",
			    getTypeCString(t->variant()),
			    integerValue(a_size));
		} else {
		    fprintf(fp, "\
    size += nexus_sizeof_int(&mbuf, 1);\n\
    n_elts++;\n\n\
    size += nexus_sizeof_%s(&mbuf, (int) A_%d_len);\n\
    n_elts += (int) A_%d_len;\n", 
			    getTypeCString(t->variant()),
			    i, i);
		}
	    } else {
		fm_internal_error("\
Do not expect port type to be %s at %s:%d",
				  getVariantByName(p_type_ref->variant()),
				  __FILE__, __LINE__);
	    }
	}
    }
    
    fprintf(fp, "\
    nexus_set_buffer_size(&mbuf, size, n_elts);\n\n");

    for (i = 0; i < inportStmt->numberOfPortTypes(); i++) {
	SgPortTypeExp     *p_type_exp;
	SgExpression      *p_type_ref;
	SgType            *t;
	
	p_type_exp        = inportStmt->porttype(i);
	p_type_ref        = p_type_exp->ref();
	t                 = p_type_exp->type();

	if (! t) {
	    fm_internal_error("\
Do not expect a null-type for a port type %s:%d", __FILE__, __LINE__);
	}

	if (! p_type_ref || t->variant() == T_INPORT || t->variant() == T_OUTPORT) {
	    fprintf(fp, "\
    nexus_put_%s(&mbuf, &dest.A_%d, 1);\n", 
		    getTypeCString(t->variant()),
		    i);
	} else if (p_type_ref->variant() == VAR_REF) {
	    fprintf(fp, "\
    nexus_put_%s(&mbuf, &dest.A_%d, 1);\n", 
		    getTypeCString(t->variant()),
		    i);
	} else if (p_type_ref->variant() == ARRAY_REF) {
	    SgArrayRefExp *a;
	    SgExpression  *a_size;
	    a             = isSgArrayRefExp(p_type_ref);
	    a_size        = &(arraySizeExp(a));
	    
	    if ( IsInteger(a_size)) {
		fprintf(fp, "\
    nexus_put_%s(&mbuf, dest.A_%d, %d);\n",
			getTypeCString(t->variant()),
			i, 
			integerValue(a_size));
	    } else {
		fprintf(fp, "\
    nexus_put_int(&mbuf, &A_%d_len, 1);\n\
    nexus_put_%s(&mbuf, dest.A_%d, (int) A_%d_len);\n",
			i,
			getTypeCString(t->variant()),
			i, i);
	    }
	} else {
	    fm_internal_error("\
Do not expect port type to be %s at %s:%d",
			      getVariantByName(p_type_ref->variant()),
			      __FILE__, __LINE__);
	}
    }

    fprintf(fp, "\
    nexus_send_remote_service_request(&mbuf);\n");

    if (n_inports || n_outports)
        fprintf(fp, "\
    _fm_synch_destroy(&synch);\n");

    fprintf(fp, "}\n\n");

    END_FOR_EACH_INPORT_DECL();
}

void gen_porttype_receive(FILE *fp, SgInportStmt *s, char *key)
{
    int         i;
    char        *prosName;
    SgStatement *proc;

    assert(s);
    proc      = whichProcAmI(s);
    
    fprintf(fp, "\
void FM_%s_RCV%s(f_integer_t *f_inport", 
	    str2upper(key), (opt_debug)? "_D" : "");
                
    for (i = 0; i < s->numberOfPortTypes(); i++) {
	SgPortTypeExp *p_type = s->porttype(i);
	SgExpression  *p_type_ref = p_type->ref();
	char *p_type_string = getTypeString((p_type->type())->variant());

	fprintf(fp,",\n\
                            f_%s_t *arg_%d", p_type_string, i);
	if (p_type_ref) {
	    if (p_type_ref->variant() == ARRAY_REF) {
		fprintf(fp, ",\n\
                            f_integer_t *arg_%d_size",
			i);
	    }
	}
    }
    fprintf(fp, ",\n\
                            f_integer_t *f_ioval");
    if (opt_debug) {
	fprintf(fp, ",\n\
                            f_character_t f_file_name,\n\
                            f_integer_t *f_file_line\n\
                            CHAR_LEN_DECL(f_file_name_len))\n");
    } else {
	fprintf(fp, ")\n");
    }

    fprintf(fp, "{\n");

    if (opt_debug) {
	fprintf(fp, "\
    char file_name[1024];\n\
    LOCAL_CHAR_LEN_DECL(f_file_name_len)\n");
    }

    fprintf(fp, "\
    %s_dest_t dest;\n\n",
	    key);

    if (opt_debug) {
	fprintf(fp, "\
    CONVERT_CHAR_FTN_TO_C_INPLACE(f_file_name, f_file_name_len, file_name);\n\n");
    }

    for (i = 0; i < s->numberOfPortTypes(); i++) {
	SgPortTypeExp *p_type = s->porttype(i);
	SgExpression  *p_type_ref = p_type->ref();
	fprintf(fp, "\
    dest.A_%d = arg_%d;\n", i, i);

	if (p_type_ref) {
	    if (p_type_ref->variant() == ARRAY_REF) {
		fprintf(fp, "\
    dest.A_%d_size = *arg_%d_size;\n",
			i, i);
	    }
	}
    }

    if (opt_debug) {
	fprintf(fp, "\
    dest.f_file_name = file_name;\n\
    dest.f_file_line = *f_file_line;\n\n");
    }

    if (opt_debug) {
	fprintf(fp, "\
    _fm_receive_debug(f_inport,(void *)&dest,f_ioval,file_name,*f_file_line);\n");
    } else {
	fprintf(fp, "\
    _fm_receive(f_inport,(void *)&dest,f_ioval);\n\n");
    }

    fprintf(fp, "}\n\n");
}

void gen_c_aux_file(char *f_name) 
{
    Pix ind;
    FILE *fp = fopen(f_name, "w");

    this_module_has_handler_table = 0;
    if (! fp) {
	fprintf(stderr, "Cannot open %s for writing: %s\n", 
		f_name, strerror(errno));
	exit(errno);
    }

    //
    // Pairing up ports to avoid the 
    // linking problem which often
    // occurs in separate compilation.
    //
    pairing_ports();

    gen_preamble(fp);

    gen_fm_args_t(fp);
    gen_fm_return_args_t(fp);

    gen_handler_defs(fp);
    gen_compiler_flavours(fp);
    gen_outport_init_func_proto(fp);
    gen_process_func_proto(fp);
    gen_module_init(fp);
    gen_fm_should_share(fp);
    gen_fm_return_copy_func(fp);
    gen_fm_return_stashed_copy_func(fp);
    gen_fm_invoke(fp);
    gen_invoke_thread(fp);
    gen_invoke_handler(fp);
    gen_destination_struct(fp);
    gen_fm_receive_copy(fp);
    gen_fm_receive_stashed_copy(fp);
    gen_fm_receive_forward(fp);
    gen_fm_port_init_func(fp);
    gen_fm_outport_init(fp);
    // 
    // do send/receive functions
    //
    for (ind = portSet->first(); ind; portSet->next(ind)) {
	String &k_string = portSet->key(ind);
	char *k = (char *)k_string.chars();
	SgStatement *s = &(portSet->contents(ind));
	
	if (s->variant() == OUTPORT_DECL) {
	    SgOutportStmt *outport_s = isSgOutportStmt(s);
	    
	    gen_porttype_send(fp, outport_s, k);
	} else {
	    SgInportStmt *inport_s = isSgInportStmt(s);
	    gen_porttype_receive(fp, inport_s, k);
	}
    }
    //
    // change each process into a subroutine
    //
    FOR_EACH_PROCESS(prosStmt, prosName);

    prosStmt->setVariant(PROC_HEDR);
    
    END_FOR_EACH_PROCESS();
}

void stmtMapDump(StringSgStatementAVLMap *map) 
{
    Pix ind;
    
    for (ind = map->first(); ind; map->next(ind)) {
	SgStatement *s = &(map->contents(ind));
	String &k_string = portSet->key(ind);
	char *k = (char *)k_string.chars();
	
	printf("%s --- ",k);
	s->unparsestdout();
    }
}

SgExpression *findSymbInDeclStmt(SgSymbol *symb)
{
    int          i;
    SgStatement *s;

    if (! symb) {
	fm_internal_error("\
Expect an non-null symbol in findSymbInDeclStmt() at %s:%d", 
			  __FILE__, __LINE__);
    }

    s = stmtMapGet(prosMap, prosMapKeyEncode(currentProcedureName,symb->identifier()));
    
    if (! s) {
	return (SgExpression *)0;
    }

    switch (s->variant()) {
      case VAR_DECL:
	SgVarDeclStmt *varDeclStmt;

	varDeclStmt = isSgVarDeclStmt(s);
	for (i = 0; i < varDeclStmt->numberOfVars(); i++) {
	    SgExpression *var_i;
	    char         *var_i_name;

	    var_i        = varDeclStmt->var(i);
	    var_i_name   = symbName(var_i);
	    
	    if (! strcmp(var_i_name, symb->identifier())) {
		return var_i;
	    }
	}
	//
	// Impossible: must contain the SYMB
	//
	fm_internal_error("\
Must contain the Symb %s,  check doProsHedrStmt() !", symb->identifier());
    
	break;
      case INPORT_DECL:
	SgInportStmt    *inportStmt;
	
	inportStmt      = isSgInportStmt(s);
	for (i = 0; i < inportStmt->numberOfNames(); i++) {
	    SgExpression *var_i;
	    char         *var_i_name;

	    var_i        = inportStmt->name(i);
	    var_i_name   = symbName(var_i);
	    
	    if (! strcmp(var_i_name, symb->identifier())) {
		return var_i;
	    }
	}
	//
	// Impossible: must contain the SYMB
	//
	fm_internal_error("\
Must contain the Symb %s,  check doProsHedrStmt() !", symb->identifier());
    
	break;
      case OUTPORT_DECL:
	SgOutportStmt    *outportStmt;
	
	outportStmt      = isSgOutportStmt(s);
	for (i = 0; i < outportStmt->numberOfNames(); i++) {
	    SgExpression *var_i;
	    char         *var_i_name;

	    var_i        = outportStmt->name(i);
	    var_i_name   = symbName(var_i);
	    
	    if (! strcmp(var_i_name, symb->identifier())) {
		return var_i;
	    }
	}
	//
	// Impossible: must contain the SYMB
	//
	fm_internal_error("\
Must contain the Symb %s,  check doProsHedrStmt() !", symb->identifier());
    
	return (SgExpression *)0;
	break;
      default:
	fm_internal_error("\
Do not expect statement type %s at %s:%d", 
			  getVariantByName(s->variant()), 
			  __FILE__, __LINE__);

	break;
    }
    return (SgExpression *)0;
}

SgType *getProsArgType(char *prosName, SgSymbol *symb)
{
    SgStatement        *stmt;
    char               *name;

    name  = symb->identifier();
    stmt  = stmtMapGet(prosMap, prosMapKeyEncode(prosName,name));

    if (! stmt ) {
	return (SgType *)0;
    }

    switch (stmt->variant()) {
      case VAR_DECL:
	SgVarDeclStmt         *varDeclStmt;

	varDeclStmt           = isSgVarDeclStmt(stmt);
	return (varDeclStmt->type());
	break;
      case INPORT_DECL:
      case OUTPORT_DECL:
	return ( integer_t );
	break;
      default:
	break;
    }
    return (SgType *)0;
}

char *str2upper(char *string)
{
    char       *theString;
    char       *p;

    theString  = strdup(string);
    for (p = theString; *p; p++) {
	if (islower(*p)) {
	    *p = toupper(*p);
	}
    }
    return theString;
}

char *str2lower(char *string)
{
    char       *theString;
    char       *p;

    theString  = strdup(string);
    for (p = theString; *p; p++) {
	if (isupper(*p)) {
	    *p = tolower(*p);
	}
    }
    return theString;
}

	
SgExpression *getFormalArgInProsHedrMap(SgExpression *e)
{
    char            *name;

    if (! e->symbol()) {
	return (SgExpression *)0;
    } 

    name = e->symbol()->identifier();
    return(exprMapGet(prosHedrMap, 
		      prosHedrMapKeyEncode(currentProcedureName,name)));
}

void gen_compute_IN_buffer_size(FILE *fp, SgProsHedrStmt *s)
{
    char       *prosName;
    int        i;
    int        should_pack_into_buffer;

    prosName   = s->symbol()->identifier();
    for (i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol     *symb;
	SgExpression *intent;
	SgType       *t;
	SgExpression *var_declared;

	symb         = s->parameter(i);
	intent       = exprMapGet(intentMap, symb->identifier());
	t            = getProsArgType(prosName, symb);
	
	if (! intent) {
	    should_pack_into_buffer = 1;
	} else if (intent->variant() != OUT_OP) {
	    should_pack_into_buffer = 1;
	} else {
	    should_pack_into_buffer = 0;
	}

	if (! t) { continue; }

	if (should_pack_into_buffer) {
	    int         port_variant;
	    
	    // get the size of the argument
	    var_declared = findSymbInDeclStmt(symb);
	    // check whether the argument is port
	    port_variant = getPortVariant(var_declared);

	    if (var_declared->variant() == VAR_REF) {
		if (port_variant >= 0) {
		    fprintf(fp, "\
    size += _fm_sizeof_%sport(&mbuf, A_%d, 1, &obj_n_elts);\n",
			    (port_variant == INPORT_DECL)? "in" : "out",
			    i);
		    fprintf(fp, "\
    n_elts += obj_n_elts;\n\n");
		} else {
		    fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, 1);\n", 
			    getTypeCString(t->variant()));
		    fprintf(fp, "\
    n_elts++;\n");
		}
	    } else if (var_declared->variant() == ARRAY_REF) {
		SgArrayRefExp  *a;
		SgExpression   *a_size;

		a              = isSgArrayRefExp(var_declared);
		a_size         = &(arraySizeExp(a));
		if (IsInteger(a_size)) {
		    if (port_variant >= 0) {
			fprintf(fp, "\
    size += _fm_sizeof_%sport(&mbuf, A_%d,%d,&obj_n_elts);\n",
				(port_variant == INPORT_DECL)? "in" : "out",
				i,
				integerValue(a_size));
			fprintf(fp, "\
    n_elts += obj_n_elts;\n");
		    } else {
			fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, %d);\n",
				getTypeCString(t->variant()),
				integerValue(a_size));
			fprintf(fp, "\
    n_elts += %d;\n",
				integerValue(a_size));
		    }
		} else {
		    if (port_variant >= 0) {
			fprintf(fp, "\
    size += nexus_sizeof_int(&mbuf, 1);\n");
			fprintf(fp, "\
    n_elts++;\n\n");
		    fprintf(fp, "\
    size += _fm_sizeof_%sport(&mbuf, A_%d,A_%d_len,&obj_n_elts);\n", 
			    (port_variant == INPORT_DECL)? "in" : "out",
			    i,i);
		    
		    fprintf(fp, "\
    n_elts += obj_n_elts;\n\n", 
			    i);
		    } else {
			fprintf(fp, "\
    size += nexus_sizeof_int(&mbuf, 1);\n");
			fprintf(fp, "\
    n_elts++;\n\n"); 
			fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, A_%d_len);\n", 
				getTypeCString(t->variant()),
				i);
			
			fprintf(fp, "\
    n_elts += A_%d_len;\n\n", 
				i);
		    }
		}
	    } else {
		fm_internal_error("\
Do not expect argument type %s for process %s (argument %d)",
				  getVariantByName(var_declared->variant()),
				  prosName, i+1);
	    }
	}
    }
    fprintf(fp, "\
    nexus_set_buffer_size(&mbuf, size, n_elts);\n");
}

void gen_compute_OUT_buffer_size(FILE *fp, SgProsHedrStmt *s)
{
    char       *prosName;
    int        i;
    int        should_pack_into_buffer;

    prosName   = s->symbol()->identifier();
    for (i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol     *symb;
	SgExpression *intent;
	SgType       *t;
	SgExpression *var_declared;

	symb         = s->parameter(i);
	intent       = exprMapGet(intentMap, symb->identifier());
	t            = getProsArgType(prosName, symb);
	
	if (! intent) {
	    should_pack_into_buffer = 1;
	} else if (intent->variant() != IN_OP) {
	    should_pack_into_buffer = 1;
	} else {
	    should_pack_into_buffer = 0;
	}

	if (! t) { continue; }

	if (should_pack_into_buffer) {
	    // get the size of the argument
	    var_declared = findSymbInDeclStmt(symb);
	    if (var_declared->variant() == VAR_REF) {
		fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, 1);\n", 
			getTypeCString(t->variant()));
		fprintf(fp, "\
    n_elts++;\n\n");
	    } else if (var_declared->variant() == ARRAY_REF) {
		SgArrayRefExp *a;
		SgExpression  *a_size;
		a             = isSgArrayRefExp(var_declared);
		a_size        = &(arraySizeExp(a));
		
		if (! IsInteger(a_size)) {
		    fprintf(fp, "\
    size += nexus_sizeof_int(&mbuf, 1);\n");
		    fprintf(fp, "\
    n_elts++;\n\n");
		    fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, args->A_%d_len);\n", 
			    getTypeCString(t->variant()),
			    i);
		    fprintf(fp, "\
    n_elts += 1;\n\n",   /* args->A_%d_len;\n\n", */
			    i);
		} else {
		    fprintf(fp, "\
    size += nexus_sizeof_%s(&mbuf, %d);\n", 
			    getTypeCString(t->variant()),
			    integerValue(a_size));
		    fprintf(fp, "\
    n_elts += 1;\n");
//			    integerValue(a_size));
		} 
	    } else {
		    fm_internal_error("\
Do not expect argument type %s for process %s (argument %d)",
				  getVariantByName(var_declared->variant()),
				  prosName, i+1);
	    }
	}
    }
    fprintf(fp, "\
    nexus_set_buffer_size(&mbuf, size, n_elts);\n");
}

void gen_pack_IN_buffer(FILE *fp, SgProsHedrStmt *s)
{
    char        *prosName;
    int         should_pack_into_buffer;

    prosName    = s->symbol()->identifier();
    
    for (int i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol     *symb;
	SgExpression *intent;
	SgType       *t;
	SgExpression *var_declared;

	symb         = s->parameter(i);
	intent       = exprMapGet(intentMap, symb->identifier());
	t            = getProsArgType(prosName, symb);
	
	if (! intent) {
	    should_pack_into_buffer = 1;
	} else if (intent->variant() != OUT_OP) {
	    should_pack_into_buffer = 1;
	} else {
	    should_pack_into_buffer = 0;
	}

	if (! t) { continue; }

	if (should_pack_into_buffer) {
	    int        port_variant;

	    // get the size of the argument
	    var_declared = findSymbInDeclStmt(symb);
	    
	    port_variant = getPortVariant(var_declared);
	    //
	    // if var_declared is a port variable
	    //
	    if (var_declared->variant() == VAR_REF) {
		if (port_variant >= 0) {
		    fprintf(fp, "\
    _fm_put_%sport(&mbuf, A_%d, 1);\n",
			    (port_variant == INPORT_DECL)? "in" : "out",
			    i);
		} else {
		    fprintf(fp, "\
    nexus_put_%s(&mbuf, A_%d, 1);\n", 
			    getTypeCString(t->variant()),
			    i);
		}
	    } else if (var_declared->variant() == ARRAY_REF) {
		SgArrayRefExp *a;
		SgExpression  *a_size;
		a             = isSgArrayRefExp(var_declared);
		a_size        = &(arraySizeExp(a));
		
		if (! IsInteger(a_size)) {
		    if (port_variant >= 0) {
			fprintf(fp, "\
    nexus_put_int(&mbuf, &A_%d_len, 1);\n", i);
			fprintf(fp, "\
    _fm_put_%sport(&mbuf, A_%d, A_%d_len);\n",
				(port_variant == INPORT_DECL)? "in" : "out",
				i, i);
		    } else {
			fprintf(fp, "\
    nexus_put_int(&mbuf, &A_%d_len, 1);\n", i);
			fprintf(fp, "\
    nexus_put_%s(&mbuf, A_%d, A_%d_len);\n",
				getTypeCString(t->variant()),
				i, i);
		    }
		} else {
		    if (port_variant >= 0) {
			fprintf(fp, "\
    _fm_put_%sport(&mbuf, A_%d, %d);\n",
				(port_variant == INPORT_DECL)? "in" : "out",
				i, integerValue(a_size));
		    } else {
			fprintf(fp, "\
    nexus_put_%s(&mbuf, A_%d, %d);\n", 
				getTypeCString(t->variant()),
				i, integerValue(a_size));
		    }
		}
	    } else {
		fm_internal_error("\
Do not expect argument type %s for process %s (argument %d)",
				  getVariantByName(var_declared->variant()),
				  prosName, i+1);
	    }
	}
    }
}

void gen_pack_OUT_buffer(FILE *fp, SgProsHedrStmt *s)
{
    char        *prosName;
    int         should_pack_into_buffer;

    prosName    = s->symbol()->identifier();
    
    for (int i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol     *symb;
	SgExpression *intent;
	SgType       *t;
	SgExpression *var_declared;

	symb         = s->parameter(i);
	intent       = exprMapGet(intentMap, symb->identifier());
	t            = getProsArgType(prosName, symb);
	
	if (! intent) {
	    should_pack_into_buffer = 1;
	} else if (intent->variant() != IN_OP) {
	    should_pack_into_buffer = 1;
	} else {
	    should_pack_into_buffer = 0;
	}

	if (! t) { continue; }

	if (should_pack_into_buffer) {
	    // get the size of the argument
	    var_declared = findSymbInDeclStmt(symb);
	    
	    if (var_declared->variant() == VAR_REF) {
		fprintf(fp, "\
    nexus_put_%s(&mbuf, &args->A_%d, 1);\n", 
			getTypeCString(t->variant()),
			i);
	    } else if (var_declared->variant() == ARRAY_REF) {
		SgArrayRefExp *a;
		SgExpression  *a_size;
		a             = isSgArrayRefExp(var_declared);
		a_size        = &(arraySizeExp(a));
		
		if (! IsInteger(a_size)) {
		    fprintf(fp, "\
    nexus_put_int(&mbuf, &args->A_%d_len, 1);\n", i);
		    fprintf(fp, "\
    nexus_put_%s(&mbuf, args->A_%d, args->A_%d_len);\n",
			    getTypeCString(t->variant()),
			    i, i);
		} else {
		    fprintf(fp, "\
    nexus_put_%s(&mbuf, args->A_%d, %d);\n", 
			    getTypeCString(t->variant()),
			    i, integerValue(a_size));
		}
	    } else {
		fm_internal_error("\
Do not expect argument type %s for process %s (argument %d)",
				  getVariantByName(var_declared->variant()),
				  prosName, i+1);
	    }
	}
    }
}

void gen_unpack_handler_buffer(FILE *fp, SgProsHedrStmt *s)
{
    char       *prosName;
    int        i;
    int        should_pack_into_buffer;

    prosName   = s->symbol()->identifier();
    for (i = 0; i < s->numberOfParameters(); i++) {
	SgSymbol     *symb;
	SgExpression *intent;
	SgType       *t;
	SgExpression *var_declared;

	symb         = s->parameter(i);
	intent       = exprMapGet(intentMap, symb->identifier());
	t            = getProsArgType(prosName, symb);
	
	if (! intent) {
	    should_pack_into_buffer = 1;
	} else if (intent->variant() != OUT_OP) {
	    should_pack_into_buffer = 1;
	} else {
	    should_pack_into_buffer = 0;
	}

	if (! t) { continue; }

	if (should_pack_into_buffer) {
	    int          port_variant;

	    // get the size of the argument
	    var_declared = findSymbInDeclStmt(symb);

	    port_variant = getPortVariant(var_declared);
	    if (var_declared->variant() == VAR_REF) {
		if (port_variant >= 0) {
		    if (port_variant == INPORT_DECL) {
			char *key = getPortSetKeyFromPortRef(var_declared);
			
			fm_assert(key);
			fprintf(fp, "\
    _fm_get_inport(buf, &args->A_%d, 1, FM_%s_INIT);\n",
				i, 
				str2upper(ipt2opt(key)));
		    } else {
			fprintf(fp, "\
    _fm_get_outport(buf, &args->A_%d, 1);\n",
				i);
		    }
		} else {
		    fprintf(fp, "\
    nexus_get_%s(buf, &args->A_%d, 1);\n", 
			    getTypeCString(t->variant()),
			    i);
		}		
	    } else if (var_declared->variant() == ARRAY_REF) {
		SgArrayRefExp *a;
		SgExpression  *a_size;
		a             = isSgArrayRefExp(var_declared);
		a_size        = &(arraySizeExp(a));
		
		if (! IsInteger(a_size)) {
		    if (port_variant >= 0) {
			if (port_variant == INPORT_DECL) {
			    char *key = getPortSetKeyFromPortRef(var_declared);
			    
			    fm_assert(key);
			    fprintf(fp, "\
    nexus_get_int(buf, &args->A_%d_len, 1);\n",
				    i);
				    
			    fprintf(fp, "\
    FM_ALLOC_ARG_SPACE(%s, f_integer_t, A_%d);\n",
				    prosName, 
				    i);
			    
			    fprintf(fp, "\
    _fm_get_inport(buf, args->A_%d, args->A_%d_len, FM_%s_INIT);\n",
				    i, i, str2upper(ipt2opt(key)));
			} else {
			    fprintf(fp, "\
    nexus_get_int(buf, &args->A_%d_len, 1);\n",
				    i);
			    fprintf(fp, "\
    FM_ALLOC_ARG_SPACE(%s, f_integer_t, A_%d);\n",
				    prosName, 
				    i);
			    
			    fprintf(fp, "\
    _fm_get_outport(buf, args->A_%d, args->A_%d_len);\n",
				    i, i);
			}
		    } else {
			fprintf(fp, "\
    nexus_get_int(buf, &args->A_%d_len, 1);\n", i);
			
			fprintf(fp, "\
    FM_ALLOC_ARG_SPACE(%s, f_%s_t, A_%d);\n",
				prosName, 
				getTypeString(t->variant()),
				i);
			
			fprintf(fp, "\
    nexus_get_%s(buf, args->A_%d, args->A_%d_len);\n",
				getTypeCString(t->variant()),
				i, i);
		    }
		} else {
		    if (port_variant >= 0) {
			if (port_variant == INPORT_DECL) {
			    char *key = getPortSetKeyFromPortRef(var_declared);
			    
			    fm_assert(key);
			    fprintf(fp, "\
    _fm_get_inport(buf, args->A_%d, %d, FM_%s_INIT);\n",
				    i, integerValue(a_size),
				    str2upper(ipt2opt(key)));
			} else {
			    fprintf(fp,"\
    _fm_get_outport(buf, args->A_%d, %d);\n",
				    i, integerValue(a_size));
			}
		    } else {
			fprintf(fp, "\
    nexus_get_%s(buf, args->A_%d, %d);\n", 
				getTypeCString(t->variant()),
				i, integerValue(a_size));
		    }
		}
	    } else {
		fm_internal_error("\
Do not expect argument type %s for process %s (argument %d)",
				  getVariantByName(var_declared->variant()),
				  prosName, i+1);
	    }
	}
    }
}

void gen_c_link_file(char *f_name)
{
    FILE *fp;

    fp = fopen(f_name, "w");
    if (! fp) {
	fprintf(stderr, "Cannot open %s for writing: %s\n", 
		f_name, strerror(errno));
	exit(errno);
    }
    FOR_EACH_PROGRAM(progStmt,progName)
    {
	fprintf(fp, "\
void _fm_linktime_init()\n{\n\
    _fm_module_init_%s();\n}",
		progName);
    }
    END_FOR_EACH_PROGRAM();
}


#define NEXUS_HANDLER_HASH_TABLE_SIZE  1021

int handler_id(char *name)
{
        int  hash;
	char *s;

	hash = 0;
	for (s = name; *s != '\0'; s++)
	    hash += (int) *s;
	hash = hash % NEXUS_HANDLER_HASH_TABLE_SIZE;
	
	return hash;
}	

void pairing_ports()
{
    Pix ind;


    FOR_EACH_INPORT_DECL(inport,key)
    {
	char *outport_key = ipt2opt(key->chars());

	if (! stmtMapGet(portSet, outport_key)) {
	    SgOutportStmt *outport;

	    outport = new SgOutportStmt(*inport->names(),*inport->porttypes());
	    stmtMapSet(portSet, outport_key, outport);
	}
    }
    END_FOR_EACH_INPORT_DECL();

    FOR_EACH_OUTPORT_DECL(outport,key)
    {
 	char *inport_key = opt2ipt(key->chars());

	if (! stmtMapGet(portSet, inport_key)) {
	    SgInportStmt *inport;

	    inport = new SgInportStmt(*outport->names(),*outport->porttypes());
	    stmtMapSet(portSet, inport_key, inport);
	}
    }
    END_FOR_EACH_OUTPORT_DECL();
//    printf("Port set dump\n\n");
//    stmtMapDump(portSet);
}    

int inport_type_has_ports(SgInportStmt *s)
{
    FOR_EACH_PORT_TYPE(_p_i,s,p_type,p_ref)
    {
	if (p_type->variant() == T_INPORT ||
	    p_type->variant() == T_OUTPORT)
	{
	    return (1);
	}
    }
    END_FOR_EACH_PORT_TYPE();

    return (0);
}

int outport_type_has_ports(SgOutportStmt *s)
{
    FOR_EACH_PORT_TYPE(_p_i,s,p_type,p_ref)
    {
	if (p_type->variant() == T_INPORT ||
	    p_type->variant() == T_OUTPORT)
	{
	    return (1);
	}
    }
    END_FOR_EACH_PORT_TYPE();

    return (0);
}





void all_processes_must_be_threadable()
{
    FOR_EACH_PROCESS(process,process_name)
    {
	if (! stmtMapGet(prosMap, directiveMapKeyEncode(process_name,"threadableProcess")))
	{
	    fm_error(process, "\
This version of Fortran M requires the use of the CFM$ThreadableProcess directive in all processes but process '%s' does not have it",
		     process_name);
	}
    }
    END_FOR_EACH_PROCESS()
}

	
