// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

/*
 * Header files
 *
 */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include "sage++user.h"

/*
 * FM compiler internal 
 * data structure
 *
 */
#include "SLL/SgStmtPtr.SLList.h"
#include "Maps/statementAVLMap.h"
#include "Maps/expressionAVLMap.h"
#include "Maps/fmProgBlkAVLMap.h"
#include "fmProgBlkdefs.h"

#ifdef GLOBAL_VARIABLES_ALLOCATED_HERE
#define EXTERNAL 
#else 
#define EXTERNAL extern
#endif /* GLOBAL_VARIABLES_ALLOCATED_HERE */

/*
 * Command Line Options
 */
EXTERNAL int                              opt_verbose;
EXTERNAL int                              opt_debug;
EXTERNAL int                              opt_use_auto_vars;
EXTERNAL int                              opt_use_private_common;
EXTERNAL int                              opt_allow_common;
EXTERNAL int                              opt_must_thread;
EXTERNAL int                              opt_i8;

/*
 * Current File that is being processed
 */
EXTERNAL char                            *currentFile;
EXTERNAL char                            *currentFileBase;
EXTERNAL SgValueExp                      *currentFileNameExp;

/*
 * Point to the current PROG_STAT, PROC_STAT, FUNC_STAT
 * and PROS_STAT 
 */
EXTERNAL SgStatement                     *currentProcedure;
EXTERNAL char                            *currentProcedureName;
EXTERNAL SgStatement                     *currentStatement;
EXTERNAL SgValueExp                      *currentLineNo;
EXTERNAL SgStatement                     *lastDeclStmt;
//
// firstExeStmt marks the point where
// fm & virtual computer initialization routines
// may be inserted
//
EXTERNAL SgStatement                     *firstExeStmt;
//
// firstUserStmt marks the point
// where port initialization calls
// may be inserted
//
EXTERNAL SgStatement                     *firstUserStmt;
/*
 * A Single-linked list that stores all the 
 * Fortran-M statments in a Project
 */
EXTERNAL SgStmtPtrSLList                 fmStmtSLL;
EXTERNAL SgStmtPtrSLList                 lbContStmtSLL;
EXTERNAL SgStmtPtrSLList                 LS_stmts;

/*
 * FM program block table Map
 */
EXTERNAL FMProgBlock                     *currentProgBlock;
EXTERNAL StringFMProgBlockAVLMap         *fmProgBlockMap;

/*
 * FM blocks generated via mpcall
 */

EXTERNAL StringSgStatementAVLMap         *mpBlock;

/*
 * portSet stores a non-duplicate set of port type information
 */
EXTERNAL StringSgStatementAVLMap         *portSet;

/*
 * portMangleMap stores the mapping between the encoded
 * port send/receive function name
 */
EXTERNAL StringSgExpressionAVLMap        *portMangleMap;

/*
 * portVarDict is an associative array with a encode port
 * variable name as the key and port type information as contents.
 */
EXTERNAL StringSgStatementAVLMap         *portVarDict;

/*
 * portTypeMap is an associative array with the type
 * as the key to get the corresponding names exprs
 */
EXTERNAL StringSgExpressionAVLMap        *portTypeMap;

/*
 * AVLMap objects that store process information
 */
EXTERNAL StringSgStatementAVLMap         *prosMap;
EXTERNAL StringSgStatementAVLMap         *prosDefMap;
EXTERNAL StringSgExpressionAVLMap        *prosHedrMap;
EXTERNAL StringSgExpressionAVLMap        *intentMap;

/*
 * A global processors decl store
 */ 
EXTERNAL StringSgExpressionAVLMap        *processorsMap;

/*
 * parameter Map
 */
EXTERNAL StringSgExpressionAVLMap        *parameterMap;

/*
 * static variable Map (variables in a SAVE statement)
 */
EXTERNAL StringSgExpressionAVLMap        *staticVarMap;

/*
 *  Types used For declaring FM data types
 */
EXTERNAL StringSgExpressionAVLMap        *typeMap;

/*
 * Distribute attribute for variables
 */

EXTERNAL StringSgExpressionAVLMap        *distributeMap;

/* 
 * MPCALL Map
 */

EXTERNAL StringSgStatementAVLMap         *mpcallMap;

/*
 * Cached types and attributes
 */

EXTERNAL SgType                          *integer_t;
EXTERNAL SgType                          *bool_t;
EXTERNAL SgType                          *char_t;
EXTERNAL SgType                          *real_t;
EXTERNAL SgType                          *double_t;
EXTERNAL SgType                          *complex_t;

EXTERNAL SgType                          *auto_t;

EXTERNAL SgType                          *inport_t;
EXTERNAL SgType                          *outport_t;

EXTERNAL SgTypeExp                       *integer_e;
EXTERNAL SgTypeExp                       *char_e;
EXTERNAL SgTypeExp                       *real_e;
EXTERNAL SgTypeExp                       *double_e;
EXTERNAL SgTypeExp                       *complex_e;

EXTERNAL SgAttributeExp                  *in_attribute;
EXTERNAL SgAttributeExp                  *out_attribute;
EXTERNAL SgAttributeExp                  *inout_attribute;

/*
 *  Symbols for FM Compiler Generated Functions
 */

// FM Number of Processors Call

EXTERNAL SgSymbol                        *fmNProcessorsCallSymb;
EXTERNAL SgSymbol                        *fmNPIntrinsicSymb;

// FM init & shutdown call

EXTERNAL SgSymbol                        *fmInitCallSymb;
EXTERNAL SgSymbol                        *fmIsMasterSymb;
EXTERNAL SgSymbol                        *fmStartCallSymb;
EXTERNAL SgSymbol                        *fmShutDownCallSymb;

// FM processors init call

EXTERNAL SgSymbol                        *fmVCShapeCallSymb;
EXTERNAL SgSymbol                        *fmChildVCOffsetCallSymb;

// Communication

EXTERNAL SgSymbol                        *fmInitInportCallSymb;
EXTERNAL SgSymbol                        *fmInitInportDebugCallSymb;
EXTERNAL SgSymbol                        *fmInportLeavingScopeCallSymb;

EXTERNAL SgSymbol                        *fmInitOutportCallSymb;
EXTERNAL SgSymbol                        *fmInitOutportDebugCallSymb;
EXTERNAL SgSymbol                        *fmOutportLeavingScopeCallSymb;

EXTERNAL SgSymbol                        *fmChannelCallSymb;
EXTERNAL SgSymbol                        *fmMoveInportCallSymb;
EXTERNAL SgSymbol                        *fmMoveOutportCallSymb;
EXTERNAL SgSymbol                        *fmErrorMessageCallSymb;
EXTERNAL SgSymbol                        *fmMergerInitCallSymb;
EXTERNAL SgSymbol                        *fmMergerAddCallSymb;
EXTERNAL SgSymbol                        *fmMergerDoneCallSymb;
EXTERNAL SgSymbol                        *fmEndChannelCallSymb;
EXTERNAL SgSymbol                        *fmProbeCallSymb;

// Concurrency

EXTERNAL SgSymbol                        *fmPblockInitCallSymb;
EXTERNAL SgSymbol                        *fmPblockSpawnCallSymb;
EXTERNAL SgSymbol                        *fmPblockResetCallSymb;
EXTERNAL SgSymbol                        *fmPblockInitInvokeCallSymb;
EXTERNAL SgSymbol                        *fmPblockInvokeCallSymb;
EXTERNAL SgSymbol                        *fmPblockContextWaitCallSymb;
EXTERNAL SgSymbol                        *fmPortCreateCallSymb;
EXTERNAL SgSymbol                        *fmNotifyPortsCallSymb;
EXTERNAL SgSymbol                        *fmPblockWaitCallSymb;

// Pablo Caller Info

EXTERNAL SgSymbol                        *fmPabloCallerInfoSymb;

// Control Flow

EXTERNAL SgSymbol                        *fmStopCallSymb;

/*
 * Regular Expressions Used to detect 
 * Compiler Directives
 *
 */

EXTERNAL Regex                           *threadableDirective;

/*
 *  Variables to store the sequence number (SN)
 *  used for Generating Unique Variable Names
 *
 */

EXTERNAL int                             iovalSN;
EXTERNAL int                             channelSN;
EXTERNAL int                             portSN;
EXTERNAL int                             pblockSN;
EXTERNAL int                             bufferSN;
EXTERNAL int                             doVarSN;
EXTERNAL int                             vecSN;
EXTERNAL int                             mpcSN;
EXTERNAL int                             emptySN;
EXTERNAL int                             mangleSN;

/*
 *  Global Constants
 */

EXTERNAL SgValueExp                     *c0;
EXTERNAL SgValueExp                     *c1;
EXTERNAL SgValueExp                     *c_minus_1;

/*
 * functions
 */


/*
 * FM Vec Annotation
 */

EXTERNAL SgValueExp                     *fmVecNoAnnValRef;
EXTERNAL SgValueExp                     *fmVecLctnValRef;
EXTERNAL SgValueExp                     *fmVecSubmValRef;

/*
 * Error & Warning Counters
 */

EXTERNAL int                             n_errors;
EXTERNAL int                             n_warnings;

/*
 * Error and Warning String
 */
EXTERNAL char                           *fm_error_string;
EXTERNAL char                           *fm_warning_string;

/*
 * MPCALL:  maximum processor counter
 *
 */

EXTERNAL int                            max_mp_processors;

/* 
 * function prototypes
 */

/*
 * Initialization
 */

void                      fm_initialize(void);

/*
 * Statement Munipulation
 */

void                      processFMStatementsInFile(SgFile *);
void                      doCanonalizeDecl(SgStatement *, SgStatement *);
void                      doVarDeclStmt(SgStatement *);
void                      doDimensionStmt(SgStatement *);
void                      doDistributeStmt(SgStatement *);
void                      doMPCallStmt(SgStatement *);
void                      doInportDeclStmt(SgStatement *);
void                      doOutportDeclStmt(SgStatement *);
void                      doChannelStmt(SgStatement *);
void                      doMergerStmt(SgStatement *);
void                      doSendStmt(SgStatement *);
void                      doReceiveStmt(SgStatement *);
void                      doProbeStmt(SgStatement *);
void                      doEndchannelStmt(SgStatement *);
void                      doProcessorsStmt(SgStatement *);
void                      doParallelBlock(SgStatement *, SgStatement **);
void                      doProsHedrStmt(SgStatement *);
void                      doProcHedrStmt(SgStatement *);
void                      doFuncHedrStmt(SgStatement *);
void                      doMoveportStmt(SgStatement *);
void                      doParametersAndSaveStmtsInProg(SgStatement *,SgStatement *);
void                      doParameterStmt(SgStatement *);
void                      doProsCommStmt(SgStatement *);
void                      doCommonStmt(SgStatement *);
void                      doDataDeclStmt(SgStatement *);
void                      doDataStmt(SgStatement *);
void                      doDataElt(SgExpression *);
void                      doDataImplDo(SgDataImpliedDoExp *);
void                      doExternalStmt(SgStatement *);
void                      doCommonStmtsInProg(SgStatement *, 
					      SgStatement *);
void                      doImplicitStmt(SgStatement *);
void                      doIntentStmt(SgStatement *);
void                      insertShutDownCall(SgStatement *,int);
void                      doStopStmt(SgStatement *);
void                      gen_c_aux_file(char *);
void                      gen_c_link_file(char *);
void                      rmFMStmts(void);
void                      restoreLBContStmts(void);

/*
 * Encoding Functions
 */

char *                    encodeStr(int);
char *                    portDictKeyEncode(SgSymbol *);
char *                    prosMapKeyEncode(char *, char *);
char *                    prosHedrMapKeyEncode(char *, char *);
char *                    prosDefMapKeyEncode(char *);
char *                    intentMapKeyEncode(char *);
char *                    parameterMapKeyEncode(SgSymbol *);
char *                    typeMapKeyEncode(char *, int);
char *                    processorsKeyEncode();
char *                    popTypeEncode(SgType *, SgExpression *);
char *                    portSetKeyEncode(SgStatement *);
char *                    directiveMapKeyEncode(char *,char *);
char *                    commonMapKeyEncode(char *);

/*
 * Map Utilities
 */

SgStatement *             stmtMapGet(StringSgStatementAVLMap *, char *); 
void                      stmtMapSet(StringSgStatementAVLMap *,
				     char *, 
				     SgStatement *);
void                      portTypeMapClear();
void                      exprMapSet(StringSgExpressionAVLMap *,
					 char *, SgExpression *);
SgExpression *            exprMapGet(StringSgExpressionAVLMap *, char *);
void                      stmtMapOverwrite(StringSgStatementAVLMap *,
					   char *, SgStatement *);
void                      exprMapOverwrite(StringSgExpressionAVLMap *,
					   char *, SgExpression *);
void                      fmProgBlockMapSet(StringFMProgBlockAVLMap *,
					    char *, FMProgBlock *);
FMProgBlock *             fmProgBlockMapGet(StringFMProgBlockAVLMap *, char *);

/*
 * Variant Utilities
 */
				     
char *                    getVariantByName(int);
void                      printVariantName(int);

/*
 * Error/Warning Reporting Procedures
 * 
 * NB: requires variable argument support
 *
 */

void                      fm_error(SgStatement *, char *, ...);
void                      fm_warning(SgStatement *, char *, ...);
void                      fm_internal_error(char *, ...);

/*
 * FM Control Information List Utilities
 *
 */

SgPortTypeExp *           getPortType(SgExpression *);
SgStatement *             getPortStatement(SgExpression *);
char *                    getPortSetKeyFromPortRef(SgExpression *);
int                       getPortVariant(SgExpression *);
int                       getNPortTypes(SgExpression *);
int                       eqPortTypeCiList(SgExpression *, SgExpression *);
int                       eqPortType(SgPortTypeExp *, SgPortTypeExp *, int);
void                      checkPortShape(SgExpression *);

int                       is_possible_port_expression(SgExpression *);
SgExpression *            portExp(SgExpression *);
SgExpression *            ioStoreExp(SgExpression *);
SgExpression *            labelExp(SgExpression *);

/*
 * Generate Statements that checks IOSTAT.
 */

SgStatement *             checkIOStat(int, SgExpression *,
				      SgExpression *, SgExpression *);
/*
 * Create temporary IOSTAT variables
 */

SgExpression *            fmIOVarExp(SgExpression *);

/*
 * Check whether the argument in the send/receive
 * match the port type specifications
 */

void                      matchPortAndArgTypes(int,SgPortTypeExp *,
					       SgExpression *);

/*
 * Virtual Computer Reference
 */

void                      checkVCRef(SgStatement *, SgExpression *);

/*
 * Type Munipulation Routines
 */

char *                    getTypeString(int);
char *                    getTypeCString(int);
char *			  getTypePrintfFormat(int type_variant);
void                      initImplicitType(SgStatement *s);
void                      newImplTypeTemplate(char *);
void                      deleteImplTypeTemplate(char *);
SgType *                  getImplicitType(char *, SgSymbol *);

/*
 * Expression 
 */

SgExpression &            arraySizeExp(SgArrayRefExp *);
int                       IsInteger(SgExpression *);
int                       integerValue(SgExpression *);
SgExpression *            integerValueExp(SgExpression *);
SgExpression &            replaceConstantInExpr(SgExpression *);
SgExpression &            replacePortProtoArgInExpr(SgExpression *);
SgExpression &            replaceFormalArgConstInExpr(SgExpression *);
void                      printExprByDeref(FILE *, SgExpression *);
SgExpression &            EPlusOne(SgExpression &);
SgExpression &            EAdd(SgExpression &, SgExpression &);
SgExpression &            EMul(SgExpression &, SgExpression &);
SgExpression &            loopSize(SgExpression *,
				   SgExpression *,
				   SgExpression *);
SgArrayRefExp *           make_array_ref(SgType *,SgExpression *,
					 SgExpression *, SgArrayRefExp *);

int                       is_f90_array_section_reference(SgExpression *);
SgExpression *            replace_DDOT_with_index(SgArrayRefExp *,int,
						  SgExpression &);
SgSubscriptExp *          get_DDOT_in_aref(SgArrayRefExp *, int);
int                       get_DDOT_position(SgArrayRefExp *, int);
SgExpression *            get_array_lbound(SgArrayRefExp *, int);
SgExpression *            get_array_ubound(SgArrayRefExp *, int);
SgExpression *            type_size_ratio(SgType *, SgType *, int *);
char *                    num_to_string(int);

/*
 * Debugging Routines
 */

void                      stmtMapDump(StringSgStatementAVLMap *);
void                      exprMapDump(char *, StringSgExpressionAVLMap *);
void                      parameterDump();

/*
 * FM Program Block Munipulation Routines
 */

FMProgBlock *             newFMProgBlock();
void                      switchToFMProgBlock(char *procName);

/*
 * Symbol Lookup
 */

SgExpression *            findSymbInDeclStmt(SgSymbol *);

/*
 * Misc Routines
 */

SgStatement *             whichProcAmI(SgStatement *);
int                       lhsId(SgExpression *);
int                       rhsId(SgExpression *);
int                       symbId(SgExpression *);
char *                    symbName(SgExpression *);
int                       typeId(SgExpression *);
SgExpression&             exprRef(SgExpression *);
char *                    ppString(char *, char *, char *);
SgStatement *             getEndParStmt(SgStatement *);
SgVarDeclStmt *           new_var_decl_stmt(SgSymbol *, 
					    SgExpression *,
					    SgType *,
					    SgStatement *);
void                      checkProsArgumentTypesIntents(SgProsHedrStmt *);
void                      declare_auto_symb(char *, SgStatement *);
void                      declare_symb_with_type(char *,SgType *,SgStatement*);
SgStatement *             declare_the_var(SgExpression *, SgType *);
SgStatement *             make_parameter_stmt(char *,
					      SgExpression *,
					      SgStatement *);
SgStatement *             make_include_stmt(char *);
SgStatement *             make_processors_stmt(SgExpression *);
SgStatement *             make_mp_ports_stmt(int, SgExpression *, int);
SgStatement *             make_mp_merger_stmt(SgExpression *);
SgStatement *             make_mp_proscall_stmt(char *, SgExpression *,
						SgMPCallSubm *,
						SgExprListExp **,
						SgExprListExp *);
SgStatement *             make_decl_stmt_with_stmt_and_var(SgStatement *,
							   SgExpression *);
SgStatement *             make_moveport_stmt(char *, char *, SgExprListExp *);
void                      scan_mpcalls(SgStatement *);
char *                    num2str(int);
SgStatement *             new_decl_stmt(SgType *, SgExpression *);
void                      copyLabelFromOriginalStmt(SgStatement*,SgStatement*);
void                      all_processes_must_be_threadable();
void                      doSaveStmt(SgStatement *);
void                      get_LS_stmts(SgStatement *);
void                      insert_port_LS_calls(SgCallStmt *);
void                      reset_LS_stmts(void);
