// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"

/*
 * stmtMapSet()
 *
 * Install the statement P with the key S into the statement AVLMap
 *
 */

void stmtMapSet(StringSgStatementAVLMap *pm, char *s, SgStatement *p)
{
    String *key = new String(s);

    if (! pm->contains(*key)) {
	(*pm)[*key] = *p;
    }
}

/*
 * stmtMapGet()
 * 
 * Retrieve the content with key S from the statement AVLMap.
 *
 */

SgStatement *stmtMapGet(StringSgStatementAVLMap *pm, char *s)
{
    String key(s);
    
    if(pm->contains(key)) {
	return( &(*pm)[key] );
    } else {
	return ( (SgStatement *)0 );
    }
}

/*
 * fmProgBlockMapSet()
 *
 * Install BLK with key S to the FM program block AVLMap.
 *
 */

void fmProgBlockMapSet(StringFMProgBlockAVLMap *map, char *s, FMProgBlock *blk)
{
    String *key = new String(s);

    if (! map->contains(*key)) {
	(*map)[*key] = *blk;
    }
}

/*
 * fmProgBlockMapGet()
 *
 * Retrieve the contents with key S from the FM program block AVLMap.
 *
 */

FMProgBlock *fmProgBlockMapGet(StringFMProgBlockAVLMap *map, char *s)
{
    String key(s);

    if (map->contains(key)) {
	return( &(*map)[key] );
    } else {
	return ( (FMProgBlock *) 0 );
    }
}

/* 
 * exprMapSet()
 *
 * Install Expression E with the key S into the expression map EM.
 *
 */

void exprMapSet(StringSgExpressionAVLMap *em, char *s, SgExpression *e)
{
    String *key = new String(s);

    if (!em->contains(*key)) {
	(*em)[*key] = *e;
    }
}

/*
 * exprMapSet()
 *
 * Retrieve the contents with the Key S from the expression map EM.
 *
 */

SgExpression *exprMapGet(StringSgExpressionAVLMap *em, char *s)
{
    String key(s);

    if (em->contains(key)) {
	return( &(*em)[key] );
    } else {
	return( (SgExpression *) 0 );
    }
}

/* 
 * portTypeMapClear()
 *
 * Delete all the contents of the global variable portTypeMap
 * 
 */

void portTypeMapClear()
{
    portTypeMap->clear();
}

void exprMapDump(char *name, StringSgExpressionAVLMap *map)
{
    printf("Map %s::\n", name);
    for (Pix ind = map->first(); ind; map->next(ind)) {
	String *key     = &(map->key(ind));
	SgExpression *e = &(map->contents(ind));

	printf("%s --- %s\n", key->chars(), e->unparse());
    }
}

