// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>

void printVariantName(int);

#define checkPortShape(x)

void doMergerStmt(SgStatement *s)
{
#ifdef _SAGEXX_1_3_
    SgMergerStmt              *mergeStmt;
#else
    SgMergeStmt               *mergeStmt;
#endif /* _SAGEXX_1_3_ */

    SgVariableSymb            *channelSymb;
    SgVarRefExp               *channelRef;
    SgVariableSymb            *doVarSymb;
    SgVarRefExp               *doVarRef;
    SgExpression              *ioStore;
    SgExpression              *ioStoreArg;
    SgCallStmt                *callStmt;
    SgInportExp               *inport;
    SgExpression              *inportExp;
    SgOutportExp              *outport;
    SgExpression              *outportExp;
    SgDataImpliedDoExp        *outports;
    SgForStmt                 *forStmt;
    SgArrayRefExp             *outpARef;
    SgArrayType               *op_array_type;
    SgSymbol                  *op_symbol;
    SgExpression              *init;
    SgExpression              *limit;
    SgExpression              *increment;
    
    char                      *externCallName;
    char                      *extern_mg_func_name;
    SgValueExp                *extern_mg_func_exp;
    SgExternalSymb            *externCallSymb;
    SgExpression              *externCallRef;
    SgDeclarationStatement    *externStmt;
    SgStatement               *scopeForDeclare;

    int                       *theSubIndex;
    char                      *ident;

    int                       mg_init_call_generated;
    SgCallStmt                *mg_init_call;

    SgExpression              *n_outports = c0;

#ifdef _SAGEXX_1_3_
    mergeStmt                 = isSgMergerStmt(s);
#else
    mergeStmt                 = isSgMergeStmt(s);
#endif /* _SAGEXX_1_3_ */

    if (! mergeStmt) {
	fm_internal_error("\
Do not expect a statement with a MERGE_STAT variant to be a non SgMergeStmt");
    }
    //
    // get CiList info
    // 
    inport         = isSgInportExp(mergeStmt->inport());

    if (! inport) {
	fm_error(s, "missing inport specifier");
	return;
    } else if (getPortVariant(inport) != INPORT_DECL) {
        fm_error(s, "%s is not an inport", 
		 symbName(portExp(inport)));
	return;
    } else {
	checkPortShape(inport);
    }

    inportExp      = portExp(inport);
    ioStore        = mergeStmt->ioStore();

    //
    // Generate declaration statement:
    // integer fm$channel, fm$i, fm$ioval
    //

    ident = encodeStr(EC_CHANNEL);
    channelSymb = new SgVariableSymb(ident, *integer_t);
    fm_assert(channelSymb);
    channelSymb->declareTheSymbol(*currentProcedure);

    if (opt_use_auto_vars) 
      declare_auto_symb(ident, currentProcedure);

    channelRef  = new SgVarRefExp(*channelSymb);

    ident = encodeStr(EC_DOVAR);
    doVarSymb   = new SgVariableSymb(ident, *integer_t);
    doVarSymb->declareTheSymbol(*currentProcedure);

    if (opt_use_auto_vars) 
      declare_auto_symb(ident, currentProcedure);

    doVarRef    = new SgVarRefExp(*doVarSymb);

    //
    // do outports
    //
    mg_init_call_generated = 0;
    for (int i = 0; i < mergeStmt->numberOfOutports(); i++)
    {
	outport     = isSgOutportExp(mergeStmt->outport(i));
	outportExp  = portExp(outport);

	if (! outport) {
	    fm_error(s, "missing outport specifier");
	    return;
	} else if (getPortVariant(outport) != OUTPORT_DECL) {
	    fm_error(s, "%s is not a outport", (outport->symbol())->identifier());
	    return;
	} else if ( ! eqPortTypeCiList(inport,outport)) {
	    fm_error(s, "inport and outport are not the same type");
	    return;
	} 
    
	if (outportExp->variant() == EXPR_LIST) {
	    SgExprListExp *eL = isSgExprListExp(outportExp);
	
	    outportExp = eL->elem(0);
	}

	// generate channel initialization call
	externCallName = ppString("fm_", portSetKeyEncode(getPortStatement(outport)), "_init");
	
	extern_mg_func_exp = isSgValueExp(exprMapGet(portMangleMap, externCallName));
	if (! extern_mg_func_exp) {
	    extern_mg_func_name = ppString(ppString("fm_", currentFileBase,""),
					   "_init", encodeStr(EC_MANGLE));
	    exprMapSet(portMangleMap, externCallName, 
		       new SgValueExp(extern_mg_func_name));
	} else {
	    extern_mg_func_name = extern_mg_func_exp->stringValue();
	}
	
	externCallSymb = new SgExternalSymb(extern_mg_func_name,*currentProcedure);
	externCallRef  = new SgVarRefExp(*externCallSymb);
	
	if (! stmtMapGet(prosMap, extern_mg_func_name)) {
	    externStmt = new SgDeclarationStatement(EXTERN_STAT);
	    externStmt->addVar(*externCallRef);
	    scopeForDeclare = currentProcedure->lastDeclaration();
	    scopeForDeclare->insertStmtBefore(*externStmt);
	    
	    stmtMapSet(prosMap, extern_mg_func_name, externStmt);
	}

	op_array_type = new SgArrayType(*integer_t);
	if (outportExp->variant() == ARRAY_REF) {
	    int found_DDOT = 0;
	    SgArrayRefExp *a;
	    a             = isSgArrayRefExp(outportExp);
	    
	    checkPortShape(outportExp);
	    fm_assert(a);
	    
	    op_symbol     = new SgVariableSymb(symbName(a), *op_array_type);
	    outpARef      = new SgArrayRefExp(*op_symbol);
	    //
	    // find the first DDOT subscripts
	    //
	    for (int dd_i = 0; dd_i < a->numberOfSubscripts(); dd_i++) {
		SgExpression *sub = a->subscript(dd_i);
		if (sub->variant() == DDOT) {
		    SgSubscriptExp *sub_exp;
		sub_exp        = isSgSubscriptExp(sub);
		init           = sub_exp->lbound();
		limit          = sub_exp->ubound();
		increment      = sub_exp->step();
                if (found_DDOT) {
                    fm_error(s,"\
Only expect one f90 array syntax, but got more than one");
                    return;
                } else {
                    found_DDOT     = 1;
                    outpARef->addSubscript(*doVarRef);
		    //
		    // if init and limit is null
		    // use the upper limit in the declaration.
		    //
		    if (! init) {
			init = get_array_lbound(a, dd_i);
		    } 
		    if (! limit) {
			limit = get_array_ubound(a, dd_i);
			if (! limit) {
			    fm_error(s, "\
%s is not an array variable\n", symbName(a));
			    return;
			}
		    }
                }
	    } else {
                outpARef->addSubscript(*sub);
	    }
	}

	if (! found_DDOT) {
	    fm_error(s,"\
expect a fortran 90 subcript or a data implied do list, but get neither");
	    return;
	}
	callStmt = new SgCallStmt(*fmMergerInitCallSymb);
	callStmt->addArg(*channelRef);
	callStmt->addArg(*inportExp);
	callStmt->addArg(*limit);
	n_outports = &(EAdd(n_outports->copy(), *limit));
    } else if (outportExp->variant() == DATA_IMPL_DO) {
	SgDataImpliedDoExp    *outports;
	int                   implDoLen;
	SgDataEltExp          *eltExp;
	SgDataSubsExp         *subs;
	SgExprListExp         *exprList;
	SgSymbol              *i_name;
	SgExprListExp         *outport_subs;

	outports = isSgDataImpliedDoExp(outport->exp());
	i_name   = outports->iname();
	fm_assert(i_name);
	
	eltExp = isSgDataEltExp(outports->dataelt(0));
	if (! eltExp->name()) {
	    fm_error(s,"\
Nested data implied do loop not supported");
	    return;
	}
	op_symbol = new SgVariableSymb((eltExp->name())->identifier(), 
				       *op_array_type);
	//
	// Set subscript of OUTPAREF, replacing the i_name symbol
	// by the temporary doVarSymb
	//
	subs         = isSgDataSubsExp(eltExp->datasubs());
	outport_subs = subs->iconexprlist();

	outpARef = new SgArrayRefExp(*op_symbol);

	for (int sub_i = 0; sub_i < outport_subs->length(); sub_i++) {
	    SgExpression *sub = outport_subs->elem(sub_i);
	    sub->replaceSymbolByExpression(*i_name, *doVarRef);
	    outpARef->addSubscript(*sub);
	}
	checkPortShape(outpARef);
//        outpARef->unparsestdout();
//        printf("\n");
//	outpARef->replaceSymbolByExpression(*i_name, *doVarRef);

	exprList  = outports->iconexprs();
	init      = exprList->elem(0);
	limit     = exprList->elem(1);
	increment = 0;
	
	callStmt = new SgCallStmt(*fmMergerInitCallSymb);
	callStmt->addArg(*channelRef);
	callStmt->addArg(*inportExp);
	callStmt->addArg(*limit);

	n_outports = &(EAdd(n_outports->copy(), *limit));
    } else {
#if ONLY_ALLOW_PORT_ARRAYS
	fm_error(s, "\
expect an array type or a data implied do for outport specifier, but get a %s",
		 getVariantByName(outportExp->variant()));
	return;
#endif /* ONLY_ALLOW_PORT_ARRAYS */
	init = limit = increment = c1;
	
	callStmt = new SgCallStmt(*fmMergerInitCallSymb);
	callStmt->addArg(*channelRef);
	fm_assert(inportExp);
	callStmt->addArg(*inportExp);
	callStmt->addArg(*c1);
	//
	// update the init counter
	//
	n_outports = &(EAdd(n_outports->copy(), *c1));
    }

    callStmt->addArg(*externCallRef);

    ioStoreArg = fmIOVarExp(ioStore);
    callStmt->addArg(*ioStoreArg);
    
//    callStmt->unparsestdout();

    if (! mg_init_call_generated) {
	s->insertStmtBefore(*callStmt);

	s->insertStmtBefore(*checkIOStat(1,ioStore,ioStoreArg,(SgExpression *)0));
	mg_init_call_generated = 1;
	mg_init_call = callStmt;
    } else {
	delete callStmt;
    }

    // 
    // Generate a do loop:
    //
    // do fm$i = 1, n
    //    call fm_merger_add(fm$channel,po(fm$i), fm$ioval)
    // enddo
    //
    //

    callStmt = new SgCallStmt(*fmMergerAddCallSymb);
    callStmt->addArg(*channelRef);
    
    if (outportExp->variant() == VAR_REF) {
	callStmt->addArg(*outportExp);
    } else {
	callStmt->addArg(*outpARef);
    }

    callStmt->addArg(*ioStoreArg);

    fm_assert(init);
    fm_assert(limit);

//    printVariantName(init->variant());
    if (! increment) {
	increment = new SgValueExp(1);
    }

    forStmt = new SgForStmt(*doVarSymb,
			    init->copy(),
			    limit->copy(),
			    *increment,
			    *callStmt);
//    forStmt->unparsestdout();
    s->insertStmtBefore(*forStmt);
    callStmt->insertStmtAfter(*checkIOStat(1,ioStore,ioStoreArg,(SgExpression *)0));
    }
    // 
    // Generate a call to fm_merger_done(fm$channel,fm$ioval);
    //

    callStmt = new SgCallStmt(*fmMergerDoneCallSymb);
    callStmt->addArg(*channelRef);
    fm_assert(ioStoreArg);
    callStmt->addArg(*ioStoreArg);
//    callStmt->unparsestdout();
    s->insertStmtBefore(*callStmt);
    s->insertStmtBefore(*checkIOStat(1,ioStore,ioStoreArg,(SgExpression *)0));

    //
    // Back Patch the MergerInit call
    // and put in the correct number of n_outport slot
    //
    if (mergeStmt->numberOfOutports() > 1) {
	callStmt = new SgCallStmt(*fmMergerInitCallSymb);
	callStmt->addArg(mg_init_call->arg(0)->copy());
	callStmt->addArg(mg_init_call->arg(1)->copy());
	// Patch the n outport slot
	callStmt->addArg(*n_outports);
	callStmt->addArg(mg_init_call->arg(3)->copy());
	callStmt->addArg(mg_init_call->arg(4)->copy());
	mg_init_call->insertStmtBefore(*callStmt);
	mg_init_call->deleteStmt();
    }

//    currentProcedure->unparsestdout();
    fmStmtSLL.append(s);
}


