// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"

#ifdef NeXT
/* __NX_CURRENT_COMPILER_RELEASE isn't defined when using g++ */
#ifdef __NX_CURRENT_COMPILER_RELEASE
extern "C" char * strdup(char *);
#endif /* __NX_CURRENT_COMPILER_RELEASE */
#endif /* NeXT */

static void gen_mpcall_funcs(SgStatement *, char *);

static void gen_ll_mp_process(SgSymbol *, 
                              SgSymbol *, 
			      SgExpression *, 
                              SgExprListExp *,
                              SgStatement *);

static void declare_args_using_distribute_specs(SgProsHedrStmt *,
						SgExprListExp *,
						SgExprListExp *,
						SgExpression *,
						SgStmtPtrSLList *);

static void gen_mp_common_block(SgStatement *);

SgExprListExp * make_formal_args(char *, int, int);
SgExprListExp * make_subm(SgExpression *);
SgExpression  * get_subm_lbound(SgMPCallSubm *);

void doMPCallStmt(SgStatement *s)
{
    if (s->variant() == MP_STAT ||
	s->variant() == MP_STAT_LCTN)
    {
	fm_error(s, "\
mpcall statement must have a submachine annotation");
	s->setVariant((s->variant() == MP_STAT)? PROS_STAT : PROS_STAT_LCTN);
	return;
    }

    if (max_mp_processors == 0) {
	scan_mpcalls(s);
    }

    if (! stmtMapGet(mpcallMap, s->symbol()->identifier())) 
    {
	SgSymbol *mp_symbol;
	char * mp_name = encodeStr(EC_MPC);
	
	mp_symbol = new SgSymbol(PROCEDURE_NAME, mp_name);

	gen_mpcall_funcs(s, mp_name);
	stmtMapSet(mpcallMap, s->symbol()->identifier(), s);
	s->setSymbol(*mp_symbol);
    }

    // change submachine notation
    // submachine (1:NP) --> submachine(1)

    isSgMPCallSubm(s)->setSubmachine(*make_subm(get_subm_lbound(isSgMPCallSubm(s))));
    s->setVariant(PROS_STAT_SUBM);
}

SgExprListExp * make_subm(SgExpression *subm_sub)
{
    SgProcessorsRefExp *processors = new SgProcessorsRefExp(*subm_sub);
    SgExprListExp *subm_eL = new SgExprListExp(*processors);

    return(subm_eL);
}


SgProcessorsRefExp * get_submachine(SgMPCallSubm *s)
{
    SgExprListExp *eL = isSgExprListExp(s->submachine());
    
    fm_assert(eL);

    return(isSgProcessorsRefExp(eL->elem(0)));
}

SgExpression * get_subm_lbound(SgMPCallSubm *s)
{
    SgProcessorsRefExp *subm;
    SgExpression *np;
    SgSubscriptExp *ddot;

    subm = get_submachine(s);
    fm_assert(subm);
    
    if (subm->numberOfSubscripts() > 1) {
	fm_error(s, "\
MPCALL only supports one dimensional processor array");
	return(SgExpression *)0;
    }
    
    np = subm->subscript(0);
    fm_assert(np);

    if (np->variant() != DDOT) {
	fm_error(s, "\
Expect an array section in the submachine specifier, but get a %s",
		 getVariantByName(np->variant()));
	return (SgExpression *)0;
    }
    
    ddot = isSgSubscriptExp(np);
    fm_assert(ddot);
    
    return(ddot->lbound());
}
    

SgExpression * get_n_processors(SgMPCallSubm *s)
{
    SgProcessorsRefExp *subm;
    SgExpression *np;
    SgSubscriptExp *ddot;

    subm = get_submachine(s);
    fm_assert(subm);
    
    if (subm->numberOfSubscripts() > 1) {
	fm_error(s, "\
MPCALL only supports one dimensional processor array");
	return(SgExpression *)0;
    }
    
    np = subm->subscript(0);
    fm_assert(np);

    if (np->variant() != DDOT) {
	fm_error(s, "\
Expect an array section in the submachine specifier, but get a %s",
		 getVariantByName(np->variant()));
	return (SgExpression *)0;
    }
    
    ddot = isSgSubscriptExp(np);
    fm_assert(ddot);
    
    if (ddot->step()) {
	return( & ((*(ddot->ubound()) - *(ddot->lbound()) + *c1) / 
		   (*(ddot->step()))));
    } else {
	return( & ((*(ddot->ubound()) - *(ddot->lbound()) + *c1)));
    }
    return (SgExpression *)0;
}

SgExpression * make_same_shape_for_symb(SgSymbol *symb, 
					SgType   *t,
					SgExpression *var)
{
    if (var->variant() == VAR_REF) {
	SgVarRefExp *new_var;
	
	symb->setType(*t);
	new_var = new SgVarRefExp(*symb);
	return (new_var);
    } else if (var->variant() == ARRAY_REF) {
	SgArrayRefExp *a;
	SgArrayRefExp *new_var;
	SgArrayType   *array_t;
	int           i;
	
	a = isSgArrayRefExp(var);
	array_t = new SgArrayType(*t);
	for (i = 0; i < a->numberOfSubscripts(); i++) {
	    array_t->addRange(*a->subscript(i));
	}
	symb->setType(*array_t);
	new_var = new SgArrayRefExp(*symb);
	for (i = 0; i < a->numberOfSubscripts(); i++) {
	    new_var->addSubscript(*a->subscript(i));
	}
	
	return new_var;
    } else {
	fm_internal_error("\
Sorry: expression not supported");
    }
    return (SgExpression *) 0;
}


static void gen_mpcall_funcs(SgStatement *s, char *mp_name)
{
    SgMPCallSubm    *mpcall;
    SgExpression    *np;
    int             np_i;
    int             i;

    // variables used for generating
    // new funcs
    SgProsHedrStmt  *topLevelPros;
    SgProsHedrStmt  *mpcPros;
    SgStatement     *where = currentProcedure->lastNodeOfStmt();

    SgInportStmt    *inport;
    SgOutportStmt   *outport;

    SgStatement     *parent_scope;
    SgStatement     *line;
    SgSymbol        *ll_pros_symbol;
    SgExprListExp   *ll_pros_args;

    SgExprListExp   *formal_args;

    mpcall = isSgMPCallSubm(s);
    fm_assert(mpcall);

    np = get_n_processors(mpcall);
    if (! IsInteger(np)) {
	fm_error(s, "\
Expect a constant number of processors, but get %s",
		 np->unparse());
	return;
    }

    np_i = integerValue(np);
    //
    // replace with constant instead
    // of the dreaded looking expression.
    //
    np = new SgValueExp (np_i);

    topLevelPros = new SgProsHedrStmt(mp_name);

    formal_args = make_formal_args("mp_user_arg", mpcall->numberOfArgs(), 0);

    // copy arguments
    for (i = 0; i < mpcall->numberOfArgs(); i++) {
	topLevelPros->AddArg(*formal_args->elem(i));
    }

    where->insertStmtAfter(*topLevelPros);
    where = topLevelPros->lexNext();

    for (i = 0; i < mpcall->numberOfArgs(); i++) {
	SgExpression *arg = mpcall->arg(i);
	SgSymbol     *arg_symb = arg->symbol();
	SgExpression *f_arg = formal_args->elem(i);
	SgSymbol     *f_arg_symb = f_arg->symbol();
	char         *arg_name ;
	SgStatement  *stmt;

	if (! arg_symb) {
	    SgType *arg_t;
	    SgVarDeclStmt *decl;
	    SgExprListExp *eL;

	    arg_t = arg->type();
	    fm_assert(arg_t);
	    eL = new SgExprListExp(*f_arg);
	    decl = new SgVarDeclStmt(*eL, *arg_t);
	    where->insertStmtBefore(*decl);
	    continue;
	}
	arg_name = arg->symbol()->identifier();
	stmt = stmtMapGet(prosMap, prosMapKeyEncode(currentProcedureName,arg_name));

	if (! stmt) {
	    // get implicit type
	    SgType *t;
	    SgVarDeclStmt *decl_s;

	    t = getImplicitType(currentProcedureName, arg_symb);
	    if (t == 0) {
		fm_error(s, "\
argument %d (%s) type unknown", i+1, arg_symb->identifier());
	    } else {
		SgExprListExp *varList;
                SgVariableSymb *varSymb =new SgVariableSymb(f_arg_symb->identifier(), *t);
		SgVarRefExp *var; 
		
		varSymb->setType(*t);
		var = new SgVarRefExp(*varSymb);
		varList = new SgExprListExp(*var);

                decl_s = new SgVarDeclStmt(*varList, *t);
		portMapSet(prosMap, 
			   prosMapKeyEncode(currentProcedureName,
					    arg_symb->identifier()),
			   decl_s);
		
		where->insertStmtBefore(*decl_s);
	    }
	} else {
	    SgStatement *decl;
	    SgExpression *var_decl = findSymbInDeclStmt(arg_symb);

	    f_arg = make_same_shape_for_symb(f_arg_symb, integer_t, var_decl);
	    decl = make_decl_stmt_with_stmt_and_var(stmt, f_arg);
	    where->insertStmtBefore(*decl);
	}
    }

    where = topLevelPros->lexNext();
    for (Pix ind = parameterMap->first(); ind; parameterMap->next(ind))
    {
	String &key = parameterMap->key(ind);
	char *param_name = (char *)key.chars();
	SgExpression *param = &(parameterMap->contents(ind));
	SgSymbol *param_symb = new SgVariableSymb(param_name, *integer_t);
	where->insertStmtBefore(*make_parameter_stmt(param_name, 
						     param, 
						     topLevelPros));
    }

    //
    // output processor definition
    // processors(<np>)
    //

    where = topLevelPros->lastDeclaration();
    where->insertStmtAfter(*make_processors_stmt(np));

    //
    // output "include 'fm_mpcall_defs.h'"
    //

    where = topLevelPros->lastDeclaration();
    gen_mp_common_block(where);
//    where->insertStmtAfter(*make_include_stmt("fm_mpcall_defs.h"));

    //
    // output the inport/outports used to 
    // hook up MP processors
    //
    
    where = topLevelPros->lastDeclaration();
    where->insertStmtAfter(*make_mp_ports_stmt(INPORT_DECL, np, 0));
    where = topLevelPros->lastDeclaration();
    where->insertStmtAfter(*make_mp_ports_stmt(OUTPORT_DECL, np, 0));
    
    parent_scope = currentProcedure;
    currentProcedure = topLevelPros;
    
    where = topLevelPros->lastDeclaration();
    line = make_mp_merger_stmt(np);
    where->insertStmtAfter(*line);

    where = line->lastNodeOfStmt();

    where->insertStmtAfter(*make_mp_proscall_stmt(mp_name, np, mpcall, 
						  &ll_pros_args, formal_args));

    currentProcedure = parent_scope;

    stmtMapSet(mpBlock, topLevelPros->symbol()->identifier(), topLevelPros);
//    topLevelPros->unparsestdout();

    ll_pros_symbol = new SgSymbol(PROCESS_NAME, ppString("mpc_",mp_name,""));


    gen_ll_mp_process(s->symbol(), ll_pros_symbol, np, ll_pros_args, 
                      topLevelPros->lastNodeOfStmt());

}

SgExprListExp * make_formal_args(char *prefix, int n, int start)
{
    SgExprListExp *formal_args;
    char arg_string[128];
    SgSymbol *arg_symb;
    SgExpression *arg_ref;
    int i;

    formal_args = 0;
    for (i = start; i < n+start; i++) {
	sprintf(arg_string, "%s_%d", prefix, i);

//	arg_symb = new SgVariableSymb(strdup(arg_string), *integer_t);
	arg_symb = new SgSymbol(VARIABLE_NAME, strdup(arg_string));
	arg_ref = new SgVarRefExp(*arg_symb);
	if (i == start) {
	    formal_args = new SgExprListExp(*arg_ref);
	} else {
	    formal_args->append(*arg_ref);
	}
    }
    return formal_args;
}

static void gen_ll_mp_process(SgSymbol      *user_func_symbol, 
			      SgSymbol      *pros_symbol,
			      SgExpression  *np, 
			      SgExprListExp *args,
                              SgStatement   *insert_point)
{
    SgProsHedrStmt *pros;
    SgStatement    *where;
    int            i;
    SgExprListExp  *formal_args = make_formal_args("mp_user_args", args->length(),-3);
    SgSymbol       *me = new SgVariableSymb("fm_mp_me", *integer_t);
    SgExpression   *me_ref = new SgVarRefExp(*me);
    SgSymbol       *call_symb;
    SgCallStmt     *call;
    SgStatement    *mp_inport_decl, *mp_outport_decl;
    SgExprListExp  *eL;
    SgStatement    *line;
    SgStatement    *do_loop;
    SgSymbol       *symb;
    SgExpression   *var;
    SgStatement    *assign_stmt;
    SgSymbol       *mp_np_symb, *mp_me_symb, *stash_symb;
    SgExpression   *mp_np_ref, *mp_me_ref;
    SgProcHedrStmt *stash_proc;
    SgStmtPtrSLList user_arg_decls;

    pros = new SgProsHedrStmt(*pros_symbol);

    insert_point->insertStmtAfter(*pros);
    
    pros->AddArg(*me_ref);
    me->declareTheSymbol(*pros);
    
    for (i = 1; i < 3; i++) {
	SgExpression *arg = args->elem(i);
	SgSymbol     *symb = arg->symbol();
	SgExpression *new_arg = new SgVarRefExp(*symb);

	pros->AddArg(*new_arg);
    }

    for (i = 3; i < formal_args->length(); i++) {
	pros->AddArg(*formal_args->elem(i));
    }

    //
    // Declaration args based on their DISTRIBUTE spec
    //
    
    declare_args_using_distribute_specs(pros, formal_args, args, np, 
					&user_arg_decls);
    where = pros->lastDeclaration();
    gen_mp_common_block(where);
//    where->insertStmtAfter(*make_include_stmt("fm_mpcall_defs.h"));
    
    call_symb = new SgSymbol(PROCEDURE_NAME, 
			     ppString("mp_stash_stuff_",
				      num_to_string(integerValue(np)),
				      user_func_symbol->identifier()));
	
    call = new SgCallStmt(*call_symb);
    call->addArg(*me_ref);

    for (i = 1; i < 3; i++) {
	SgExpression *arg = args->elem(i);
	SgSymbol     *symb = arg->symbol();
	SgExpression *new_arg = new SgVarRefExp(*symb);

	call->addArg(*new_arg);
    }

    stash_symb = call_symb;
    where = pros->lastDeclaration();
    where->insertStmtAfter(*call);
    
    where = call;

    call_symb = new SgSymbol(PROCEDURE_NAME, user_func_symbol->identifier());
    call = new SgCallStmt(*call_symb);
    
    for (i = 3; i < formal_args->length(); i++) {
        call->addArg(*formal_args->elem(i));
    }
    where->insertStmtAfter(*call);

    stmtMapSet(mpBlock, pros->symbol()->identifier(), pros);
//    pros->unparsestdout();


    if (! stmtMapGet(mpBlock, ppString(user_func_symbol->identifier(),
				       num_to_string(integerValue(np)),
				       "")))
    {
	stmtMapSet(mpBlock, ppString(user_func_symbol->identifier(),
				     num_to_string(integerValue(np)),
				     ""),
		   call);
    } else {
	return;
    }

    //
    // define subroutine mp_stash_stuff_<np>
    //
    stash_proc = new SgProcHedrStmt(*stash_symb);
    
    stash_proc->AddArg(*me_ref);

    for (i = 1; i < 3; i++) {
	SgExpression *arg = args->elem(i);
	SgSymbol     *symb = arg->symbol();
	SgExpression *new_arg = new SgVarRefExp(*symb);

	call->addArg(*new_arg);
    }
    
    where = pros->lastNodeOfStmt();
    where->insertStmtAfter(*stash_proc);

    me->declareTheSymbol(*stash_proc);
    where = stash_proc->lastDeclaration();
    mp_inport_decl = make_mp_ports_stmt(INPORT_DECL, np, 1);
    where->insertStmtAfter(*mp_inport_decl);

    where = stash_proc->lastDeclaration();
    mp_outport_decl = make_mp_ports_stmt(OUTPORT_DECL, np, 1);
    where->insertStmtAfter(*mp_outport_decl);

    //
    // add args: copying it from the last one
    //
    where = stash_proc->lastDeclaration();
    if (! user_arg_decls.empty()) {
	Pix ind;

	for (ind = user_arg_decls.first(); ind; user_arg_decls.next(ind))
	{
	    SgStatement *stmt = user_arg_decls(ind);
//	    SgStatement *after = stmt->lexNext();

//	    stmt->deleteStmt();
	    where->insertStmtBefore(*stmt);
//	    after->insertStmtBefore(*stmt);
	}
    }

    where = stash_proc->lastDeclaration();
    gen_mp_common_block(where);
//    where->insertStmtAfter(*make_include_stmt("fm_mpcall_defs.h"));
    
    where = stash_proc->lastDeclaration();

    mp_np_symb = new SgVariableSymb("fm_mp_np", *integer_t);
    mp_np_ref  = new SgVarRefExp(*mp_np_symb);
    mp_me_symb = new SgVariableSymb("mp_me", *integer_t);
    mp_me_ref  = new SgVarRefExp(*mp_me_symb);

    assign_stmt = new SgAssignStmt(*mp_np_ref, *np);
    where->insertStmtAfter(*assign_stmt);

    where = assign_stmt;
    assign_stmt = new SgAssignStmt(*mp_me_ref, *me_ref);
    where->insertStmtAfter(*assign_stmt);

    where = assign_stmt;

    symb = new SgVariableSymb("i", *integer_t);
    var  = new SgVarRefExp(*symb);
    symb->declareTheSymbol(*stash_proc);

    // 
    // gen moveport(from=fm_mp_Pi, to=mp_pi)
    //
    
    line = make_moveport_stmt("fm_mp_Pi", "mp_pi", (SgExprListExp *) 0);
    where->insertStmtAfter(*line);
    where = line;
    
    eL = new SgExprListExp(*var);
    line = make_moveport_stmt("fm_mp_Po", "mp_pos", eL);

    do_loop = new SgForStmt(*symb, *c1, *np, *line);
    where->insertStmtAfter(*do_loop);
    where = do_loop->lastNodeOfStmt();
    
    symb = new SgVariableSymb("mp_nmsgs", *integer_t);
    var = new SgVarRefExp(*symb);
    assign_stmt = new SgAssignStmt(*var, *c0);
    where->insertStmtAfter(*assign_stmt);

    stmtMapSet(mpBlock, stash_proc->symbol()->identifier(), stash_proc);
//    stash_proc->unparsestdout();
}

SgStatement * make_moveport_stmt(char *from_port_name, 
				 char *to_port_name,
				 SgExprListExp *dims)
{
    SgSymbol       *from_symb;
    SgSymbol       *to_symb;
    SgFromportExp  *from_ref;
    SgToportExp    *to_ref;
    SgArrayType    *array_t;
    int            i;
    SgExpression   *var;
    SgExprListExp   *eL;
    SgMoveportStmt *move_port;

    if (dims) {
	array_t = new SgArrayType(*integer_t);
	for (i = 0; i < dims->length(); i++) {
	    array_t->addRange(*c1);
	}
    }

    if (dims) {
        from_symb = new SgVariableSymb(from_port_name, *array_t);
        to_symb = new SgVariableSymb(to_port_name, *array_t);
    } else {
        from_symb = new SgVariableSymb(from_port_name, *integer_t);
        to_symb = new SgVariableSymb(to_port_name, *integer_t);
    }   
     
    if (dims) {
	var = new SgArrayRefExp(*from_symb);
	for (i = 0; i < dims->length(); i++) {
	    isSgArrayRefExp(var)->addSubscript(*dims->elem(i));
	}

        eL = new SgExprListExp(*var);
	from_ref = new SgFromportExp((SgExprListExp&)*var);
	
	var = new SgArrayRefExp(*to_symb);
	for (i = 0; i < dims->length(); i++) {
	    isSgArrayRefExp(var)->addSubscript(*dims->elem(i));
	}

        eL = new SgExprListExp(*var);
	to_ref = new SgToportExp((SgExprListExp&)*var);
    } else {
	var = new SgVarRefExp(*from_symb);
        eL = new SgExprListExp(*var);
	from_ref = new SgFromportExp((SgExprListExp&)*var);
	
	var = new SgVarRefExp(*to_symb);
        eL = new SgExprListExp(*var);
	to_ref = new SgToportExp((SgExprListExp&)*var);
    }	

    move_port = new SgMoveportStmt(*from_ref, *to_ref);
    return (move_port);
}


static void declare_args_using_distribute_specs(SgProsHedrStmt *s, 
						SgExprListExp *formal_args,
						SgExprListExp *args,
						SgExpression *np,
						SgStmtPtrSLList *ua_decls)
{
    int         i;
    SgStatement *where;

    //
    // first three are fixed (ME, PI, PO)
    // integer me
    // inport(integer n, integer, integer, character msg(n)) Pi
    // outport(integer n, integer, integer, character msg(n)) Po(np)
    //
    if (args->length() < 3) {
	fm_internal_error("\
low level MP process should have at least 3 arguments but only get %d",
			  args->length());
    }

    for (i = 1; i < 3; i++) {
	SgExpression *arg = args->elem(i);
	SgExpression *f_arg = formal_args->elem(i);
	SgSymbol *f_symb = f_arg->symbol();
	SgStatement *decl;

	fm_assert(f_symb);

        if (i == 1) {
            where = s->lastDeclaration();
	    decl = make_mp_ports_stmt(INPORT_DECL, np, 1);
	    where->insertStmtBefore(*decl);
	} else if (i = 2) {
	    decl = make_mp_ports_stmt(OUTPORT_DECL, np, 1);
	    where->insertStmtBefore(*decl);
	}
    }
    
    //
    // Process user arguments
    // 

    
    for (i = 3; i < args->length(); i++) {
	SgExpression *arg = args->elem(i);
	SgExpression *f_arg = formal_args->elem(i);
	SgSymbol     *arg_symb = arg->symbol();
	SgSymbol     *f_arg_symb = f_arg->symbol();
	SgStatement  *stmt, *theDecl;

	if (! arg_symb) {
	    SgType *arg_t;
	    SgVarDeclStmt *decl;
	    SgExprListExp *eL;

	    arg_t = arg->type();
	    fm_assert(arg_t);
	    eL = new SgExprListExp(*f_arg);
	    decl = new SgVarDeclStmt(*eL, *arg_t);
	    where->insertStmtBefore(*decl);
	    decl = new SgVarDeclStmt(*eL, *arg_t);
	    ua_decls->append((SgStmtPtr&)decl);
	    continue;
	}

	stmt = stmtMapGet(prosMap, prosMapKeyEncode(currentProcedureName,symbName(arg)));

	if (! stmt) {
	    // get implicit type
	    SgType *t;
	    SgVarDeclStmt *decl_s;

	    t = getImplicitType(currentProcedureName, arg_symb);
	    if (t == 0) {
		fm_error(s, "\
argument %d (%s) type unknown", i+1, arg_symb->identifier());
	    } else {
		SgExprListExp *varList;

		varList = new SgExprListExp(*f_arg);

                decl_s = new SgVarDeclStmt(*varList, *t);
		where->insertStmtBefore(*decl_s);
                decl_s = new SgVarDeclStmt(*varList, *t);
		ua_decls->append((SgStmtPtr&)decl_s);
	    }
	} else {
	    SgExpression *var_decl;
	    
	    var_decl = findSymbInDeclStmt(arg_symb);
	    
	    fm_assert(var_decl);
	    if (! exprMapGet(distributeMap, symbName(arg)) ||
		var_decl->variant() != ARRAY_REF) 
	    {
		SgStatement *decl;
		SgExpression *var_decl;

		var_decl = findSymbInDeclStmt(arg_symb);
		
		f_arg = make_same_shape_for_symb(f_arg_symb, integer_t, var_decl);
		decl = make_decl_stmt_with_stmt_and_var(stmt, f_arg);
		where->insertStmtBefore(*decl);
		decl = make_decl_stmt_with_stmt_and_var(stmt, f_arg);
		ua_decls->append(decl);
		
		continue;
	    } else {
		//
		// size down the variable according 
		// to their distribute attribute
		//
		SgArrayRefExp *a = isSgArrayRefExp(var_decl);
		SgArrayRefExp *new_a;
		SgArrayRefExp *dist;
		SgArrayType   *array_t;
		int k;
		
		dist = isSgArrayRefExp(exprMapGet(distributeMap, symbName(arg)));
		
		if (dist->numberOfSubscripts() != a->numberOfSubscripts()) {
		    fm_error(currentStatement, "\
distribute attribute does not match the number of subscripts for variable %s",
			     symbName(arg));
		    return;
		}
		
		array_t = new SgArrayType(*integer_t);
		for (k = 0; k < a->numberOfSubscripts(); k++) {
		    array_t->addRange(*c1);
		}
		
		f_arg->symbol()->setType(*array_t);
		new_a = new SgArrayRefExp(*f_arg->symbol());

		for (k = 0; k < a->numberOfSubscripts(); k++) {
		    if (dist->subscript(k)->variant() == BLOCK_OP) {
			SgExpression *dim;

			if (! IsInteger(a->subscript(k))) {
			    fm_error(currentStatement,"\
array %s dimension %d: unknown size %s\n",
				     symbName(arg), k+1, 
				     a->subscript(k)->unparse());
			    return;
			}
				     
			dim = new SgValueExp(integerValue(a->subscript(k))/
					     integerValue(np));
			new_a->addSubscript(*dim);
		    } else {
			new_a->addSubscript(*a->subscript(k));
		    }
		}
		//
		// insert this var into the declaration 
		// statement.
		// 
		theDecl = make_decl_stmt_with_stmt_and_var(stmt, new_a);
		where->insertStmtBefore(*theDecl);
		theDecl = make_decl_stmt_with_stmt_and_var(stmt, new_a);
		ua_decls->append(theDecl);
	    }
	}
    }
}


SgStatement * make_mp_proscall_stmt(char *pros_name, 
                                    SgExpression *np,
                                    SgMPCallSubm *mpcall,
				    SgExprListExp **ll_pros_args,
				    SgExprListExp *formal_args)
{
    SgProsCallSubm *pros_call;
    SgExprListExp  *pros_args;
    SgExprListExp  *subm;
    SgSymbol       *ll_pros_symbol;
    SgProcessDoStmt *pros_do;
    SgSymbol        *pros_do_symbol;
    SgVarRefExp     *pros_do_var;
    int i;
    SgSymbol        *port_symbol;
    SgArrayRefExp   *port_ref;
    SgArrayType     *port_array_t;
    SgProcessorsRefExp *processors_ref;

    ll_pros_symbol = new SgSymbol(PROCESS_NAME, ppString("mpc_",pros_name,""));

    pros_do_symbol = new SgVariableSymb("fm_mp_prosdo_i",  *integer_t);

    pros_do_var    = new SgVarRefExp(*pros_do_symbol);
    pros_do_symbol->declareTheSymbol(*currentProcedure);

    if (opt_use_auto_vars) 
      declare_auto_symb("fm_mp_prosdo_i", currentProcedure);

    // make args
    pros_args = new SgExprListExp(*pros_do_var - (*c1));
    *ll_pros_args = new SgExprListExp(*pros_do_var - (*c1));

    port_array_t = new SgArrayType(*integer_t);
    port_array_t->addRange(*np);
    
    port_symbol = new SgVariableSymb("fm_mp_Pi", *port_array_t);
    port_ref    = new SgArrayRefExp(*port_symbol);
    port_ref->addSubscript(*pros_do_var);

    pros_args->append(*port_ref);
    (*ll_pros_args)->append(*port_ref);

    port_array_t = new SgArrayType(*integer_t);
    port_array_t->addRange(*np);
    port_array_t->addRange(*np);
    
    port_symbol = new SgVariableSymb("fm_mp_Po", *port_array_t);
    port_ref    = new SgArrayRefExp(*port_symbol);
    port_ref->addSubscript(*c1);
    port_ref->addSubscript(*pros_do_var);
    
    pros_args->append(*port_ref);
    (*ll_pros_args)->append(*port_ref);

    for (i = 0; i < mpcall->numberOfArgs(); i++) {
        SgExpression *f_arg = formal_args->elem(i);
	SgExpression *arg = mpcall->arg(i);
	SgSymbol     *arg_symb = arg->symbol();
	SgStatement *stmt;

	if (! arg->symbol())
	{
	    pros_args->append(*f_arg);
	    (*ll_pros_args)->append(*mpcall->arg(i));
	    continue;
	}
	stmt = stmtMapGet(prosMap, prosMapKeyEncode(currentProcedureName,symbName(arg)));

	if (! stmt) {
	    pros_args->append(*f_arg);
	    (*ll_pros_args)->append(*mpcall->arg(i));
	    continue;
	} else {
	    SgExpression *var_decl;
	    
	    var_decl = findSymbInDeclStmt(arg_symb);
	    
	    fm_assert(var_decl);
	    if (! exprMapGet(distributeMap, symbName(arg)) ||
		var_decl->variant() != ARRAY_REF) 
	    {
		pros_args->append(*f_arg);
		(*ll_pros_args)->append(*mpcall->arg(i));
		continue;
	    } else {
		//
		// size down the variable according 
		// to their distribute attribute
		//
		SgArrayRefExp *a = isSgArrayRefExp(var_decl);
		SgArrayRefExp *new_a;
		SgArrayRefExp *dist;
		SgArrayType   *array_t;
		int k;
		
		dist = isSgArrayRefExp(exprMapGet(distributeMap, symbName(arg)));
		
		if (dist->numberOfSubscripts() != a->numberOfSubscripts()) {
		    fm_error(currentStatement, "\
distribute attribute does not match the number of subscripts for variable %s",
			     symbName(arg));
		    break;
		}
		
		array_t = new SgArrayType(*integer_t);
		for (k = 0; k < a->numberOfSubscripts(); k++) {
		    array_t->addRange(*c1);
		}
		
		f_arg->symbol()->setType(*array_t);
		new_a = new SgArrayRefExp(*f_arg->symbol());

		for (k = 0; k < a->numberOfSubscripts(); k++) {
		    if (dist->subscript(k)->variant() == BLOCK_OP) {
			SgExpression *dim;

			if (! IsInteger(a->subscript(k))) {
			    fm_error(currentStatement,"\
array %s dimension %d: unknown size %s\n",
				     symbName(arg), k+1, 
				     a->subscript(k)->unparse());
			    break;
			}
				     
			dim = new SgValueExp(integerValue(a->subscript(k))/
					     integerValue(np));
			dim = &((*pros_do_var-(*c1))*(dim->copy())+(*c1));
			new_a->addSubscript(*dim);
		    } else {
			new_a->addSubscript(*c1);
		    }
		}
		pros_args->append(*new_a);
		(*ll_pros_args)->append(*arg);
	    }
	}
    }
    
    //
    // make submachine expression
    //
    processors_ref = new SgProcessorsRefExp(*pros_do_var);
    subm  = new SgExprListExp(*processors_ref);

    pros_call = new SgProsCallSubm(*ll_pros_symbol, *pros_args, *subm);
    
    pros_do = new SgProcessDoStmt(*pros_do_symbol, *c1, *np, *pros_call);
    return (pros_do);
}




SgStatement * make_mp_ports_stmt(int variant, SgExpression *np, int phase)
{
    SgType         *t;
    SgPortTypeExp  *port_t;
    SgArrayRefExp  *msg_buffer;
    SgArrayType    *array_t;
    SgSymbol       *msg_buffer_symbol;
    SgSymbol       *size_symbol;
    SgVarRefExp    *size_ref;
    char           *port_ident;

    size_symbol = new SgVariableSymb("n", *integer_t);
    size_ref    = new SgVarRefExp(*size_symbol);

    array_t = new SgArrayType(*char_t);
    array_t->addRange(*size_ref);
    
    msg_buffer_symbol = new SgVariableSymb("msg", *array_t);
    msg_buffer = new SgArrayRefExp(*msg_buffer_symbol, *size_ref);
    if (variant == INPORT_DECL) {
	if (phase == 2) {
	    port_ident = strdup("mp_pi");
	    phase = 1;
	} else {
	    port_ident = strdup("fm_mp_Pi");
	}
    } else {
	if (phase == 2) {
	    port_ident = strdup("mp_pos");
	    phase = 1;
	} else {
	    port_ident = strdup("fm_mp_Po");
	}
    }
	
    t = new SgType(T_STRING);
    if (variant == INPORT_DECL) {
	SgInportStmt *inportStmt;
	SgExprListExp *port_eL;
	SgSymbol *port_symbol;
	SgExpression  *port_ref;
	SgArrayType   *port_array_t;

	port_array_t = new SgArrayType(*integer_t);
	port_array_t->addRange(*np);
	
	if (phase == 0) {
	    port_symbol = new SgVariableSymb(port_ident, *array_t);
	    port_ref    = new SgArrayRefExp(*port_symbol);
	    isSgArrayRefExp(port_ref)->addSubscript(*np);
	} else if (phase == 1) {
	    port_symbol = new SgVariableSymb(port_ident, *integer_t);
	    port_ref = new SgVarRefExp(*port_symbol);
	} else {
	    fm_internal_error("\
Illegal phase number %d occurs at %s:%d",
			      phase, __FILE__, __LINE__);
	}

	port_eL     = new SgExprListExp(*port_ref);

	port_t = new SgPortTypeExp(*integer_t, *size_ref);
	inportStmt = new SgInportStmt(*port_eL, *port_t);
	port_t = new SgPortTypeExp(*integer_t);
	inportStmt->addporttype(*port_t);
	port_t = new SgPortTypeExp(*integer_t);
	inportStmt->addporttype(*port_t);
	port_t = new SgPortTypeExp(*t, *msg_buffer);
	inportStmt->addporttype(*port_t);
	return (inportStmt);
    } else if (variant == OUTPORT_DECL) {
	SgOutportStmt *outportStmt;
	SgExprListExp *port_eL;
	SgSymbol *port_symbol;
	SgArrayRefExp *port_ref;
	SgArrayType   *port_array_t;

	port_array_t = new SgArrayType(*integer_t);
	port_array_t->addRange(*np);
	port_array_t->addRange(*np);
	
	port_symbol = new SgVariableSymb(port_ident, *array_t);
	port_ref    = new SgArrayRefExp(*port_symbol);
	port_ref->addSubscript(*np);
	if (phase == 0) {
	    port_ref->addSubscript(*np);
	}

	port_eL     = new SgExprListExp(*port_ref);

	port_t = new SgPortTypeExp(*integer_t, *size_ref);
	outportStmt = new SgOutportStmt(*port_eL, *port_t);
	port_t = new SgPortTypeExp(*integer_t);
	outportStmt->addporttype(*port_t);
	port_t = new SgPortTypeExp(*integer_t);
	outportStmt->addporttype(*port_t);
	port_t = new SgPortTypeExp(*t, *msg_buffer);
	outportStmt->addporttype(*port_t);
	return (outportStmt);
    } else {
	fm_internal_error("\
Expect either INPORT_DECL or OUTPORT_DECL at %s:%d but get %d",
			  __FILE__, __LINE__, variant);
    }
    return (SgStatement *)0;
}

SgStatement * make_mp_merger_stmt(SgExpression *np)
{
    SgForStmt      *do_loop;
    SgSymbol       *do_var_symbol;
    SgVarRefExp    *do_var_ref;
    SgMergerStmt   *merger_stmt;

    SgSymbol       *inport_symbol;
    SgArrayRefExp  *inport_ref;
    SgExprListExp  *inport_eL;

    SgSymbol       *outport_symbol;
    SgArrayRefExp  *outport_ref;
    SgExprListExp  *outport_eL;
    SgArrayType    *array_t;
    SgInportExp    *inport;
    SgOutportExp   *outport;
    SgSubscriptExp *ddot;

    do_var_symbol = new SgVariableSymb("fm_mp_i",  *integer_t);
    do_var_symbol->declareTheSymbol(*currentProcedure);

    if (opt_use_auto_vars) 
      declare_auto_symb("fm_mp_i", currentProcedure);

    do_var_ref    = new SgVarRefExp(*do_var_symbol);

    array_t = new SgArrayType(*integer_t);
    array_t->addRange(*np);

    inport_symbol = new SgVariableSymb("fm_mp_Pi", *array_t);
    inport_ref = new SgArrayRefExp(*inport_symbol);
    inport_ref->addSubscript(*do_var_ref);
    inport_eL = new SgExprListExp(*inport_ref);

    inport   = new SgInportExp(*inport_eL);

    outport_symbol = new SgVariableSymb("fm_mp_Po", *array_t);
    outport_ref = new SgArrayRefExp(*outport_symbol);
    outport_ref->addSubscript(*do_var_ref);
    ddot = new SgSubscriptExp(*c1, *np);
    outport_ref->addSubscript(*ddot);
    outport_eL = new SgExprListExp(*outport_ref);

    outport   = new SgOutportExp(*outport_eL);
    
    merger_stmt = new SgMergerStmt(*outport, *inport);
    
    do_loop = new SgForStmt(*do_var_symbol, *c1, *np, *merger_stmt);
    return (do_loop);
}

SgStatement * make_include_stmt(char *incl_file_name)
{
    SgIncludeStmt *includeStmt;
    SgValueExp    *file = new SgValueExp(incl_file_name);

    includeStmt = new SgIncludeStmt(*file);
    return (includeStmt);
}

static void gen_mp_common_block(SgStatement *where)
{
    SgSymbol          *symb;
    SgExpression      *var;
    SgExprListExp     *var_eL, *int_var_eL;
    SgSymbol          *comm_symb;
    SgObjectListExp   *obj_list;
    SgNestedVarListDeclStmt *pros_comm;
    SgStatement       *stmt;
    SgExpression      *np;
    SgArrayType       *array_t;
    SgExpression      *range;
    

    np = new SgValueExp(max_mp_processors);
    comm_symb = new SgSymbol(VARIABLE_NAME, "mp_state");

    symb = new SgVariableSymb("mp_nmsgs", *integer_t);
    var = new SgVarRefExp(*symb);
    var_eL = new SgExprListExp(*var);
    int_var_eL = new SgExprListExp(*var);

    symb = new SgVariableSymb("mp_buff", *char_t);
    var = new SgVarRefExp(*symb);
    var_eL->append(*var);

    symb = new SgVariableSymb("mp_np", *integer_t);
    var = new SgVarRefExp(*symb);
    var_eL->append(*var);
    int_var_eL = new SgExprListExp(*var);

    symb = new SgVariableSymb("mp_me", *integer_t);
    var = new SgVarRefExp(*symb);
    var_eL->append(*var);
    int_var_eL = new SgExprListExp(*var);

    symb = new SgVariableSymb("mp_pi", *integer_t);
    var = new SgVarRefExp(*symb);
    var_eL->append(*var);

    symb = new SgVariableSymb("mp_pos", *integer_t);
    var = new SgVarRefExp(*symb);
    var_eL->append(*var);

    obj_list = new SgObjectListExp(COMM_LIST, *comm_symb, *var_eL);
    pros_comm = new SgNestedVarListDeclStmt(PROS_COMM, *obj_list);
//    pros_comm->unparsestdout();
    where->insertStmtAfter(*pros_comm);

    where = pros_comm;
    stmt = new SgVarDeclStmt(*int_var_eL, *integer_t);
    where->insertStmtAfter(*stmt);

    where = stmt;
    stmt = make_mp_ports_stmt(INPORT_DECL, np, 2);
    where->insertStmtAfter(*stmt);

    where = stmt;
    stmt = make_mp_ports_stmt(OUTPORT_DECL, np, 2);
    where->insertStmtAfter(*stmt);

    where = stmt;
    array_t = new SgArrayType(*char_t);
    range = new SgValueExp(MP_BUFFER_DEFAULT_SIZE);
    array_t->addRange(*range);
    symb = new SgVariableSymb("mp_buff", *array_t);
    var = new SgArrayRefExp(*symb);
    isSgArrayRefExp(var)->addSubscript(*range);
    var_eL = new SgExprListExp(*var);
    stmt = new SgVarDeclStmt(*var_eL, *char_t);
    
    where->insertStmtAfter(*stmt);
}



SgStatement * make_processors_stmt(SgExpression *np)
{
    SgProcessorsRefExp *processors_ref;
    SgDeclarationStatement  *processors_stmt;
    SgExprListExp      *processors_eL;

    processors_ref = new SgProcessorsRefExp();
    processors_ref->addSubscript(*np);
    processors_eL = new SgExprListExp(*processors_ref);

    processors_stmt = new SgDeclarationStatement(PROCESSORS_STAT);
    processors_stmt->addVar(*processors_ref);

    return (processors_stmt);
}


SgStatement * make_decl_stmt_with_stmt_and_var(SgStatement *s, 
					       SgExpression *var)
{
    SgExprListExp *var_eL = new SgExprListExp(*var);

    if (s->variant() == VAR_DECL) {
	SgType        *t;
	SgVarDeclStmt *decl, *new_decl;
	
	decl = isSgVarDeclStmt(s);
	t = decl->type();

	new_decl = new SgVarDeclStmt(*var_eL, *t);
	return (new_decl);
    } else if (s->variant() == INPORT_DECL) {
	SgInportStmt  *inport_stmt, *new_inport_stmt;
	SgPortTypeExp *port_t;

	inport_stmt = isSgInportStmt(s);
	port_t = inport_stmt->porttypes();
	new_inport_stmt = new SgInportStmt(*var_eL, *port_t);

	return (new_inport_stmt);
    } else if (s->variant() == OUTPORT_DECL) {
	SgOutportStmt  *outport_stmt, *new_outport_stmt;
	SgPortTypeExp *port_t;

	outport_stmt = isSgOutportStmt(s);
	port_t = outport_stmt->porttypes();
	new_outport_stmt = new SgOutportStmt(*var_eL, *port_t);

	return (new_outport_stmt);
    } else {
	fm_internal_error("\
Expect an declaration statement but get an %s in %s:%d",
			  getVariantByName(s->variant()),
			  __FILE__, __LINE__);
    }
    return (SgStatement *)0;
}


void scan_mpcalls(SgStatement *s)
{
    SgStatement  *line;
    SgExpression *np;
    int          n_processors;

    for (line = s; line; line = line->lexNext()) {
	if (line->variant() == MP_STAT_SUBM) {
	    np = get_n_processors(isSgMPCallSubm(s));
	    if (! IsInteger(np)) {
		continue;
	    }
	    
	    n_processors = integerValue(np);

	    // update MAX_MP_PROCESSORS variable
	    if (max_mp_processors < n_processors) { 
		max_mp_processors = n_processors;
	    }
	}
    }
}
