// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <String.h>
#include <string.h>
#include <assert.h>



void doProcessorsStmt(SgStatement *s)
{
    char               *key;
    SgVarListDeclStmt  *varList;
    SgProcessorsRefExp *processorsSpec; 
    SgProcessorsRefExp *newSpec;
    SgArrayRefExp      *fm_vc_ref;
    SgVariableSymb     *fm_vc_symb;
    SgVarRefExp        *fm_vc_name_ref;
    SgArrayType        *arrayType = new SgArrayType(*integer_t);
    SgAssignStmt       *assignStmt;
    char               *fm_vc_string = encodeStr(EC_VEC);
    SgValueExp         *fm_vc_range_exp;
    SgCallStmt         *callStmt;
    SgVarRefExp        *is_master_reference;
    SgIfStmt           *ifStmt;
    SgFunctionCallExp  *func_call;
    SgAssignStmt       *fm_init_call_assign;
    SgContinueStmt     *cont;
    
    // check for multiple processors statements
    key                = processorsKeyEncode();

    if (processorsMapGet(processorsMap, key)) {
	fm_warning(s, "multiply defined PROCESSOR statement, use the first declaration");
	free(key);
	return;
    }

    if (currentProcedure->variant() == PROG_HEDR) {
	is_master_reference = new SgVarRefExp(*fmIsMasterSymb);
    }

    varList            = isSgVarListDeclStmt(s);
    processorsSpec     = isSgProcessorsRefExp(varList->var(0));
    newSpec            = new SgProcessorsRefExp;
    fm_vc_range_exp    = new SgValueExp(processorsSpec->numberOfSubscripts());
    arrayType->addRange(*fm_vc_range_exp);
    fm_vc_symb         = new SgVariableSymb(fm_vc_string, *arrayType);
    fm_vc_symb->declareTheSymbol(*currentProcedure);
    
    if (opt_use_auto_vars)
      declare_auto_symb(fm_vc_string, currentProcedure);

    fm_vc_name_ref     = new SgVarRefExp(*fm_vc_symb);

    //
    // generate an if statement for PROGRAM block
    // 
    // if (fm$_is_master .ne. 0) then
    //   ...shape-argument-init...
    //   fm_vc_shape()
    // endif
    //
    //
    
    if (currentProcedure->variant() == PROG_HEDR) {
	cont       = new SgContinueStmt;
	ifStmt     = new SgIfStmt(*is_master_reference > *c0, 
				  *cont);

	firstExeStmt->insertStmtBefore(*ifStmt);
    }

    for (int i = 0; i < processorsSpec->numberOfSubscripts(); i++) {
	SgValueExp *f_ind = new SgValueExp(i+1);
	//
	// check whether the dimension is integer and non-zero
	//
	SgExpression *sub = processorsSpec->subscript(i);

#if 0
	if (IsInteger(sub)) {
	    SgExpression *val = integerValueExp(sub);

	    newSpec->addSubscript(*integerValueExp(sub));
	} else {
	    fm_warning(s, "\
dimension %d (%s) is not a constant expression", i+1, sub->unparse());
	    newSpec->addSubscript(*sub);
	}
#endif /* 0 */

	if (sub->variant() == FUNC_CALL) {
	    String func_name(sub->symbol()->identifier());

	    // convert the function name to upper case
	    //func_name.upcase();
	    if (! strcmp((char *)func_name.chars(), FM_N_PROCESSORS_PROC))
	    {
		SgFunctionCallExp *fm_call;

//		fmNProcessorsCallSymb->declareTheSymbol(*currentProcedure);
		fm_call = new SgFunctionCallExp(*fmNProcessorsCallSymb);
		sub = fm_call;
	    } else {
		fm_error(s, "\
Unknown intrinsic %s() inside the processors statement",
			 func_name.chars());
	    }
	} 
	newSpec->addSubscript(*sub);

	// fill in fm_vc_shape() array
	fm_vc_ref = new SgArrayRefExp(*fm_vc_symb, *f_ind);
	assignStmt = new SgAssignStmt(*fm_vc_ref, *sub);
	if (currentProcedure->variant() == PROG_HEDR) {
	    cont->insertStmtBefore(*assignStmt);
	} else {
	    firstExeStmt->insertStmtBefore(*assignStmt);
	}
    }

    callStmt = new SgCallStmt(*fmVCShapeCallSymb);
    callStmt->addArg(*fm_vc_name_ref);
    callStmt->addArg(*fm_vc_range_exp);

    if (currentProcedure->variant() == PROG_HEDR) {
	cont->insertStmtBefore(*callStmt);
	cont->deleteStmt();
    } else {
	firstExeStmt->insertStmtBefore(*callStmt);
    }

    processorsMapSet(processorsMap, key, newSpec);

    fmStmtSLL.append(s);
}


void checkVCRef(SgStatement *s, SgExpression *e)
{
    SgExprListExp *e_list  = isSgExprListExp(e);
    SgProcessorsRefExp *vc_ref = isSgProcessorsRefExp(e_list->elem(0));
    int i;

    // get the processors statement in 
    // the current procedure
    SgProcessorsRefExp *vc_shape
      = isSgProcessorsRefExp(processorsMapGet(processorsMap, 
					      processorsKeyEncode()));

    //
    // do not check VC shape but rather let the
    // runtime find out the absense of the shape info.
    // see fm_runtime/vc.c for further info.
    //
    if (! vc_shape) { return; }

    if (vc_shape->numberOfSubscripts() != vc_ref->numberOfSubscripts()) {
	fm_error(s, "\
Invalid virtual computer reference: %d-dimensional but the shape is %d-dimentional",
		 vc_ref->numberOfSubscripts(), vc_shape->numberOfSubscripts());
    }
}   


