// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>

char *str2lower(char *);

/*
 * genPortInitCall()
 *
 * Return a procedure call fm_<proc>_port_init() statement
 * with the entire argument list being passed along.
 */

SgStatement *genPortInitCall(SgStatement *s, SgVarRefExp *fm_pblock)
{
    char                    *prosName;
    char                    *callName;
    SgSymbol                *callSymb;
    SgCallStmt              *callStmt;
    
    prosName = (s->symbol())? s->symbol()->identifier() : "unknown_pros";
    callName = ppString("fm_", prosName, "_port_init");
    callSymb = new SgSymbol(PROCEDURE_NAME,callName);
    callStmt = new SgCallStmt(*callSymb);
    
    if (s->variant() == PROS_STAT) {
	SgProsCallStmt *prosStmt = isSgProsCallStmt(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_LCTN) {
	SgProsCallLctn *prosStmt = isSgProsCallLctn(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_SUBM) {
	SgProsCallSubm *prosStmt = isSgProsCallSubm(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else {
	fm_internal_error("\
Dont know what porsstatement %s is\n", getVariantByName(s->variant()));
    }
    callStmt->addArg(*fm_pblock);
    return (callStmt);
}

/*
 * genInvokeCall()
 *
 * Return a procedure call (fm_<proc>_invoke()) statement
 * with the entire argument list being passed along. 
 *
 */

SgStatement *genInvokeCall(SgStatement *s, 
			   SgExpression *child_vc_offset,
			   SgVarRefExp *fm_pblock)
{
    char                    *prosName;
    char                    *callName;
    SgSymbol                *callSymb;
    SgCallStmt              *callStmt;
    int                     n_args;

    prosName = (s->symbol())? s->symbol()->identifier() : "unknown_pros";

    if (s->variant() == PROS_STAT) {
	SgProsCallStmt *prosStmt = isSgProsCallStmt(s);

	n_args = prosStmt->numberOfArgs();
    } else if (s->variant() == PROS_STAT_LCTN) {
	SgProsCallLctn *prosStmt = isSgProsCallLctn(s);

	n_args = prosStmt->numberOfArgs();
    } else if (s->variant() == PROS_STAT_SUBM) {
	SgProsCallSubm *prosStmt = isSgProsCallSubm(s);

	n_args = prosStmt->numberOfArgs();
    } else {
	n_args = 0;
    }

    callName = ppString("fm_", prosName, 
			ppString("_", num2str(n_args),"_invoke"));
    callSymb = new SgSymbol(PROCEDURE_NAME,callName);
    callStmt = new SgCallStmt(*callSymb);
    
    if (s->variant() == PROS_STAT) {
	SgProsCallStmt *prosStmt = isSgProsCallStmt(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_LCTN) {
	SgProsCallLctn *prosStmt = isSgProsCallLctn(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_SUBM) {
	SgProsCallSubm *prosStmt = isSgProsCallSubm(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else {
	fm_internal_error("\
Dont know what porsstatement %s is\n", getVariantByName(s->variant()));
    }
    callStmt->addArg(*child_vc_offset);
    callStmt->addArg(*fm_pblock);
    return (callStmt);
}


SgStatement *genProcPackArgsCall(SgStatement *s, 
			     SgVarRefExp *fm_buffer)
{
    char *prosName       = (s->symbol())->identifier();
    int  packCallNameLen = 3+strlen(prosName)+10;
    char *packCallName   = (char *)malloc(packCallNameLen+1);
    SgSymbol *packCallSymb;
    SgCallStmt *callStmt;

    assert(packCallName);
    sprintf(packCallName,"fm_%s_pack_args", prosName);
    packCallSymb = new SgSymbol(PROCEDURE_NAME,packCallName);
    callStmt = new SgCallStmt(*packCallSymb);

    if (s->variant() == PROS_STAT) {
	SgProsCallStmt *prosStmt = isSgProsCallStmt(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_LCTN) {
	SgProsCallLctn *prosStmt = isSgProsCallLctn(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_SUBM) {
	SgProsCallSubm *prosStmt = isSgProsCallSubm(s);
	for (int i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else {
	// fm_interal_error()
	fprintf(stderr,
		"Dont know what portstament %s\n", 
		getVariantByName(s->variant()));
    }
    callStmt->addArg(*fm_buffer);
    return (callStmt);
}

SgStatement *genProcSpawnSetupCall(SgStatement *s, 
                                   SgVarRefExp *fm_pblock)
{
    int i;
    char *prosName = (s->symbol())->identifier();
    int  spawnCallNameLen = 3 + strlen(prosName) + 12;
    char *spawnCallName   = (char *)malloc(spawnCallNameLen+1);
    SgSymbol *spawnCallSymb;
    SgCallStmt *callStmt;

    assert(spawnCallName);
    sprintf(spawnCallName, "fm_%s_spawn_setup", prosName);
    spawnCallSymb = new SgSymbol(PROCEDURE_NAME,spawnCallName);
    callStmt = new SgCallStmt(*spawnCallSymb);

    if (s->variant() == PROS_STAT) {
	SgProsCallStmt *prosStmt = isSgProsCallStmt(s);

	for (i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_LCTN) {
	SgProsCallLctn *prosStmt = isSgProsCallLctn(s);

	for (i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else if (s->variant() == PROS_STAT_SUBM) {
	SgProsCallSubm *prosStmt = isSgProsCallSubm(s);

	for (i = 0; i < prosStmt->numberOfArgs(); i++) {
	    callStmt->addArg(*prosStmt->arg(i));
	}
    } else {
	// fm_internal_error
	fprintf(stderr, "dont know what port this is\n");
    }

    callStmt->addArg(*fm_pblock);
    return(callStmt);
}
    
SgStatement *genPblockInitCall(int size, SgVarRefExp *fm_pblock)
{
    SgValueExp *n = new SgValueExp(size);
    SgCallStmt *callStmt = new SgCallStmt(*fmPblockInitCallSymb);
    
    callStmt->addArg(*n);
    callStmt->addArg(*fm_pblock);

    return(callStmt);
}

SgStatement *genPblockSpawnCall(SgVarRefExp        *fm_vector,
				SgValueExp         *fm_vector_length,
				SgVarRefExp        *fm_pblock,
				SgValueExp	   *fm_pname)
{
    SgCallStmt *callStmt = new SgCallStmt(*fmPblockSpawnCallSymb);
    
    callStmt->addArg(*fm_vector);
    callStmt->addArg(*fm_vector_length);
//    callStmt->addArg(*fm_should_share_func_call);
    callStmt->addArg(*fm_pblock);
    callStmt->addArg(*fm_pname);

    return (callStmt);
}

SgStatement *genPblockResetCall(SgVarRefExp *fm_pblock)
{
    SgCallStmt *callStmt = new SgCallStmt(*fmPblockResetCallSymb);

    callStmt->addArg(*fm_pblock);
    
    return(callStmt);
}


SgStatement *genPblockInitInvokeCall(SgVarRefExp *fm_pblock,
				 SgVarRefExp *fm_buffer,
                                 SgSymbol    *procSymb)
{
    SgCallStmt *callStmt = new SgCallStmt(*fmPblockInitInvokeCallSymb);
    int startStrLen      = 8 + strlen(procSymb->identifier());
    char *startStr        = (char *) malloc(startStrLen + 1);   
    SgValueExp *startRef;

    assert(startStr);
    
    callStmt->addArg(*fm_pblock);
    callStmt->addArg(*fm_buffer);
    sprintf(startStr, "FMStart_%s", procSymb->identifier());
    startRef = new SgValueExp(startStr);
    callStmt->addArg(*startRef);

    return(callStmt);
}

SgStatement *genPblockInvokeCall(SgVarRefExp *fm_pblock,
				 SgVarRefExp *fm_buffer)
{
    SgCallStmt *callStmt = new SgCallStmt(*fmPblockInvokeCallSymb);
    
    callStmt->addArg(*fm_pblock);
    callStmt->addArg(*fm_buffer);
    
    return(callStmt);
}


SgStatement *genPblockInitCall(SgVarRefExp *size,
                               SgVarRefExp *fm_pblock)
{
    SgCallStmt *callStmt = new SgCallStmt(*fmPblockInitCallSymb);

    callStmt->addArg(*size);
    callStmt->addArg(*fm_pblock);

    return(callStmt);
}

SgStatement *genPblockWaitCall(SgVarRefExp *fm_pblock)
{
    SgCallStmt *callStmt = new SgCallStmt(*fmPblockWaitCallSymb);

    callStmt->addArg(*fm_pblock);

    return(callStmt);
}

#if 0
void doProsCallStmt(SgStatement *s, 
		    SgStatement *where,
		    int phase, 
		    SgVarRefExp *fm_pblock, 
		    SgVarRefExp *fm_buffer,
		    SgVarRefExp *fm_ports)
#endif /* 0 */

void doProsCallStmt(SgStatement *s, 
		    SgStatement *where,
		    int phase, 
		    SgVarRefExp *fm_pblock)
{
    SgSymbol              *procSymb;
    char                  *prosName;
    //
    // Variables for generating fm_vector initialization
    //
    SgSymbol              *fm_vector;
    SgArrayRefExp         *fm_vector_ref;
    SgVariableSymb        *fm_vector_name_symbol;
    SgVarRefExp           *fm_vector_name_ref;
    SgAssignStmt          *assignStmt;
    SgArrayType           *arrayType;
    SgProsCallLctn        *lctnCall;
    SgProsCallSubm        *submCall;
    char                  *fm_vector_string = encodeStr(EC_VEC);
    SgValueExp            *fm_vector_range_v_exp;
    SgValueExp		  *fm_pname;
    //
    // variables used for generating 
    // fm_<proc>_should_share() function calls
    //
    char                  *funcName;
    SgSymbol              *funcSymb;
    SgFunctionCallExp     *funcCall;

    int i;

    static SgExprListExp  *vec_pool = 0;
    static int which_vec = 1;

    if (! vec_pool) {
	vec_pool = new SgExprListExp(*c1);
    }
	
    procSymb              = s->symbol();
    prosName              = (procSymb) ? procSymb->identifier():"unknown_proc";
    arrayType             = new SgArrayType(*integer_t);
    fm_vector_name_symbol = new SgVariableSymb(fm_vector_string);
    fm_vector_name_ref    = new SgVarRefExp(*fm_vector_name_symbol);

    
    if (phase == 0) {
#if OLD_VERSION_RUN_TIME
	where->insertStmtBefore(*genProcSpawnSetupCall(s,fm_pblock));
#endif /* OLD_VERSION_RUN_TIME */
	if (lctnCall = isSgProsCallLctn(s)) {
	    SgExprListExp *lctn     = isSgExprListExp(lctnCall->location());
	    SgProcessorsRefExp *vec = isSgProcessorsRefExp(lctn->elem(0));
	    int fm_vector_range     = vec->numberOfSubscripts() + 1;
	    fm_vector_range_v_exp   = new SgValueExp(fm_vector_range);

	    arrayType->addRange(*fm_vector_range_v_exp);

	    fm_vector = new SgVariableSymb(fm_vector_string, *arrayType);
	    fm_vector->declareTheSymbol(*currentProcedure);

	    vec_pool->append(*fm_vector_name_ref);
	    vec_pool->append(*fm_vector_range_v_exp);

	    if (opt_use_auto_vars) 
	      declare_auto_symb(fm_vector_string, currentProcedure);
	    
	    fm_vector_ref = new SgArrayRefExp(*fm_vector, *c1);
	    assignStmt = new SgAssignStmt(*fm_vector_ref, *fmVecLctnValRef);
	    where->insertStmtBefore(*assignStmt);
	    
	    for (i = 0; i < vec->numberOfSubscripts(); i++) {
		SgValueExp *ind = new SgValueExp(i+2);
		fm_vector_ref = new SgArrayRefExp(*fm_vector, *ind);
		assignStmt = new SgAssignStmt(*fm_vector_ref, 
					      *vec->subscript(i));
		where->insertStmtBefore(*assignStmt);
	    }

	} else if (submCall = isSgProsCallSubm(s)) {
	    SgExprListExp *subm     = isSgExprListExp(submCall->submachine());
	    SgProcessorsRefExp *vec = isSgProcessorsRefExp(subm->elem(0));
	    int fm_vector_range     = vec->numberOfSubscripts() + 1;

	    fm_vector_range_v_exp = new SgValueExp(fm_vector_range);
	    arrayType->addRange(*fm_vector_range_v_exp);

	    fm_vector = new SgVariableSymb(fm_vector_string, *arrayType);
	    fm_vector->declareTheSymbol(*currentProcedure);
	    
	    vec_pool->append(*fm_vector_name_ref);
	    vec_pool->append(*fm_vector_range_v_exp);

	    fm_vector_ref = new SgArrayRefExp(*fm_vector, *c1);
	    assignStmt = new SgAssignStmt(*fm_vector_ref, *fmVecSubmValRef);
	    where->insertStmtBefore(*assignStmt);

	    for (i = 0; i < vec->numberOfSubscripts(); i++) {
		SgValueExp *ind = new SgValueExp(i+2);
		fm_vector_ref = new SgArrayRefExp(*fm_vector, *ind);
		assignStmt = new SgAssignStmt(*fm_vector_ref, 
					      *vec->subscript(i));
		where->insertStmtBefore(*assignStmt);
	    }
	} else {
	    int fm_vector_range = 1;
	    
	    fm_vector_range_v_exp = new SgValueExp(fm_vector_range);
	    arrayType->addRange(*fm_vector_range_v_exp);

	    fm_vector = new SgVariableSymb(fm_vector_string, *arrayType);
	    fm_vector->declareTheSymbol(*currentProcedure);
	    
	    vec_pool->append(*fm_vector_name_ref);
	    vec_pool->append(*fm_vector_range_v_exp);

	    fm_vector_ref = new SgArrayRefExp(*fm_vector, *c1);
	    assignStmt = new SgAssignStmt(*fm_vector_ref, *fmVecNoAnnValRef);
	    where->insertStmtBefore(*assignStmt);
	}

	//
	// function name = "fm_" . $proc . "_should_share"
	//
	funcName = (char *)malloc(3 + strlen(prosName) + 14);
	if (! funcName) {
	    fm_internal_error("\
Unable to allocate memory at %s:%d", __FILE__, __LINE__);
	}
	sprintf(funcName, "fm_%s_should_share", prosName);
	
	funcSymb = new SgSymbol(FUNCTION_NAME,funcName);
	funcSymb->setType(*integer_t);
	//
	// check whether this symbol has
	// been declared, if not declare it
	//
	if (! stmtMapGet(prosMap,funcName)) {
	    funcSymb->declareTheSymbol(*currentProcedure);
	    stmtMapSet(prosMap,funcName,currentProcedure);
	}
	
	funcCall = new SgFunctionCallExp(*funcSymb);
	char *s = str2lower(prosName);
	fm_pname = new SgValueExp(s);
//	free(s);

	where->insertStmtBefore(*genPblockSpawnCall(fm_vector_name_ref,
						    fm_vector_range_v_exp,
						    fm_pblock,
						    fm_pname));
    } else if (phase == 1) {
#if OLD_VERSION_RUN_TIME
	where->insertStmtBefore(*genPblockInitInvokeCall(fm_pblock,
							 fm_buffer,
							 procSymb));
	where->insertStmtBefore(*genProcPackArgsCall(s, 
						      fm_buffer));
	where->insertStmtBefore( *genPblockInvokeCall(fm_pblock,
						      fm_buffer));
#else /* OLD_VERSION_RUN_TIME */
	
	where->insertStmtBefore( *genPortInitCall(s, fm_pblock) );

#endif /* OLD_VERSION_RUN_TIME */
    } else if (phase == 2) {
	SgCallStmt *call;
	
	//
	// Generate calls to FM_CHILD_VC_OFFSET()
	//
	call = new SgCallStmt(*fmChildVCOffsetCallSymb);
	call->addArg(*vec_pool->elem(which_vec));
	call->addArg(*vec_pool->elem(which_vec+1));
	call->addArg(*fm_vector_name_ref);

	fm_vector_name_symbol->setType(*integer_t);
	fm_vector_name_symbol->declareTheSymbol(*currentProcedure);
	where->insertStmtBefore(*call);

#ifdef FM_PABLO_CALLER_INFO
	//
	// Generate Call to FM_PABLO_CALLER_INFO()
	//
	// Args: (process_name, caller_name, file_name, line_number)
	//
	if (opt_debug) {
	    SgValueExp *arg;

	    call = new SgCallStmt(*fmPabloCallerInfoSymb);
	    arg = new SgValueExp(prosName);
	    call->addArg(*arg);
	    arg = new SgValueExp(currentProcedureName);
	    call->addArg(*arg);
	    call->addArg(*currentFileNameExp);
	    arg = new SgValueExp(s->lineNumber());
	    call->addArg(*arg);
	    
	    where->insertStmtBefore(*call);
	}
#endif /* FM_PABLO_CALLER_INFO */
	//
	// Generate FM_<PROC>_<N_ARGS>_INVOKE() call
	// 
	where->insertStmtBefore( *genInvokeCall(s, fm_vector_name_ref,
						fm_pblock) );
	which_vec += 2;
    } else {
	// fm_error
	fprintf(stderr, "Huh, phase = %d\n", phase);
    }
}




