// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

// Hacked from the instrument code by  Ming Q. Xu

/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <stdio.h>
#ifndef NeXT
#include <malloc.h>
#endif
#include <stdlib.h>
#include <stdarg.h>
#define DISTRIBUTION
#include "sage++user.h"

#if defined(__GNUC__)
extern "C" {
    int getopt(int, char **, char *);
}
#endif /* __GNUC__ */

#define MAXSUBS 100

struct ftable {
  char *function_name;
  SgStatement *function;
} ftable;

struct ftable function_table[MAXSUBS];
int totalfunctions;
static int event = 0;
SgProject *project;

SgType      *auto_t;
SgStatement *currentProcedure;
SgValueExp  *currentProcedureName;
SgSymbol    *sentinel_symbol;
SgVarRefExp *sentinel_reference;
SgDeclarationStatement *external_statement;
SgSymbol    *first_var, *last_var;
SgExpression *first_v_ref, *last_v_ref, *last_v_list;
SgStatement *elseIf, *elseIfCP;

// Command Line arguments

int         opt_use_auto_vars;       /* -a */

// Fortran program transformations

SgType *SgTypeInt();

SgSymbol *entry, *eexit, *loop_entry, *loop_exit;
char *fnames[] = {"sentinel_push", 
		  "sentinel_pop", 
		  "loop_entry_event", 
		  "loop_exit_event"};

#define ENTRY_EVENT 0
#define EXIT_EVENT  1
#define LOOP_ENTRY  2
#define LOOP_EXIT   3

#define AFTER  0
#define BEFORE 1

void declare_auto_symb(char *, SgStatement *);

// Initialize the function symbols (used in the instrumentation)
void 
InitSymbols(SgProject *p)
{
  SgFile *f;
  SgExpression *call_reference;

  f = &(p->file(0));
  entry = new SgFunctionSymb(PROCEDURE_NAME,fnames[0],
			     *SgTypeInt(),*(f->firstStatement()));
  eexit  = new SgFunctionSymb(PROCEDURE_NAME,fnames[1],
			     *SgTypeInt(), *(f->firstStatement()));
  loop_entry= new SgFunctionSymb(PROCEDURE_NAME,fnames[2],
			      *SgTypeInt(), *(f->firstStatement()));
  loop_exit = new SgFunctionSymb(PROCEDURE_NAME,fnames[3],
			     *SgTypeInt(), *(f->firstStatement()));
  sentinel_symbol = new SgVariableSymb("fm__sentinel_var", *SgTypeInt());
  sentinel_reference = new SgVarRefExp(*sentinel_symbol);

  external_statement = new SgDeclarationStatement(EXTERN_STAT);
  call_reference = new SgVarRefExp(*entry);
  external_statement->addVar(*call_reference);
  call_reference = new SgVarRefExp(*eexit);
  external_statement->addVar(*call_reference);
  first_var = new SgVariableSymb("fm__first_var", *SgTypeInt());
  first_v_ref = new SgVarRefExp(*first_var);
  last_var = new SgVariableSymb("fm__last_var", *SgTypeInt());
  last_v_ref = new SgVarRefExp(*last_var);
  last_v_list = new SgExprListExp(*last_v_ref);
}


// Initialize the function table (used for searching for functions).
void
InitFunctionTable(SgProject *project)
{

  int Files = project->numberOfFiles();
  int k = 0;

  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;
    
    f = &(project->file(i));
    
    int num_routines;
    
    num_routines = f->numberOfFunctions();
    for (int j = 0; j < num_routines; j++){
      SgStatement *sub;
      SgSymbol *subsym;

      sub = function_table[k].function = f->functions(j);
      subsym = sub->symbol();
      function_table[k++].function_name = subsym->identifier();
#if 0
      printf("Function %d's name is %s\n", i*Files+j, subsym->identifier());
#endif
    }
  }

  totalfunctions = k;
}


// Inserting a call to the instrumentation library
void InsertFCallNode(int event_type, SgStatement *s, int how)
{
  SgSymbol *symb;
       
  switch (event_type) {
    case ENTRY_EVENT: symb = entry;
                    break;
    case EXIT_EVENT:  symb = eexit;
                    break;
    case LOOP_ENTRY:  symb = loop_entry;
                    break;
    case LOOP_EXIT :  symb = loop_exit;
                    break;
  }


  SgCallStmt *c;
  SgValueExp *ev;

  c = new SgCallStmt(*symb);
//  ev = new SgValueExp(event);
//  c->addArg(*ev);
  c->addArg(*sentinel_reference);
  c->addArg(*currentProcedureName);
  c->addArg(*first_v_ref);
  c->addArg(*last_v_ref);
//  delete(ev);

//  event++;

  switch (how) {
    case BEFORE: 
         s->insertStmtBefore(*c);
         break;
    case AFTER:
         s->insertStmtAfter(*c);
         break;
  }
}


// Fixing the loops in the subroutine (massaging the program)
void 
FixLoops(SgStatement* s)
{
  SgStatement *br;
  SgForStmt *f;

  for (br = s->lexNext(); ; br = br->lexNext()) {
    if (f = isSgForStmt(br))
      if (! f->isEnddoLoop())
	f->convertLoop();
   if (br == s->lastNodeOfStmt())
     break;
    
  }
}

// Instrumentation running within a subroutine.
void 
FInstrumentSub(SgStatement* s)
{
    SgStatement *br;
    
    for (br = s->lexNext(); ; br = br->lexNext()) {
//	printf("line %d: variant=%d\n", br->lineNumber(), br->variant());
	switch (br->variant()){
#if 0
	  case FOR_NODE:
	    InsertFCallNode(LOOP_ENTRY, br, BEFORE);
	    InsertFCallNode(LOOP_EXIT, br->lastNodeOfStmt(), AFTER);
	    break;
#endif /* 0 */
	  case ENTRY_STAT:
	    InsertFCallNode(ENTRY_EVENT, br, AFTER);
	    break;
	  case RETURN_STAT:
	  case EXIT_STMT:
	    if (br != s->lastNodeOfStmt()) {
		InsertFCallNode(EXIT_EVENT, br, BEFORE);
	    }
	    break;
	  default:
	    break;
	}

	if (br == s->lastNodeOfStmt()) 
	  break;
    }
}

// Basic initialization
extern "C"
void
FInitialize(SgProject *project)
{
  InitSymbols(project);
  InitFunctionTable(project);
}

// Instrumentation call (search for function symbol based on function name)
extern "C"
void
FInstrument(char *name)
{
  SgStatement *sub;
  int i;

  for(i=0; i < totalfunctions; i++) {
    if (strcmp(name, function_table[i].function_name) == 0)
      break;
  }

  sub = function_table[i].function;
  FixLoops(sub);
  InsertFCallNode(ENTRY_EVENT, sub->lastDeclaration(), AFTER);
  InsertFCallNode(EXIT_EVENT, sub->lastExecutable(), AFTER);
  FInstrumentSub(function_table[i].function);
}

extern "C"
void
ProjectUnparse(SgProject *project)
{
  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;
    
    f = &(project->file(i));
    f->unparsestdout();
  }
}

extern "C"
void
UnparseSub(char *name){
  SgStatement *sub;
  int i;

  for(i=0; i < totalfunctions; i++) {
    if (strcmp(name, function_table[i].function_name) == 0)
      break;
  }

  sub = function_table[i].function;
  sub->unparsestdout();
}


// needs to be filled in (used for instrumentation for animation)
void
VistaInstrument(SgStatement *s)
{
  SgStatement *br;
  SgExprListExp *l;
  SgExpression *e;
  SgExpression *a;

  for (br = s->lexNext(); ; br = br->lexNext()) {
    for (l = isSgExprListExp(br->expr(0)); l ; l = l->next()){
      e = l->value();
    }  
 
   if (br == s->lastNodeOfStmt())
     break;
  }  
}

// Instrumenting the entire program.


void
FAnalyze(SgProject *project)
{
  char s[32];
  SgDeclarationStatement *decl;


//  printf ("Project has %d number of files\n", project->numberOfFiles());
  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;

    f = &(project->file(i));
//    printf("Source file %s has", project->fileName(i));
    
    int num_routines;

    num_routines = f->numberOfFunctions();
//    printf ("%d routines\n", num_routines);
    for (int j = 0; j < num_routines; j++) {
      SgStatement *sub;
      SgStatement *exe;

      sub = f->functions(j);
      elseIf = 0;
      // keep track of where we are

      currentProcedure = sub;
      currentProcedureName = new SgValueExp(sub->symbol()->identifier());

      if (currentProcedure->variant() == PROG_HEDR) {
	  continue;
      }
      // generate local variable 
      sentinel_symbol->declareTheSymbol(*currentProcedure);

      if (opt_use_auto_vars) 
	declare_auto_symb(sentinel_symbol->identifier(), currentProcedure);

      exe = sub->lastDeclaration();

      InsertFCallNode(ENTRY_EVENT, exe, AFTER);
//      FixLoops(sub);
      FInstrumentSub(sub);

      if (((sub->lastExecutable())->controlParent())->variant() == ELSEIF_NODE)
      {
	  elseIf = sub->lastExecutable()->controlParent();
	  elseIfCP = elseIf->controlParent();
	  elseIf->setVariant(IF_NODE);
	  elseIf->setControlParent(*sub);
	  InsertFCallNode(EXIT_EVENT, sub->lastExecutable(), AFTER);
      } else if (sub->lastExecutable()->variant() != RETURN_STAT &&
		 sub->lastExecutable()->variant() != EXIT_STMT)
      {
	  InsertFCallNode(EXIT_EVENT, sub->lastNodeOfStmt(), BEFORE);
      }

      first_var->declareTheSymbol(*currentProcedure);

      if (opt_use_auto_vars) 
	declare_auto_symb(first_var->identifier(), currentProcedure);

      decl = new SgVarDeclStmt(*last_v_list, *SgTypeInt());

      exe = sub->lastDeclaration();
      while (exe) {
	  if (exe->variant() != DATA_DECL &&
	      exe->variant() != STMTFN_STAT) {
	      break;
	  }
	  exe = exe->lexPrev();
      }
      if (! exe) {
	  fprintf(stderr, "Expect an non-null statement pointer %s:%d\n",
		  __FILE__, __LINE__);
	  exit(1);
      }

      exe->insertStmtBefore(external_statement->copy());
      exe->insertStmtAfter(*decl);
      
      if (opt_use_auto_vars)
	declare_auto_symb(last_var->identifier(), currentProcedure);

      if (elseIf) {
	  elseIf->setVariant(ELSEIF_NODE);
	  elseIf->setControlParent(*elseIfCP);
      }
    }
    f->unparsestdout();
//    printf ("\n Done unparsing\n");
#if 0
    f->saveDepFile(sprintf(s,"debug%d.dep",i));
#endif
  }
#if 0
  for (i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;

    f = &(project->file(i));
    printf("Source file %s has", project->fileName(i));
    
    int num_routines;

    num_routines = f->numberOfFunctions();
    printf ("%d routines\n", num_routines);
    for (int j = 0; j < num_routines; j++) {
      SgStatement *sub;

      sub = f->functions(j);
      VistaInstrument(sub);
    }
    f->unparsestdout();
    printf ("\n Done unparsing\n");

    f->saveDepFile(sprintf(s,"debug%d.dep",i));

  }
#endif /* 0 */
}

int main(int argc, char **argv)
{
    SgProject *project;
    int optionChar;
    int  i = 1;

    while ((optionChar = getopt(argc, argv,"a")) != EOF) 
    {
	switch (optionChar) {
	  case 'a':
	    opt_use_auto_vars = 1;
	    i++;
	    break;
	  default:
	    i++;
	    break;
	}
    }

    if (i >= argc) {
	fprintf(stderr, "\
Usage: %s [-a] file\n", argv[0]);
	exit (1);
    } 

    project = new SgProject(argv[i]);

    auto_t = new SgType(T_AUTO);
    InitSymbols(project);
    InitFunctionTable(project);
    FAnalyze(project);
    return 0;
}


// pC++ program transformations.

SgSymbol *timer_clear, *timer_start, *timer_stop, *timer_elapsed;

char *cnames[] = {"Test_CMMD_node_timer_clear",
		  "Test_CMMD_node_timer_start",
		  "Test_CMMD_node_timer_stop",
		  "Test_CMMD_node_timer_elapsed"};

#define TIMER_CLEAR 0
#define TIMER_START 1
#define TIMER_STOP  2
#define TIMER_ELAPSED   3


// Basic help routines (all obtained from the compiler)
// Should be made part of the library.
SgSymbol *
isReferenceToMethodOfElement(SgExpression *expr)
{
   SgVarRefExp *varef;
   SgFunctionCallExp *fcall;
   SgSymbol *symb;
   SgMemberFuncSymb *symbm;
   SgFieldSymb *symbf;
   if (!expr)
     return NULL;
   if (varef = isSgVarRefExp(expr))
     {
       symb = varef->symbol();
     }
   if (fcall = isSgFunctionCallExp(expr))
     {
       symb = fcall->symbol();
     }
   if (symbm = isSgMemberFuncSymb(symb))
     {
       if (symbm->isMethodOfElement())
         return symbm;
       else
         return NULL;
     }
   if (symbf = isSgFieldSymb(symb))
     {
       if (symbf->isMethodOfElement())
         return symbf;
       else
         return NULL;
     }
   return 0;
}

SgSymbol *
isReferenceToClassOfElement(SgDerivedCollectionType *type,SgExpression *expr)
{
  SgVarRefExp *varef;
  SgFunctionCallExp *fcall;
  SgSymbol *symb,*elsymb;
  SgMemberFuncSymb *symbm;
  SgFieldSymb *symbf;
  SgType *elmetype;
  SgDerivedType *eltype;

   if (!expr || !type)
     return NULL;
  elmetype = type->elementClass();
  if (!elmetype)
    return NULL;
  if (elmetype->isTheElementType())
    return NULL;
  
  if (eltype = isSgDerivedType(elmetype))
    {
      if (!(elsymb = eltype->typeName()))
        {
          return NULL;
        }
    } else
      return NULL;

   if (varef = isSgVarRefExp(expr))
     {
       symb = varef->symbol();
     }
   if (fcall = isSgFunctionCallExp(expr))
     {
       symb = fcall->symbol();
     }

   if (symbm = isSgMemberFuncSymb(symb))
     {
       if (symbm->className() == elsymb)
         return symbm;
       else
         return NULL;
     }
   if (symbf = isSgFieldSymb(symb))
     {
       if (symbf->structureName() == elsymb)
         return symbf;
       else
         return NULL;
     }
   return NULL;
}


int
inMethodOfTheElement(SgStatement *where)
{
  SgFuncHedrStmt *fh;
  SgMemberFuncSymb *symb;

  if (!where)
    return 0;

  if (where->variant() == GLOBAL)
    return 0;

  if (fh = isSgFuncHedrStmt(where))
    {
      if (symb = isSgMemberFuncSymb(fh->symbol()))
        {
          return symb->isMethodOfElement();
        }
    }
  return inMethodOfTheElement(where->controlParent());
}


SgFuncHedrStmt *
whichFunctionAmI(SgStatement *where)
{
  SgFuncHedrStmt *fh;
  SgMemberFuncSymb *symb;

  if (!where)
    return NULL;

  if (where->variant() == GLOBAL)
    return NULL;

  if (fh = isSgFuncHedrStmt(where))
    {
      return fh;
    }
  return whichFunctionAmI(where->controlParent());
}


//
// where is used to disambiguate this
//
SgDerivedCollectionType *
isReferenceToCollection(SgExpression *expr, SgStatement *where)
 {
   SgType *type = NULL;
   SgExpression *lhs;
   SgVarRefExp *varef;
   SgPointerType *tpt;
   SgReferenceType *tref;
   SgDerivedCollectionType *dstype;
   SgArrayRefExp *array;
   SgSymbol *symb;
   SgFunctionCallExp *func;

   if (!expr)
     return NULL;
   switch (expr->variant())
     {
     case RECORD_REF:
     case POINTST_OP:
     case DEREF_OP:
       lhs = expr->lhs();
       break;
     case VAR_REF:
     case ARRAY_REF:
     case FUNC_CALL:
       lhs = expr;
       break;
       default :
// default should be to try to generate a copy for the expression
// and see if expression becomes call the function again.
         return NULL;
     }
   if (!lhs)
     return NULL;
   if (varef = isSgVarRefExp(lhs))
     {
       if (varef->symbol())
         type = varef->symbol()->type();
     } else
       if (isSgThisExp(lhs))
         {
           if (inMethodOfTheElement(where))
             return 0; // this refer to the element 
           type = lhs->type();
         } else
           if (array = isSgArrayRefExp(lhs))
             {
               symb = array->symbol();
               if (symb)
                 type = symb->type();
               if (type)
                 type = type->internalBaseType();
             } else  
               if (func = isSgFunctionCallExp(lhs))
                 {
                   symb = func->symbol();
                   if (symb)
                     type = symb->type();
                 } else
                   if  (isSgPointerDerefExp(lhs))
                     {
                       type = isReferenceToCollection(lhs,where);
                     } else
                       if  (isSgExprListExp(lhs))
                         {
                           type = isReferenceToCollection(lhs->lhs(),where);
                         } else
                           return NULL;
   if (!type)
     return NULL;
   switch (expr->variant())
     {
     case FUNC_CALL:
     case ARRAY_REF:
     case VAR_REF:
     case RECORD_REF:
       if (dstype = isSgDerivedCollectionType(type))
         return dstype;
       if (tref = isSgReferenceType(type))
         {
           type = tref->baseType();
           if (dstype = isSgDerivedCollectionType(type))
             return dstype;
         }
       break;
     case DEREF_OP:
     case POINTST_OP: 
       if (tpt = isSgPointerType(type))
         {
           type = tpt->baseType();
           if (dstype = isSgDerivedCollectionType(type))
             return dstype;
           if (tref = isSgReferenceType(type))
             {
               type = tref->baseType();
               if (dstype = isSgDerivedCollectionType(type))
                 return dstype;
             }
         }
       break;
     }
   return NULL;
 }


// Simple routines to look into a pC++ program.
void 
ListCollections(SgFile *file)
{
  SgStatement *s;
  SgCollectionStmt *c;

  for (s = file->firstStatement(); s ; s = s->lexNext()){
    if (c = isSgCollectionStmt(s)){
      c->unparsestdout();
    }
  }
}

void 
ListCollectionInstances(SgFile *file)
{
  SgType *type;
  SgDerivedCollectionType *c;

  for (type = file->firstType(); type ; type = type->next()){
    if (c = isSgDerivedCollectionType(type)){
      SgStatement *s;

      printf("%s<%s>\n",c->collectionName()->identifier(),
	               c->elementClass()->symbol()->identifier());
    }
  }
}


void
ListCollectionInvocations(SgFile *f)
{
  SgStatement *br;
  SgExprListExp *l;
  SgExpression *e;
  SgStatement *s;
  SgDerivedCollectionType *a;
  SgSymbol *b;

  s = f->firstStatement();
  for (br = s; br; br = br->lexNext())
    for (l = isSgExprListExp(br->expr(0)); l ; l = l->next()){
      e = l->value();
      switch (e->variant()) {
        case DEREF_OP:
        case RECORD_REF:
        case POINTST_OP:
	  a = isReferenceToCollection(e, br);
	  b = isReferenceToMethodOfElement(e->rhs());
	  if (a && !b)
	    b = isReferenceToClassOfElement(a, e->rhs());
	  if (a && b)
	    br->unparsestdout();
      }
    }
}

void
CAnalyze(SgProject *project){

  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;
    
    f = &(project->file(i));
    ListCollections(f);
    ListCollectionInstances(f);
    printf("Source file %s has ", project->fileName(i));
    
    int num_routines;
    
    num_routines = f->numberOfFunctions();
    printf ("%d routines\n", num_routines);
    ListCollectionInvocations(f);
  }
}


// Initializing the timer function symbols.
void CInitSymbols(SgProject *p)
{
  SgFile *f;

  f = &(p->file(0));
  timer_clear = new SgFunctionSymb(FUNCTION_NAME,cnames[0],
				   *SgTypeInt(),*(f->firstStatement()));
  timer_start  = new SgFunctionSymb(FUNCTION_NAME,cnames[1],
				    *SgTypeInt(), *(f->firstStatement()));
  timer_stop = new SgFunctionSymb(FUNCTION_NAME,cnames[2],
				  *SgTypeInt(), *(f->firstStatement()));
  timer_elapsed = new SgFunctionSymb(FUNCTION_NAME,cnames[3],
				     *SgTypeInt(), *(f->firstStatement()));
}

extern "C"
void
CInitialize(SgProject *project)
{
  CInitSymbols(project);
  InitFunctionTable(project);
}


// Insert a C function call statement
void 
InsertCCallNode(int event_type, SgStatement *s, int how)
{
  SgSymbol *symb;
       
  switch (event_type) {
    case TIMER_CLEAR: symb = timer_clear;
                      break;
    case TIMER_START: symb = timer_start;
                      break;
    case TIMER_STOP:  symb = timer_stop;
                      break;
    case TIMER_ELAPSED: symb = timer_elapsed;
                       break;
  }

  SgFunctionCallExp *func;
  SgExprListExp *exp;
  SgCExpStmt *c;

  func = new SgFunctionCallExp(*symb);
  exp = new SgExprListExp();
  exp->setValue(*func);
  c = new SgCExpStmt (*exp);


  switch (how) {
    case BEFORE: 
         s->insertStmtBefore(*c);
         break;
    case AFTER:
         s->insertStmtAfter(*c);
         break;
  }
}

// Instrument a pC++ function for timing collection invocations.

void
CTimingInstrumentSub(SgStatement *s){

  SgStatement *br;
  SgExprListExp *l;
  SgExpression *e;
  SgDerivedCollectionType *a;
  SgSymbol *b;

  for (br = s->lastDeclaration(); ; br = br->lexNext()) {
    for (l = isSgExprListExp(br->expr(0)); l ; l = l->next()){
      e = l->value();

      switch(e->variant()){
      case DEREF_OP:
      case POINTST_OP:
      case RECORD_REF:
	a = isReferenceToCollection(e, br);
	b = isReferenceToMethodOfElement(e->rhs());
	if (a && !b)
	  b = isReferenceToClassOfElement(a, e->rhs());

	if (a && isSgCExpStmt(br)) {
	  InsertCCallNode(TIMER_START, br, BEFORE);
	  InsertCCallNode(TIMER_STOP, br, AFTER);
          InsertCCallNode(TIMER_ELAPSED, br->lexNext(), AFTER);
	}
      }
    }
    if (br == s->lastNodeOfStmt())
      break;
  }
}

// Instrumentation based on function name.

extern "C"
void
CTimingInstrument(char *name){
  int i;

  for(i=0; i < totalfunctions; i++) {
    if (strcmp(name, function_table[i].function_name) == 0)
      break;
  }

  CTimingInstrumentSub(function_table[i].function);
}

  
extern "C" void OpenProject(char *proj); 

void OpenProject(char *proj) {
  project = new SgProject (proj);
}

#define TRUE  1
#define FALSE 0

extern "C"
int
IsFortran() {
  SgFile *file;

  file = &(project->file(0));
  if (file->languageType() == ForSrc)
    return TRUE;
  else 
    return FALSE;
}

extern "C"
int
IsC() {
  SgFile *file;

  file = &(project->file(0));
  if (file->languageType() == CSrc)
    return TRUE;
  else 
    return FALSE;
}

void declare_auto_symb(char *name, SgStatement *scope)
{
    SgSymbol *symb;

    symb = new SgVariableSymb(name, *auto_t);
    symb->declareTheSymbol(*scope);
}
