// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"

SgStatement * transform_type_length_operator(SgVarDeclStmt *);

void doVarDeclStmt(SgStatement *s)
{
    int           i;
    SgVarDeclStmt *varDeclStmt = isSgVarDeclStmt(s);
    SgExpression  *theVar;
    SgType        *theType;

    fm_assert(varDeclStmt->numberOfVars() == 1);

    // check whether the var is in the common block or not
    theVar = varDeclStmt->var(0);
    theType = varDeclStmt->type();

    if (opt_use_auto_vars &&
	! stmtMapGet(prosMap, commonMapKeyEncode(symbName(theVar))) &&
	! exprMapGet(prosHedrMap, prosHedrMapKeyEncode(currentProcedureName,
						       symbName(theVar))) &&
	! exprMapGet(parameterMap,symbName(theVar)) &&
	! exprMapGet(staticVarMap, symbName(theVar)) &&
	strcmp(symbName(theVar), "fm_number_of_processors"))
    {
	// if the symbol is a function name
	// do not change it to automatic
	if ((theVar->symbol()->variant() != FUNCTION_NAME &&
	     theVar->symbol()->variant() != ROUTINE_NAME) &&
	    theType->variant() != T_AUTO) 
	{
	    SgSymbol *symb_copy = new SgVariableSymb(theVar->symbol()->identifier(),
						    *auto_t);
	    SgExpression *var_copy = new SgVarRefExp(*symb_copy);
	    s->insertStmtBefore(*declare_the_var(var_copy, auto_t));

#if 0
	    switch (theType->variant()) {
	      case T_INT:
		varDeclStmt->setType(*auto_integer_t);
		break;
	      case T_FLOAT:
		varDeclStmt->setType(*auto_real_t);
		break;
	      case T_DOUBLE:
		varDeclStmt->setType(*auto_double_t);
		break;
	      case T_CHAR:
		varDeclStmt->setType(*auto_char_t);
		break;
	      case T_BOOL:
		varDeclStmt->setType(*auto_bool_t);
		break;
	      case T_STRING:
		varDeclStmt->setType(*auto_string_t);
		break;
	      case T_COMPLEX:
		varDeclStmt->setType(*auto_complex_t);
		break;
	      case T_DCOMPLEX:
		varDeclStmt->setType(*auto_dcomplex_t);
		break;
	      case T_AUTO_INT:
	      case T_AUTO_FLOAT:
	      case T_AUTO_DOUBLE:
	      case T_AUTO_CHAR:
	      case T_AUTO_BOOL:
	      case T_AUTO_STRING:
	      case T_AUTO_COMPLEX:
	      case T_AUTO_DCOMPLEX:
		break;
	      default:
		fm_internal_error("\
Automatic type for %s not yet supported",
				  getVariantByName(theType->variant()));
		break;
	    }
#endif /* 0 */
	}
    }

    for (i = 0; i < varDeclStmt->numberOfVars(); i++) {
	SgExpression *var = varDeclStmt->var(i);
	SgSymbol *symb = varDeclStmt->var(i)->symbol();
	SgStatement *decl_s;
	
	decl_s = portMapGet(prosMap, prosMapKeyEncode(currentProcedureName, symb->identifier()));
	if (! decl_s) {
	    SgStatement *transformedDecl;

	    transformedDecl = transform_type_length_operator(varDeclStmt);
//	    if (transformedDecl != varDeclStmt) {
//		s->insertStmtBefore(*transformedDecl);
//		fmStmtSLL.append(s);
//	    }
	    portMapSet(prosMap, prosMapKeyEncode(currentProcedureName, symb->identifier()), transformedDecl);
	} else if (decl_s->variant() == VAR_DECL) {
	    //
	    // check whether the variable
	    // is a scalar, warn the user if it isn't
	    //
	    SgVarDeclStmt *varDecl = isSgVarDeclStmt(decl_s);
	    SgVarDeclStmt *n_decl;
	    SgStatement   *transformedDecl;

	    n_decl = new SgVarDeclStmt(*varDecl->varList(),*varDeclStmt->type());
	    transformedDecl = transform_type_length_operator(n_decl);
	    if (transformedDecl != n_decl) {
		n_decl->replaceWithStmt(*transformedDecl);
	    }

	    stmtMapOverwrite(prosMap,prosMapKeyEncode(currentProcedureName,symbName(var)), transformedDecl);
	    
	} else {
	    fm_internal_error("\
Do not expect statement type: %s at %s:%d",
			      getVariantByName(decl_s->variant()),
			      __FILE__, __LINE__);
	}
    }
}
    
SgStatement *transform_type_length_operator(SgVarDeclStmt *s)
{
    SgType             *t;
    SgExpression       *length_expression;
    SgType             *new_t;
    SgStatement        *new_declaration;

    fm_assert(s);

    t                  = s->type();
    fm_assert(t);

    length_expression  = t->length();

    if (length_expression) {
	SgStringLengthExp     *length_operator;
	SgExpression          *length;
	SgExprListExp         *var_list;
	SgArrayType           *array_t;
	SgSymbol              *array_symbol;
	int                   i, j;

	length_operator       = isSgStringLengthExp(length_expression);
	if (! length_operator) {
	    return (s);
	}
//	fm_assert(length_operator);
	
	length                = length_operator->length();

//	if (length->variant() == KEYWORD_VAL) {
//	    length = new SgExpression(STAR_RANGE);
//	}

	if (! IsInteger(length) && length->variant() != KEYWORD_VAL) {
	    fm_warning(s, "\
expect an constant for a length operator, but get %s(%s)",
		     length->unparse(),
		     getVariantByName(length->variant()));
	    //
	    // do not do any transformations
	    //
	    return s;
	}
	
	if (t->variant() == T_STRING) {
	    new_t       = new SgType(T_CHAR);
	} else {
	    new_t           = new SgType(t->variant());
	}
	
	array_t         = new SgArrayType(*new_t);

	for (i = 0; i < s->numberOfVars(); i++) {
	    SgExpression  *var;
	    SgArrayRefExp *a;

	    array_t->addRange(*length);
	    
	    var = s->var(i);
	    if (var->variant() == ARRAY_REF) {
		//
		// append the subscripts of VAR
		//
		SgArrayRefExp *var_a = isSgArrayRefExp(var);

		for (j = 0; j < var_a->numberOfSubscripts(); j++) {
		    array_t->addRange(*var_a->subscript(j));
		}
	    }
	    array_symbol = new SgVariableSymb(symbName(var), *array_t);
	    a            = new SgArrayRefExp(*array_symbol);
	    a->addSubscript(*length);

	    if (var->variant() == ARRAY_REF) {
		//
		// append the subscripts of VAR
		//
		SgArrayRefExp *var_a = isSgArrayRefExp(var);

		for (j = 0; j < var_a->numberOfSubscripts(); j++) {
		    a->addSubscript(*var_a->subscript(j));
		}
	    }
	    if (i == 0) {
		var_list        = new SgExprListExp(*a);
	    } else {
		var_list->append(*a);
	    }
	}
	new_declaration = new SgVarDeclStmt(*var_list, *new_t);
	return (new_declaration);
    } else {
	return s;
    }
}

void doDimensionStmt(SgStatement *s)
{
    //
    // if the variable is already declared
    // change the declaration to an array one
    //
    SgVarListDeclStmt *dim_stat;
    int                i;

    dim_stat          = isSgVarListDeclStmt(s);
    for (i = 0; i < dim_stat->numberOfVars(); i++) {
	SgExpression *var;
	SgStatement  *decl_s;
	SgSymbol     *symb;
	var          = dim_stat->var(i);
	symb         = var->symbol();
	
	fm_assert(symb);
	decl_s       = stmtMapGet(prosMap,prosMapKeyEncode(currentProcedureName,symbName(var)));
	if (decl_s) {
	    //
	    // then build a new SgVarDeclStmt
	    // which merges the dimension info with
	    // the type info in DECL_S
	    //
	    SgVarDeclStmt *n_decl;
	    
	    if (decl_s->variant() == VAR_DECL) {
		SgVarDeclStmt *varDecl;
		SgExpression  *theVar;
		
		varDecl = isSgVarDeclStmt(decl_s);
		theVar = findSymbInDeclStmt(symb);
		n_decl  = new_var_decl_stmt(symb,theVar, varDecl->type(),dim_stat);
		if (n_decl) {
		    //			    dim_stat->insertStmtBefore(*n_decl);
		    stmtMapOverwrite(prosMap,prosMapKeyEncode(currentProcedureName,symbName(var)),n_decl);
		}
	    }
	} else {
	    // 
	    // find implicit type
	    //
	    SgVarDeclStmt *n_decl;
	    SgType *t = getImplicitType(currentProcedureName,symb);
	    SgVarRefExp *var;
	    SgSymbol *new_symbol = symb;

	    new_symbol->setType(*t);

	    var = new SgVarRefExp(*new_symbol);
	    n_decl = new_var_decl_stmt(symb,var, t, dim_stat);
	    if (n_decl) {
		//			dim_stat->insertStmtBefore(*n_decl);
		stmtMapOverwrite(prosMap,prosMapKeyEncode(currentProcedureName,symbName(var)),n_decl);
	    }
	}
    }
}    

void doExternalStmt(SgStatement *s)
{
    SgNestedVarListDeclStmt *nd_stmt;
    
    int           i;
    int           j;

    nd_stmt = isSgNestedVarListDeclStmt(s);

    for (i = 0; i < nd_stmt->numberOfLists(); i++) {
	SgObjectListExp *list;

	list = isSgObjectListExp(nd_stmt->list(i));

	fm_assert(list);

	for (j = 0; j < list->listLength(); j++) {
	    SgExpression *var = list->objectRef(j);

	    while (var->variant() == EXPR_LIST) {
		var = isSgExprListExp(var)->elem(0);
	    }
	    //
	    // put into current block's common list
	    //
	    stmtMapSet(prosMap, 
		       ppString("extern_",symbName(var),""), 
		       s);
	}
    }
}

SgVarDeclStmt *new_var_decl_stmt(SgSymbol *symb, 
				 SgExpression *theVar,
				 SgType *t, 
				 SgStatement *dim)
{
    SgVarDeclStmt     *var_decl;
    SgExprListExp     *var_list;
    SgArrayRefExp     *var_reference;
    SgArrayType       *array_t;
    SgSymbol          *var_symbol;
    SgExpression      *dim_expression;
    SgVarListDeclStmt *decl_s;
    int               i;
    SgArrayRefExp     *a;
    SgExpression      *var_declared;

    fm_assert(dim->variant() == DIM_STAT);
    fm_assert(symb);
    fm_assert(t);

    decl_s = isSgVarListDeclStmt(dim);
    fm_assert(decl_s);

    for (i = 0; i < decl_s->numberOfVars(); i++) {
	char *var_name;
	
	var_declared = decl_s->var(i);
	var_name = symbName(var_declared);

	if (! strcmp(var_name, symb->identifier())) {
	    dim_expression = decl_s->var(i);
	    break;
	}
    }

    fm_assert(dim_expression);
    if (dim_expression->variant() == ARRAY_REF) {
	a = isSgArrayRefExp(dim_expression);
    } else {
	fm_internal_error("\
Do not expect a non array reference at %s:%d",
			  __FILE__, __LINE__);
    }
    //
    // Build an array reference
    //
    array_t       = new SgArrayType(*t);
	
    if (theVar->variant() == ARRAY_REF) {
	SgArrayRefExp *var_a;
	
	var_a = isSgArrayRefExp(theVar);
	for (i = 0; i < var_a->numberOfSubscripts(); i++) {
	    array_t->addRange(*a->subscript(i));
	}
    }

    for (i = 0; i < a->numberOfSubscripts(); i++) {
	array_t->addRange(*a->subscript(i));
    }

    var_symbol    = new SgVariableSymb(symb->identifier(), *array_t);
    var_reference = new SgArrayRefExp(*var_symbol);
    
    if (theVar->variant() == ARRAY_REF) {
	SgArrayRefExp *var_a;
	
	var_a = isSgArrayRefExp(theVar);
	for (i = 0; i < var_a->numberOfSubscripts(); i++) {
	    var_reference->addSubscript(*a->subscript(i));
	}
    }

    for (i = 0; i < a->numberOfSubscripts(); i++) {
        var_reference->addSubscript(*a->subscript(i));
    }
    var_list      = new SgExprListExp(*var_reference);
    var_decl      = new SgVarDeclStmt(*var_list, *t);
//    printf ("new var decl stmt = ");
//    var_decl->unparsestdout();
    return (var_decl);
}

/*
 * doCanonalizeDecl()
 *
 * change the declaration statement into
 * canonalized form
 *
 */

void doCanonalizeDecl(SgStatement *start, SgStatement *end)
{
    SgStatement       *line;
    SgStatement       *lsp;
    SgVarDeclStmt     *var_decl;
    int                i;

    lsp = start;
    for (line = start; lsp != end; line = lsp) {
	lsp = line->lexNext();
	switch (line->variant()) {
	  case VAR_DECL:
	    var_decl = isSgVarDeclStmt(line);
	    
	    if (var_decl->numberOfVars() == 1) {
		// 
		// Already in Canonalized form; try next line
		//
		continue;
	    }
	    
	    for (i = 0; i < var_decl->numberOfVars(); i++) {
		SgExpression *var = var_decl->var(i);
		SgType       *t   = var_decl->type();

		line->insertStmtAfter(*new_decl_stmt(t, var));
	    }
	    line->deleteStmt();
	    break;
	  default:
	    lsp = line->lexNext();
	    break;
	}
    }
}


SgStatement *new_decl_stmt(SgType *t, SgExpression *var)
{
    SgVarDeclStmt *decl;
    SgExprListExp *var_eL;

    var_eL = new SgExprListExp(*var);

    decl = new SgVarDeclStmt(*var_eL, *t);
    return (decl);
}

/*
 * doDataStmt(SgStatement *s)
 *    
 * Keep track of the variables in a Data Statement
 * These variables cannot be declared as automatic
 * as the SunOSa F77 compiler does not like to initialize
 * automatic variables with data statement.
 *
 */

void doDataStmt(SgStatement *s)
{
    SgDataStmt *data_stmt;
    int i, j;

    /*
     * variables in a data statement are
     * treated as static variables
     */

    data_stmt = isSgDataStmt(s);

    for (i = 0; i < data_stmt->numberOfDataPair(); i++)
    {
	SgDataPairExp *data_pair = isSgDataPairExp(data_stmt->dataPair(i));

	for (j = 0; j < data_pair->numberOfDataelt(); j++)
	{
	    SgExpression *elt = data_pair->dataelt(j);

	    doDataElt(elt);
	}
    }
}

void doDataElt(SgExpression *e)
{
    SgDataEltExp *d_elt = isSgDataEltExp(e);
    SgSymbol *name = d_elt->name();

    fm_assert(d_elt);

    if (! name)
        doDataImplDo(isSgDataImpliedDoExp(d_elt->dataimplieddo()));
    else 
    {
//	printf("%s\n", name->identifier());
	exprMapSet(staticVarMap, name->identifier(), d_elt);
    }
}

void doDataImplDo(SgDataImpliedDoExp *e)
{
    int i;
    int n_elts = e->numberOfDataelt();

    for (i = 0; i < n_elts; i++)
        doDataElt(e->dataelt(i));
}
