/**********************************************************************
$Id: debug_defs.h,v 1.6 1995/04/17 06:29:53 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/debug_defs.h,v $

$Log: debug_defs.h,v $
Revision 1.6  1995/04/17  06:29:53  kohr
Doc. added, moved migration functions to migrate.c.

1.5  1995/04/06  06:31:57  kohr
documentation, cleaned up a bit.

FM runtime debugging support.
**********************************************************************/


#ifndef _fm_debug_defs_h
#define _fm_debug_defs_h

/*
 * Support for conditional execution based on debug level.
 *
 * We keep a table of symbolic subsystem name, defined constant
 * for the subsystem, and debug level for that subsystem.
 *
 * The defined constant is also the index into the subsystem debug
 * table.
 */

#ifdef FM_DEBUG

#define FMDebug(subsystem, level) \
    (_fm_subsystem_debug_table[FM_SUBSYSTEM_ ## subsystem].debug_level >= (level))

#endif /* FM_DEBUG */

/*
 * Magic numbers for debugging.
 */
#ifdef FM_DEBUG
#define FM_INPORT_START_MAGIC 0x11223344
#define FM_INPORT_END_MAGIC 0x12233445

#define FM_OUTPORT_START_MAGIC 0x43215678
#define FM_OUTPORT_END_MAGIC 0x56781234

#define FM_SYNCH_START_MAGIC 0x99887766
#define FM_SYNCH_END_MAGIC 0x55443322

#define FM_CHANNEL_START_MAGIC 0x15151515
#define FM_CHANNEL_END_MAGIC 0x43434343

#define FM_PBLOCK_START_MAGIC 0x91919191
#define FM_PBLOCK_END_MAGIC 0x81818181

#define FM_PBLOCK_ENTRY_START_MAGIC 0x72727272
#define FM_PBLOCK_ENTRY_END_MAGIC 0x37373737

#define StartMagicNumber int start_magic;
#define EndMagicNumber int end_magic;


#define MagicCheck(what, obj, start, end)				      \
    if ((obj) == (what *) NULL)						      \
      nexus_fatal("Got NULL %s at %s:%d\n", 				      \
                    #what, __FILE__, __LINE__);				      \
    if ((obj)->start_magic != start)					      \
      nexus_fatal("Magic check (start) for %s %x failed at %s:%d: %x != %x\n",\
                    #what, (obj), __FILE__, __LINE__, (obj)->start_magic, start); \
    if ((obj)->end_magic != end)						      \
	nexus_fatal("Magic check (end) for %s %x failed at %s:%d: %x != %x\n" \
                    #what, (obj), __FILE__, __LINE__, (obj)->end_magic, end);

#define SynchMagicCheck(obj) \
    MagicCheck(fm_synch_t, obj, FM_SYNCH_START_MAGIC, FM_SYNCH_END_MAGIC)

#define ChannelMagicCheck(obj) \
    MagicCheck(fm_channel_t, obj, FM_CHANNEL_START_MAGIC, FM_CHANNEL_END_MAGIC)

#define InportMagicCheck(obj) \
    MagicCheck(fm_inport_t, obj, FM_INPORT_START_MAGIC, FM_INPORT_END_MAGIC)

#define OutportMagicCheck(obj) \
    MagicCheck(fm_outport_t, obj, FM_OUTPORT_START_MAGIC, FM_OUTPORT_END_MAGIC)

#define PblockMagicCheck(obj) \
    MagicCheck(fm_pblock_t, obj, FM_PBLOCK_START_MAGIC, FM_PBLOCK_END_MAGIC)

#define PblockEntryMagicCheck(obj) \
    MagicCheck(fm_pblock_entry_t, obj, FM_PBLOCK_ENTRY_START_MAGIC, FM_PBLOCK_ENTRY_END_MAGIC)

#define MagicInit(obj, start, end)					      \
    (obj)->start_magic = start;						      \
    (obj)->end_magic = end;

#define SynchMagicInit(obj) \
    MagicInit(obj, FM_SYNCH_START_MAGIC, FM_SYNCH_END_MAGIC)

#define ChannelMagicInit(obj) \
    MagicInit(obj, FM_CHANNEL_START_MAGIC, FM_CHANNEL_END_MAGIC)

#define InportMagicInit(obj) \
    MagicInit(obj, FM_INPORT_START_MAGIC, FM_INPORT_END_MAGIC)

#define OutportMagicInit(obj) \
    MagicInit(obj, FM_OUTPORT_START_MAGIC, FM_OUTPORT_END_MAGIC)

#define PblockMagicInit(obj) \
    MagicInit(obj, FM_PBLOCK_START_MAGIC, FM_PBLOCK_END_MAGIC)

#define PblockEntryMagicInit(obj) \
    MagicInit(obj, FM_PBLOCK_ENTRY_START_MAGIC, FM_PBLOCK_ENTRY_END_MAGIC)

#else /* FM_DEBUG */

#define StartMagicNumber
#define EndMagicNumber

#define SynchMagicCheck(obj) 
#define ChannelMagicCheck(obj) 
#define InportMagicCheck(obj) 
#define OutportMagicCheck(obj) 
#define PblockMagicCheck(obj) 
#define PblockEntryMagicCheck(obj) 

#define SynchMagicInit(obj) 
#define ChannelMagicInit(obj) 
#define InportMagicInit(obj) 
#define OutportMagicInit(obj) 
#define PblockMagicInit(obj) 
#define PblockEntryMagicInit(obj) 

#endif /* FM_DEBUG */ 

#endif /* _fm_debug_defs_h */
