/**********************************************************************
$Id: defs.h,v 1.22 1995/04/17 06:29:55 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/defs.h,v $

$Log: defs.h,v $
Revision 1.22  1995/04/17  06:29:55  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.21  1995/04/06  06:32:00  kohr
Added documentation, cleaned up a bit.


FM runtime general definitions.
**********************************************************************/


#ifndef _fm_defs_h
#define _fm_defs_h

/*
Nexus doesn't actually have these functions.
*/

#define nexus_put_inport _fm_put_inport
#define nexus_put_outport _fm_put_outport
#define nexus_get_inport _fm_get_inport
#define nexus_get_outport _fm_get_outport
#define nexus_get_stashed_inport _fm_get_stashed_inport
#define nexus_get_stashed_outport _fm_get_stashed_outport
#define nexus_sizeof_inport _fm_sizeof_inport
#define nexus_sizeof_outport _fm_sizeof_outport

#define FM_MAX_PROCESS_NAME_LEN 256

/*
Macros to abstract channel & port locking, incorporating state stuff.
*/

#define InportLock(INPORT) nexus_mutex_lock(&((INPORT)->lock))
#define InportUnlock(INPORT) nexus_mutex_unlock(&((INPORT)->lock))
#define InportSetState(INPORT, STATE) (INPORT)->state = STATE
#define InportLockAndSetState(INPORT, STATE)				      \
    InportLock(INPORT);						      	      \
    InportSetState(INPORT, STATE);					      \
    InportUnlock(INPORT);
#define InportState(INPORT) ((INPORT)->state)
#define InportWait(INPORT) nexus_cond_wait(&((INPORT)->cond), &((INPORT)->lock))
#define InportSignal(INPORT) nexus_cond_signal(&((INPORT)->cond))

#define OutportLock(OUTPORT) nexus_mutex_lock(&((OUTPORT)->lock))
#define OutportUnlock(OUTPORT) nexus_mutex_unlock(&((OUTPORT)->lock))
#define OutportSetState(OUTPORT, STATE) (OUTPORT)->state = STATE
#define OutportLockAndSetState(OUTPORT, STATE)				      \
    OutportLock(OUTPORT);						      \
    OutportSetState(OUTPORT, STATE);					      \
    OutportUnlock(OUTPORT);
#define OutportState(OUTPORT) ((OUTPORT)->state)
#define OutportWait(OUTPORT) nexus_cond_wait(&((OUTPORT)->cond), &((OUTPORT)->lock))
#define OutportSignal(OUTPORT) nexus_cond_signal(&((OUTPORT)->cond))

#define ChannelLock(CHANNEL) nexus_mutex_lock(&((CHANNEL)->lock))
#define ChannelUnlock(CHANNEL) nexus_mutex_unlock(&((CHANNEL)->lock))
#define ChannelSetState(CHANNEL, STATE) (CHANNEL)->state = STATE
#define ChannelLockAndSetState(CHANNEL, STATE)				      \
    ChannelLock(CHANNEL);						      \
    ChannelSetState(CHANNEL, STATE);					      \
    ChannelUnlock(CHANNEL);
#define ChannelState(CHANNEL) ((CHANNEL)->state)
#define ChannelWait(CHANNEL) nexus_cond_wait(&((CHANNEL)->cond), &((CHANNEL)->lock))
/* #define ChannelTimedWait(CHANNEL, TIMEOUT) nexus_macro_cond_timedwait(&((CHANNEL)->cond), &((CHANNEL)->lock), TIMEOUT) */
#define ChannelSignal(CHANNEL) nexus_cond_signal(&((CHANNEL)->cond))

/*
IOVAL convenience macros. Assumes an argument f_integer_t *f_ioval.
*/

#define FM_SET_IOVAL_EOF(f_ioval) (*f_ioval) = (f_integer_t) -1;

#define FM_SET_IOVAL(f_ioval, ioval) \
	(*f_ioval) = (f_integer_t) FM_IOVAL_ ## ioval;

#define FM_RETURN_IOVAL(ioval) \
{ \
    FM_SET_IOVAL(f_ioval, ioval); \
    return; \
}
 

/*
Simple instrumentation used within the FM system.
*/

#define SI_SYS_N_TIMER 0
#define SI_SYS_N_COUNTER 2
#define SI_SYS_N_TIMERHIST 0
#define SI_SYS_N_INTHIST 0

#define SI_SYS_COUNTER_FM_POLL 0
#define SI_SYS_COUNTER_FM_YIELD 1

/*
State values
*/

#define FM_STATE_UNDEFINED		1
#define FM_STATE_AWAITING_ACTIVATION	2
#define FM_STATE_NORMAL			3
#define FM_STATE_WAITING_FOR_NOTIFY_PORTS_ACK	4
#define FM_STATE_CLOSING_OUTPORT	5

	/* How many queue entries do we allocate at a time? */

#define FM_QUEUE_ENTRY_ALLOC_SIZE 16
#define FM_PORT_ALLOC_SIZE 16

	/* Handler types used in fmc-generated code */
#define FM_SHOULD_SHARE_HANDLER_TYPE NEXUS_HANDLER_TYPE_NON_THREADED
#define FM_INVOKE_HANDLER_TYPE NEXUS_HANDLER_TYPE_NON_THREADED

/*
FM Handlers.
*/

#define FM_VC_INIT_NAME "_FM_VCInit"
#define FM_VC_INIT_HASH 894

#define FM_VC_SHUTDOWN_CONTEXTS_NAME "_FM_VCShutdownContexts"
#define FM_VC_SHUTDOWN_CONTEXTS_HASH 164

#define FM_VC_SHUTDOWN_CONTEXTS_ACK_NAME "_FM_VCShutdownContextsAck"
#define FM_VC_SHUTDOWN_CONTEXTS_ACK_HASH 435

#define FM_CREATE_PROCESS_NAME "_FM_CreateProcess"
#define FM_CREATE_PROCESS_HASH 647

#define FM_CREATE_PROCESS_RETURN_NAME "_FM_CreateProcessReturn"
#define FM_CREATE_PROCESS_RETURN_HASH 266

#define FM_SHOULD_SHARE_RETURN_NAME "_FM_ShouldShareReturn"
#define FM_SHOULD_SHARE_RETURN_HASH 57

#define FM_PORT_SEND_SETUP_NAME "_FM_PortSendSetup"
#define FM_PORT_SEND_SETUP_HASH 660

#define FM_PORT_SEND_SETUP_RETURN_NAME "_FM_PortSendSetupReturn"
#define FM_PORT_SEND_SETUP_RETURN_HASH 279

#define FM_PROCESS_RETURN_NAME "_FM_ProcessReturn"
#define FM_PROCESS_RETURN_HASH 691

#define FM_PORT_SEND_NAME "_FM_PortSend"
#define FM_PORT_SEND_HASH 131

#define FM_END_CHANNEL_NAME "_FM_EndChannel"
#define FM_END_CHANNEL_HASH 292

#define FM_END_CHANNEL_ACK_NAME "_FM_EndChannelAck"
#define FM_END_CHANNEL_ACK_HASH 563

#define FM_CHANNEL_FORWARD_BUFFER_NAME "_FM_ChannelForwardBuffer"
#define FM_CHANNEL_FORWARD_BUFFER_HASH 319

#define FM_CHANNEL_MIGRATING_NAME "_FM_ChannelMigrating"
#define FM_CHANNEL_MIGRATING_HASH 943

#define FM_CHANNEL_MIGRATING_ACK_NAME "_FM_ChannelMigratingAck"
#define FM_CHANNEL_MIGRATING_ACK_HASH 193

#define FM_FWD_CHANNEL_MIGRATING_NAME "_FM_FwdChannelMigrating"
#define FM_FWD_CHANNEL_MIGRATING_HASH 211

#define FM_FWD_CHANNEL_MIGRATING_ACK_NAME "_FM_FwdChannelMigratingAck"
#define FM_FWD_CHANNEL_MIGRATING_ACK_HASH 482

#define FM_OUTPORT_MIGRATING_NAME "_FM_OutportMigrating"
#define FM_OUTPORT_MIGRATING_HASH 1011

#define FM_OUTPORT_MIGRATING_ACK_NAME "_FM_OutportMigratingAck"
#define FM_OUTPORT_MIGRATING_ACK_HASH 261

#define FM_FWD_OUTPORT_MIGRATING_NAME "_FM_FwdOutportMigrating"
#define FM_FWD_OUTPORT_MIGRATING_HASH 279

#define FM_FWD_OUTPORT_MIGRATING_ACK_NAME "_FM_FwdOutportMigratingAck"
#define FM_FWD_OUTPORT_MIGRATING_ACK_HASH 550

	/* Definitions of constants that define annotation types. */
#define FM_ANNOTATION_LOCAL		0
#define FM_ANNOTATION_LOCATION		1
#define FM_ANNOTATION_SUBMACHINE	2


/*
Macros for a dynamically allocated list of global pointers, useful for
packing messages that contain a runtime-determined numbers of gps.
*/

#define FM_GPL_DEFS \
    nexus_global_pointer_t *gp_list;  \
    int gp_list_next, gp_list_size;

#define FM_GPL_INIT(FUNC, SIZE) 			\
    gp_list_next = 0;					\
    gp_list_size = SIZE;				\
    NexusMalloc(FUNC, gp_list, nexus_global_pointer_t *,\
		sizeof(nexus_global_pointer_t) * SIZE);

#define FM_GPL_ADD(GP) \
    NexusAssert(gp_list_next < gp_list_size); \
    GP = &(gp_list[gp_list_next++]);

#define FM_GPL_FREE \
{									      \
     int ii;								      \
     for (ii = 0; ii < gp_list_next; ii++)				      \
     {									      \
	 nexus_destroy_global_pointer(&gp_list[ii]);			      \
     }									      \
     NexusFree(gp_list);						      \
}
    

/* 
Following are a number of macros for use within compiler-generated code.
*/

	/* Invoker code */
#define FM_ARGS_T(NAME) fm_ ## NAME ## _args_t
#define FM_RETURN_ARGS_T(NAME) fm_ ## NAME ## _return_args_t
#define FM_RETURN_COPY_FUNC_NAME(NAME) NAME ## _return_copy_func
#define FM_RETURN_STASHED_COPY_FUNC_NAME(NAME) NAME ## _return_stashed_copy_func
#define FM_INVOKER_NAME(NAME) FM_## NAME ## _INVOKE

#define FM_RETURN_COPY_FUNC_DECL(NAME) \
static void FM_RETURN_COPY_FUNC_NAME(NAME)(nexus_buffer_t *buffer, void *dest)

#define FM_RETURN_COPY_FUNC_CODE(NAME) 					      \
    FM_RETURN_ARGS_T(NAME) *args;					      \
									      \
    args = (FM_RETURN_ARGS_T(NAME) *) dest;

#define FM_RETURN_STASHED_COPY_FUNC_DECL(NAME)				      \
	static void FM_RETURN_STASHED_COPY_FUNC_NAME(NAME)(		      \
		nexus_stashed_buffer_t *buffer, void *dest)

#define FM_RETURN_STASHED_COPY_FUNC_CODE(NAME) 				      \
    FM_RETURN_ARGS_T(NAME) *args;					      \
									      \
    args = (FM_RETURN_ARGS_T(NAME) *) dest;

#define FM_INVOKER_DECLS(NAME) 						      \
    nexus_buffer_t mbuf;						      \
    fm_pblock_t *pblock;						      \
    fm_pblock_entry_t *entry;						      \
    nexus_global_pointer_t entry_gp;					      \
    int size, n_elts, obj_n_elts;					      \
    FM_RETURN_ARGS_T(NAME) *return_args;

#define FM_INVOKER_CODE_1(NAME, INVOKE_HANDLER_NAME, INVOKE_HANDLER_HASH)     \
    pblock = (fm_pblock_t *) (*f_pblock);				      \
									      \
    NexusAssert(pblock->current_entry >= 0 &&				      \
		pblock->current_entry < pblock->n_entries);		      \
    									      \
    entry = &(pblock->entries[pblock->current_entry]);			      \
									      \
    entry->return_copy_func = FM_RETURN_COPY_FUNC_NAME(NAME);		      \
    entry->return_stashed_copy_func = FM_RETURN_STASHED_COPY_FUNC_NAME(NAME); \
									      \
    nexus_init_remote_service_request(&mbuf,				      \
				      &entry->dest_gp,			      \
				      INVOKE_HANDLER_NAME,		      \
				      INVOKE_HANDLER_HASH);		      \
									      \
    NexusMalloc(FM_INVOKER_NAME(NAME)(),				      \
		return_args,						      \
		FM_RETURN_ARGS_T(NAME) *,				      \
		sizeof(FM_RETURN_ARGS_T(NAME)));

#define FM_INVOKER_CODE_2(NAME)						      \
    entry->process_return_data = (void *) return_args;			      \
									      \
    nexus_global_pointer(&entry_gp, entry);				      \
									      \
    size = nexus_sizeof_global_pointer(&mbuf, &entry_gp, 1, &n_elts);

#define FM_INVOKER_CODE_3(NAME)						      \
    nexus_put_global_pointer(&mbuf, &entry_gp, 1);			      \
									      \
    _fm_synch_clear(&entry->synch);					      \
    _fm_synch_increment(&entry->synch);					      \
    									      \
    nexus_send_remote_service_request(&mbuf);				      \
									      \
    pblock->current_entry++;

#define FM_INVOKE_THREAD_DECL(NAME) \
	static void *invoke_ ## NAME ## _thread(void *t_args)

#ifdef FM_DEBUG
#define FM_THREAD_START_MESSAGE(NAME) \
    if (FMDebug(VC, 1)) \
        nexus_printf("Thread for " #NAME " starting\n"); 
#else
#define FM_THREAD_START_MESSAGE(NAME)
#endif

#ifdef FM_SIMPLE_INST
#define FM_SIMPLE_DECL si_instrument_block_t instruments;

#ifndef SIF_N_TIMER
#define SIF_N_TIMER 0
#endif

#ifndef SIF_N_TIMERHIST
#define SIF_N_TIMERHIST 0
#endif

#ifndef SIF_N_COUNTER
#define SIF_N_COUNTER 0
#endif

#ifndef SIF_N_INTHIST
#define SIF_N_INTHIST 0
#endif

#ifndef SIF_HIST_INIT
#define SIF_HIST_INIT
#endif

#define FM_SIMPLE_INIT if (_fm_simple_inst) { si_instrument_init(&instruments, SIF_N_TIMER, SIF_N_TIMERHIST, SIF_N_COUNTER, SIF_N_INTHIST); SIF_HIST_INIT } 
#define FM_SIMPLE_DUMP(NAME) if (_fm_simple_inst) si_instrument_dump(#NAME); si_instrument_free();

#else /* FM_SIMPLE_INST */

#define FM_SIMPLE_DECL
#define FM_SIMPLE_INIT
#define FM_SIMPLE_DUMP(NAME)

#endif /* FM_SIMPLE_INST */

#define FM_INVOKE_THREAD_CODE_1(NAME)					      \
    FM_ARGS_T(NAME) *args;						      \
    nexus_buffer_t mbuf;						      \
    int size, n_elts;							      \
    FM_SIMPLE_DECL \
    FM_SIMPLE_INIT \
    args = (FM_ARGS_T(NAME) *) t_args;					      \
    FM_THREAD_START_MESSAGE(NAME)

#define FM_INVOKE_THREAD_CODE_2(NAME)					      \
    nexus_init_remote_service_request(&mbuf,				      \
				      &args->entry_gp,			      \
				      FM_PROCESS_RETURN_NAME,		      \
				      FM_PROCESS_RETURN_HASH);		      \
									      \
    size = n_elts = 0;

/*
FM_ALLOC_ARG_SPACE() : Allocate the space in the args buf for a variable
length array argument.
*/

#define FM_ALLOC_ARG_SPACE(name, type, arg)    			              \
    NexusMalloc(invoke_ ## name ## _thread(),				      \
		args->arg,						      \
		type *,							      \
		(args->arg ## _len) * sizeof(type));

#if 0
/* SJT: Hack to make processes go away */
#ifdef FM_HACK_THREADABLE_PROCESSES
#define FM_HACK_DESTROY_CONTEXT
#else
#define FM_HACK_DESTROY_CONTEXT nexus_destroy_current_context(NEXUS_TRUE);
#endif
#endif

#ifdef FM_DEBUG
#define FM_DO_PROCESS_EXIT_DEBUG(NAME) 					      \
    if (FMDebug(VC, 1))							      \
	nexus_printf("Thread for " #NAME " exiting. "			      \
		     "_fm_is_node_context=%d\n", _fm_is_node_context);
#else
#define FM_DO_PROCESS_EXIT_DEBUG(NAME)
#endif

#define FM_INVOKE_THREAD_CODE_3(NAME) 					      \
    nexus_send_remote_service_request(&mbuf);           		      \
    FM_SIMPLE_DUMP(NAME) 						      \
    FM_DO_PROCESS_EXIT_DEBUG(NAME) 					      \
    if (!_fm_is_node_context) 						      \
        nexus_destroy_current_context(NEXUS_FALSE); 			      \
    else 								      \
	nexus_thread_exit((void *) 0);

#define FM_INVOKE_HANDLER_DECL(NAME) \
static void invoke_## NAME ##_handler(void *address, nexus_buffer_t *buf)

#define FM_INVOKE_HANDLER_CODE_1(NAME)					      \
    FM_ARGS_T(NAME) *args;						      \
    nexus_thread_t thread;                                                    \
									      \
    NexusMalloc(invoke_## NAME ## _handler(), args, FM_ARGS_T(NAME) *,	      \
		sizeof(FM_ARGS_T(NAME)));

#ifdef FM_DEBUG
#define FM_THREAD_CREATE_MESSAGE(NAME) \
    if (FMDebug(VC, 1)) \
	nexus_printf("Creating thread for " # NAME "\n");
#else
#define FM_THREAD_CREATE_MESSAGE(NAME)
#endif

#define FM_INVOKE_HANDLER_CODE_2(NAME)					      \
    nexus_get_global_pointer(buf, &args->entry_gp, 1);			      \
									      \
    FM_THREAD_CREATE_MESSAGE(NAME) \
    nexus_thread_create(&thread, (nexus_thread_attr_t *) NULL,                \
                        invoke_## NAME ##_thread, args);

#define FM_PORT_INIT_CODE_1						      \
    fm_pblock_t *pblock;						      \
    fm_pblock_entry_t *entry;						      \
    									      \
    pblock = (fm_pblock_t *) (*f_pblock);				      \
									      \
    PblockMagicCheck(pblock);						      \
    NexusAssert(pblock->current_entry >= 0 &&				      \
		pblock->current_entry < pblock->n_entries);		      \
									      \
    entry = &(pblock->entries[pblock->current_entry]);

#define FM_PORT_INIT_CODE_2  \
    _fm_pblock_finish_port_init(pblock, entry);

#define FM_SHOULD_SHARE_HANDLER_FUNC(NAME) should_share_## NAME ##_handler
#define FM_SHOULD_SHARE_HANDLER_NAME_PREFIX "_FMH_ShouldShare_"
#define FM_SHOULD_SHARE_HANDLER_NAME(NAME) FM_SHOULD_SHARE_NAME_PREFIX #NAME

#define FM_SHOULD_SHARE_HANDLER(NAME, VALUE)				      \
static void FM_SHOULD_SHARE_HANDLER_FUNC(NAME)                                \
(void *address, nexus_buffer_t *buf)                                          \
{									      \
    nexus_buffer_t mbuf;						      \
    nexus_global_pointer_t return_gp;					      \
    int size, val;							      \
									      \
    nexus_get_global_pointer(buf, &return_gp, 1);			      \
    nexus_init_remote_service_request(&mbuf,				      \
				      &return_gp,			      \
				      FM_SHOULD_SHARE_RETURN_NAME,	      \
				      FM_SHOULD_SHARE_RETURN_HASH);	      \
    size = nexus_sizeof_int(&mbuf, 1);					      \
    nexus_set_buffer_size(&mbuf, size, 1);				      \
    val = VALUE;							      \
    nexus_put_int(&mbuf, &val, 1);					      \
    nexus_send_remote_service_request(&mbuf);				      \
}


#ifndef TRUE
#define TRUE 1
#endif /* TRUE */

#ifndef FALSE
#define FALSE 0
#endif /* FALSE */

#endif /* _fm_defs_h */

