/**********************************************************************
$Id: error.c,v 1.6 1995/06/14 04:40:58 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/error.c,v $

$Log: error.c,v $
Revision 1.6  1995/06/14  04:40:58  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.5  1995/04/17  06:29:57  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.4  1995/04/06  06:32:02  kohr
Added documentation, cleaned up a bit.


FM runtime support for error messages based on IOVAL.
**********************************************************************/


#include "fm.h"


typedef struct _ioval_def_t
{
    fm_ioval_t ioval;
    fm_bool_t is_fatal;
    char *message;
} ioval_def_t;

#undef FM_IOVAL
#define FM_IOVAL(ioval, fatal, symbol, message) \
	{FM_IOVAL_ ## symbol, fatal, #message } ,

static ioval_def_t ioval_defs[] = {
#include "ioval.def"
{-1, -1, (char *) NULL }
};

/*
FM_ERROR_MESSAGE() : Print an error message for the given ioval. If it is a
fatal error, abort out of the computation.
*/

void FM_ERROR_MESSAGE(f_integer_t *f_ioval)
{
    int ioval, i;
    ioval_def_t *def;

    ioval = (int) *f_ioval;

    if (ioval <= 0)
	return;

    for (i = 0; ioval_defs[i].message != (char *) NULL; i++)
    {
	if (ioval_defs[i].ioval == ioval)
	    break;
    }
    def = &ioval_defs[i];
    if (def->message == (char *) NULL)
    {
	nexus_fatal("FM Runtime error: Unknown IOSTAT %d\n", ioval);
    }
    else if (def->is_fatal)
    {
	nexus_fatal("FM Runtime error: %s\n", def->message);
    }
    else
    {
	nexus_warning("FM Runtime error: %s\n", def->message);
    }
} /* FM_ERROR_MESSAGE() */
