/**********************************************************************
$Id: freelist.h,v 1.5 1995/04/17 06:30:06 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/freelist.h,v $

$Log: freelist.h,v $
Revision 1.5  1995/04/17  06:30:06  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.4  1995/04/06  06:32:10  kohr
Added documentation, cleaned up a bit.


Macros for template functions for freelists for various datatypes.
**********************************************************************/


#ifndef _freelist_h_
#define _freelist_h_


#ifdef FM_FREELIST_DONT_PREPEND_FM
#define TYPENAME(TYPE) TYPE##_t
#else
#define TYPENAME(TYPE) fm_##TYPE##_t
#endif


#define FM_FREE_LIST_INIT(TYPE)						      \
    nexus_mutex_init(&TYPE##_freelist_lock, (nexus_mutexattr_t *) NULL);      \
    TYPE##_freelist = (TYPENAME(TYPE) *) NULL;				      \
    alloc_##TYPE##_entries();


#define FM_FREE_LIST_CODE(TYPE)						      \
									      \
static TYPENAME(TYPE) *TYPE##_freelist;					      \
static nexus_mutex_t TYPE##_freelist_lock;				      \
									      \
									      \
static void alloc_## TYPE ##_entries()					      \
{									      \
    TYPENAME(TYPE) *entries;						      \
    int i;								      \
									      \
    NexusMalloc(alloc_## TYPE ## _entries(), entries, TYPENAME(TYPE) *,	      \
		sizeof(TYPENAME(TYPE)) * FM_PORT_ALLOC_SIZE);		      \
    for (i = 0; i < FM_PORT_ALLOC_SIZE - 1; i++)			      \
    {									      \
	init_##TYPE(&entries[i]);  					      \
	entries[i].next = &(entries[i+1]);				      \
    }									      \
    init_##TYPE(&entries[i]);  						      \
    entries[i].next = TYPE ## _freelist;				      \
									      \
    TYPE ## _freelist = entries;					      \
}									      \
									      \
									      \
TYPENAME(TYPE) *_fm_alloc_ ## TYPE ()					      \
{									      \
    TYPENAME(TYPE) *entry;						      \
									      \
    nexus_mutex_lock(&TYPE ## _freelist_lock);				      \
    if (TYPE ##_freelist == (TYPENAME(TYPE) *) NULL)			      \
	alloc_## TYPE ##_entries();					      \
    NexusAssert(TYPE ##_freelist != (TYPENAME(TYPE) *) NULL);		      \
									      \
    entry = TYPE ##_freelist;						      \
    TYPE##_freelist = TYPE##_freelist->next;				      \
    nexus_mutex_unlock(&TYPE##_freelist_lock);				      \
									      \
    entry->next = (TYPENAME(TYPE) *) NULL;				      \
    reinit_##TYPE(entry);  						      \
									      \
    return entry;							      \
}									      \
									      \
									      \
void _fm_free_##TYPE(TYPENAME(TYPE) *entry)				      \
{									      \
    nexus_mutex_lock(&TYPE##_freelist_lock);				      \
									      \
    entry->next = TYPE##_freelist;					      \
    TYPE##_freelist = entry;						      \
									      \
    nexus_mutex_unlock(&TYPE##_freelist_lock);				      \
}


#endif	/* _freelist_h_ */
