/**********************************************************************
$Id: funcs.h,v 1.18 1995/07/07 22:05:42 mqxu Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/funcs.h,v $

$Log: funcs.h,v $
Revision 1.18  1995/07/07  22:05:42  mqxu
Version 2.4b7

Revision 1.17  1995/06/14  04:41:00  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.16  1995/05/12  20:26:17  tuecke
Updated to new Nexus.
Added AIX shared library support:
	- initwrap.c is necessary on AIX to avoid undefined external symbols
	- Moved sentinel code into here from ../sentinel

Revision 1.15  1995/04/17  06:30:09  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.14  1995/04/06  06:32:13  kohr
Added documentation, cleaned up a bit.


FM runtime global function declarations.
**********************************************************************/


#ifndef _fm_funcs_h
#define _fm_funcs_h


			/* Shouldn't this be in nexus.h?--DRK */
extern char *nexus_global_pointer_string(nexus_global_pointer_t *g);

extern void _fm_linktime_init();

extern int _fm_real_init(nexus_node_t *nodes, int n_nodes);

/*
Synch object functions.
*/

extern void _fm_synch_init(fm_synch_t *);
extern void _fm_synch_clear(fm_synch_t *);
extern int _fm_synch_count(fm_synch_t *);
extern void _fm_synch_increment(fm_synch_t *);
extern void _fm_synch_decrement(fm_synch_t *);
extern void _fm_synch_wait(fm_synch_t *);
extern void _fm_synch_destroy(fm_synch_t *);

/*
Nexus Handlers.
*/

extern void _fm_vc_init_handler(void *address, nexus_buffer_t *buffer);
extern void _fm_vc_shutdown_contexts_handler(void *address,
					     nexus_buffer_t *buffer);
extern void _fm_vc_shutdown_contexts_ack_handler(void *address,
						 nexus_buffer_t *buffer);
extern void _fm_create_process_handler(void *address, nexus_buffer_t *buffer);
extern void _fm_create_process_return_handler(void *address,
					      nexus_buffer_t *buffer);
extern void _fm_should_share_return_handler(void *addr, nexus_buffer_t *buf);
extern void _fm_port_send_handler(void *address, nexus_buffer_t *buffer);
extern void _fm_port_send_setup_handler(void *address, nexus_buffer_t *buffer);
extern void _fm_port_send_setup_return_handler(void *add, nexus_buffer_t *buf);
extern void _fm_process_return_handler(void *address, nexus_buffer_t *buffer);

extern void _fm_channel_migrating_handler(void *address, nexus_buffer_t *buf);
extern void _fm_channel_migrating_ack_handler(void *address,
					      nexus_buffer_t *buf);
extern void _fm_outport_migrating_handler(void *address, nexus_buffer_t *buf);
extern void _fm_outport_migrating_ack_handler(void *address,
					      nexus_buffer_t *buf);
extern void _fm_fwd_channel_migrating_handler(void *address,
					      nexus_buffer_t *buf);
extern void _fm_fwd_channel_migrating_ack_handler(void *address,
					      nexus_buffer_t *buf);
extern void _fm_fwd_outport_migrating_handler(void *address,
					      nexus_buffer_t *buf);
extern void _fm_fwd_outport_migrating_ack_handler(void *address,
					      nexus_buffer_t *buf);
extern void _fm_end_channel_handler(void *address, nexus_buffer_t *buf);
extern void _fm_end_channel_ack_handler(void *address,
					nexus_buffer_t *buf);
extern void _fm_channel_forward_buffer_handler(void *address,
					       nexus_buffer_t *buf);

extern int _fm_sizeof_outport(nexus_buffer_t *buffer,
                              f_integer_t *f_outports,
                              int n_outports,
                              int *n_elts);

extern void _fm_put_outport(nexus_buffer_t *buffer,
                           f_integer_t *f_outports,
                           int n_outports);

extern void _fm_get_outport(nexus_buffer_t *buffer,
                            f_integer_t *dest,
                            int n_outports);

extern int _fm_sizeof_inport(nexus_buffer_t *buffer,
			     f_integer_t *f_inports,
			     int n_inports,
			     int *n_elts);

extern void _fm_put_inport(nexus_buffer_t *buffer,
			   f_integer_t *f_inports,
			   int n_inports);

extern void _fm_get_inport(nexus_buffer_t *buffer,
			   f_integer_t *inports,
			   int n_inports,
			   void (*init_func)(fm_channel_t *));

extern void _fm_migrate_outport(f_integer_t *f_outport_list,
				int n_outports,
				nexus_global_pointer_t *dest_gp);

extern void _fm_create_ports(int n_ipls, fm_ipl_t *ipls, int n_total_inports,
			     int n_opls, fm_opl_t *opls, int n_total_outports,
			     nexus_global_pointer_t *dest_gp,
			     fm_synch_t *synch);

extern void _fm_notify_ports(int n_ipls, fm_ipl_t *ipls, int n_total_inports,
			     int n_opls, fm_opl_t *opls, int n_total_outports);

extern void
_fm_do_outport_migration_notification(fm_channel_t *channel,
				      nexus_global_pointer_t *old_outport,
				      nexus_global_pointer_t *new_outport);
extern void _fm_send_outport_migrating_notification(fm_outport_t *outport);

extern void
_fm_acknowledge_outport_migration_request(nexus_global_pointer_t *op_gp);

extern void _fm_invalidate_outport(fm_outport_t *outport);
extern void _fm_invalidate_inport(fm_inport_t *inport);

extern fm_inport_t *_fm_alloc_inport();
extern fm_outport_t *_fm_alloc_outport();
extern fm_channel_t *_fm_alloc_channel();

extern void _fm_free_inport(fm_inport_t *inport);
extern void _fm_free_outport(fm_outport_t *outport);
extern void _fm_free_channel(fm_channel_t *channel);

extern void _fm_set_channel_size(fm_channel_t *c, int n_outports);

extern void
_fm_channel_queueing_send(fm_channel_t *channel, nexus_buffer_t *buffer);

extern void
_fm_channel_direct_send(fm_channel_t *channel, nexus_buffer_t *buffer);


/*
FM message queues.
*/

extern void _fm_queue_module_init();
extern void _fm_ports_module_init();

extern void _fm_init_queue(fm_queue_t *queue);
extern void _fm_reinit_queue(fm_queue_t *queue);
extern void _fm_destroy_queue(fm_queue_t *queue);
void _fm_stash_and_enqueue(fm_queue_t *queue, nexus_buffer_t *buf);
void _fm_stash_and_enqueue_at_head(fm_queue_t *queue, nexus_buffer_t *buf);
extern fm_queue_entry_t *_fm_dequeue(fm_queue_t *queue);
extern fm_bool_t _fm_queue_empty(fm_queue_t *queue);
extern void _fm_free_queue_entry(fm_queue_entry_t *entry);


/*
Port-based send/receive
*/

extern void _fm_send(nexus_buffer_t *buffer, f_integer_t *f_ioval);
extern void _fm_init_send(fm_outport_t *outport, nexus_buffer_t *buffer,
			  f_integer_t *f_ioval);

extern void _fm_receive(f_integer_t *f_inport, void *destination,
			f_integer_t *f_ioval);


/*
Process block stuff.
*/

extern void _fm_pblock_finish_port_init(fm_pblock_t *pblock,
					fm_pblock_entry_t *entry);


/*
Fortran interface code.
*/

extern int FM_INIT();
extern void FM_START();
extern void FM_INIT_INPORT(f_integer_t *f_inports,
			   f_integer_t *f_n_inports);
extern void FM_INIT_INPORT_DEBUG(f_integer_t *f_inports,
				 f_integer_t *f_n_inports,
				 f_character_t port_name
				 CHAR_LEN_DECL(port_name_len));
extern void FM_INIT_OUTPORT(f_integer_t *f_outports,
			    f_integer_t *f_n_outports);
extern void FM_INIT_OUTPORT_DEBUG(f_integer_t *f_outports,
				  f_integer_t *f_n_outports,
				  f_character_t port_name
				  CHAR_LEN_DECL(port_name_len));
extern void FM_CHANNEL(f_integer_t *f_inport,
		       f_integer_t *f_outport,
		       void (*init_func)(fm_channel_t *),
		       f_integer_t *f_ioval);
extern void FM_MERGER_INIT(f_integer_t *f_channel,
			   f_integer_t *f_inport,
			   f_integer_t *f_size,
			   void (*init_func)(fm_channel_t *),
			   f_integer_t *f_ioval);
extern void FM_MERGER_ADD(f_integer_t *f_channel,
			  f_integer_t *f_outport,
			  f_integer_t *f_ioval);
extern void FM_MERGER_DONE(f_integer_t *f_channel,
			   f_integer_t *f_ioval);
extern void FM_MOVE_INPORT(f_integer_t *f_src, f_integer_t *f_dst,
			   f_integer_t *f_ioval);
extern void FM_MOVE_OUTPORT(f_integer_t *f_src, f_integer_t *f_dst,
			    f_integer_t *f_ioval);
extern void FM_INVALIDATE_INPORT(f_integer_t *f_inports,
				 f_integer_t *f_n_inports);
extern void FM_INVALIDATE_OUTPORT(f_integer_t *f_outports,
				  f_integer_t *f_n_outports);


/*
Virtual computers.
*/

extern nexus_global_pointer_t
    _fm_vc_compute_destination(f_integer_t *f_anno, f_integer_t *f_anno_len);
extern void _fm_vc_alloc();
extern void _fm_vc_init(nexus_global_pointer_t *node_gps, int n_node_gps);
extern void _fm_vc_table_offset(int);
extern void FM_VC_SHAPE(f_integer_t *f_shape, f_integer_t *f_shape_len);

extern fm_create_process_info_t *_fm_alloc_create_process_info();
extern void _fm_free_create_process_info(fm_create_process_info_t *);


/*
Utilities.
*/

extern void _fm_util_init();
extern char *_fm_util_get_string();
extern char *_fm_inport_string(fm_inport_t *);
extern char *_fm_outport_string(fm_outport_t *);
extern char *_fm_channel_string(fm_channel_t *);


/*
Debugging support for FM programs.
*/

#ifdef FM_DEBUG
extern void FM_PAUSE(f_integer_t *f_id,
		     f_character_t f_msg CHAR_LEN_DECL(f_msg_len));
extern void _fm_pause(int id, char *msg);
extern void _fm_pause_init();
extern int _fm_pause_params(int n, char *buf, int size);
extern void _fm_set_show_pausepoints(fm_bool_t val);
extern void
_fm_sizeof_pausepoints(nexus_buffer_t *buf, int *size, int *count);
extern void _fm_put_pausepoints(nexus_buffer_t *buf);
extern void _fm_get_pausepoints(nexus_buffer_t *buf);
extern void _fm_add_pausepoint(int id);
extern void _fm_add_sleeppoint(int id, int secs);
#endif


#endif /* _fm_funcs_h */
