/**********************************************************************
$Id: migrate.c,v 1.6 1995/04/17 06:30:16 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/migrate.c,v $

$Log: migrate.c,v $
Revision 1.6  1995/04/17  06:30:16  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.5  1995/04/06  06:32:20  kohr
Added documentation, cleaned up a bit.


FM runtime port migration code.
**********************************************************************/


#include "fm.h"

static void notify_for_outport(fm_outport_t *outport);
static void notify_for_inport(fm_inport_t *inport);
static void forward_channel_contents(fm_inport_t *inport);

/*
_fm_migrate_outport() : obsolete version of port migration started by Bob,
not finished yet.--DRK
*/

#if 0
void _fm_migrate_outport(f_integer_t
*f_outport_list,
			 int n_outports,
			 nexus_global_pointer_t *dest_gp)
{
    nexus_buffer_t mbuf;
    int size, n_elts, n_elts;

    nexus_init_remote_service_request(&mbuf,
				      &entry->dest_gp,
				      FM_PORT_SEND_SETUP_NAME,
				      FM_PORT_SEND_SETUP_HASH,
				      0);

    size = n_elts = 0;

#ifdef FM_DEBUG
    n_ints = 3;
#else
    n_ints = 2;
#endif

    size += nexus_sizeof_int(&mbuf, n_ints);
    n_elts += n_ints;

    nexus_global_pointer(&synch_gp, (void *) &synch);

    size += nexus_sizeof_global_pointer(&mbuf, &synch_gp, 1, &obj_n_elts);
    n_elts += obj_n_elts;

    

} /* _fm_migrate_outport() */
#endif


/*
Overview of what port migr. functions are where should go here.  Also
description of what still needs to be changed for new protocol--DRK
*/

/*
_fm_create_ports() : Create a set of ports on the same context as a given
global pointer.  The set of ports could either be sent over a channel, or
passed as arguments to a new process (i.e., the set of ports in one entry
of a pblock).
*/

void _fm_create_ports(int n_ipls, fm_ipl_t *ipls, int n_total_inports,
		      int n_opls, fm_opl_t *opls, int n_total_outports,
		      nexus_global_pointer_t *dest_gp,
		      fm_synch_t *synch)
{
    nexus_buffer_t mbuf;
    int size, n_elts, obj_n_elts;
    int i, j;
    fm_ipl_t *ipl;
    fm_opl_t *opl;
    fm_inport_t *inport;
    fm_channel_t *channel;
    fm_outport_t *outport;
    int n_ints, n_gps, gp_size, gp_n_elts;
    nexus_global_pointer_t *gp_list, *synch_gp, *gp;
    int gp_list_next, gp_list_size;
#ifdef FM_DEBUG
    int name_len;
#endif

#ifdef FM_DEBUG
    if (FMDebug(Migration, 1))
	nexus_printf("_fm_create_ports(): n_ipls=%d n_total_inports=%d n_opls=%d n_total_outports=%d dest_gp=%s\n",
		     n_ipls, n_total_inports, n_opls, n_total_outports,
		     nexus_global_pointer_string(dest_gp));
#endif

    nexus_init_remote_service_request(&mbuf,
				      dest_gp,
				      FM_PORT_SEND_SETUP_NAME,
				      FM_PORT_SEND_SETUP_HASH);

    size = n_elts = 0;

    /*
     * Construct the message. For each inport we send
     *    inport name
     *    channel name
     *    gp to original inport
     *    gp to original channel
     *    number of outport references in channel
     *    outport references
     *    number of messages in the channel
     *    repacked messages
     *
     * (The latter two may actually get sent as separate messages so
     *       
     * as to allow the reuse of the existing port send code)
     *
     * The size of each inport is
     *  sizeof(int) * 2 +		[for string lengths]	DEBUG
     *  strlen(inport_name) + 					DEBUG
     *  strlen(channel_name) +					DEBUG
     *  sizeof(gp to dest) * 2 +
     *  sizeof(int) + sizeof(outport ref list)
     *
     * For each outport we send
     *    outport name
     *    gp to original outport
     *    channel reference
     *
     * There is also a gp to a synch object passed, once in the message,
     * as well as integers for the number of inports & number of 
     * outports.
     *
     * The size of each outport is
     *  sizeof(int) + 			[for string length] 	DEBUG
     *  strlen(outport_name) +					DEBUG
     *  sizeof(gp to dest) +
     *  sizeof(channel ref)
     *
     * Hence the total message size for N inports and M outports is
     * (for the debugging case)
     *
     * sum[1,N](strlen(inport names) + strlen(channel names)) +
     * sum[1,M](strlen(outport names)) +
     * sizeof(int) * (3N + M + 2) +
     * sizeof(gp to dest) * (2N + M) +
     * sum[1,N] sizeof(outport ref list) +
     * sum[1,M] sizeof(channel ref) 
     *
     * or (for the non-debugging case)
     *
     * sizeof(int) * (N + 2)+
     * sizeof(gp to dest) * (2N + M) +
     * sum[1,N] sizeof(outport ref list) +
     * sum[1,M] sizeof(channel ref) 
     *
     * We set the object's forwarded flag. If a channel send comes in
     * when the forwarded flag is set, it is repacked and sent on
     * to the channel's future home. We can ignore the forwarded flag
     * once the migration notification ack has been received.
     *
     * What if we get a migration notification for an outport or
     * inport after the new object has been created but before the
     * notification has been sent? We need to again forward the
     * notification to the new data structure. 
     * 
     * 
     */

    size = n_elts = 0;

#ifdef FM_DEBUG
    n_ints = 3 * n_total_inports + n_total_outports + 2;
#else
    n_ints = n_total_inports + 2;
#endif

    size += nexus_sizeof_int(&mbuf, n_ints);
    n_elts += n_ints;

    n_gps = 2 * n_total_inports + n_total_outports + 1;

    gp_list_next = 0;
    gp_list_size = n_gps;
    NexusMalloc(create_ports_for_entry(),
		gp_list,
		nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t) * gp_list_size);

    synch_gp = &(gp_list[gp_list_next++]);
    nexus_global_pointer(synch_gp, (void *) synch);
    
    gp_size = nexus_sizeof_global_pointer(&mbuf, synch_gp, 1, &gp_n_elts);
    size += gp_size * n_gps;
    n_elts += gp_n_elts * n_gps;


    for (i = 0; i < n_ipls; i++)
    {
	ipl = &(ipls[i]);
	for (j = 0; j < ipl->n_entries; j++)
	{
	    inport = ipl->entries[j];
	    InportMagicCheck(inport);

	    channel = (fm_channel_t *)
		nexus_convert_global_pointer_address(&inport->channel_gp);
	    ChannelMagicCheck(channel);

#ifdef FM_DEBUG
	    if (FMDebug(Migration, 2))
	    {
		nexus_printf("_fm_create_ports(): i=%d j=%d inport=%s channel=%s\n",
			     i, j,
			     _fm_inport_string(inport),
			     _fm_channel_string(channel));
	    }
#endif

	    size += nexus_sizeof_global_pointer(&mbuf,
						channel->outports,
						channel->n_outports,
						&obj_n_elts);
	    n_elts += obj_n_elts;

#ifdef FM_DEBUG
	    size += nexus_sizeof_char(&mbuf,
				      strlen(inport->name) +
				      strlen(channel->name));
	    n_elts += 2;
#endif
	}
    }

    for (i = 0; i < n_opls; i++)
    {
	opl = &(opls[i]);
	for (j = 0; j < opl->n_entries; j++)
	{
	    outport = opl->entries[j];
	    OutportMagicCheck(outport);


#ifdef FM_DEBUG
	    if (FMDebug(Migration, 2))
	    {
		nexus_printf("_fm_create_ports(): i=%d j=%d outport=%s\n",
			     i, j,
			     _fm_outport_string(outport));
	    }
#endif
	    size += nexus_sizeof_global_pointer(&mbuf,
						&outport->channel_gp,
						1,
						&obj_n_elts);
	    n_elts += obj_n_elts;

#ifdef FM_DEBUG
	    size += nexus_sizeof_char(&mbuf, strlen(outport->name));
	    n_elts++;
#endif
	}
    }

    nexus_set_buffer_size(&mbuf, size, n_elts);

    /*
     * The buffer is packed as follows:
     * 
     *    number of inports
     *    inports packed as:
     *       inport name len
     *       inport name
     *       channel name len
     *       channel name
     *       gp to inport
     *       gp to channel
     *       number of outports in channel
     *       channel outports
     *     number of outports
     *     outports packed as:
     *       outport name len
     *       outport name
     *       gp to outport
     *       channel reference
     *     gp to synch
     *
     */

    nexus_put_int(&mbuf, &n_total_inports, 1);

    for (i = 0; i < n_ipls; i++)
    {
	ipl = &(ipls[i]);
	for (j = 0; j < ipl->n_entries; j++)
	{
	    inport = ipl->entries[j];
	    InportMagicCheck(inport);

	    channel = (fm_channel_t *)
		nexus_convert_global_pointer_address(&inport->channel_gp);
	    ChannelMagicCheck(channel);

#ifdef FM_DEBUG
	    name_len = strlen(inport->name);
	    nexus_put_int(&mbuf, &name_len, 1);
	    nexus_put_char(&mbuf, inport->name, name_len);

	    name_len = strlen(channel->name);
	    nexus_put_int(&mbuf, &name_len, 1);
	    nexus_put_char(&mbuf, channel->name, name_len);
#endif
	    gp = &(gp_list[gp_list_next++]);
	    nexus_global_pointer(gp, (void *) inport);
	    nexus_put_global_pointer(&mbuf, gp, 1);

	    gp = &(gp_list[gp_list_next++]);
	    nexus_global_pointer(gp, (void *) channel);
	    nexus_put_global_pointer(&mbuf, gp, 1);

	    nexus_put_int(&mbuf, &channel->n_outports, 1);
	    nexus_put_global_pointer(&mbuf,
				     channel->outports,
				     channel->n_outports);

	}
    }

    nexus_put_int(&mbuf, &n_total_outports, 1);

    for (i = 0; i < n_opls; i++)
    {
	opl = &(opls[i]);
	for (j = 0; j < opl->n_entries; j++)
	{
	    outport = opl->entries[j];
	    OutportMagicCheck(outport);

#ifdef FM_DEBUG
	    name_len = strlen(outport->name);
	    nexus_put_int(&mbuf, &name_len, 1);
	    nexus_put_char(&mbuf, outport->name, name_len);
#endif

	    NexusAssert(gp_list_next < gp_list_size);
	    gp = &(gp_list[gp_list_next++]);
	    nexus_global_pointer(gp, (void *) outport);
	    nexus_put_global_pointer(&mbuf, gp, 1);
	    nexus_put_global_pointer(&mbuf, &outport->channel_gp, 1);
	}
    }
    
    
    _fm_synch_clear(synch);
    _fm_synch_increment(synch);

    nexus_put_global_pointer(&mbuf, synch_gp, 1);

    nexus_send_remote_service_request(&mbuf);

    for (i = 0; i < gp_list_next; i++)
	nexus_destroy_global_pointer(&gp_list[i]);

    NexusFree(gp_list);
} /* _fm_create_ports() */


/*
_fm_port_send_setup_handler() : Process a request to set up a set of ports
and channels.
*/

void _fm_port_send_setup_handler(void *address, nexus_buffer_t *buffer)
{
    int i;
    int n_inports, n_outports;
    nexus_global_pointer_t synch_gp;
    fm_inport_t *inport;
    fm_channel_t *channel;
    fm_outport_t *outport;

    fm_inport_t **inport_list;
    fm_channel_t **channel_list;
    fm_outport_t **outport_list;
    nexus_buffer_t mbuf;
    int size, n_elts, obj_n_elts;
    int n_gps, gp_size, gp_n_elts;
    int channel_n_outports;
    nexus_global_pointer_t lgp, *gp;

    FM_GPL_DEFS
#ifdef FM_DEBUG
    int name_len;
#endif

    nexus_get_int(buffer, &n_inports, 1);

    NexusMalloc(_fm_port_send_setup_handler(), inport_list,
		fm_inport_t **, sizeof(fm_inport_t *) * n_inports);

    NexusMalloc(_fm_port_send_setup_handler(), channel_list,
		fm_channel_t **, sizeof(fm_channel_t *) * n_inports);

    for (i = 0; i < n_inports; i++)
    {
	inport_list[i] = inport = _fm_alloc_inport();
	channel_list[i] = channel = _fm_alloc_channel();

#ifdef FM_DEBUG
	nexus_get_int(buffer, &name_len, 1);
	nexus_get_char(buffer, inport->name, name_len);
	inport->name[name_len] = '\0';

	nexus_get_int(buffer, &name_len, 1);
	nexus_get_char(buffer, channel->name, name_len);
	channel->name[name_len] = '\0';
#endif

	nexus_get_global_pointer(buffer, &inport->old_home, 1);
	nexus_get_global_pointer(buffer, &channel->old_home, 1);

	nexus_get_int(buffer, &channel_n_outports, 1);
	_fm_set_channel_size(channel, channel_n_outports);

	nexus_get_global_pointer(buffer,
				 channel->outports,
				 channel_n_outports);
	channel->n_outports = channel_n_outports;

	channel->send_func = _fm_channel_queueing_send;

	nexus_global_pointer(&channel->inport,
			     (void *) inport);
	
	nexus_global_pointer(&inport->channel_gp,
			     (void *) channel);

	ChannelSetState(channel,  FM_STATE_AWAITING_ACTIVATION);
	InportSetState(inport,  FM_STATE_AWAITING_ACTIVATION);
    }

    nexus_get_int(buffer, &n_outports, 1);

    NexusMalloc(_fm_port_send_setup_handler(), outport_list,
		fm_outport_t **, sizeof(fm_outport_t *) * n_outports);

    for (i = 0; i < n_outports; i++)
    {
	outport_list[i] = outport = _fm_alloc_outport();
	
#ifdef FM_DEBUG
	nexus_get_int(buffer, &name_len, 1);
	nexus_get_char(buffer, outport->name, name_len);
	outport->name[name_len] = '\0';
#endif

	nexus_get_global_pointer(buffer, &outport->old_home, 1);
	nexus_get_global_pointer(buffer, &outport->channel_gp, 1);

	OutportSetState(outport, FM_STATE_AWAITING_ACTIVATION);
    }

    nexus_get_global_pointer(buffer, &synch_gp, 1);

    /*
     * Pack up the return message:
     *
     *    n_inports
     *    for each inport,
     *       original inport gp
     *       new inport gp
     *       new channel gp
     *    n_outports
     *    for each outport
     *       original outport gp
     *       new outport gp
     */

    n_gps = 2 * n_inports + n_outports;
    FM_GPL_INIT(_fm_port_send_setup_handler(), n_gps);

    nexus_init_remote_service_request(&mbuf,
				      &synch_gp,
				      FM_PORT_SEND_SETUP_RETURN_NAME,
				      FM_PORT_SEND_SETUP_RETURN_HASH);

    size = n_elts = 0;

    nexus_global_pointer(&lgp, (void *) &size);

    gp_size = nexus_sizeof_global_pointer(&mbuf, &lgp, 1, &gp_n_elts);

    nexus_destroy_global_pointer(&lgp);

    size += gp_size * n_gps;
    n_elts += gp_n_elts * n_gps;

    size += nexus_sizeof_int(&mbuf, 2);	/* n_inports & n_outports */
    n_elts += 2;

    for (i = 0; i < n_inports; i++)
    {
	size += nexus_sizeof_global_pointer(&mbuf,
					    &inport_list[i]->old_home,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;
    }	    

    for (i = 0; i < n_outports; i++)
    {
	size += nexus_sizeof_global_pointer(&mbuf,
					    &outport_list[i]->old_home,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;
    }	    

    nexus_set_buffer_size(&mbuf, size, n_elts);

    nexus_put_int(&mbuf, &n_inports, 1);

    for (i = 0; i < n_inports; i++)
    {
	nexus_put_global_pointer(&mbuf, &inport_list[i]->old_home, 1);
	FM_GPL_ADD(gp);
	nexus_global_pointer(gp, (void *) inport_list[i]);
	nexus_put_global_pointer(&mbuf, gp, 1);

	FM_GPL_ADD(gp);
	nexus_global_pointer(gp, (void *) channel_list[i]);
	nexus_put_global_pointer(&mbuf, gp, 1);
    }	    

    nexus_put_int(&mbuf, &n_outports, 1);

    for (i = 0; i < n_outports; i++)
    {
	nexus_put_global_pointer(&mbuf, &outport_list[i]->old_home, 1);
	FM_GPL_ADD(gp);
	nexus_global_pointer(gp, (void *) outport_list[i]);
	nexus_put_global_pointer(&mbuf, gp, 1);
    }

    nexus_send_remote_service_request(&mbuf);

    FM_GPL_FREE;
    
} /* _fm_port_send_setup_handler() */


/*
_fm_port_send_setup_return_handler() : Handle the return from the port send
setup.  Unpack the buffer and set the future home fields accordingly.
*/

void _fm_port_send_setup_return_handler(void *address, nexus_buffer_t *buffer)
{
    fm_synch_t *synch;
    nexus_global_pointer_t old_inport_gp, new_inport_gp, new_channel_gp;
    nexus_global_pointer_t old_outport_gp, new_outport_gp;

    fm_inport_t *inport;
    fm_outport_t *outport;
    fm_channel_t *channel;
    int i;

    int n_inports, n_outports;

    synch = (fm_synch_t *) address;

    SynchMagicCheck(synch);

#ifdef FM_DEBUG
    if (FMDebug(Ports, 1))
	nexus_printf("port_send_setup_return_handler(): synch=%x\n", synch);
#endif

    nexus_get_int(buffer, &n_inports, 1);

    for (i = 0; i < n_inports; i++)
    {
	nexus_get_global_pointer(buffer, &old_inport_gp, 1);
	nexus_get_global_pointer(buffer, &new_inport_gp, 1);
	nexus_get_global_pointer(buffer, &new_channel_gp, 1);

	inport = (fm_inport_t *)
	    nexus_convert_global_pointer_address(&old_inport_gp);

	InportMagicCheck(inport);
	
#ifdef FM_DEBUG
	if (FMDebug(Migration, 2))
	    nexus_printf("port_send_setup_return_handler(): got inport=%s new_inport_gp=%s new_channel_gp=%s\n",
			 _fm_inport_string(inport),
			 nexus_global_pointer_string(&new_inport_gp),
			 nexus_global_pointer_string(&new_channel_gp));
#endif

	channel = (fm_channel_t *)
	    nexus_convert_global_pointer_address(&inport->channel_gp);

	ChannelMagicCheck(channel);

	InportLock(inport);
	ChannelLock(channel);

	inport->future_home = new_inport_gp;

	channel->future_home = new_channel_gp;

	ChannelUnlock(channel);
	InportUnlock(inport);
    }

    nexus_get_int(buffer, &n_outports, 1);

    for (i = 0; i < n_outports; i++)
    {
	nexus_get_global_pointer(buffer, &old_outport_gp, 1);
	nexus_get_global_pointer(buffer, &new_outport_gp, 1);

	outport = (fm_outport_t *)
	    nexus_convert_global_pointer_address(&old_outport_gp);

	OutportMagicCheck(outport);

#ifdef FM_DEBUG
	if (FMDebug(Migration, 2))
	    nexus_printf("port_send_setup_return_handler(): got outport=%s new_outport_gp=%s\n",
			 _fm_outport_string(outport),
			 nexus_global_pointer_string(&new_outport_gp));
#endif
	OutportLock(outport);
	outport->future_home = new_outport_gp;
	OutportUnlock(outport);
    }

    _fm_synch_decrement(synch);
    
} /* _fm_port_send_setup_return_handler() */


/*
_fm_notify_ports() : Perform the appropriate notifications for a set of
ports. We assume that each migrating port has a valid future home.
*/

void _fm_notify_ports(int n_ipls, fm_ipl_t *ipls, int n_total_inports,
		      int n_opls, fm_opl_t *opls, int n_total_outports)
{
    int i, j;
    fm_inport_t *inport;
    fm_outport_t *outport;
    fm_ipl_t *ipl;
    fm_opl_t *opl;

    for (i = 0; i < n_ipls; i++)
    {
	ipl = &(ipls[i]);

	for (j = 0; j < ipl->n_entries; j++)
	{
	    inport = ipl->entries[j];
	    notify_for_inport(inport);
	    forward_channel_contents(inport);
	}
    }
    for (i = 0; i < n_opls; i++)
    {
	opl = &(opls[i]);

	for (j = 0; j < opl->n_entries; j++)
	{
	    outport = opl->entries[j];
	    notify_for_outport(outport);
	}
    }
} /* _fm_notify_ports() */


/*
notify_for_inport() : Notify the outports for the channel associated with
this inport that the inport & channel are migrating.
*/

static void notify_for_inport(fm_inport_t *inport)
{
    fm_channel_t *channel;
    nexus_global_pointer_t *gp;
    int i;
    nexus_buffer_t mbuf;
    int size, n_elts, obj_n_elts;

    NexusAssert(nexus_global_pointer_to_current_context(&inport->channel_gp));

    channel = (fm_channel_t *)
	nexus_convert_global_pointer_address(&inport->channel_gp);
    ChannelMagicCheck(channel);

#ifdef FM_DEBUG
    if (FMDebug(Ports, 1))
	nexus_printf("notify_for_inport(): inport=%s channel=%s\n",
		     _fm_inport_string(inport),
		     _fm_channel_string(channel));
#endif

    /*
     * We can leave the channel locked thru the sending of the
     * migration notifications because the migration ack handler
     * doesn't need to lock the channel, just bump the sych.
     *
     * In order for this to work, we need to check for the
     * FM_STATE_WAITING_FOR_NOTIFY_PORTS_ACK in the outport-migrating
     * handler *before* we lock the channel there.
     *
     * However, that leaves an unlocked check on a state variable.
     * The other solution is to unlock the channel before sending the
     * rsr, and locking again afterward.
     */

    ChannelLock(channel);
    ChannelSetState(channel, FM_STATE_WAITING_FOR_NOTIFY_PORTS_ACK);

    _fm_synch_clear(&channel->synch);

    for (i = 0; i < channel->n_outports; i++)
    {
	gp = &(channel->outports[i]);

	nexus_init_remote_service_request(&mbuf,
					  gp,
					  FM_CHANNEL_MIGRATING_NAME,
					  FM_CHANNEL_MIGRATING_HASH);

	size = n_elts = 0;

	size += nexus_sizeof_global_pointer(&mbuf, &inport->channel_gp,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;

	size += nexus_sizeof_global_pointer(&mbuf, &channel->future_home,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;

	nexus_set_buffer_size(&mbuf, size, n_elts);

	nexus_put_global_pointer(&mbuf, &inport->channel_gp, 1);
	nexus_put_global_pointer(&mbuf, &channel->future_home, 1);

	_fm_synch_increment(&channel->synch);

#ifdef FM_DEBUG
	if (FMDebug(Migration, 2))
	{
	    nexus_printf("sending CHANNEL_MIGRATING rsr to %s, %s -> %s\n",
			 nexus_global_pointer_string(gp),
			 nexus_global_pointer_string(&inport->channel_gp),
			 nexus_global_pointer_string(&channel->future_home));
	}
#endif
	ChannelUnlock(channel);
	nexus_send_remote_service_request(&mbuf);
	ChannelLock(channel);
    }

    ChannelUnlock(channel);

    /*
     * For now we will wait here. Eventually this wait can be
     * shifted until after all notifications have been sent.
     */

#ifdef FM_DEBUG
    if (FMDebug(Migration, 2))
	nexus_printf("notify_for_inport(): Waiting for synch %x\n",
		     &channel->synch);
#endif

    _fm_synch_wait(&channel->synch);

    ChannelLockAndSetState(channel, FM_STATE_NORMAL);

#ifdef FM_DEBUG
    if (FMDebug(Migration, 2))
	nexus_printf("notify_for_inport(): Got synch %x\n", &channel->synch);
#endif
    
} /* notify_for_inport */


/*
_fm_channel_migrating_handler() : This handler is invoked on an outport
when its channel is migrating. If the state of the outport is
WaitingForNotifyChannelAck, then we know that our channel notification
request has been dropped by the channel (because it was waiting for a
channel migrating ack) and we need to resend the notification after acking
the channel migration.
*/

void _fm_channel_migrating_handler(void *address, nexus_buffer_t *buffer)
{
    fm_outport_t *outport;
    nexus_global_pointer_t gps[2], *channel_gp, *channel_fh_gp;
    nexus_buffer_t mbuf;
    int size, n_elts, obj_n_elts;

    nexus_get_global_pointer(buffer, gps, 2);

    channel_gp = &(gps[0]);
    channel_fh_gp = &(gps[1]);

    outport = (fm_outport_t *) address;
    OutportMagicCheck(outport);

    OutportLock(outport);

#ifdef FM_DEBUG
    if (FMDebug(Migration, 1))
    {
	nexus_printf("ChannelMigrating: channel_gp=%s\n",
		     nexus_global_pointer_string(channel_gp));
	nexus_printf("                  channel_fh_gp=%s\n",
		     nexus_global_pointer_string(channel_fh_gp));
	nexus_printf("                  outport=%s\n",
		     _fm_outport_string(outport));
    }
#endif

    /*
     * Verify that we have the right channel.
     */

    NexusAssert(nexus_same_global_pointer(channel_gp,
					  &outport->channel_gp));

    outport->channel_gp = *channel_fh_gp;

    /*
     * If there is a future home, then the update that we need to
     * do is really on the future home side.
     *
     * We'll do the update locally, too, so that the notification that
     * is going to happen (but hasn't happened yet) goes to the
     * right place.
     *
     * How do we know it hasn't happened yet? If it had, we wouldn't
     * get this message.
     */
    if (!nexus_is_null_global_pointer(&outport->future_home))
    {
	nexus_buffer_t mbuf;
	nexus_global_pointer_t outport_gp;

	_fm_synch_increment(&outport->busy_synch);
	
#ifdef FM_DEBUG
	if (FMDebug(Migration, 1))
	    nexus_printf("_fm_channel_migrating_handler(): forwarding channel_migrating message to future home\n");
#endif

	/*
	 * Send off a forwarded channel migrating message to the outport's
	 * future home.
	 *
	 * This contains the old and new channel gps, as well as the
	 * outport's current gp. (tho we could prolly find that as
	 * the outport's old_home)
	 *
	 * The new outport will reply with a FWD_CHANNEL_MIGRATING_ACK
	 * whose handler sends the CHANNEL_MIGRATING_ACK and resets
	 * the outport's busy flag.
	 */

	nexus_init_remote_service_request(&mbuf,
					  &outport->future_home,
					  FM_FWD_CHANNEL_MIGRATING_NAME,
					  FM_FWD_CHANNEL_MIGRATING_HASH);

	nexus_global_pointer(&outport_gp, (void *) outport);

	size = n_elts = 0;

	size += nexus_sizeof_global_pointer(&mbuf, channel_gp,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;

	size += nexus_sizeof_global_pointer(&mbuf, channel_fh_gp,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;

	size += nexus_sizeof_global_pointer(&mbuf, &outport_gp,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;

	nexus_set_buffer_size(&mbuf, size, n_elts);

	nexus_put_global_pointer(&mbuf, channel_gp, 1);
	nexus_put_global_pointer(&mbuf, channel_fh_gp, 1);	
	nexus_put_global_pointer(&mbuf, &outport_gp, 1);

	OutportUnlock(outport);
	
	nexus_send_remote_service_request(&mbuf);

	nexus_destroy_global_pointer(&outport_gp);
    }
    else
    {
	nexus_init_remote_service_request(&mbuf,
					  channel_gp,
					  FM_CHANNEL_MIGRATING_ACK_NAME,
					  FM_CHANNEL_MIGRATING_ACK_HASH);

	nexus_send_remote_service_request(&mbuf);

	/*
	 * Check to see if we're waiting for a notify ports acknowledgement.
	 * If we are are, we need to resend the request because the
	 * channel dropped the message.
	 */
	if (OutportState(outport) == FM_STATE_WAITING_FOR_NOTIFY_PORTS_ACK)
	{
#ifdef FM_DEBUG
	    if (FMDebug(Ports, 1))
		nexus_printf("_fm_channel_migrating_handler(): resending outport migrating notification for %s\n",
			     _fm_outport_string(outport));
#endif
	    _fm_send_outport_migrating_notification(outport);
	}

	OutportUnlock(outport);
    }
} /* _fm_channel_migrating_handler() */


/*
_fm_fwd_channel_migrating_handler() : Handle a forwarded channel migration
message. This is received by a new outport which hasn't been completely
migrated to yet. Do the update and reply to the old outport with a
FWD_MIGRATING_ACK.
*/

void _fm_fwd_channel_migrating_handler(void *address, nexus_buffer_t *buffer)
{
    nexus_global_pointer_t channel_gp, channel_fh_gp, outport_gp;
    fm_outport_t *outport;
    nexus_buffer_t mbuf;
    int size, n_elts, obj_n_elts;

    outport = (fm_outport_t *) address;
    OutportMagicCheck(outport);
    
    nexus_get_global_pointer(buffer, &channel_gp, 1);
    nexus_get_global_pointer(buffer, &channel_fh_gp, 1);
    nexus_get_global_pointer(buffer, &outport_gp, 1);

#ifdef FM_DEBUG
    if (FMDebug(Migration, 1))
	nexus_printf("_fm_fwd_channel_migrating_handler(): got outport=%s channel_gp=%s channel_fh_gp=%s outport_gp=%s\n",
		     _fm_outport_string(outport),
		     nexus_global_pointer_string(&channel_gp),
		     nexus_global_pointer_string(&channel_fh_gp),
		     nexus_global_pointer_string(&outport_gp));
#endif

    OutportLock(outport);

    NexusAssert(nexus_same_global_pointer(&channel_gp, &outport->channel_gp));
    NexusAssert(nexus_same_global_pointer(&outport_gp, &outport->old_home));

    outport->channel_gp = channel_fh_gp;

    nexus_init_remote_service_request(&mbuf,
				      &outport_gp, 
				      FM_FWD_CHANNEL_MIGRATING_ACK_NAME,
				      FM_FWD_CHANNEL_MIGRATING_ACK_HASH);

    size = n_elts = 0;
    
    size += nexus_sizeof_global_pointer(&mbuf, &channel_gp, 1, &obj_n_elts);
    n_elts += obj_n_elts;

    nexus_set_buffer_size(&mbuf, size, n_elts);

    nexus_put_global_pointer(&mbuf, &channel_gp, 1);

    OutportUnlock(outport);
    
    nexus_send_remote_service_request(&mbuf);
} /* _fm_fwd_channel_migrating_handler() */


/*
_fm_fwd_channel_migrating_ack_handler() : stuff goes here--DRK
*/

void _fm_fwd_channel_migrating_ack_handler(void *address,
					   nexus_buffer_t *buffer)
{
    fm_outport_t *outport;
    nexus_global_pointer_t channel_gp;
    nexus_buffer_t mbuf;
    
    outport = (fm_outport_t *) address;
    OutportMagicCheck(outport);

    nexus_get_global_pointer(buffer, &channel_gp, 1);

#ifdef FM_DEBUG
    if (FMDebug(Migration,1 ))
	nexus_printf("_fm_fwd_channel_migrating_ack_handler(): got outport=%s channel_gp=%s\n",
		     _fm_outport_string(outport),
		     nexus_global_pointer_string(&channel_gp));
#endif
    
    OutportLock(outport);
    NexusAssert(_fm_synch_count(&outport->busy_synch) > 0);
    
    nexus_init_remote_service_request(&mbuf,
				      &channel_gp,
				      FM_CHANNEL_MIGRATING_ACK_NAME,
				      FM_CHANNEL_MIGRATING_ACK_HASH);

    nexus_send_remote_service_request(&mbuf);

    /*
     * Check to see if we're waiting for a notify ports acknowledgement.
     * If we are are, we need to resend the request because the
     * channel dropped the message.
     */
    if (OutportState(outport) == FM_STATE_WAITING_FOR_NOTIFY_PORTS_ACK)
    {
#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_fwd_channel_migrating_ack_handler(): resending outport migrating notification for %s\n",
			 _fm_outport_string(outport));
#endif
	_fm_send_outport_migrating_notification(outport);
    }

    _fm_synch_decrement(&outport->busy_synch);
    OutportSignal(outport);
    
    OutportUnlock(outport);

} /* _fm_fwd_channel_migrating_ack_handler() */


/*
_fm_channel_migrating_ack_handler() : Handler for the acknowledgement of a
channel_migrating message. Just decrement the synch in the channel. 
*/

void _fm_channel_migrating_ack_handler(void *address, nexus_buffer_t *buf)
{
    fm_channel_t *channel;

    channel = (fm_channel_t *) address;

    ChannelMagicCheck(channel);

    SynchMagicCheck(&channel->synch);

    _fm_synch_decrement(&channel->synch);
    
} /* _fm_channel_migrating_ack_handler() */


/*
notify_for_outport() : Notify the channel associated with this outport that
the outport is migrating.
*/

static void notify_for_outport(fm_outport_t *outport)
{
    OutportMagicCheck(outport);

#ifdef FM_DEBUG
    if (FMDebug(Ports, 1))
	nexus_printf("notify_for_outport(): outport=%s\n",
		     _fm_outport_string(outport));
#endif

    NexusAssert(!nexus_is_null_global_pointer(&outport->future_home));

    _fm_synch_wait(&outport->busy_synch);
    
    OutportLock(outport);

    OutportSetState(outport, FM_STATE_WAITING_FOR_NOTIFY_PORTS_ACK);

    _fm_send_outport_migrating_notification(outport);
    OutportUnlock(outport);

    _fm_synch_wait(&outport->synch);

    OutportLockAndSetState(outport, FM_STATE_NORMAL);

} /* notify_for_outport */


/*
_fm_send_outport_migrating_notification() : Pack up and send an
OUTPORT_MIGRATING handler to the outport's channel. This is a utility
routine; assumes the outport is locked and everything is valid.
*/

void _fm_send_outport_migrating_notification(fm_outport_t *outport)
{
    nexus_buffer_t mbuf;
    nexus_global_pointer_t outport_gp;
    int size, n_elts, obj_n_elts;

    nexus_init_remote_service_request(&mbuf,
				      &outport->channel_gp,
				      FM_OUTPORT_MIGRATING_NAME,
				      FM_OUTPORT_MIGRATING_HASH);

    nexus_global_pointer(&outport_gp, outport);

    size = n_elts = 0;
    size += nexus_sizeof_global_pointer(&mbuf, &outport_gp,
					1, &obj_n_elts);
    n_elts += obj_n_elts;
    
    size += nexus_sizeof_global_pointer(&mbuf, &outport->future_home,
					1, &obj_n_elts);
    n_elts += obj_n_elts;

    nexus_set_buffer_size(&mbuf, size, n_elts);

    nexus_put_global_pointer(&mbuf, &outport_gp, 1);
    nexus_put_global_pointer(&mbuf, &outport->future_home, 1);

    _fm_synch_clear(&outport->synch);
    _fm_synch_increment(&outport->synch);
    nexus_send_remote_service_request(&mbuf);

    nexus_destroy_global_pointer(&outport_gp);
}


/*
_fm_outport_migrating_handler() :
     lock channel state
     if state == WAITING_FOR_NOTIFY_PORTS_ACK
 	  unlock state & return from handler
 	  (we are dropping this message)
     unpack message into outport_old and outport_new
     find outport reference in list
     replace outport ref with outport_new
     init OUTPORT_MIGRATING_ACK rsr to outport_old
     invoke rsr
     unlock channel state
*/

void _fm_outport_migrating_handler(void *address, nexus_buffer_t *buffer)
{
    fm_channel_t *channel;
    nexus_global_pointer_t outport_gp, outport_fh_gp;

    channel = (fm_channel_t *) address;

    ChannelMagicCheck(channel);

    ChannelLock(channel);

    if (ChannelState(channel) == FM_STATE_WAITING_FOR_NOTIFY_PORTS_ACK)
    {
#ifdef FM_DEBUG
	if (FMDebug(Migration, 2))
	    nexus_printf("_fm_outport_migrating_handler(): dropping message\n");
#endif
	ChannelUnlock(channel);
	return;
    }

    nexus_get_global_pointer(buffer, &outport_gp, 1);
    nexus_get_global_pointer(buffer, &outport_fh_gp, 1);

    _fm_do_outport_migration_notification(channel,
					  &outport_gp,
					  &outport_fh_gp);

    /*
     * If the channel has a future home (meaning that it has send
     * a create_ports message and gotten an ack, and that we it hasn't
     * yet gone thru its own migration (since if it had this message would
     * either have been dropped above, or wouldn't have arrived here),
     * mark the channel as busy, forward the message to the future home,
     * wait for the ack to arrive here, send an ack, and unmark the
     * channel as busy.
     */
    if (!nexus_is_null_global_pointer(&channel->future_home))
    {
	nexus_buffer_t mbuf;
	int size, n_elts, obj_n_elts;
	nexus_global_pointer_t channel_gp;

#ifdef FM_DEBUG
	if (FMDebug(Migration, 1))
	    nexus_printf("_fm_outport_migrating_handler(): forwarding outport migration to future home\n");
#endif

	nexus_global_pointer(&channel_gp, (void *) channel);

	nexus_init_remote_service_request(&mbuf,
					  &channel->future_home,
					  FM_FWD_OUTPORT_MIGRATING_NAME,
					  FM_FWD_OUTPORT_MIGRATING_HASH);

	size = n_elts = 0;
	size += nexus_sizeof_global_pointer(&mbuf, &outport_gp,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;
	
	size += nexus_sizeof_global_pointer(&mbuf, &outport_fh_gp,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;

	size += nexus_sizeof_global_pointer(&mbuf, &channel_gp,
					    1, &obj_n_elts);
	n_elts += obj_n_elts;

	nexus_set_buffer_size(&mbuf, size, n_elts);

	nexus_put_global_pointer(&mbuf, &outport_gp, 1);
	nexus_put_global_pointer(&mbuf, &outport_fh_gp, 1);
	nexus_put_global_pointer(&mbuf, &channel_gp, 1);

	_fm_synch_increment(&channel->busy_synch);
 	ChannelUnlock(channel);

	nexus_send_remote_service_request(&mbuf);

	nexus_destroy_global_pointer(&channel_gp);
    }
    else
    {
	_fm_acknowledge_outport_migration_request(&outport_gp);

	ChannelUnlock(channel);
    }
    
} /* _fm_outport_migrating_handler() */


/*
_fm_do_outport_migration_notification() : Actually update the outport field
in the channel data structure as a result of an outport migration. Assume
that the channel is locked and any assertion testing has already been done.
*/

void
_fm_do_outport_migration_notification(fm_channel_t *channel,
				      nexus_global_pointer_t *old_outport,
				      nexus_global_pointer_t *new_outport)
{
    int idx, i;
    nexus_global_pointer_t *gp;

    /*
     * Find which outport in the channel's list
     * of outports corresponds to this one.
     */
    idx = -1;
    for (i = 0; i < channel->n_outports; i++)
    {
	gp = &(channel->outports[i]);

	/* This should be nexus_same_global_pointer() */

	if (nexus_same_context(gp, old_outport) &&
	    gp->address == old_outport->address)
	{
	    idx = i;
	    break;
	}
    }
    if (idx == -1)
    {
#ifdef FM_DEBUG
	nexus_fatal("_fm_do_outport_migration_notification(): Did not find outport %s in channel %s\n",
		    nexus_global_pointer_string(old_outport),
		    _fm_channel_string(channel));
#else /* FM_DEBUG */
	nexus_fatal("_fm_do_outport_migration_notification(): Did not find outport in channel %x\n",
		    channel);
#endif /* FM_DEBUG */
    }
#ifdef FM_DEBUG
    if (FMDebug(Ports, 1))
	nexus_printf("_fm_do_outport_migration_notification: channel=%s outport_gp=%s outport_fh_gp=%s index=%d\n",
		     _fm_channel_string(channel),
		     nexus_global_pointer_string(old_outport),
		     nexus_global_pointer_string(new_outport),
		     idx);
#endif

    /*
     * Got it, change the pointer.
     */
    channel->outports[idx] = *new_outport;
}


void
_fm_acknowledge_outport_migration_request(nexus_global_pointer_t *outport_gp)
{
    nexus_buffer_t mbuf;
    
    nexus_init_remote_service_request(&mbuf,
				      outport_gp,
				      FM_OUTPORT_MIGRATING_ACK_NAME,
				      FM_OUTPORT_MIGRATING_ACK_HASH);
    nexus_send_remote_service_request(&mbuf);
}


/*
_fm_outport_migrating_ack_handler()
*/

void _fm_outport_migrating_ack_handler(void *address, nexus_buffer_t *buf)
{
    fm_outport_t *outport;

    outport = (fm_outport_t *)  address;

    OutportMagicCheck(outport);

    _fm_synch_decrement(&outport->synch);
    
} /* _fm_outport_migrating_ack_handler() */


/*
_fm_fwd_outport_migrating_handler()
*/

void _fm_fwd_outport_migrating_handler(void *address, nexus_buffer_t *buffer)
{
    nexus_global_pointer_t channel_gp, outport_fh_gp, outport_gp;
    fm_channel_t *channel;
    nexus_buffer_t mbuf;
    int size, n_elts, obj_n_elts;

    channel = (fm_channel_t *) address;
    ChannelMagicCheck(channel);
    
    nexus_get_global_pointer(buffer, &outport_gp, 1);
    nexus_get_global_pointer(buffer, &outport_fh_gp, 1);
    nexus_get_global_pointer(buffer, &channel_gp, 1);

#ifdef FM_DEBUG
    if (FMDebug(Migration, 1))
	nexus_printf("_fm_fwd_outport_migrating_handler(): got channel=%s outport_gp=%s outport_fh_gp=%s channel_gp=%s\n",
		     _fm_channel_string(channel),
		     nexus_global_pointer_string(&outport_gp),
		     nexus_global_pointer_string(&outport_fh_gp),
		     nexus_global_pointer_string(&channel_gp));
#endif

    ChannelLock(channel);

    NexusAssert(nexus_same_global_pointer(&channel_gp, &channel->old_home));

    _fm_do_outport_migration_notification(channel,
					  &outport_gp,
					  &outport_fh_gp);

    nexus_init_remote_service_request(&mbuf,
				      &channel_gp, 
				      FM_FWD_OUTPORT_MIGRATING_ACK_NAME,
				      FM_FWD_OUTPORT_MIGRATING_ACK_HASH);

    size = n_elts = 0;
    
    size += nexus_sizeof_global_pointer(&mbuf, &outport_gp, 1, &obj_n_elts);
    n_elts += obj_n_elts;

    nexus_set_buffer_size(&mbuf, size, n_elts);

    nexus_put_global_pointer(&mbuf, &outport_gp, 1);

    ChannelUnlock(channel);
    
    nexus_send_remote_service_request(&mbuf);
} /* _fm_fwd_outport_migrating_handler() */


/*
_fm_fwd_outport_migrating_ack_handler()
*/

void _fm_fwd_outport_migrating_ack_handler(void *address,
					   nexus_buffer_t *buffer)
{
    fm_channel_t *channel;
    nexus_global_pointer_t outport_gp;
    nexus_buffer_t mbuf;

    channel = (fm_channel_t *) address;
    ChannelMagicCheck(channel);

    /*
     * the outport in the message is old ptr to the one that migrated
     * and started this whole mess.
     */
    nexus_get_global_pointer(buffer, &outport_gp, 1);

#ifdef FM_DEBUG
    if (FMDebug(Migration,1 ))
	nexus_printf("_fm_fwd_outport_migrating_ack_handler(): got channel=%s outport_gp=%s\n",
		     _fm_channel_string(channel),
		     nexus_global_pointer_string(&outport_gp));
#endif
    
    OutportLock(channel);
    NexusAssert(_fm_synch_count(&channel->busy_synch) > 0);
    
    nexus_init_remote_service_request(&mbuf,
				      &outport_gp,
				      FM_OUTPORT_MIGRATING_ACK_NAME,
				      FM_OUTPORT_MIGRATING_ACK_HASH);
    nexus_send_remote_service_request(&mbuf);

    _fm_synch_decrement(&channel->busy_synch);
    
    ChannelUnlock(channel);

} /* _fm_fwd_outport_migrating_ack_handler() */


/*
forward_channel_contents() : Send the contents (if any) of the inport's
channel to the channel's future home, where they are placed at the
beginning of the queue. We make the order turn out right by sending the
buffers in reverse order. We don't have to worry about someone dequeuing
from the new channel since this routine is called after the notification
and before the new process is given the channel pointer.
*/

static void forward_channel_contents(fm_inport_t *inport)
{
    fm_channel_t *channel;
    fm_queue_entry_t **queue_entries, *ent;
    int n_queue_entries;
    int i;

    InportMagicCheck(inport);
    NexusAssert(nexus_global_pointer_to_current_context(&inport->channel_gp));

    channel = (fm_channel_t *)
	nexus_convert_global_pointer_address(&inport->channel_gp);
    
    ChannelMagicCheck(channel);

    /*
     * Shouldn't even have to do this, as other processes will now
     * go to the channel's future home for any messaging. 
     */
    ChannelLock(channel);

    nexus_mutex_lock(&channel->queue.lock);

    n_queue_entries = 0;
    for (ent = channel->queue.head.next; ent != &channel->queue.head;
	 ent = ent->next)
    {
	n_queue_entries++;
    }

    if (n_queue_entries == 0)
    {
	ChannelUnlock(channel);
	return;
    }

    NexusAssert(channel->forward_func != (fm_channel_forward_func_t) NULL);

    NexusMalloc(forward_channel_contents(), queue_entries,
		fm_queue_entry_t **,
		sizeof(fm_queue_entry_t *) * n_queue_entries);
    
    for (ent = channel->queue.head.next, i = 0;
	 ent != &channel->queue.head;
	 ent = ent->next, i++)
    {
	queue_entries[i] = ent;
    }

    for (i = n_queue_entries - 1; i >= 0; i--)
    {
	(channel->forward_func)(&queue_entries[i]->buffer,
				&channel->future_home);
	_fm_free_queue_entry(queue_entries[i]);
    }

    NexusFree(queue_entries);
		
} /* forward_channel_contents() */


/*
_fm_channel_forward_buffer_handler() : This hadnler is invoked to forward
the contents of a migrated channel. Stash the buffer and insert it at the
head of the queue. The address passed is that of the channel.
*/

void _fm_channel_forward_buffer_handler(void *address,
					nexus_buffer_t *buffer)
{
    fm_channel_t *channel;

    channel = (fm_channel_t *) address;
    ChannelMagicCheck(channel);

#ifdef FM_DEBUG
    if (FMDebug(Migration, 1))
	nexus_printf("_fm_channel_forward_buffer_handler(): got buffer for %s\n",
		     _fm_channel_string(channel));
#endif

    _fm_stash_and_enqueue_at_head(&channel->queue, buffer);
    
} /* _fm_channel_forward_buffer_handler() */
