/**********************************************************************
$Id: pblock.c,v 1.17 1995/04/17 06:30:21 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/pblock.c,v $

$Log: pblock.c,v $
Revision 1.17  1995/04/17  06:30:21  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.16  1995/04/06  06:32:25  kohr
Added documentation, cleaned up a bit.


FM runtime process blocks for process spawning and tracking.
**********************************************************************/


#include "fm.h"


static void create_ports_for_entry(fm_pblock_entry_t *entry,
				   fm_pblock_t *pblock);
static void notify_ports_for_entry(fm_pblock_entry_t *entry,
				   fm_pblock_t *pblock);


/*
FM_PBLOCK_INIT() : Allocate and inititialize a process block of size
f_size.
*/

void FM_PBLOCK_INIT(f_integer_t *f_size,
		    f_integer_t *f_pblock)
{
    fm_pblock_t *pblock;
    fm_pblock_entry_t *entry;
    int size, i;

    size = (int) (*f_size);

    NexusAssert(size >= 0);

    NexusMalloc(FM_PBLOCK_INIT(), pblock, fm_pblock_t *, sizeof(fm_pblock_t));

    *f_pblock = (f_integer_t) pblock;

    if (size > 0)
    {
	NexusMalloc(FM_PBLOCK_INIT(), pblock->entries, fm_pblock_entry_t *,
		    sizeof(fm_pblock_entry_t) * size);
    }
    else
	pblock->entries = (fm_pblock_entry_t *) NULL;

    PblockMagicInit(pblock);
    
    pblock->n_entries = size;

    pblock->current_entry = 0;
    pblock->n_unfinished_processes = 0;

    _fm_synch_init(&pblock->synch);


    for (i = 0; i < size; i++)
    {
	entry = &(pblock->entries[i]);

	PblockEntryMagicInit(entry);
	
	entry->ipls = (fm_ipl_t *) NULL;
	entry->n_ipls = 0;
	entry->opls = (fm_opl_t *) NULL;
	entry->n_opls = 0;

	entry->n_total_inports = 0;
	entry->n_total_outports = 0;

	entry->process_return_data = (void *) NULL;

	nexus_null_global_pointer(&entry->dest_gp);

	_fm_synch_init(&entry->synch);
    }
} /* FM_PBLOCK_INIT() */


/*
FM_SPAWN() : Spawn the context for a new FM process.
*/

void FM_SPAWN(f_integer_t *f_anno,
	      f_integer_t *f_anno_len,
	      f_integer_t *f_pblock,
	      f_character_t f_process_name
	      CHAR_LEN_DECL(f_process_name_len))
{
    fm_pblock_t *pblock;
    fm_pblock_entry_t *entry;
    nexus_global_pointer_t dest;
    LOCAL_CHAR_LEN_DECL(f_process_name_len)
    char process_name[FM_MAX_PROCESS_NAME_LEN], *s;
    int process_name_len;
    nexus_global_pointer_t pblock_gp;
    int size, n_elts, obj_n_elts;
    nexus_buffer_t mbuf;

    CONVERT_CHAR_FTN_TO_C_INPLACE(f_process_name, f_process_name_len,
				  process_name);

    /*
     * Strip out any space-padding in the name.
     */
    if ((s = strchr(process_name, ' ')) != (char *) NULL)
	*s = '\0';
    process_name_len = strlen(process_name);

    pblock = (fm_pblock_t *) (*f_pblock);

    PblockMagicCheck(pblock);

    NexusAssert(pblock->n_entries == 0 ||
                (pblock->current_entry >= 0 &&
                 pblock->current_entry < pblock->n_entries));
    
    entry = &(pblock->entries[pblock->current_entry]);

    PblockEntryMagicCheck(entry);

    nexus_global_pointer(&pblock_gp, (void *) pblock);

    dest = _fm_vc_compute_destination(f_anno, f_anno_len);

    nexus_init_remote_service_request(&mbuf,
				      &dest,
				      FM_CREATE_PROCESS_NAME,
				      FM_CREATE_PROCESS_HASH);
    size = n_elts = 0;

    size += nexus_sizeof_int(&mbuf, 2);
    n_elts += 2;

    size += nexus_sizeof_char(&mbuf, process_name_len);
    n_elts++;

    size += nexus_sizeof_global_pointer(&mbuf, &pblock_gp, 1, &obj_n_elts);
    n_elts += obj_n_elts;

    nexus_set_buffer_size(&mbuf, size, n_elts);

    nexus_put_int(&mbuf, &process_name_len, 1);
    nexus_put_char(&mbuf, process_name, process_name_len);
    nexus_put_global_pointer(&mbuf, &pblock_gp, 1);
    nexus_put_int(&mbuf, &pblock->current_entry, 1);

    _fm_synch_increment(&pblock->synch);

    nexus_send_remote_service_request(&mbuf);

    nexus_destroy_global_pointer(&pblock_gp);
					
    pblock->current_entry++;
} /* FM_SPAWN() */


/*
_fm_create_process_return_handler() : Handle the return message for the
CREATE_PROCESS handler.
*/

void _fm_create_process_return_handler(void *address, nexus_buffer_t *buffer)
{
    int entry_idx;
    nexus_global_pointer_t dest_gp;
    fm_pblock_t *pblock;
    fm_pblock_entry_t *entry;

    pblock = (fm_pblock_t *) address;

    PblockMagicCheck(pblock);

    nexus_get_int(buffer, &entry_idx, 1);
    nexus_get_global_pointer(buffer, &dest_gp, 1);

#ifdef FM_DEBUG
    if (FMDebug(PBlock,1 ))
	nexus_printf("_fm_create_process_return_handler(): got pblock=%x entry_idx=%d dest_gp=%s\n",
		     (unsigned long) pblock,
		     entry_idx,
		     nexus_global_pointer_string(&dest_gp));
#endif

    entry = &(pblock->entries[entry_idx]);

    PblockEntryMagicCheck(entry);

    entry->dest_gp = dest_gp;

    _fm_synch_decrement(&pblock->synch);

} /* _fm_create_process_return_handler() */



/*
FM_PBLOCK_RESET() : Reset the current_pblock pointer to zero.
*/

void FM_PBLOCK_RESET(f_integer_t *f_pblock)
{
    fm_pblock_t *pblock;

    pblock = (fm_pblock_t *) (*f_pblock);

    PblockMagicCheck(pblock);
    
    NexusAssert(pblock->n_entries == 0 ||
                (pblock->current_entry >= 0 &&
                 pblock->current_entry < pblock->n_entries + 1));
    
    pblock->current_entry = 0;

} /* FM_PBLOCK_RESET() */


/*
_fm_process_return_handler() : Handler invoked at the completion of a FM
process.
*/

void _fm_process_return_handler(void *address, nexus_buffer_t *buffer)
{
    fm_pblock_entry_t *entry;

    entry = (fm_pblock_entry_t *) address;

    PblockEntryMagicCheck(entry);
#ifdef FM_DEBUG
    if (FMDebug(PBlock, 1))
	nexus_printf("Got process return handler for %x synch %x\n",
		     entry, &entry->synch);
#endif

    nexus_stash_buffer(buffer, &entry->return_buffer);
    _fm_synch_decrement(&entry->synch);
}	/* _fm_process_return_handler() */


/*
FM_PBLOCK_CONTEXT_WAIT() : Wait for the processes created by FM_SPAWN to
return.
*/

void FM_PBLOCK_CONTEXT_WAIT(f_integer_t *f_pblock)
{
    fm_pblock_t *pblock;

    pblock = (fm_pblock_t *) (*f_pblock);

    PblockMagicCheck(pblock);

    NexusAssert(pblock->n_entries == 0 ||
                (pblock->current_entry >= 0 &&
                 pblock->current_entry < pblock->n_entries));

#ifdef FM_DEBUG
    if (FMDebug(PBlock, 1))
	nexus_printf("waiting for contexts...\n");
#endif

    _fm_synch_wait(&pblock->synch);
    
#ifdef FM_DEBUG
    if (FMDebug(PBlock, 1))
	nexus_printf("done waiting for contexts\n");
#endif
    
} /* FM_PBLOCK_CONTEXT_WAIT() */


/*
FM_PORT_CREATE() : Initiate migration of port arguments in a pblock.  Just
calls create_ports_for_entry() for each entry.  Updates pblock's
current_entry iteration field.

Need to add "PBLOCK" somewhere in the name--DRK
*/

void FM_PORT_CREATE(f_integer_t *f_pblock)
{
    fm_pblock_t *pblock;
    int i;

    pblock = (fm_pblock_t *) (*f_pblock);

    PblockMagicCheck(pblock);

    NexusAssert(pblock->n_entries == 0 ||
                (pblock->current_entry >= 0 &&
                 pblock->current_entry < pblock->n_entries));

    for (i = 0; i < pblock->n_entries; i++)
    {
	create_ports_for_entry(&(pblock->entries[i]), pblock);
    }

    /*
     * Be sure to reset the current_entry!
     */
    pblock->current_entry = 0;
} /* FM_PORT_CREATE() */


/*
create_ports_for_entry() : Create the ports on the target node for this
pblock entry, by calling _fm_create_ports().

This trivial routine can be merged into FM_PORT_CREATE()--DRK
*/

static void create_ports_for_entry(fm_pblock_entry_t *entry,
				   fm_pblock_t *pblock)
{
    PblockEntryMagicCheck(entry);

    if (entry->n_total_inports == 0 && entry->n_total_outports == 0)
	return;

    _fm_create_ports(entry->n_ipls, entry->ipls, entry->n_total_inports,
		     entry->n_opls, entry->opls, entry->n_total_outports,
		     &entry->dest_gp, &entry->synch);
} /* create_ports_for_entry */


/*
_fm_pblock_finish_port_init() : General code for finishing up the port_init
stage of a process block entry. Looks for pairs of ports being sent in an
entry.
*/

void _fm_pblock_finish_port_init(fm_pblock_t *pblock,
				 fm_pblock_entry_t *entry)
{
    PblockMagicCheck(pblock);
    PblockEntryMagicCheck(entry);

    pblock->current_entry++;
} /* _fm_pblock_finish_port_init */


/*
FM_NOTIFY_PORTS() : Complete migration of port arguments in a pblock.  Just
For each entry, waits for port creation to complete, then calls
notify_ports_for_entry().

Need to add "PBLOCK" somewhere in the name--DRK
*/

void FM_NOTIFY_PORTS(f_integer_t *f_pblock)
{
    fm_pblock_t *pblock;
    fm_pblock_entry_t *entry;
    int i;

    pblock = (fm_pblock_t *) (*f_pblock);

    PblockMagicCheck(pblock);

    /*
     * First wait for the port creates to all finish
     */

    for (i = 0; i < pblock->n_entries; i++)
    {
	entry = &(pblock->entries[i]);

	PblockEntryMagicCheck(entry);

	if (entry->n_total_inports == 0 && entry->n_total_outports == 0)
	{
#ifdef FM_DEBUG
	    if (FMDebug(PBlock, 1))
		nexus_printf("FM_NOTIFY_PORTS(): skipping wait due to no ports\n");
#endif
	    continue;
	}

#ifdef FM_DEBUG
	if (FMDebug(PBlock, 1))
	    nexus_printf("Waiting for port creation on entry %d\n", i);
#endif
	
	_fm_synch_wait(&entry->synch);
	
#ifdef FM_DEBUG
	if (FMDebug(PBlock, 1))
	    nexus_printf("Got port creation on entry %d\n", i);
#endif
    }

    /*
     * At this point, all future homes have been assigned.
     */
    
    for (i = 0; i < pblock->n_entries; i++)
    {
	entry = &(pblock->entries[i]);

	PblockEntryMagicCheck(entry);

	if (entry->n_total_inports == 0 && entry->n_total_outports == 0)
	{
#ifdef FM_DEBUG
	    if (FMDebug(PBlock, 1))
		nexus_printf("FM_NOTIFY_PORTS(): skipping notify due to no ports\n");
#endif
	    continue;
	}

	notify_ports_for_entry(entry, pblock);
    }

    pblock->current_entry = 0;
} /* FM_NOTIFY_PORTS() */


/*
notify_ports_for_entry() : Notify the ports on the target node for this
pblock entry.

This trivial routine can be merged into FM_NOTIFY_PORTS()--DRK
*/

static void notify_ports_for_entry(fm_pblock_entry_t *entry,
				   fm_pblock_t *pblock)
{
    
    PblockEntryMagicCheck(entry);
    PblockMagicCheck(pblock);

    _fm_notify_ports(entry->n_ipls, entry->ipls, entry->n_total_inports,
		     entry->n_opls, entry->opls, entry->n_total_outports);
    
} /* notify_ports_for_entry() */


/*
FM_PBLOCK_WAIT() : Wait for the process block to complete.
*/

void FM_PBLOCK_WAIT(f_integer_t *f_pblock)
{
    int i;
    fm_pblock_t *pblock;
    fm_pblock_entry_t *entry;

    pblock = (fm_pblock_t *) (*f_pblock);

    PblockMagicCheck(pblock);

    NexusAssert(pblock->n_entries == 0 ||
                (pblock->current_entry >= 0 &&
                 pblock->current_entry < pblock->n_entries+1));


    for (i = 0; i < pblock->n_entries; i++)
    {
	entry = &(pblock->entries[i]);

	PblockEntryMagicCheck(entry);
	
#ifdef FM_DEBUG
	if (FMDebug(PBlock, 1))
	    nexus_printf("Waiting for pblock entry %d %x synch=%x\n",
			 i, entry, &entry->synch);
#endif
	_fm_synch_wait(&entry->synch);

#ifdef FM_DEBUG
	if (FMDebug(PBlock, 1))
	    nexus_printf("Got pblock entry %d\n", i);
#endif

	(entry->return_stashed_copy_func)(&entry->return_buffer,
					   entry->process_return_data);
	nexus_free_stashed_buffer(&entry->return_buffer);
    }
} /* FM_PBLOCK_WAIT() */
