/**********************************************************************
$Id: simple_inst.c,v 1.6 1995/05/12 20:26:28 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/simple_inst.c,v $

$Log: simple_inst.c,v $
Revision 1.6  1995/05/12  20:26:28  tuecke
Updated to new Nexus.
Added AIX shared library support:
	- initwrap.c is necessary on AIX to avoid undefined external symbols
	- Moved sentinel code into here from ../sentinel

Revision 1.5  1995/04/17  06:30:31  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.4  1995/04/06  06:32:34  kohr
Added documentation, cleaned up a bit.


FM runtime support for simple system and user-defined instrumentation.
**********************************************************************/


#include "fm.h"
#include "simple_inst.h"


#ifdef FM_SIMPLE_INST

#ifdef FM_NEED_TRAILING_UNDERSCORES

#define SIF_COUNTER_INCREMENT sif_counter_increment_
#define SIF_CTR_INC_TIMER_START sif_ctr_inc_timer_start_
#define SIF_TIMER_START sif_timer_start_
#define SIF_TIMER_STOP sif_timer_stop_
#define SIF_INTHIST_ADD sif_inthist_add_

#else

#define SIF_COUNTER_INCREMENT sif_counter_increment
#define SIF_CTR_INC_TIMER_START sif_ctr_inc_timer_start
#define SIF_TIMER_START sif_timer_start
#define SIF_TIMER_STOP sif_timer_stop
#define SIF_INTHIST_ADD sif_inthist_add

#endif

#ifndef SI_SYS_N_TIMER
#define SI_SYS_N_TIMER 0
#endif

#ifndef SI_SYS_N_COUNTER
#define SI_SYS_N_COUNTER 0
#endif

#ifndef SI_SYS_N_TIMERHIST
#define SI_SYS_N_TIMERHIST 0
#endif

#ifndef SI_SYS_N_INTHIST
#define SI_SYS_N_INTHIST 0
#endif

static nexus_thread_key_t si_instrument_key;

static void si_once_init()
{
    nexus_thread_key_create(&si_instrument_key,
			    (nexus_thread_key_destructor_func_t) NULL);
}

void si_init()
{
    nexus_thread_once_t once_control = NEXUS_THREAD_ONCE_INIT;

    nexus_thread_once(&once_control, si_once_init);
}


/*
 * si_program_init()
 * 
 * Invoked after nexus_start() returns, this function is invoked only
 * in the main program. Dump the table of instrument definitions defined
 * by gen_inst. Note that we define si_dump_inst_names() in
 * simple_inst_dump.c; that definition is only linked in if the user
 * does not define the routine.
 */
void si_program_init()
{
    char path[1024];
    FILE *fp;

    strcpy(path, _fm_simple_inst_log_dir);
    strcat(path, "/si.defs");

#ifdef FM_DEBUG
    if (FMDebug(SI, 1))
	nexus_printf("si_program_init(): dumping to %s\n", path);
#endif

    if ((fp = fopen(path, "w")) == (FILE *) NULL)
    {
	nexus_warning("si_program_init(): Cannot open file %s\n", path);
	return;
    }

    si_dump_inst_names(fp);

    fclose(fp);
    
} /* si_program_init() */




void si_instrument_init(si_instrument_block_t *block,
			int n_timers,
			int n_timerhists,
			int n_counters,
			int n_inthists)
{

#ifdef FM_DEBUG
    if (FMDebug(SI, 1))
	nexus_printf("si_instrument_init(): n_counters=%d n_timers=%d\n",
		     n_counters, n_timers);
#endif

    si_instrument_set_init(&block->user,
			   n_timers, n_timerhists,
			   n_counters, n_inthists);

    si_instrument_set_init(&block->sys,
			   SI_SYS_N_TIMER,
			   SI_SYS_N_TIMERHIST,
			   SI_SYS_N_COUNTER,
			   SI_SYS_N_INTHIST);

    nexus_thread_setspecific(si_instrument_key, (void *) block);
}

void si_instrument_set_init(si_instrument_set_t *set,
			int n_timers,
			int n_timerhists,
			int n_counters,
			int n_inthists)
{			    
    int i;
    
    NexusMalloc(si_instrument_init(),
		set->counters,
		si_counter_t *,
		sizeof(si_counter_t) * n_counters);
    set->n_counters = n_counters;

    NexusMalloc(si_instrument_init(),
		set->timers,
		si_timer_t *,
		sizeof(si_timer_t) * n_timers);
    set->n_timers = n_timers;

    NexusMalloc(si_instrument_init(),
		set->inthists,
		si_inthist_t *,
		sizeof(si_inthist_t) * n_inthists);
    set->n_inthists = n_inthists;

    NexusMalloc(si_instrument_init(),
		set->timerhists,
		si_timerhist_t *,
		sizeof(si_timerhist_t) * n_timerhists);
    set->n_timerhists = n_timerhists;

    for (i = 0; i < n_counters; i++)
	set->counters[i].count = 0;

    for (i = 0; i < n_timers; i++)
    {
	set->timers[i].accumulated_time = 0.0;
	set->timers[i].running = 0;
    }

    for (i = 0; i < n_timerhists; i++)
    {
	set->timerhists[i].bins = (int *) NULL;
	set->timerhists[i].n_bins = 0;
	set->timerhists[i].bottom = 0.0;
	set->timerhists[i].top = 0.0;
    }

    for (i = 0; i < n_inthists; i++)
    {
	set->inthists[i].bins = (int *) NULL;
	set->inthists[i].n_bins = 0;
	set->inthists[i].bottom = 0;
	set->inthists[i].top = 0;
    }

}

void si_inthist_init(int id, int n_bins, int bottom, int top)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_INTHIST_INIT(si_inthist_init, &block->user, id, n_bins, bottom, top);

}

void si_timerhist_init(int id, int n_bins, double bottom, double top)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_TIMERHIST_INIT(si_timerhist_init, &block->user, id, n_bins,
		       bottom, top);

}

void SIF_COUNTER_INCREMENT(f_integer_t *id)
{
    si_counter_increment(*id);
}

void SIF_TIMER_START(f_integer_t *id)
{
    si_timer_start(*id);
}

void SIF_TIMER_STOP(f_integer_t *id)
{
    si_timer_stop(*id);
}

void SIF_CTR_INC_TIMER_START(f_integer_t *id)
{
    si_counter_increment(*id);
    si_timer_start(*id);
}

void SIF_INTHIST_ADD(f_integer_t *id, f_integer_t *value)
{
    si_inthist_add(*id, *value);
}

void si_counter_increment(int id)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_COUNTER_INCREMENT(si_counter_increment, &block->user, id);
}

void si_timer_start(int id)
{
    si_instrument_block_t *block;
    si_timer_t *timer;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_TIMER_START(si_timer_start, &block->user, id);
}

void si_timer_stop(int id)
{
    si_instrument_block_t *block;
    si_timer_t *timer;
    double stop_time;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_TIMER_STOP(si_instrument_key, &block->user, id);
}

void si_timerhist_add(int id, int timer_id)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_TIMERHIST_ADD(si_timerhist_add, &block->user, id, timer_id);
}

void si_inthist_add(int id, int value)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_INTHIST_ADD(si_inthist_add, &block->user, id, value);
}

void si_sys_counter_increment(int id)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_COUNTER_INCREMENT(si_sys_counter_increment, &block->sys, id);
}

void si_sys_timer_start(int id)
{
    si_instrument_block_t *block;
    si_timer_t *timer;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_TIMER_START(si_sys_timer_start, &block->sys, id);
}

void si_sys_timer_stop(int id)
{
    si_instrument_block_t *block;
    si_timer_t *timer;
    double stop_time;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_TIMER_STOP(si_sys_timer_stop, &block->sys, id);
}

void si_sys_timerhist_add(int id, int timer_id)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_TIMERHIST_ADD(si_sys_timerhist_add, &block->sys, id, timer_id);
}

void si_sys_inthist_add(int id, int value)
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    SII_INTHIST_ADD(si_sys_inthist_add, &block->sys, id, value);
}

void si_instrument_dump(char *process_name)
{
    char path[1024];
    si_instrument_block_t *block;
    int node_id, context_id, thread_id;
    FILE *fp;

    nexus_ids(&node_id, &context_id, &thread_id);
    sprintf(path, "%s/si_n%d_c%d_t%d.log",
	    _fm_simple_inst_log_dir, node_id, context_id, thread_id);

#ifdef FM_DEBUG
    if (FMDebug(SI, 1))
	nexus_printf("si_instrument_dump(%s): dumping to %s\n",
		     process_name, path);
#endif

    if ((fp = fopen(path, "w")) == (FILE *) NULL)
    {
	nexus_warning("si_instrument_dump(): Cannot open file %s\n", path);
	return;
    }

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

    fprintf(fp, "P %s\n", process_name);
    fprintf(fp, "I %d %d %d\n", node_id, context_id, thread_id);

    si_instrument_set_dump(fp, &block->user, "user");
    si_instrument_set_dump(fp, &block->sys, "sys");

    fclose(fp);
}

void si_instrument_set_dump(FILE *fp, si_instrument_set_t *set, char *prefix)
{
    int i, j;

    for (i = 0; i < set->n_counters; i++)
	fprintf(fp, "C %s %d %d\n", prefix, i, set->counters[i].count);

    for (i = 0; i < set->n_timers; i++)
	fprintf(fp, "T %s %d %f\n", prefix, i,
		set->timers[i].accumulated_time);

    for (i = 0; i < set->n_timerhists; i++)
    {
	si_timerhist_t *timerhist = &(set->timerhists[i]);
	fprintf(fp, "TH %s %d %d %f %f",
		prefix, i, timerhist->n_bins, timerhist->bottom,
		timerhist->top);
	for (j = 0; j < timerhist->n_bins; j++)
	{
	    fprintf(fp, " %d", timerhist->bins[j]);
	}
	fprintf(fp, "\n");
    }
    for (i = 0; i < set->n_inthists; i++)
    {
	si_inthist_t *inthist = &(set->inthists[i]);
	fprintf(fp, "IH %s %d %d %d %d",
		prefix, i, inthist->n_bins, inthist->bottom, inthist->top);
	for (j = 0; j < inthist->n_bins; j++)
	{
	    fprintf(fp, " %d", inthist->bins[j]);
	}
	fprintf(fp, "\n");
    }
}

void si_instrument_free()
{
    si_instrument_block_t *block;

    block=(si_instrument_block_t *)nexus_thread_getspecific(si_instrument_key);

/*
    NexusFree(block->counters);
    NexusFree(block->timers);
*/    
}


#endif /* FM_SIMPLE_INST */
