/**********************************************************************
$Id: sr.c,v 1.20 1995/04/17 06:30:37 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/sr.c,v $

$Log: sr.c,v $
Revision 1.20  1995/04/17  06:30:37  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.19  1995/04/06  06:32:42  kohr
Added documentation, cleaned up a bit.


FM runtime support for sending/receiving data through ports.
**********************************************************************/


#include "fm.h"


#ifdef FM_DEBUG

void _fm_init_send_debug(fm_outport_t *outport,
			 nexus_buffer_t *buffer,
			 f_integer_t *f_ioval,
			 char *file,
			 int line)
{
    outport->file = file;
    outport->line = line;

    _fm_init_send(outport, buffer, f_ioval);
}

void _fm_send_debug(nexus_buffer_t *buffer,
		    f_integer_t *f_ioval,
		    char *file,
		    int line)
{
    _fm_send(buffer, f_ioval);
}

void _fm_receive_debug(f_integer_t *f_inport,
		       void *destination,
		       f_integer_t *f_ioval,
		       char *file,
		       int line)
{
    _fm_receive(f_inport, destination, f_ioval);
}

#endif


/*
 * _fm_init_send()
 * 
 * Set up to send a message on a port.
 * 
 */
void _fm_init_send(fm_outport_t *outport,
		   nexus_buffer_t *buffer,
		   f_integer_t *f_ioval)
{
    nexus_global_pointer_t *channel;

    OutportMagicCheck(outport);

    channel = &outport->channel_gp;

#ifdef FM_DEBUG
    if (FMDebug(SR, 1))
	nexus_printf("SEND: initializing send on outport=%s channel=%s\n",
		     _fm_outport_string(outport),
		     nexus_global_pointer_string(channel));
#endif /* FM_DEBUG */

    if (nexus_is_null_global_pointer(channel))
    {
#ifdef FM_DEBUG
	nexus_printf("(%s:%d) _fm_init_send(): got null channel gp, "
		     "outport=%s\n", 
		     outport->file,
		     outport->line,
		     outport->name);
#else
	nexus_printf("_fm_init_send(): got null channel gp\n");
#endif /* FM_DEBUG */
    }
    NexusAssert(!nexus_is_null_global_pointer(channel));

    /* Check whether outport is in a valid state --- Ming */
    if (OutportState(outport) == FM_STATE_CLOSING_OUTPORT ||
	OutportState(outport) == FM_STATE_UNDEFINED )
    {
#ifdef FM_DEBUG	
	nexus_fatal("(%s:%d) _fm_init_send(): outport(%s) is in an "
		    "invalid state.  An endchannel operation has probably "
		    "been performed on the outport\n",
		    outport->file,
		    outport->line,
		    outport->name);
#else
	nexus_fatal("_fm_init_send(): outport is in an invalid state\n");
#endif /* FM_DEBUG */
    }

    nexus_init_remote_service_request(buffer,
				      channel,
				      FM_PORT_SEND_NAME,
				      FM_PORT_SEND_HASH);

    *f_ioval = 0;

} /* _fm_init_send() */


/*
 * _fm_send()
 * 
 * Send off the message.
 * 
 */
void _fm_send(nexus_buffer_t *buffer, f_integer_t *f_ioval)
{
#ifdef FM_DEBUG
    if (FMDebug(SR, 1))
	nexus_printf("SEND: sending\n");
#endif
    
    nexus_send_remote_service_request(buffer);
    *f_ioval = 0;
} /* _fm_send() */


/*
 * _fm_receive()
 * 
 * Pluck a message off the inport's channel. To be replaced with the
 * ready-receiver code.
 * 
 */
void _fm_receive(f_integer_t *f_inport,
		 void *destination,
		 f_integer_t *f_ioval)
{
    fm_inport_t *inport;
    fm_channel_t *channel;

    *f_ioval = 0;

    inport = (fm_inport_t *) *f_inport;

    InportMagicCheck(inport);

#ifdef FM_DEBUG
    if (FMDebug(SR, 1))
	nexus_printf("RECEIVE %s\n", _fm_inport_string(inport));
#endif

    if (nexus_is_null_global_pointer(&inport->channel_gp))
	FM_RETURN_IOVAL(INPORT_NULL_CHANNEL);

    if (!nexus_global_pointer_to_current_context(&inport->channel_gp))
	FM_RETURN_IOVAL(INPORT_NONLOCAL_CHANNEL);

    channel = (fm_channel_t *) nexus_convert_global_pointer_address(
    	&inport->channel_gp);

    ChannelMagicCheck(channel);

    ChannelLock(channel);

    if (_fm_queue_empty(&channel->queue))
    {
	if (channel->n_outports == 0)
	{
	    *f_ioval = -1;
	    ChannelUnlock(channel);
	    return;
	}
#ifdef FM_DEBUG
	if (FMDebug(SR, 1))
	    nexus_printf("RECEIVE installing direct copy for channel %s\n",
			 _fm_channel_string(channel));
#endif
	channel->direct_copy_destination = destination;
	channel->send_func = _fm_channel_direct_send;
	channel->waiting_for_direct_send = NEXUS_TRUE;
	while (channel->waiting_for_direct_send)
	{
	    ChannelWait(channel);
	}
#ifdef FM_DEBUG
	if (FMDebug(SR, 1))
	    nexus_printf("RECEIVE got direct copy iostat=%d\n",
			 channel->direct_send_iostat);
#endif
	*f_ioval = channel->direct_send_iostat;
    }
    else
    {
	fm_queue_entry_t *entry;

#ifdef FM_DEBUG
	if (FMDebug(SR, 1))
	    nexus_printf("RECEIVE dequeuing from channel %s\n",
			 _fm_channel_string(channel));
#endif
	entry = _fm_dequeue(&channel->queue);
	(channel->stashed_copy_func)(&entry->buffer, destination);
	_fm_free_queue_entry(entry);
    }

    ChannelUnlock(channel);
} /* _fm_receive() */


/*
 * FM_PROBE()
 *
 * Test an inport to see if the channel is empty.
 * 
 */
void FM_PROBE(f_integer_t *f_inport,
	      f_integer_t *f_ioval,
	      f_logical_t *f_empty)
{
    fm_inport_t *inport;
    fm_channel_t *channel;

    *f_ioval = 0;

    inport = (fm_inport_t *) *f_inport;

    InportMagicCheck(inport);

#ifdef FM_DEBUG
    if (FMDebug(SR, 1))
	nexus_printf("PROBE %s\n", _fm_inport_string(inport));
#endif

    if (nexus_is_null_global_pointer(&inport->channel_gp))
	FM_RETURN_IOVAL(INPORT_NULL_CHANNEL);

    if (!nexus_global_pointer_to_current_context(&inport->channel_gp))
	FM_RETURN_IOVAL(INPORT_NONLOCAL_CHANNEL);

    channel = (fm_channel_t *) nexus_convert_global_pointer_address(
    	&inport->channel_gp);
    ChannelMagicCheck(channel);

    ChannelLock(channel);

    if (_fm_queue_empty(&channel->queue))
    {
	if (channel->n_outports == 0)
	{
	    FM_SET_IOVAL_EOF(f_ioval);
	    *f_empty = FTN_LOGICAL_FALSE;
	}
	else
	    *f_empty = FTN_LOGICAL_TRUE;

    }
    else
    {
	*f_empty = FTN_LOGICAL_FALSE;
    }
    ChannelUnlock(channel);

} /* FM_PROBE() */


/*
 * port_send_handler()
 * 
 * Handler for sending on a port.
 * 
 */
void _fm_port_send_handler(void *address, nexus_buffer_t *buffer)
{
    fm_channel_t *channel;

    channel = (fm_channel_t *) address;

    ChannelMagicCheck(channel);
    
#ifdef FM_DEBUG
    if (FMDebug(SR, 2))
    {
	nexus_printf("PortSendHandler: buffer=%x *buffer=%x channel=%s\n",
		     (unsigned long) buffer,
		     (unsigned long) *buffer,
		     _fm_channel_string(channel));
    }
#endif
    
    ChannelLock(channel);

    NexusAssert(channel->send_func != (fm_channel_send_func_t) NULL);

    (channel->send_func)(channel, buffer);
} /* port_send_handler() */


/*
 * _fm_channel_queueing_send()
 *
 * Send function for non-ready receivers. Enqueues the buffer on the
 * channel.
 */
void _fm_channel_queueing_send(fm_channel_t *channel, nexus_buffer_t *buffer)
{
#ifdef FM_DEBUG
    if (FMDebug(SR, 2))
	nexus_printf("QueueingSend(): channel=%s\n",
		     _fm_channel_string(channel));
#endif

    _fm_stash_and_enqueue(&channel->queue, buffer);
    ChannelUnlock(channel);

    /*
     * No need to signal the condition, since we know there is
     * no one waiting. (If there were, we'd be in the direct send
     * function)
     */
	
} /* _fm_channel_queueing_send */


/*
 * _fm_channel_direct_send()
 *
 * Send function for ready receivers.
 */
void _fm_channel_direct_send(fm_channel_t *channel, nexus_buffer_t *buffer)
{

#ifdef FM_DEBUG
    if (FMDebug(SR, 2))
	nexus_printf("DirectSend(): channel=%s\n",
		     _fm_channel_string(channel));
#endif
    
    /* Copy the data... */

    (channel->copy_func)(buffer, channel->direct_copy_destination);
    channel->direct_copy_destination = NULL;
    channel->waiting_for_direct_send = NEXUS_FALSE;
    ChannelSignal(channel);

    /*
     * Reset the send func. Note: if we had a queue of direct copy
     * destinations, at this point we would look at the rest of the
     * queue. If there were no more destinations, we would reset the
     * send_func to queueing_send; otherwise, we'd leave it as
     * a direct_send.
     */
    channel->send_func = _fm_channel_queueing_send;

    channel->direct_send_iostat = 0;

    ChannelUnlock(channel);

} /* _fm_channel_direct_send */


/*
 * FM_END_CHANNEL()
 * 
 * Close a channel. This is done by invoking the CloseChannel
 * rsr on the channel associated with this outport. The outport is
 * placed in a state where any further actions on it must wait
 * for the endchannel acknowledgement to be returned.
 *
 * What happens if we get a channel migration request while
 * we are in the after-endchannel state? Yuck!
 *what to do???--DRK
 * 
 */
void FM_END_CHANNEL(f_integer_t *f_outport,
		    f_integer_t *f_ioval)
{
    nexus_global_pointer_t *channel;
    fm_outport_t *outport;
    nexus_buffer_t mbuf;
    nexus_global_pointer_t outport_gp;
    int size, n_elts, obj_n_elts;
    
    outport = (fm_outport_t *) *f_outport;

    OutportMagicCheck(outport);

    nexus_global_pointer(&outport_gp, (void *) outport);

    OutportLock(outport);
    OutportSetState(outport, FM_STATE_CLOSING_OUTPORT);

    channel = &outport->channel_gp;

#ifdef FM_DEBUG
    if (FMDebug(SR, 1))
	nexus_printf("ENDCHANNEL: outport=%s channel=%s\n",
		     _fm_outport_string(outport),
		     nexus_global_pointer_string(channel));
#endif

/*
    NexusAssert(!nexus_is_null_global_pointer(channel));
*/
    if (nexus_is_null_global_pointer(channel))
    {
	*f_ioval = 0;
	OutportUnlock(outport);
	return;
    }

    nexus_init_remote_service_request(&mbuf,
				      channel,
				      FM_END_CHANNEL_NAME,
				      FM_END_CHANNEL_HASH);
    size = n_elts = 0;

    size += nexus_sizeof_global_pointer(&mbuf, &outport_gp, 1, &obj_n_elts);
    n_elts += obj_n_elts;

    nexus_set_buffer_size(&mbuf, size, n_elts);

    nexus_put_global_pointer(&mbuf, &outport_gp, 1);

    _fm_synch_clear(&outport->synch);
    _fm_synch_increment(&outport->synch);

    OutportUnlock(outport);
    
    nexus_send_remote_service_request(&mbuf);

    _fm_synch_wait(&outport->synch);

    OutportLockAndSetState(outport, FM_STATE_UNDEFINED);
    /* 
     * Set outport channel_gp to null --- Ming
     *
     */
    OutportLock(outport);
    /* 
     * TODO: need to free the memory occupied by
     *       the channel
     *
     */
    nexus_null_global_pointer(&outport->channel_gp);
    OutportUnlock(outport);

    *f_ioval = 0;

} /* FM_END_CHANNEL() */


/*
 * _fm_end_channel_handler()
 * 
 * Handle an endchannel rsr.
 * 
 */
void _fm_end_channel_handler(void *address, nexus_buffer_t *buffer)
{
    fm_channel_t *channel;
    nexus_global_pointer_t outport_gp, *gp;
    int idx, i;
    nexus_buffer_t mbuf;

    channel = (fm_channel_t *) address;

    ChannelMagicCheck(channel);

    nexus_get_global_pointer(buffer, &outport_gp, 1);

    ChannelLock(channel);

#ifdef FM_DEBUG
    if (FMDebug(SR, 1))
        nexus_printf("_fm_end_channel_handler(): channel=%s outport_gp=%s\n",
                     _fm_channel_string(channel),
                     nexus_global_pointer_string(&outport_gp));
#endif

    /*
     * Find which outport in the channel's list
     * of outports corresponds to this one.
     */
    idx = -1;
    for (i = 0; i < channel->n_outports; i++)
    {
	gp = &(channel->outports[i]);

	/* This should be nexus_same_global_pointer() */

	if (nexus_same_context(gp, &outport_gp) &&
	    gp->address == outport_gp.address)
	{
	    idx = i;
	    break;
	}
    }
    if (idx == -1)
    {
#ifdef FM_DEBUG
	nexus_warning("_fm_end_channel_handler(): Did not find outport %s "
		      "in channel %s. Outport list:\n",
		    nexus_global_pointer_string(&outport_gp),
		    _fm_channel_string(channel));
	for (i = 0; i < channel->n_outports; i++)
	    nexus_printf("   %d: %s\n",
			 i,
			 nexus_global_pointer_string(&(channel->outports[i])));
#else /* FM_DEBUG */
	nexus_warning("_fm_end_channel_handler(): Did not find outport "
		      "in channel %x\n", (unsigned long) channel);
#endif /* FM_DEBUG */
    }
    else
    {
#ifdef FM_DEBUG
	if (FMDebug(SR, 1))
	    nexus_printf("_fm_end_channel_handler(): Removing outport %d %s\n",
			 idx, nexus_global_pointer_string(&outport_gp));

	if (FMDebug(SR, 2))
	{
	    nexus_printf("Channel outport list before shift:\n");
	    for (i = 0; i < channel->n_outports; i++)
		nexus_printf("   %d: %s\n",
			     i,
			     nexus_global_pointer_string(
			     	&(channel->outports[i])));
	}
#endif

	for (i = idx; i < channel->n_outports - 1; i++)
	{
	    channel->outports[i] = channel->outports[i+1];
	}
	channel->n_outports--;
	
#ifdef FM_DEBUG
	if (FMDebug(SR, 2))
	{
	    nexus_printf("Channel outport list after shift:\n");
	    for (i = 0; i < channel->n_outports; i++)
		nexus_printf("   %d: %s\n",
			     i,
			     nexus_global_pointer_string(
			     	&(channel->outports[i])));
	}
#endif
    }

    nexus_init_remote_service_request(&mbuf,
				      &outport_gp,
				      FM_END_CHANNEL_ACK_NAME,
				      FM_END_CHANNEL_ACK_HASH);
    nexus_set_buffer_size(&mbuf, 0, 0);

    /*
     * If there is a thread blocked in a direct send handler (on a
     * receive), and there are no more outports, wake it up
     * so that it can get the -1 iostat.
     */
    if (channel->n_outports == 0 && channel->waiting_for_direct_send)
    {
	channel->direct_send_iostat = -1;
	channel->waiting_for_direct_send = NEXUS_FALSE;
	ChannelSignal(channel);
    }
    ChannelUnlock(channel);

    /*
     * Wait to send until the channel is awakened; otherwise
     * we can deadlock.
     */
    
    nexus_send_remote_service_request(&mbuf);
} /* _fm_end_channel_handler() */


/*
 * _fm_end_channel_ack_handler()
 * 
 * Handle an endchannel acknowledgement.
 * 
 */
void _fm_end_channel_ack_handler(void *address, nexus_buffer_t *buffer)
{
    fm_outport_t *outport;

    outport = (fm_outport_t *) address;

    OutportMagicCheck(outport);

#ifdef FM_DEBUG
    if (FMDebug(SR, 1))
	nexus_printf("_fm_end_channel_ack_handler(): got ack for %s\n",
		     _fm_outport_string(outport));
#endif
    _fm_synch_decrement(&outport->synch);

} /* _fm_end_channel_ack_handler() */
