/**********************************************************************
$Id: synch.c,v 1.9 1995/07/07 22:05:44 mqxu Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/synch.c,v $

$Log: synch.c,v $
Revision 1.9  1995/07/07  22:05:44  mqxu
Version 2.4b7

Revision 1.8  1995/04/17  06:30:40  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.7  1995/04/06  06:32:44  kohr
Added documentation, cleaned up a bit.


FM runtime synchronization object operations.
**********************************************************************/


#include "fm.h"


/*
 * _fm_synch_init()
 * 
 * Initialize a FM synch object.
 * 
 */
void _fm_synch_init(fm_synch_t *synch)
{
    SynchMagicInit(synch);
    nexus_mutex_init(&synch->lock, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&synch->cond, (nexus_condattr_t *) NULL);
    synch->count = 0;
} /* _fm_synch_init() */


/*
 * _fm_synch_clear()
 * 
 * Reinitialize a synch object.
 * 
 */
void _fm_synch_clear(fm_synch_t *synch)
{
    synch->count = 0;
} /* _fm_synch_clear() */


/*
 * _fm_synch_count()
 * 
 * Return the synch object's count.
 * 
 */
int _fm_synch_count(fm_synch_t *synch)
{
    return synch->count;
} /* _fm_synch_count() */


/*
 * _fm_synch_increment()
 * 
 * Increment a synch object.
 * 
 */
void _fm_synch_increment(fm_synch_t *synch)
{
    nexus_mutex_lock(&synch->lock);
    synch->count++;
    nexus_mutex_unlock(&synch->lock);
} /* _fm_synch_bump() */


/*
 * _fm_synch_decrement()
 * 
 * Decrement a synch object.
 * 
 */
void _fm_synch_decrement(fm_synch_t *synch)
{
    nexus_mutex_lock(&synch->lock);
    synch->count--;
    if (synch->count == 0)
	nexus_cond_signal(&synch->cond);
    nexus_mutex_unlock(&synch->lock);
} /* _fm_synch_bump() */


/*
 * _fm_synch_wait()
 * 
 * Wait for the count on the synch to go to zero.
 * 
 */
void _fm_synch_wait(fm_synch_t *synch)
{
    nexus_mutex_lock(&synch->lock);
#ifdef FM_DEBUG
    if (FMDebug(Synch, 1))
	nexus_printf("Got lock for synch %x count=%d\n",
		     synch, synch->count);
#endif
    while (synch->count > 0)
    {
	nexus_cond_wait(&synch->cond, &synch->lock);
#ifdef FM_DEBUG
	if (FMDebug(Synch, 1))
	    nexus_printf("Wait returns for synch %x count=%d\n",
			 synch, synch->count);
#endif
    }
#ifdef FM_DEBUG
    if (FMDebug(Synch, 1))
	nexus_printf("Wait finished for synch %x count=%d\n",
		     synch, synch->count);
#endif
    nexus_mutex_unlock(&synch->lock);
	
} /* _fm_synch_wait() */

void _fm_synch_destroy(fm_synch_t *synch)
{
} /* _fm_sync_destroy */

