/**********************************************************************
$Id: types.h,v 1.16 1995/06/14 04:41:10 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/types.h,v $

$Log: types.h,v $
Revision 1.16  1995/06/14  04:41:10  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.15  1995/04/17  06:30:42  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.14  1995/04/06  06:32:47  kohr
Added documentation, cleaned up a bit.


FM runtime datatype declarations.
**********************************************************************/

#ifndef _fm_types_h
#define _fm_types_h

typedef nexus_bool_t fm_bool_t;


/*
Forward decls.
*/

typedef struct _fm_pblock_t fm_pblock_t;
typedef struct _fm_pblock_entry_t fm_pblock_entry_t;

typedef struct _fm_inport_t fm_inport_t;
typedef struct _fm_outport_t fm_outport_t;
typedef struct _fm_channel_t fm_channel_t;

typedef struct _fm_queue_t fm_queue_t;

typedef struct _fm_port_pair_t fm_port_pair_t;


/*
Synch objects.
*/

typedef struct _fm_synch_t
{
    StartMagicNumber
    nexus_mutex_t lock;
    nexus_cond_t cond;
    int count;
    EndMagicNumber
} fm_synch_t;


/*
Inport/outport lists.
*/

typedef struct _fm_ipl_t
{
    fm_inport_t **entries;
    int n_entries;
} fm_ipl_t;

typedef struct _fm_opl_t
{
    fm_outport_t **entries;
    int n_entries;
} fm_opl_t;


/*
Process blocks.
*/

struct _fm_pblock_t
{
    StartMagicNumber
    fm_pblock_entry_t *entries;
    int n_entries;
    int current_entry;

    int n_unfinished_processes;

		/* Used to wait for spawn of remote contexts for all
		   entries in this pblock. */
    fm_synch_t synch;
    EndMagicNumber
};

struct _fm_pblock_entry_t
{
    StartMagicNumber
    fm_ipl_t *ipls;
    int n_ipls;

    fm_opl_t *opls;
    int n_opls;

    int n_total_inports;
    int n_total_outports;

    int create_return_code;
    nexus_global_pointer_t dest_gp;

    void *process_return_data;
    void (*return_copy_func)(nexus_buffer_t *buffer, void *destination);
    void (*return_stashed_copy_func)(nexus_stashed_buffer_t *buffer,
				     void *destination);
    nexus_stashed_buffer_t return_buffer;
		/* Used to wait for the following: completion of setup for
		   sending ports (using _fm_create_ports()) during creation
		   of the process for this entry; completion of the process
		   associated with this entry. */
    fm_synch_t synch;
    EndMagicNumber
};


/*
Ports.
*/

struct _fm_inport_t
{
    StartMagicNumber

#ifdef FM_DEBUG
    char name[128];
#endif
    
    nexus_global_pointer_t channel_gp;

    nexus_global_pointer_t future_home;
    nexus_global_pointer_t old_home;

		/* Gain mutex over entire data structure. */
    nexus_mutex_t lock;
		/* This appears to be unused.--DRK */
    nexus_cond_t cond;
    int state;
		/* This appears to be unused.--DRK */
    fm_synch_t busy_synch;

    fm_inport_t *next;		/*  Used in freelists */

    /* Location of the FM statement which operates on this inport--Ming  */
#ifdef FM_DEBUG
    char *file;
    int  line;
#endif /* FM_DEBUG */

    EndMagicNumber
};

struct _fm_outport_t
{
    StartMagicNumber

#ifdef FM_DEBUG
    char name[128];
#endif

    nexus_global_pointer_t channel_gp;

    nexus_global_pointer_t future_home;
    nexus_global_pointer_t old_home;

    /* The following are for enqueuing a notification on the port. We don't
       really need a queue since only one notification will ever end up
       here. */
    nexus_global_pointer_t channel_future_home;

		/* Gain mutex over entire data structure. */
    nexus_mutex_t lock;
		/* Used only in _fm_fwd_channel_migrating_ack_handler() in
		   migrate.c, which will get nuked using the new migration
		   protocol.  Hence probably unnecessary.--DRK */
    nexus_cond_t cond;
    int state;
		/* Used only to wait for completion of forwarding of
		   channel-migrating messages, which don't exist in the
		   new migration protocol.  Hence probably unnecessary.
		   --DRK */
    fm_synch_t busy_synch;

		/* Used to wait for completion of migration notification
		   and ENDCHANNEL notification.  Currently a thread can
		   only have one such operation outstanding, so this field
		   is not really necessary.--DRK */
    fm_synch_t synch;

    fm_outport_t *next;		/*  Used in freelists */

    /* Location of the FM statement which operates on this outport--Ming */
#ifdef FM_DEBUG
    char *file;
    int  line;
#endif /* FM_DEBUG */

    EndMagicNumber
};

typedef struct _fm_queue_entry_t
{
    nexus_stashed_buffer_t buffer;
    struct _fm_queue_entry_t *next;
} fm_queue_entry_t;

struct _fm_queue_t
{
    fm_queue_entry_t head;
    fm_queue_entry_t *tail;
    nexus_mutex_t lock;
    nexus_cond_t cond;
    int n_waiting;
};

typedef void (*fm_channel_send_func_t)(fm_channel_t *channel,
				       nexus_buffer_t *buffer);
typedef void (*fm_channel_close_func_t)(fm_channel_t *channel,
					nexus_buffer_t *buffer);
typedef void (*fm_channel_copy_func_t)(nexus_buffer_t *buffer,
				       void *destination);
typedef void (*fm_channel_stashed_copy_func_t)(nexus_stashed_buffer_t *buffer,
					       void *destination);
typedef void (*fm_channel_forward_func_t)
	(nexus_stashed_buffer_t *buffer,
	 nexus_global_pointer_t *dest_channel_gp);

struct _fm_channel_t
{
    StartMagicNumber

#ifdef FM_DEBUG
    char name[128];
#endif
    
    nexus_global_pointer_t *outports;
    int n_outports, n_outports_alloc;
    nexus_global_pointer_t inport;

    nexus_global_pointer_t future_home;
    nexus_global_pointer_t old_home;

    fm_queue_t queue;

		/* Used to wait for completion of migration notification.
		   Currently a thread can only have one such operation
		   outstanding, so this field is not necessary.--DRK */
    fm_synch_t synch;

    fm_channel_send_func_t send_func;
    fm_channel_close_func_t close_func;
    fm_channel_copy_func_t copy_func;
    fm_channel_stashed_copy_func_t stashed_copy_func;
    fm_channel_forward_func_t forward_func;

    void *direct_copy_destination;
    fm_bool_t waiting_for_direct_send;
    int direct_send_iostat;

		/* Gain mutex over entire data structure. */
    nexus_mutex_t lock;
		/* Used for "ready" receiver to wait for "direct" send.
		   Since a channel always has a queue, which in turn has a
		   condition variable, and the cond. var. is not used at
		   all during a "direct" send, perhaps the queue's cond.
		   var. could be used instead?--DRK */
    nexus_cond_t cond;
    int state;
		/* Used only to wait for completion of forwarding of
		   outport-migrating messages, which don't exist in the
		   new migration protocol.  Hence probably unnecessary.
		   --DRK */
    fm_synch_t busy_synch;

	/* Used for determining paired ports in a process call or port
	   send. */
    int index;

    fm_channel_t *next;		/* Used in freelists */

    EndMagicNumber
};


/*
Type used in process creation.
*/

typedef struct _fm_create_process_info_t
{
    char process_name[FM_MAX_PROCESS_NAME_LEN];
    nexus_global_pointer_t pblock_gp;
    int entry_idx;
    struct _fm_create_process_info_t *next;
} fm_create_process_info_t;


/*
Debugging stuff...

Some of this looks like it duplicates what's in debug_defs.h and
global.h--DRK
*/

#ifdef FM_DEBUG

#undef FM_SUBSYSTEM
#define FM_SUBSYSTEM(option_key, desc, id) FM_SUBSYSTEM_ ## id,
typedef enum _fm_subsystem_t
{
#include "subsystems.def"
    FM_SUBSYSTEM_LAST
} fm_subsystem_t;

typedef struct _fm_subsystem_debug_entry_t
{
    fm_subsystem_t subsystem;
    int debug_level;
    char *option;
    char *description;
} fm_subsystem_debug_entry_t;


/*
Pause point support.
*/

typedef struct _fm_pausepoint_t
{
    int id;
    fm_bool_t pause_me;
    int sleep_me;
    struct _fm_pausepoint_t *next;
} *fm_pausepoint_t;

#endif /* FM_DEBUG */


/*
IOVAL designators.
*/

#undef FM_IOVAL
#define FM_IOVAL(ioval, fatal, symbol, message) \
	FM_IOVAL_ ## symbol = ioval,
typedef enum _fm_ioval_t
{
#include "ioval.def"
FM_IOVAL_LAST    
} fm_ioval_t;


#endif /* _fm_types_h */
