#include "server.h"
#include "nexus_config.h"
#include "ports0_config.h"
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/time.h>

#ifdef NEXUS_ARCH_NEXT040
#include <libc.h>
#endif

#if defined(NEXUS_ARCH_AIX) 
#include <sys/select.h>
#endif

static int start_secure_remote_node(char *host, int port, char *username, char *dir, char *cmd, char *args);
static int connect_to_server(char *host, int port);
static void send_string(int sock, char *str);
static void recv_string(int sock, char *str, int len);
static void setup_anon_listener(int backlog, int *port, int *sock);
static char *getpw_ss(char *host, char *username);
static int echo(void);

main(int argc, char **argv)
{
int i;
    start_secure_remote_node(argv[1], atoi(argv[2]), argv[3], argv[4], argv[5], argv[6]);
}


/*
 * start_secure_remote_node()
 */
static int start_secure_remote_node(char *host, int port, char *remote_username, char *dir, char *cmd, char *args)
{
    char myhost[256];
    int conn;
    struct passwd *passwd;
    char *local_username;
    struct timeval timeout;
    fd_set recv_fds;
    char buf[250];
    char *pw_string;
    int pid;
    int new_port, new_fd;
    char port_string[50];
    int stdout_fd;
    struct sockaddr_in temp;
    int templen;
    int n;
    char msg[500];

    /* practically copied from p4_secure.c in the MPIch release */

    myhost[0] = '\0';
    gethostname(myhost, sizeof(myhost));

    conn = connect_to_server(host, port);
    if (conn < 0)
    {
	fprintf(stderr, "start_secure_node(): connect failed %s:%d\n", host, port);
	return 0;
    }

#if defined(NEXUS_DEBUG) && defined(NEXUS_LISTENER_DEBUG)
    fprintf(stderr, ("start_secure_node(): Connected\n"));
#endif

    passwd = getpwuid(geteuid());
    if (passwd == NULL)
    {
	local_username = getlogin();
	if (local_username == NULL)
	{
	    fprintf(stderr, "No local username found\n");
	    return 0;
	}
    }
    else
    {
	local_username = passwd->pw_name;
    }

    send_string(conn, local_username);
    if (strcmp(remote_username, "NEXUS") == 0)
    {
	send_string(conn, local_username);
    }
    else
    {
        send_string(conn, remote_username);
    }

    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    FD_ZERO(&recv_fds);
    FD_SET(conn, &recv_fds);
/* Check out if this can be changed back to the original code. */
/* JGG
    n = select(conn+1, &recv_fds, NULL, NULL, &timeout);
JGG */ n = 1;
    if (n == 1 || (n == -1 && errno == EINTR)) 
    {
	recv_string(conn, buf, sizeof(buf));
    }
    else
    {
	if (n > 0)
	{
	    fprintf(stderr, "start_secure_remote_node(): wrong number of new connections: %d instead of 1\n", n);
	    return 0;
	}
	else
	{
	    fprintf(stderr, "start_secure_remote_node(): select() failed--errno = %d\n", errno);
	}
    }

#if defined(NEXUS_DEBUG) && defined(NEXUS_LISTENER_DEBUG)
    fprintf(stderr, ("start_secure_remote_node(): Got reply1 %s\n", buf));
#endif

    if (strncmp(buf, "Password", 8) == 0)
    {
        pw_string = getpw_ss(host, remote_username);

	send_string(conn, pw_string);
	recv_string(conn, buf, sizeof(buf));
	free(pw_string);

#if defined(NEXUS_DEBUG) && defined(NEXUS_LISTENER_DEBUG)
	fprintf(stderr, ("start_secure_remote_node(): Got reply2 %s\n", buf));
#endif

	if (strncmp(buf, "Proceed", 7) != 0)
	{
	    return 0;
	}
    }
    else
    {
	if (strncmp(buf, "Proceed", 7) != 0)
	{
	    return 0;
	}
    }

#if defined(NEXUS_DEBUG) && defined(NEXUS_LISTENER_DEBUG)
    fprintf(stderr, ("start_secure_remote_node(): Validated on remote system\n"));
#endif

    send_string(conn, SERVER_CD_NOTIFIER);
    send_string(conn, dir);
    send_string(conn, cmd);
    send_string(conn, args);

    if ((pid = fork()) == 0)
    {
	setup_anon_listener(10, &new_port, &new_fd);
	sprintf(port_string, "%d", new_port);
	send_string(conn, port_string);
	fflush(stdout);
	templen = sizeof(temp);

	stdout_fd = accept(new_fd, (struct sockaddr *)&temp, &templen);
	close(new_fd);

	n = 1;
	while (n > 0)
	{
	    n = read(stdout_fd, msg, sizeof(msg)-1);
	    if (n > 0)
	    {
		msg[n] = '\0';
		printf("%s\n", msg);
	    }
	}
        exit(0);
    }

    recv_string(conn, buf, sizeof(buf));

#if defined(NEXUS_DEBUG) && defined(NEXUS_LISTENER_DEBUG)
    fprintf(stderr, ("start_secure_remote_node(): Got reply3 %s\n", buf));
#endif

    if (strncmp(buf, "Success", 7) != 0)
    {
	kill(pid, SIGKILL);
	return 0;
    }

    close(conn);

    return 1;
} /* start_secure_remote_node() */

static int connect_to_server(char *host, int server_port)
{
    int conn;
    struct hostent *hostent;
    struct sockaddr_in addr;
    int rc;

    conn = socket(AF_INET, SOCK_STREAM, 0);
    if (conn < 0)
    {
	return -1;
    }

    hostent = gethostbyname(host);

    addr.sin_family = hostent->h_addrtype;
    addr.sin_port = htons(server_port);
    memcpy(&addr.sin_addr, hostent->h_addr, hostent->h_length);

    rc = connect(conn, (struct sockaddr *)&addr, sizeof(addr));
    if (rc < 0)
    {
	return -1;
    }

    return conn;
}

static void send_string(int sock, char *str)
{
    int rc;
    int len = strlen(str);
    char nl = 10;

    rc = write(sock, str, len);
    if (rc < 0)
    {
	fprintf(stderr, "send_string():  write to secure server failed\n");
    }

    rc = write(sock, &nl, 1);
    if (rc < 0)
    {
	fprintf(stderr, "send_string():  write to secure server failed\n");
    }
    
#if defined(NEXUS_DEBUG) && defined(NEXUS_LISTENER_DEBUG)
    fprintf(stderr, ("send_string(): sent \"%s\" successfully\n", str));
#endif
}

static void recv_string(int sock, char *buf, int len)
{
    char *bptr;
    int n;

    bptr = buf;
    while(1)
    {
	n = read(sock, bptr, 1);
	if (n < 0)
	{
	    fprintf(stderr, "recv_string(): read from secure server failed\n");
	}

	if (*bptr == '\n') break;

	bptr++;
	if (bptr - buf >= len) break;
    }
    *bptr = 0;
#if defined(NEXUS_DEBUG) && defined(NEXUS_LISTENER_DEBUG)
    fprintf(stderr, ("recv_string(): read \"%s\" successfully\n", buf));
#endif
}

static void setup_anon_listener(int backlog, int *port, int *sock)
{
    int sin_len;
    struct sockaddr_in sin;
    int rc;

    *sock = socket(AF_INET, SOCK_STREAM, 0);
    if (*sock < 0)
    {
	fprintf(stderr, "setup_anon_listener(): unable to setup a listener socket\n");
    }

    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(0);

    sin_len = sizeof(sin);

    rc = bind (*sock, (struct sockaddr *)&sin, sin_len);
    if (rc < 0)
    {
	fprintf(stderr, "setup_anon_listener(): unable to bind socket\n");
    }

    rc = listen (*sock, backlog);
    if (rc < 0)
    {
	fprintf(stderr, "setup_anon_listener(): unable to listen on to socket\n");
    }

    getsockname(*sock, (struct sockaddr *)&sin, &sin_len);
    *port = ntohs(sin.sin_port);

}

static char *getpw_ss(char *host, char *username)
{
    char *buf;
    char *s;

    buf = (char *)malloc(sizeof(char)*100);
    echo();
    printf("Password for %s@%s:  ", username, host);
    fflush(stdout);
    fgets(buf, 100, stdin);
    echo();
    printf("\n");

    for (s = buf; *s; s++)
    {
	if (*s == '\n')
	{
	    *s = '\0';
	    break;
	}
    }

    return buf;
}

#if defined(NEXUS_ARCH_NEXT040)

#include <sys/ioctl.h>

static int echo()
{
    struct sgttyb orig_tty, tty_new;

    if (ioctl(0, TIOCGETP, &orig_tty) < 0)
    {
	fprintf(stderr, "ioctl() TIOCGETP failed\n");
	return -1;
    }
    
    tty_new = orig_tty;
    tty_new.sg_flags &= ~(ECHO);

    if (ioctl(0, TIOCSETP, &tty_new) < 0)
    {
	fprintf(stderr, "ioctl() TIOSETP failed\n");
	return -1;
    }
    return 0;
}
#elif defined(NEXUS_ARCH_SUNOS41) || defined(NEXUS_ARCH_HPUX) \
     || defined(NEXUS_ARCH_AIX) || defined(NEXUS_ARCH_SGI) \
     || defined(NEXUS_ARCH_SOLARIS) 

#include <termio.h>

static int echo()
{
    struct termio tty_orig, tty_new;

    if (ioctl(0, TCGETA, &tty_orig) < 0)
    {
	fprintf(stderr, "ioctl() TCGETA failed\n");
	return -1;
    }

    tty_new = tty_orig;
    tty_new.c_lflag &= ~(ECHO);

    if (ioctl(0, TCSETA, &tty_new) < 0)
    {
	fprintf(stderr, "ioctl() TCSETA failed\n");
	return -1;
    }

    return 0;
}
#endif
