/*
 * macros.h
 *
 * rcsid = "$Header: /home/compose/Master/FM/src/nexus/macros.h,v 1.28 1995/08/22 20:31:33 tuecke Exp $"
 * nexus_rcsid = "Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/macros.h,v 1.17 1995/08/01 01:26:19 tuecke Exp "
 */

#ifndef _NEXUS_INCLUDE_MACROS_H
#define _NEXUS_INCLUDE_MACROS_H

/*
 * _nx_context()
 *
 * Get the context for the current thread and save it
 * into *Context (nexus_context_t **).
 */
#ifdef NEXUS_LITE
#define _nx_context(Context) \
    *(Context) = _nx_lite_context
#else
#define _nx_context(Context) \
    *(Context) = (nexus_context_t *)nexus_thread_getspecific(_nx_context_key)
#endif


/*
 * _nx_set_context()
 *
 * Set Context to be the context of the current thread.
 */
#ifdef NEXUS_LITE
#define _nx_set_context(Context) \
    _nx_lite_context = (Context)
#else
#define _nx_set_context(Context) \
    nexus_thread_setspecific(_nx_context_key, (void *) (Context));
#endif


#ifndef NEXUS_LITE

/*
 * _nx_thread_context()
 *
 * Set *Context (nexus_context_t **) to be the context for the passed thread.
 */
#define _nx_thread_context(Thread, Context) \
    *(Context) = ((Thread)->context)

/*
 * _nx_thread_set_context()
 *
 * Set the context (nexus_context_t *) for the passed thread.
 *
 * This may be used by a message handling thread to set itself
 * to the appropriate context before handling a rsr.
 */
#define _nx_thread_set_context(Thread, Context) \
    (Thread)->context = (Context)

#endif /* NEXUS_LITE */


/*
 * _nx_context_id()
 *
 * Set *Context_ID (int *) to be the context id for the context
 * of the calling thread.
 */
#define _nx_context_id(Context_ID) \
{ \
    nexus_context_t *__context; \
    _nx_context(&__context); \
    *(Context_ID) = __context->id; \
}


/*
 * _nx_node_id()
 *
 * Set *Node_ID (int *) to be the node id for the node
 * of the calling thread.
 */
#define _nx_node_id(Node_ID) \
{ \
    *(Node_ID) = _nx_my_node_id; \
}


#ifdef NEXUS_DEBUG
nexus_bool_t NexusDbgState( unsigned long, unsigned long,
			      unsigned long, unsigned long );
#define NexusDebug(Level) (_nx_debug_level >= (Level) || NexusDbgState( 0xffffffff, 0xffffffff, 0xffffffff, Level ))
#endif

#ifdef NEXUS_DEBUG
#define nexus_debug_printf(level, message) \
do { \
    if (NexusDebug(level)) \
    { \
	nexus_printf message; \
    } \
} while (0)
#define nexus_new_debug_printf(catagory, module, operation, level, message) \
do { \
    if (NexusDbgState(catagory, module, operation, level)) \
    { \
	nexus_printf message; \
    } \
} while (0)
#else
#define nexus_debug_printf(level, message)
#define nexus_new_debug_printf(level, message)
#endif

#ifdef NEXUS_DEBUG
#define NexusBufferMagicCheck(func, buf) \
{ \
    if (*(buf) == (nexus_buffer_t) NULL) \
	nexus_fatal(#func ": NULL nexus_buffer_t\n"); \
    if ((*(buf))->magic != NEXUS_BUFFER_MAGIC) \
	nexus_fatal(#func ": Bad magic number %x on buffer %x\n", \
		    (*(buf))->magic, buf); \
}
#else
#define NexusBufferMagicCheck(func, buf)
#endif


#ifdef NEXUS_LITE
/*
 * We need to spoof a few functions in the NexusLite version.
 */
#define _nx_thread_id(Thread_ID) *(Thread_ID) = 0;
#endif /* NEXUS_LITE */

#endif /* _NEXUS_INCLUDE_MACROS_H */
